#ifndef __MLIB_BIT_SET_H__
#define __MLIB_BIT_SET_H__

#include <mlib.h>

M_BEGIN_HEADER

 /*************************************************************************
  *
  * @section: bitset
  *
  * @title: Bit Sets
  *
  * @abstract: Fast bit sets
  *
  * @description:
  *   an MLib @M_Bitset is a dynamically growable set of bits
  */

 /***********************************************************************
  *
  * @type: M_Bitset
  *
  * @description:
  *    handle to a @M_BitsetRec object, used to model a given bitset
  */
  typedef struct M_BitsetRec_*          M_Bitset;


 /***********************************************************************
  *
  * @function: m_bitset_new
  *
  * @description:
  *    create a new (empty) bitset
  *
  * @input:
  *    memory :: current memory handler
  *    push   :: boolean. when set, the new bitset is pushed on the
  *              cleanup stack
  * @return:
  *    handle to new bitset, cannot be NULL
  *
  * @throws: out-of-memory
  *
  * @also:
  *    @m_bitset_destroy, @m_bitset_pop, @m_bitset_push
  *
  * @throws: out-of-memory
  */
  MLIB_API(M_Bitset)   m_bitset_new( M_Memory  memory,
                                     M_Bool    push );


 /***********************************************************************
  *
  * @function: m_bitset_copy
  *
  * @description:
  *    returns the copy of a given bitset
  *
  * @input:
  *    source  :: source bitset handle
  *    push    :: boolean. when set, the copy is pushed on the cleanup stack
  *
  * @return:
  *    handle to bitset copy
  *
  * @throws: out-of-memory
  */
  MLIB_API(M_Bitset)   m_bitset_copy( M_Bitset  source,
                                      M_Bool    push );


 /***********************************************************************
  *
  * @function: m_bitset_destroy
  *
  * @description:
  *    destroy a given bitset (you must ensure that the bitset is not
  *    on the cleanup stack)
  *
  * @input:
  *    bitset  :: target bitset handle
  */
  MLIB_API(void)       m_bitset_destroy( M_Bitset  bitset );


 /***********************************************************************
  *
  * @function: m_bitset_is_empty
  *
  * @description:
  *    a function that returns true iff a bitset is empty (i.e. all bits
  *    to 0)
  *
  * @input:
  *    bitset  :: source bitset handle
  *
  * @return:
  *    boolean
  */
  MLIB_API(M_Bool)     m_bitset_is_empty( M_Bitset  bitset );


 /***********************************************************************
  *
  * @function: m_bitset_ensure
  *
  * @description:
  *    reallocates a bitset block pointers table in order to ensure that
  *    it can hold at least "count" bits.
  *
  *    this function is used internally, and shouldn't be called directly
  *    by client applications, except in certain rare cases..
  *
  * @input:
  *    source  :: source bitset handle
  *    count   :: minimum bit count to ensure.
  *
  * @throws: out-of-memory
  */
  MLIB_API(void)       m_bitset_ensure( M_Bitset  source,
                                        M_UInt    count );


 /***********************************************************************
  *
  * @function: m_bitset_get
  *
  * @description:
  *    return the value of a given bit in a bitset
  *
  * @input:
  *    bitset  :: source bitset handle
  *    index   :: bit position, starting from 0. there is no upper
  *               limit but bits are set to 0 by default.
  *
  * @return:
  *    bit value (0 or 1)
  */
  MLIB_API(M_UInt)     m_bitset_get ( M_Bitset  bitset,
                                      M_UInt    index );


 /***********************************************************************
  *
  * @function: m_bitset_get2
  *
  * @description:
  *    return the value of a given 2-bit cell in a bitset
  *
  * @input:
  *    bitset  :: source bitset handle
  *    index   :: cell position, starting from 0. there is no upper
  *               limit but cells are set to 0 by default.
  *
  * @return:
  *    cell value (0 to 3)
  *
  * @note:
  *    this functions reads the bits at [2*index..2*index+1]
  */
  MLIB_API(M_UInt)     m_bitset_get2( M_Bitset  bitset,
                                      M_UInt    index );

 /***********************************************************************
  *
  * @function: m_bitset_get4
  *
  * @description:
  *    return the value of a given 4-bit cell in a bitset
  *
  * @input:
  *    bitset  :: source bitset handle
  *    index   :: cell position, starting from 0. there is no upper
  *               limit but cells are set to 0 by default.
  *
  * @return:
  *    cell value (0 to 15)
  *
  * @note:
  *    this functions reads the bits at [4*index..4*index+3]
  */
  MLIB_API(M_UInt)     m_bitset_get4( M_Bitset  bitset,
                                      M_UInt    index );

 /***********************************************************************
  *
  * @function: m_bitset_get8
  *
  * @description:
  *    return the value of a given 8-bit cell in a bitset
  *
  * @input:
  *    bitset  :: source bitset handle
  *    index   :: cell position, starting from 0. there is no upper
  *               limit but cells are set to 0 by default.
  *
  * @return:
  *    cell value (0 to 255)
  *
  * @note:
  *    this functions reads the bits at [8*index..8*index+7]
  */
  MLIB_API(M_UInt)     m_bitset_get8( M_Bitset  bitset,
                                      M_UInt    index );


 /***********************************************************************
  *
  * @function: m_bitset_set
  *
  * @description:
  *    set the value of a given bit in a bitset
  *
  * @input:
  *    bitset  :: target bitset handle
  *    index   :: bit position, starting from 0. there is no upper
  *               limit as the bitset is extended accordingly
  *    value   :: bit value (note that is is anded with 1)
  *
  * @throws: out-of-memory
  */
  MLIB_API(void)       m_bitset_set ( M_Bitset   bitset,
                                      M_UInt     index,
                                      M_UInt     value );


 /***********************************************************************
  *
  * @function: m_bitset_set2
  *
  * @description:
  *    set the value of a given 2-bit cell in a bitset
  *
  * @input:
  *    bitset  :: target bitset handle
  *    index   :: bit position, starting from 0. there is no upper
  *               limit as the bitset is extended accordingly
  *    value   :: new cell value
  *
  * @note:
  *    the value is anded with 3 before being applied
  *    this function sets the bits at [2*index..2*index+1]
  *
  * @throws: out-of-memory
  */
  MLIB_API(void)       m_bitset_set2( M_Bitset   bitset,
                                      M_UInt     index,
                                      M_UInt     value );

 /***********************************************************************
  *
  * @function: m_bitset_set4
  *
  * @description:
  *    set the value of a given 4-bit cell in a bitset
  *
  * @input:
  *    bitset  :: target bitset handle
  *    index   :: bit position, starting from 0. there is no upper
  *               limit as the bitset is extended accordingly
  *    value   :: new cell value
  *
  * @note:
  *    the value is anded with 15 before being applied
  *    this function sets the bits at [4*index..4*index+3]
  *
  * @throws: out-of-memory
  */
  MLIB_API(void)       m_bitset_set4( M_Bitset   bitset,
                                      M_UInt     index,
                                      M_UInt     value );

 /***********************************************************************
  *
  * @function: m_bitset_set8
  *
  * @description:
  *    set the value of a given 8-bit cell in a bitset
  *
  * @input:
  *    bitset  :: target bitset handle
  *    index   :: bit position, starting from 0. there is no upper
  *               limit as the bitset is extended accordingly
  *    value   :: new cell value
  *
  * @note:
  *    the value is anded with 255 before being applied
  *    this function sets the bits at [8*index..8*index+7]
  *
  * @throws: out-of-memory
  */
  MLIB_API(void)       m_bitset_set8( M_Bitset   bitset,
                                      M_UInt     index,
                                      M_UInt     value );


 /***********************************************************************
  *
  * @function: m_bitset_set_range
  *
  * @description:
  *    this function is used to set a range of bit cells to a single
  *    value.
  *
  * @input:
  *    bitset  :: target bitset handle
  *    start   :: start bit/cell position
  *    count   :: number of bits/cells to set
  *    size    :: size of bit cell (1,2,4 or 8)
  *    value   :: new bit/cell value
  *
  * @note:
  *    this function will panic if "size" is not 1, 2, 4 or 8
  *
  *    the value is anded with (1 << size)-1 before being applied
  *    this function sets the bits at [size*index..size*index+size-1]
  *
  * @throws: out-of-memory
  */
  MLIB_API(void)       m_bitset_set_range( M_Bitset  bitset,
                                           M_UInt    start,
                                           M_UInt    count,
                                           M_UInt    size,
                                           M_UInt    value );

 /***********************************************************************
  *
  * @function: m_bitset_test_intersect
  *
  * @description:
  *    returns true iff two bitsets intersect
  *
  * @input:
  *    bitset1 :: first bitset
  *    bitset2 :: second bitset
  *
  * @note:
  *   this function is a quick way to test if two bitsets intersect.
  *   however, if you need the intersection itself, consider using
  *   @m_bitset_intersect directly..
  */
  MLIB_API(M_Bool)     m_bitset_test_intersect( M_Bitset  bitset1,
                                                M_Bitset  bitset2 );

 /***********************************************************************
  *
  * @function: m_bitset_clear
  *
  * @description:
  *    clear a bitset, i.e. set all bits/cells to 0
  *
  * @input:
  *    bitset :: target bitset handle
  */
  MLIB_API(void)       m_bitset_clear( M_Bitset  bitset );


 /***********************************************************************
  *
  * @function: m_bitset_union
  *
  * @description:
  *    computes the union of two bitsets
  *
  * @input:
  *    bitset1 :: first bitset
  *    bitset2 :: second bitset
  *    push    :: boolean. if set, the result is pushed on the cleanup stack
  *
  * @return:
  *    union of the two bitsets
  *
  * @throws: out-of-memory
  */
  MLIB_API(M_Bitset)   m_bitset_union( M_Bitset  bitset1,
                                       M_Bitset  bitset2,
                                       M_Bool    push );


 /***********************************************************************
  *
  * @function: m_bitset_intersect
  *
  * @description:
  *    computes the intersection of two bitsets
  *
  * @input:
  *    bitset1 :: first bitset
  *    bitset2 :: second bitset
  *    push    :: boolean. if set, the result is pushed on the cleanup stack
  *
  * @return:
  *    intersection of the two bitsets
  *
  * @throws: out-of-memory
  */
  MLIB_API(M_Bitset)   m_bitset_intersect( M_Bitset  bitset1,
                                           M_Bitset  bitset2,
                                           M_Bool    push );


 /***********************************************************************
  *
  * @function: m_bitset_substract
  *
  * @description:
  *    computes the substraction of two bitsets
  *
  * @input:
  *    bitset1 :: first bitset
  *    bitset2 :: second bitset
  *    push    :: boolean. if set, the result is pushed on the cleanup stack
  *
  * @return:
  *    bitset1 - bitset2
  *
  * @throws: out-of-memory
  */
  MLIB_API(M_Bitset)   m_bitset_substract( M_Bitset  bitset1,
                                           M_Bitset  bitset2,
                                           M_Bool    push );


 /***********************************************************************
  *
  * @function: m_bitset_compress
  *
  * @description:
  *    compress a given bitset in memory.
  *
  * @input:
  *    bitset :: handle to target bitset
  */
  MLIB_API(void)       m_bitset_compress( M_Bitset  bitset );


 /***********************************************************************
  *
  * @function: m_bitset_push
  *
  * @description:
  *    push a given bitset on the cleanup stack
  *
  * @input:
  *    bitset  :: target bitset handle
  */
  MLIB_API(void)       m_bitset_push( M_Bitset  set );


 /***********************************************************************
  *
  * @function: m_bitset_pop
  *
  * @description:
  *    pops a given bitset from the cleanup stack
  *
  * @input:
  *    bitset  :: target bitset handle
  *    action  :: action to perform on pop
  */
  MLIB_API(void)       m_bitset_pop( M_Bitset     bitset,
                                     M_PopAction  action );

  /* */

M_END_HEADER

#endif /* __MLIB_SET_H__ */

