#ifndef __MLIB_BUFFER_H__
#define __MLIB_BUFFER_H__

#include <mlib.h>
#include <mlib/bytes.h>

M_BEGIN_HEADER

 /*************************************************************************
  *
  * @type: M_Buffer
  *
  * @description:
  *   handle to a @M_BufferRec structure used to model a memory block
  *   used for buffering reads or writes.
  *
  *   note that @M_Buffer structures do not own the memory they address,
  *   they only provide mechanisms to read, write and peek in/to it..
  */
  typedef struct M_BufferRec_*   M_Buffer;


 /*************************************************************************
  *
  * @functype: M_Buffer_FeedFunc
  *
  * @description:
  *   the type of a function used to "feed" a buffer with new data.
  *
  * @input:
  *   buffer :: handle to target buffer
  *   wanted :: the minimal number of bytes we want to access from the
  *             buffer after the feed.
  *
  * @return:
  *   the number of bytes available in the buffer for further reads or
  *   writes. This must be "buffer.limit - buffer.cur"
  *
  * @note:
  *   the feed function should behave as follows. let 'b' be the buffer
  *   structure (i.e. "*buffer")
  *
  *   - if this is a write buffer, it should "flush" the data between
  *     "b.base" and "b.cur" to the corresponding output stream
  *
  *   - it should preserve all data between "b.cur" and "b.limit" if any,
  *     to ensure that they're still available after the feed..
  *
  *   - it should update "b.cur" and "b.pos" correctly. It might eventually
  *     also update "b.base" and "b.limit"
  *
  *   - it should always return "b.limit - b.cur"
  *
  *   - it should always ensure that "b.base <= b.cur <= b.limit"
  *
  *   - end of file is noticed by "b.base == b.cur == b.limit" and
  *     a return value of 0
  */
  typedef M_Size  (*M_Buffer_FeedFunc)( M_Buffer  buffer,
                                        M_Size    wanted );


 /*************************************************************************
  *
  * @struct: M_BufferRec
  *
  * @description:
  *   a structure used to model a memory buffer.
  *
  * @fields:
  *   cur   :: current cursor position within buffer
  *   base  :: buffer memory address
  *   limit :: buffer memory limit (i.e. first byte after the buffer)
  *
  *   pos   :: gives the offset corresponding to the data addressed
  *            by 'base' in the buffer's corresponding stream
  *
  *   feed  :: the buffer feed function
  *   data  :: generic pointer, can be used by custom feed functions
  *
  * @note:
  *   the condition 'base' <= 'cur' <= 'limit' must _always_ be ensured
  *   by a valid buffer
  */
  typedef struct M_BufferRec_
  {
    M_ByteP            cur;
    M_ByteP            limit;
    M_ByteP            base;
    M_Size             pos;
    M_Buffer_FeedFunc  feed;
    M_Pointer          data;

  } M_BufferRec;


 /*************************************************************************
  *
  * @function: m_buffer_init
  *
  * @description:
  *   initialize a buffer structure
  *
  * @input:
  *   buffer :: target buffer handle
  *   feed   :: buffer feed function
  *   data   :: generic feed data
  *
  * @note:
  *   all fields in the structure are set to 0 or NULL, it's up to the
  *   feed function to set 'base', 'cur' and 'limit' the first time it
  *   is called..
  */
  MLIB_API( void )
  m_buffer_init( M_Buffer           buffer,
                 M_Buffer_FeedFunc  feed,
                 M_Pointer          data );


 /*************************************************************************
  *
  * @function: m_buffer_init_fixed
  *
  * @description:
  *   initialize a buffer structure to use a fixed-size memory block as
  *   the input/output stream. this is generally useful when you want to
  *   use the buffer methods to fill local or static byte arrays..
  *
  * @input:
  *   buffer :: target buffer handle
  *   base   :: fixed-size memory block address
  *   size   :: size of memory block in bytes
  */
  MLIB_API( void )
  m_buffer_init_fixed( M_Buffer   buffer,
                       M_Byte*    base,
                       M_Size     size );


 /*************************************************************************
  *
  * @function: m_buffer_init_full
  *
  * @description:
  *   perform a full initialization of a buffer structure
  *
  * @input:
  *   buffer :: target buffer handle
  *   base   :: memory buffer address
  *   size   :: size of memory buffer in bytes
  *   feed   :: buffer feed function
  *   data   :: generic feed data
  *
  * @note:
  *   this function sets the buffer's position to zero.
  */
  MLIB_API( void )
  m_buffer_init_full( M_Buffer           buffer,
                      M_ByteP            base,
                      M_Size             size,
                      M_Buffer_FeedFunc  feed,
                      M_Pointer          data );


 /*************************************************************************
  *
  * @function: m_buffer_offset
  *
  * @description:
  *   return the buffer's current position within the corresponding
  *   stream.
  *
  * @input:
  *   buffer :: target buffer handle
  *
  * @return:
  *   current position in buffer's stream
  */
  MLIB_API( M_Size )
  m_buffer_offset( M_Buffer  buffer );


 /*************************************************************************
  *
  * @function: m_buffer_read
  *
  * @description:
  *   try to read data from a buffer
  *
  * @input:
  *   buffer :: target buffer handle
  *   dest   :: destination address
  *   wanted :: number of bytes to read
  *
  * @return:
  *   number of bytes really read, less or equal to "wanted".
  *   0 means "end of stream"
  *
  * @throws: any exception raised by the buffer feed function
  */
  MLIB_API( M_Size )
  m_buffer_read( M_Buffer  buffer,
                 M_ByteP   dest,
                 M_Size    wanted );


 /*************************************************************************
  *
  * @function: m_buffer_skip
  *
  * @description:
  *   try to skip data in a buffer
  *
  * @input:
  *   buffer :: target buffer handle
  *   wanted :: number of bytes to skip
  *
  * @return:
  *   number of bytes really skipped, less or equal to "wanted"
  *   0 means "end of stream"
  *
  * @throws: any exception raised by the buffer feed function
  */
  MLIB_API( M_Size )
  m_buffer_skip( M_Buffer  buffer,
                 M_Size    wanted );


 /*************************************************************************
  *
  * @function: m_buffer_write
  *
  * @description:
  *   try to write bytes to a buffer
  *
  * @input:
  *   buffer :: target buffer handle
  *   src    :: source address
  *   wanted :: number of bytes to write
  *
  * @return:
  *   number of bytes really written, less or equal to "wanted".
  *   0 means "not enough space"
  *
  * @throws: any exception raised by the buffer feed function
  */
  MLIB_API( M_Size )
  m_buffer_write( M_Buffer  buffer,
                  M_CByteP  src,
                  M_Size    wanted );


 /*************************************************************************
  *
  * @function: m_buffer_fill
  *
  * @description:
  *   try to fill a buffer with a byte value
  *
  * @input:
  *   buffer :: target buffer handle
  *   value  :: fill value
  *   wanted :: number of bytes to write
  *
  * @return:
  *   number of bytes really written, less or equal to "wanted".
  *   0 means "not enough space"
  *
  * @throws: any exception raised by the buffer feed function
  */
  MLIB_API( M_Size )
  m_buffer_fill( M_Buffer  buffer,
                 M_Byte    value,
                 M_Size    wanted );


 /*************************************************************************
  *
  * @function: m_buffer_flush
  *
  * @description:
  *   flush a buffer
  *
  * @input:
  *   buffer :: target buffer handle
  *
  * @return:
  *   number of bytes flushed.
  *
  * @throws: any exception raised by the buffer feed function
  *
  * @note:
  *   this simply calls 'buffer->feed(buffer,0)' and will only produce
  *   something for output/write buffers.
  */
  MLIB_API( M_Size )
  m_buffer_flush( M_Buffer  buffer );


 /*************************************************************************
  *
  * @function: m_buffer_peek
  *
  * @description:
  *   checks wether there is at least "wanted" bytes available in a buffer
  *   for reading or writing. If it does, increment the buffer's cursor
  *   returning the available "place"..
  *
  * @input:
  *   buffer :: target buffer handle
  *   wanted :: number of bytes to look for
  *
  * @return:
  *   in case of success, the address of the buffer's cursor before the
  *   function was called. Otherwise, NULL to indicate that there isn't
  *   enough room..
  *
  *   note that in case of success, the function also increments the
  *   buffer's cursor..
  *
  * @throws: any exception raised by the buffer feed function
  *
  * @note:
  *  this function might call the buffer feed function.
  */
  MLIB_API( M_Byte* )
  m_buffer_peek( M_Buffer  buffer,
                 M_Size    wanted );

 /*************************************************************************
  *
  * @function: m_buffer_get
  *
  * @description:
  *   read one byte from a buffer. return 0 in case of end-of-stream
  *
  * @input:
  *   buffer :: target buffer handle
  *
  * @return:
  *   byte value.
  *
  * @throws: any exception raised by the buffer feed function
  */
  MLIB_API( M_Byte )
  m_buffer_get( M_Buffer  buffer );


 /*************************************************************************
  *
  * @function: m_buffer_get_line
  *
  * @description:
  *   read one line of text from a buffer. this allocates a new string
  *   in the heap and skips the newline when found.
  *
  * @input:
  *   buffer :: target buffer handle
  *   memory :: memory manager handle
  *
  * @output:
  *   alen   :: if not NULL, the address of an integer that will receive
  *             the length of the string (excluding the trailing zero)
  *
  * @return:
  *   string handle. NULL in case of error (i.e. end of stream)
  *
  * @throws: any exception raised by the buffer feed function
  *
  * @note:
  *   this function detects newlines and filters them from the buffer
  *   (be they in Unix/PC/Mac format).
  *
  *   you can thus call this function repeatedly to read a text file line
  *   by line easily.
  */
  MLIB_API( M_String )
  m_buffer_get_line( M_Buffer   buffer,
                     M_Memory   memory,
                     M_Size    *alen );


 /*************************************************************************
  *
  * @function: m_buffer_get_string
  *
  * @description:
  *   read a string from a buffer. the string is delimited by a '\0'
  *   and all newlines are included
  *
  * @input:
  *   buffer :: target buffer handle
  *   memory :: memory manager handle
  *
  * @output:
  *   alen   :: if not NULL, the address of an integer that will receive
  *             the length of the string (excluding the trailing zero)
  *
  * @return:
  *   string handle. NULL in case of error (i.e. end of stream)
  *
  * @throws: any exception raised by the buffer feed function
  */
  MLIB_API( M_String )
  m_buffer_get_string( M_Buffer   buffer,
                       M_Memory   memory,
                       M_Size    *alen );


 /*************************************************************************
  *
  * @function: m_buffer_put
  *
  * @description:
  *   try to write a byte value to a buffer
  *
  * @input:
  *   buffer :: target buffer handle
  *   v      :: byte value
  *
  * @return:
  *   number of bytes really written (0 or 1)
  *
  * @throws: any exception raised by the buffer feed function
  */
  MLIB_API( M_Size )
  m_buffer_put( M_Buffer  buffer,
                M_Byte    v );


 /*************************************************************************
  *
  * @function: m_buffer_put_string
  *
  * @description:
  *   try to write a string to a buffer (including the trailing zero)
  *
  * @input:
  *   buffer :: target buffer handle
  *   str    :: input string
  *
  * @return:
  *   number of bytes really written, including the trailing zero
  *
  * @throws: any exception raised by the buffer feed function
  */
  MLIB_API( M_Size )
  m_buffet_put_string( M_Buffer   buffer,
                       M_CString  str );


 /*************************************************************************
  *
  * @function: m_buffer_put16_ube
  *
  * @description:
  *   try to write a 16-bit unsigned value in big-endian format to a
  *   buffer.
  *
  * @input:
  *   buffer :: target buffer handle
  *   v      :: value
  *
  * @return:
  *   number of bytes really written
  *
  * @throws: any exception raised by the buffer feed function
  */
  MLIB_API( M_Size )
  m_buffer_put16_ube( M_Buffer  buffer,
                      M_UInt16  value );


 /*************************************************************************
  *
  * @function: m_buffer_put16_ule
  *
  * @description:
  *   try to write a 16-bit unsigned value in little-endian format to a
  *   buffer.
  *
  * @input:
  *   buffer :: target buffer handle
  *   v      :: value
  *
  * @return:
  *   number of bytes really written
  *
  * @throws: any exception raised by the buffer feed function
  */
  MLIB_API( M_Size )
  m_buffer_put16_ule( M_Buffer  buffer,
                      M_UInt16  value );


 /*************************************************************************
  *
  * @function: m_buffer_put24_ube
  *
  * @description:
  *   try to write a 24-bit unsigned value in big-endian format to a
  *   buffer.
  *
  * @input:
  *   buffer :: target buffer handle
  *   v      :: value
  *
  * @return:
  *   number of bytes really written
  *
  * @throws: any exception raised by the buffer feed function
  */
  MLIB_API( M_Size )
  m_buffer_put24_ube( M_Buffer  buffer,
                      M_UInt32  value );


 /*************************************************************************
  *
  * @function: m_buffer_put24_ule
  *
  * @description:
  *   try to write a 24-bit unsigned value in little-endian format to a
  *   buffer.
  *
  * @input:
  *   buffer :: target buffer handle
  *   v      :: value
  *
  * @return:
  *   number of bytes really written
  *
  * @throws: any exception raised by the buffer feed function
  */
  MLIB_API( M_Size )
  m_buffer_put24_ule( M_Buffer  buffer,
                      M_UInt32  value );


 /*************************************************************************
  *
  * @function: m_buffer_put32_ube
  *
  * @description:
  *   try to write a 32-bit unsigned value in big-endian format to a
  *   buffer.
  *
  * @input:
  *   buffer :: target buffer handle
  *   v      :: value
  *
  * @return:
  *   number of bytes really written
  *
  * @throws: any exception raised by the buffer feed function
  */
  MLIB_API( M_Size )
  m_buffer_put32_ube( M_Buffer  buffer,
                      M_UInt32  value );


 /*************************************************************************
  *
  * @function: m_buffer_put32_ule
  *
  * @description:
  *   try to write a 32-bit unsigned value in little-endian format to a
  *   buffer.
  *
  * @input:
  *   buffer :: target buffer handle
  *   v      :: value
  *
  * @return:
  *   number of bytes really written
  *
  * @throws: any exception raised by the buffer feed function
  */
  MLIB_API( M_Size )
  m_buffer_put32_ule( M_Buffer  buffer,
                      M_UInt32  value );

 /*************************************************************************
  *
  * @function: m_buffer_put64_ube
  *
  * @description:
  *   try to write a 64-bit unsigned value in big-endian format to a
  *   buffer.
  *
  * @input:
  *   buffer :: target buffer handle
  *   v      :: value
  *
  * @return:
  *   number of bytes really written
  *
  * @throws: any exception raised by the buffer feed function
  */
#ifdef M_CONFIG_HAS_INT64

  MLIB_API( M_Size )
  m_buffer_put64_ube( M_Buffer  buffer,
                      M_UInt64  value );

#endif

 /*************************************************************************
  *
  * @function: m_buffer_put64_ule
  *
  * @description:
  *   try to write a 64-bit unsigned value in little-endian format to a
  *   buffer.
  *
  * @input:
  *   buffer :: target buffer handle
  *   v      :: value
  *
  * @return:
  *   number of bytes really written
  *
  * @throws: any exception raised by the buffer feed function
  */
#ifdef M_CONFIG_HAS_INT64

  MLIB_API( M_Size )
  m_buffer_put64_ule( M_Buffer  buffer,
                      M_UInt64  value );

#endif


 /*************************************************************************
  *
  * @function: m_buffer_get16_ube
  *
  * @description:
  *   retrieve a 16-bit unsigned integer in big-endian format from a
  *   buffer
  *
  * @input:
  *   buffer :: target buffer handle
  *
  * @return:
  *   integer value. will be 0 if not enough bytes could be read from
  *   the buffer
  *
  * @throws: any exception raised by the buffer feed function
  */
  MLIB_API( M_UInt16 )
  m_buffer_get16_ube( M_Buffer  buffer );


 /*************************************************************************
  *
  * @function: m_buffer_get16_ule
  *
  * @description:
  *   retrieve a 16-bit unsigned integer in little-endian format from a
  *   buffer
  *
  * @input:
  *   buffer :: target buffer handle
  *
  * @return:
  *   integer value. will be 0 if not enough bytes could be read from
  *   the buffer
  *
  * @throws: any exception raised by the buffer feed function
  */
  MLIB_API( M_UInt16 )
  m_buffer_get16_ule( M_Buffer  buffer );


 /*************************************************************************
  *
  * @function: m_buffer_get24_ube
  *
  * @description:
  *   retrieve a 24-bit unsigned integer in big-endian format from a
  *   buffer
  *
  * @input:
  *   buffer :: target buffer handle
  *
  * @return:
  *   integer value. will be 0 if not enough bytes could be read from
  *   the buffer
  *
  * @throws: any exception raised by the buffer feed function
  */
  MLIB_API( M_UInt32 )
  m_buffer_get24_ube( M_Buffer  buffer );


 /*************************************************************************
  *
  * @function: m_buffer_get24_ule
  *
  * @description:
  *   retrieve a 24-bit unsigned integer in little-endian format from a
  *   buffer
  *
  * @input:
  *   buffer :: target buffer handle
  *
  * @return:
  *   integer value. will be 0 if not enough bytes could be read from
  *   the buffer
  *
  * @throws: any exception raised by the buffer feed function
  */
  MLIB_API( M_UInt32 )
  m_buffer_get24_ule( M_Buffer  buffer );


 /*************************************************************************
  *
  * @function: m_buffer_get32_ube
  *
  * @description:
  *   retrieve a 32-bit unsigned integer in big-endian format from a
  *   buffer
  *
  * @input:
  *   buffer :: target buffer handle
  *
  * @return:
  *   integer value. will be 0 if not enough bytes could be read from
  *   the buffer
  *
  * @throws: any exception raised by the buffer feed function
  */
  MLIB_API( M_UInt32 )
  m_buffer_get32_ube( M_Buffer  buffer );


 /*************************************************************************
  *
  * @function: m_buffer_get32_ule
  *
  * @description:
  *   retrieve a 32-bit unsigned integer in little-endian format from a
  *   buffer
  *
  * @input:
  *   buffer :: target buffer handle
  *
  * @return:
  *   integer value. will be 0 if not enough bytes could be read from
  *   the buffer
  *
  * @throws: any exception raised by the buffer feed function
  */
  MLIB_API( M_UInt32 )
  m_buffer_get32_ule( M_Buffer  buffer );


 /*************************************************************************
  *
  * @function: m_buffer_get64_ube
  *
  * @description:
  *   retrieve a 64-bit unsigned integer in big-endian format from a
  *   buffer
  *
  * @input:
  *   buffer :: target buffer handle
  *
  * @return:
  *   integer value. will be 0 if not enough bytes could be read from
  *   the buffer
  *
  * @throws: any exception raised by the buffer feed function
  */
#ifdef M_CONFIG_HAS_INT64

  MLIB_API( M_UInt64 )
  m_buffer_get64_ube( M_Buffer  buffer );

#endif

 /*************************************************************************
  *
  * @function: m_buffer_get64_ule
  *
  * @description:
  *   retrieve a 64-bit unsigned integer in little-endian format from a
  *   buffer
  *
  * @input:
  *   buffer :: target buffer handle
  *
  * @return:
  *   integer value. will be 0 if not enough bytes could be read from
  *   the buffer
  *
  * @throws: any exception raised by the buffer feed function
  */
#ifdef M_CONFIG_HAS_INT64

  MLIB_API( M_UInt64 )
  m_buffer_get64_ule( M_Buffer  buffer );

#endif

/* */

#define  m_assert_is_buffer(x)   \
            m_assert( (x) && (x)->feed &&   \
                      (x)->cur >= (x)->base && (x)->cur <= (x)->limit )

M_END_HEADER

#endif /* __MLIB_BUFFER_H__ */
