#ifndef __MLIB_CALC_H__
#define __MLIB_CALC_H__

#include <mlib.h>

M_BEGIN_HEADER

 /***********************************************************************/
 /***********************************************************************/
 /*****                                                             *****/
 /*****                  useful integer computations                *****/
 /*****                                                             *****/
 /***********************************************************************/
 /***********************************************************************/

 /**************************************************************************
  *
  * @function: m_imuldiv
  *
  * @description:
  *   perform the computation "a*b/c" where all parameters are signed 32-bit
  *   integers, with maximum accuracy
  *
  * @input:
  *   a  :: first value
  *   b  :: second value
  *   c  :: third value (divisor)
  *
  * @return:
  *   result of 'a*b/c'. this will be @M_INT32_MIN or @M_INT32_MAX in case
  *   of overflow, including division by zero
  */
  MLIB_API( M_Int32  )    m_imuldiv( M_Int32   a,
                                     M_Int32   b,
                                     M_Int32   c );

 /**************************************************************************
  *
  * @function: m_imuldiv_round
  *
  * @description:
  *   same as @m_imuldiv with rounding
  *
  * @input:
  *   a  :: first value
  *   b  :: second value
  *   c  :: third value (divisor)
  *
  * @return:
  *   result of 'a*b/c' with rounding (i.e. "(a*b + c/2)/c" when c is positive,
  *   and (a*b - c/2)/c" when c is negative). this will be @M_INT32_MIN or
  *   @M_INT32_MAX in case of overflow, including division by zero
  */
  MLIB_API( M_Int32  )    m_imuldiv_round( M_Int32   a,
                                           M_Int32   b,
                                           M_Int32   c );

 /**************************************************************************
  *
  * @function: m_umuldiv
  *
  * @description:
  *   a variant of @m_imuldiv where all parameters and the result are
  *   unsigned 32-bit integers.
  *
  * @input:
  *   a  :: first value
  *   b  :: second value
  *   c  :: third value (divisor)
  *
  * @return:
  *   result of 'a*b/c'. can be @M_UINT32_MAX in case of overflow, including
  *   division by zero
  */
  MLIB_API( M_UInt32 )    m_umuldiv( M_UInt32  a,
                                     M_UInt32  b,
                                     M_UInt32  c );

 /**************************************************************************
  *
  * @function: m_umuldiv_round
  *
  * @description:
  *   same as @m_umuldiv with rounding
  *
  * @input:
  *   a  :: first value
  *   b  :: second value
  *   c  :: third value (divisor)
  *
  * @return:
  *   result of 'a*b/c' with rounding (i.e. "(a*b + c/2)/c").
  *   this will be @M_INT32_MIN or @M_INT32_MAX in case of overflow,
  *   including division by zero
  */
  MLIB_API( M_UInt32 )    m_umuldiv_round( M_UInt32  a,
                                           M_UInt32  b,
                                           M_UInt32  c );

 /***********************************************************************/
 /***********************************************************************/
 /*****                                                             *****/
 /*****                unsigned 64-bit management                   *****/
 /*****                                                             *****/
 /***********************************************************************/
 /***********************************************************************/

 /***********************************************************************
  *
  * @type: M_Card64
  *
  * @description:
  *  pointer to a @M_Card64Rec structure used to model unsigned 64-bit
  *  integers as a structure of two 32-bit integers.
  */
  typedef struct M_Card64Rec_*   M_Card64;


 /***********************************************************************
  *
  * @type: M_Card64Rec
  *
  * @description:
  *  a structure used to model an unsigned 64-bit integer. this is useful
  *  on platforms that do not provide a @M_Int64 type (see @M_CONFIG_HAS_INT64)
  *
  * @fields:
  *   hi  :: high 32-bits
  *   lo  :: low 32-bits
  */
  typedef struct M_Card64Rec_
  {
    M_UInt32  hi;
    M_UInt32  lo;

  } M_Card64Rec;


 /***********************************************************************
  *
  * @macro: M_CARD64
  *
  * @description:
  *  a convenient macro to cast any pointer to @M_Card64
  */
#define  M_CARD64(x)   ((M_Card64)(x))


 /***********************************************************************
  *
  * @function: m_card64_compare
  *
  * @description:
  *  compare two @M_Card64 values
  *
  * @fields:
  *   a  :: address of first value
  *   b  :: address of second value
  *
  * @return:
  *   -1, 0 or +1 depending on the values of 'a' and 'b'
  */
  MLIB_API( M_Int )
  m_card64_compare( const M_Card64  a,
                    const M_Card64  b );


 /***********************************************************************
  *
  * @function: m_card64_add
  *
  * @description:
  *  add a (hi,lo) pair of 32-bit integers to a @M_Card64
  *
  * @fields:
  *   z  :: address of target value
  *   hi :: high part of addition
  *   lo :: low part of addition
  */
  MLIB_API( void )
  m_card64_add( M_Card64   z,
                M_UInt32   hi,
                M_UInt32   lo );


 /***********************************************************************
  *
  * @function: m_card64_sub
  *
  * @description:
  *  substract a (hi,lo) pair of 32-bit integers to a @M_Card64
  *
  * @fields:
  *   z  :: address of target value
  *   hi :: high part of substract
  *   lo :: low part of substract
  */
  MLIB_API( void )
  m_card64_sub( M_Card64   z,
                M_UInt32   hi,
                M_UInt32   lo );


 /***********************************************************************
  *
  * @function: m_card64_mul
  *
  * @description:
  *  multiply two 32-bit unsigned integer and store the result in a @M_Card64
  *
  * @fields:
  *   z  :: address of target value
  *   a  :: first operand
  *   b  :: second operand
  */
  MLIB_API( void )
  m_card64_mul( M_Card64   z,
                M_UInt32   a,
                M_UInt32   b );


 /***********************************************************************
  *
  * @function: m_card64_div
  *
  * @description:
  *  divide a @M_Card64 by a 32-bit unsigned integer, return a
  *  32-bit unsigned quotient (remainder is in 'z')
  *
  * @fields:
  *   z       :: address of target value
  *   divisor :: divisor
  *
  * @return:
  *   division value, bounded to 0..M_UINT32_MAX in case of overflow
  *   (including division by zero)
  */
  MLIB_API( M_UInt32 )
  m_card64_div( M_Card64   z,
                M_UInt32   divisor );


 /***********************************************************************
  *
  * @function: m_card64_shl
  *
  * @description:
  *  shift the bits of a @M_Card64 to the left.
  *
  * @fields:
  *   z     :: address of target value
  *   shift :: shit in bits, can be < 0 for right shifts
  */
  MLIB_API( void )
  m_card64_shl( M_Card64   z,
                M_Int      shift );

 /***********************************************************************
  *
  * @function: m_card64_shr
  *
  * @description:
  *  shift the bits of a @M_Card64 to the right.
  *
  * @fields:
  *   z     :: address of target value
  *   shift :: shit in bits, can be < 0 for left shifts
  */
  MLIB_API( void )
  m_card64_shr( M_Card64   z,
                M_Int      shift );



 /***********************************************************************/
 /***********************************************************************/
 /*****                                                             *****/
 /*****                  signed 64-bit management                   *****/
 /*****                                                             *****/
 /***********************************************************************/
 /***********************************************************************/


 /***********************************************************************
  *
  * @type: M_Calc64
  *
  * @description:
  *  pointer to a @M_Calc64Rec structure used to model signed 64-bit
  *  integers as a structure of two 32-bit integers.
  */
  typedef struct M_Calc64Rec_*   M_Calc64;


 /***********************************************************************
  *
  * @type: M_Calc64Rec
  *
  * @description:
  *  a structure used to model a signed 64-bit integer. this is useful
  *  on platforms that do not provide a @M_Int64 type (see @M_CONFIG_HAS_INT64)
  *
  * @fields:
  *   hi  :: high 32-bits
  *   lo  :: low 32-bits
  */
  typedef struct M_Calc64Rec_
  {
    M_Int32   hi;
    M_UInt32  lo;

  } M_Calc64Rec;


 /***********************************************************************
  *
  * @macro: M_CALC64 (pointer)
  *
  * @description:
  *  a convenient macro to cast anything to a pointer
  */
#define  M_CALC64(x)  ((M_Calc64)(x))


 /***********************************************************************
  *
  * @function: m_calc64_compare
  *
  * @description:
  *  compare two 64-bit values
  *
  * @input:
  *   a  :: address of first value
  *   b  :: address of second value
  *
  * @return:
  *   -1, 0 or +1 depending on their values
  */
  MLIB_API( M_Int )
  m_calc64_compare( const M_Calc64  a,
                    const M_Calc64  b );


 /***********************************************************************
  *
  * @function: m_calc64_neg
  *
  * @description:
  *  negate a 64-bit signed value
  *
  * @input:
  *   z :: address of value
  */
  MLIB_API( void )
  m_calc64_neg( M_Calc64  z );


 /***********************************************************************
  *
  * @function: m_calc64_shl
  *
  * @description:
  *  shift a @M_Calc64 to the left
  *
  * @input:
  *   z     :: address of value
  *   shift :: number of bits to shift to the left, can be < 0
  *            (in which case, shift happens to the right)
  */
  MLIB_API( void )
  m_calc64_shl( M_Calc64  z,
                M_Int     shift );


 /***********************************************************************
  *
  * @function: m_calc64_shr
  *
  * @description:
  *  shift a @M_Calc64 to the right. All new bits are set to 0
  *
  * @input:
  *   z     :: address of value
  *   shift :: number of bits to shift to the right, can be < 0
  *            (in which case, shift happens to the right)
  *
  * @also: @m_calc64_sar
  */
  MLIB_API( void )
  m_calc64_shr( M_Calc64  z,
                M_Int     shift );


 /***********************************************************************
  *
  * @function: m_calc64_sar
  *
  * @description:
  *  shift a @M_Calc64 to the right. sign bit preserved
  *
  * @input:
  *   z     :: address of value
  *   shift :: number of bits to shift to the right, can be < 0
  *            (in which case, shift happens to the left)
  */
  MLIB_API( void )
  m_calc64_sar( M_Calc64  z,
                M_Int     shift );


 /***********************************************************************
  *
  * @function: m_calc64_add
  *
  * @description:
  *  add a (hi,lo) pair of 32-bit values a @M_Calc64. Note that 'hi' is
  *  a signed integer
  *
  * @input:
  *   z  :: address of target 64-bit variable
  *   hi :: hight part of addition (signed)
  *   lo :: low part of addition
  */
  MLIB_API( void )
  m_calc64_add( M_Calc64  z,
                M_Int32   hi,
                M_UInt32  lo );


 /***********************************************************************
  *
  * @function: m_calc64_sub
  *
  * @description:
  *  sustract a (hi,lo) pair of 32-bit values to a 64-bit variable
  *
  * @input:
  *   z  :: address of target variable
  *   hi :: high part of substraction (signed)
  *   lo :: low part of substration
  */
  MLIB_API( void )
  m_calc64_sub( M_Calc64  z,
                M_Int32   hi,
                M_UInt32  lo );


 /***********************************************************************
  *
  * @function: m_calc64_mul
  *
  * @description:
  *  computes "a = b*c" where 'b' and 'c' are 32-bit signed integers, and
  *  'a' is a @M_Calc64
  *
  * @input:
  *   z  :: address of result
  *   a  :: first multiplication argument
  *   b  :: second multiplication argument
  */
  MLIB_API( void )
  m_calc64_mul( M_Calc64  z,
                M_Int32   a,
                M_Int32   b );



 /***********************************************************************
  *
  * @function: m_calc64_div
  *
  * @description:
  *  divide a @M_Calc64 by a 32-bit value
  *
  * @input:
  *   z       :: address of value to divide
  *   divisor :: 32-bit divisor
  *
  * @return:
  *   the value of 'a/divisor', bounded to @M_INT32_MIN .. @M_INT32_MAX
  *   remainder is in 'a->lo'
  */
  MLIB_API( M_Int32 )
  m_calc64_div( M_Calc64  z,
                M_Int32   divisor );


 /***********************************************************************/
 /***********************************************************************/
 /*****                                                             *****/
 /*****                  16.16 fixed point numbers                  *****/
 /*****                                                             *****/
 /***********************************************************************/
 /***********************************************************************/

 /*************************************************************************
  *
  * @section: fixed_16
  *
  * @title: Fixed Point
  *
  * @abstract: 16.16 fixed point routines
  *
  * @description:
  *   the @M_Fix16 type is used to model a fixed-point number with 16 bits
  *   of fractional bits. This section presents macros, types and functions
  *   used to manage these numbers.
  */

 /*************************************************************************
  *
  * @type: M_Fix16
  *
  * @description:
  *   this type is used to store and operate on 16.16 fixed float numbers.
  */
  typedef M_Int32   M_Fix16;


 /*************************************************************************
  *
  * @macro: M_FIX16 (value)
  *
  * @description:
  *   a convenient macro to cast any value to a @M_Fix16.
  */
#define  M_FIX16(x)  ((M_Fix16)(x))

 /*************************************************************************
  *
  * @constant: M_FIX16_MAX
  *
  * @description:
  *   the biggest @M_Fix16 number
  */
#define  M_FIX16_MAX    ((M_Fix16)0x7FFFFFF)

 /*************************************************************************
  *
  * @constant: M_FIX16_MIN
  *
  * @description:
  *   the smallest @M_Fix16 number
  */
#define  M_FIX16_MIN    ((M_Fix16)0x8000001)


 /*************************************************************************
  *
  * @macro: M_FIX16_C (constant)
  *
  * @description:
  *   macro used to express a constant value as a @M_Fix16. see
  *   @M_INT32_C for more details..
  */
#define  M_FIX16_C(x)   M_INT32_C(x)


 /*************************************************************************
  *
  * @macro: m_fix16_add (x,y)
  *
  * @description:
  *   returns the sum of two fix16 numbers
  *
  * @note:
  *   this macro doesn't check overflows
  */
#define  m_fix16_add(x,y)    ((x)+(y))


 /*************************************************************************
  *
  * @macro: m_fix16_sub (x,y)
  *
  * @description:
  *   returns the difference of two fix16 numbers
  *
  * @note:
  *   this macro doesn't check underflows
  */
#define  m_fix16_sub(x,y)    ((x)-(y))


 /*************************************************************************
  *
  * @function: m_fix16_mul
  *
  * @description:
  *   returns the product of two @M_Fix16 number.
  *
  * @input:
  *    a :: first number
  *    b :: second number
  *
  * @result:
  *    the product of "a" and "b".
  *
  * @note:
  *    this really computes 'a*b/0x10000' with 64 bits of accuracy.
  *
  *    this function returns M_FIX16_MAX or M_FIX16_MIN in case of
  *    overflow..
  */
  MLIB_API(M_Fix16)   m_fix16_mul( M_Fix16   a,
                                   M_Fix16   b );


 /*************************************************************************
  *
  * @function: m_fix16_mul_round
  *
  * @description:
  *   same as @m_fix16_mul with rounding
  *
  * @input:
  *    a :: first number
  *    b :: second number
  *
  * @result:
  *    the product of "a" and "b".
  */
  MLIB_API(M_Fix16)   m_fix16_mul_round( M_Fix16   a,
                                         M_Fix16   b );


 /*************************************************************************
  *
  * @function: m_fix16_div
  *
  * @description:
  *   returns the division of two @M_Fix16 number.
  *
  * @input:
  *    a :: first number
  *    b :: second number
  *
  * @result:
  *    the value of "a" divided by "b" in 16.16 format
  *
  * @note:
  *    this really computes 'a*0x10000/b' with 64 bits of accuracy
  *
  *    this function returns M_FIX16_MAX or M_FIX16_MIN in case of
  *    overflow (e.g. divide by 0)
  */
  MLIB_API(M_Fix16)   m_fix16_div( M_Fix16   a,
                                   M_Fix16   b );


 /*************************************************************************
  *
  * @function: m_fix16_div_round
  *
  * @description:
  *   same as @m_fix16_div with rounding
  *
  * @input:
  *    a :: first number
  *    b :: second number
  *
  * @result:
  *    the value of "a" divided by "b" in 16.16 format
  */
  MLIB_API(M_Fix16)   m_fix16_div_round( M_Fix16   a,
                                         M_Fix16   b );



 /*************************************************************************
  *
  * @function: m_fix16_muldiv
  *
  * @description:
  *   returns the value of "a*b/c", computed with maximal accuracy
  *
  * @input:
  *    a :: first number
  *    b :: second number
  *    c :: third number
  *
  * @result:
  *    the value of "a*b/c" in 16.16 format
  *
  * @note:
  *    this really computes 'a*b/c" with 64 bits of accuracy
  *
  *    this function returns M_FIX16_MAX or M_FIX16_MIN in case of
  *    overflow (e.g. divide by 0)
  */
  MLIB_API(M_Fix16)   m_fix16_muldiv( M_Fix16    a,
                                      M_Fix16    b,
                                      M_Fix16    c );


 /*************************************************************************
  *
  * @function: m_fix16_muldiv_round
  *
  * @description:
  *   same as @m_fix16_muldiv with rounding
  *
  * @input:
  *    a :: first number
  *    b :: second number
  *    c :: third number
  *
  * @result:
  *    the value of "a*b/c" with rounding in 16.16 format
  */
  MLIB_API(M_Fix16)   m_fix16_muldiv_round( M_Fix16    a,
                                            M_Fix16    b,
                                            M_Fix16    c );


 /*************************************************************************
  *
  * @function: m_fix16_sqrt
  *
  * @description:
  *   returns the square root of a @M_Fix16 number
  *
  * @input:
  *    a :: number
  *
  * @result:
  *    square root. 0 if "a" was <= 0
  */
  MLIB_API(M_Fix16)   m_fix16_sqrt( M_Fix16  a );


  /* */

M_END_HEADER

#endif /* __MLIB_CALC_H__ */
