/* system-specific configuration macros */

#ifndef __MLIB_CONFIG_SYSTEM_H__
#define __MLIB_CONFIG_SYSTEM_H__

  /* We use <limits.h> values to know the sizes of the types.  */
#include <limits.h>


/* include compiler-specific headers to enable certain optional advanced */
/* features..                                                            */

#include <mlib/config/compiler/gnuc.h>
#include <mlib/config/compiler/visualc.h>
#include <mlib/config/compiler/borlandc.h>
#include <mlib/config/compiler/codewarrior.h>
#include <mlib/config/compiler/watcom.h>

/*****************************
 *
 * @section: macros
 *
 */

/************************************************************************
 *
 *  @macro: M_SIZEOF_INT
 *
 *  @description:
 *     returns the number of bytes in an "int", normally 2, 4 or 8
 *     depending on the system
 *
 *  @also:
 *     @M_Int32
 */
#ifndef M_SIZEOF_INT
#  if   UINT_MAX == 0xFFFFFFFFU
#    define M_SIZEOF_INT  4
#  elif UINT_MAX == 0xFFFFU
#    define M_SIZEOF_INT  2
#  elif UINT_MAX > 0xFFFFFFFFU && UINT_MAX == 0xFFFFFFFFFFFFFFFFU
#    define M_SIZEOF_INT  8
#  else
#    error "Unsupported number of bytes in `int' type!"
#  endif
#endif

/************************************************************************
 *
 *  @macro: M_SIZEOF_LONG
 *
 *  @description:
 *     returns the number of bytes in an "long", normally 4 or 8
 *     depending on the system
 *
 *  @also:
 *     @M_Int32
 */
#ifndef M_SIZEOF_LONG
#  if   ULONG_MAX == 0xFFFFFFFFU
#    define M_SIZEOF_LONG  4
#  elif ULONG_MAX > 0xFFFFFFFFU && ULONG_MAX == 0xFFFFFFFFFFFFFFFFU
#    define M_SIZEOF_LONG  8
#  else
#    error "Unsupported number of bytes in `long' type!"
#  endif
#endif


/************************************************************************
 *
 *  @macro: M_ALIGNMENT
 *
 *  @description:
 *     preferred alignment values for data in memory. Must be at least
 *     8 bytes for 64-bit systems..
 */
#ifndef   M_ALIGNMENT
#  define M_ALIGNMENT  8
#endif

/* If M_DLL_IMPORT and M_DLL_EXPORT are not defined, define them */
/* as simple externs..                                           */

/**************************************************************************
 *
 * @macro: M_DLL_IMPORT (x)
 *
 * @description:
 *    a macro used to _declare_ a DLL entry point. Certain compilers
 *    need a special pragma or keyword to notify client applications
 *    that an external function is defined in a DLL.
 *
 *    You should not use this macro directly in your code. Rather,
 *    use it when defining your project-specific API export macros,
 *    as in the following fragment:
 *
 *     {
 *       #ifdef PROJECT_BUILD_DLL
 *       #
 *       #  define  PROJECT_API(x)     M_DLL_EXPORT(x)
 *       #  define  PROJECT_APIDEF(x)  x
 *       #
 *       #elif defined(PROJECT_USE_DLL)
 *       #
 *       #  define  PROJECT_API(x)     M_DLL_IMPORT(x)
 *       #  define  PROJECT_APIDEF(x)  x               // should not happen
 *       #
 *       #else // use static library
 *       #
 *       #  define  PROJECT_API(x)     M_EXTERN(x)
 *       #  define  PROJECT_APIDEF(x)  M_EXTERNDEF(x)
 *       #
 *       #endif
 *     }
 *
 *    then, you can use PROJECT_API and PROJECT_APIDEF to respectively
 *    declare your exported functions (in public header files), or
 *    define them (in project private source files).
 *
 * @also:
 *    the macros @M_DLL_EXPORT and @M_DLL_EXPORT_LAZY are also used to
 *    _define_ DLL entry points in your code.
 *
 *    see also the definition of @MLIB_API and @MLIB_APIDEF for more
 *    details..
 */
#ifndef M_DLL_IMPORT
#  define  M_DLL_IMPORT(x)   extern x
#endif

/**************************************************************************
 *
 * @macro: M_DLL_EXPORT (x)
 *
 * @description:
 *    a macro used to _define_ a DLL entry point in your project private
 *    source code. Certain compilers need a special pragma or keyword to
 *    generate appropriate code when a dynamic library is built.
 *
 * @note:
 *    On certain platforms ( Windows & OS/2 ), you'll need an additional
 *    .DEF file to build the DLL when this macro is used. The file is used
 *    to list the exported functions and their ordinal number.
 *
 *    a .DEF file is convenient because it lets you upgrade a DLL without
 *    running the risk of breaking all applications that use it.
 *
 * @also:
 *    the function @M_DLL_IMPORT is used to declare an exported DLL entry
 *    point in client applications, while @M_DLL_EXPORT_LAZY is used for
 *    quick'n dirty exports on Windows..
 */
#ifndef M_DLL_EXPORT
#  define  M_DLL_EXPORT(x)   extern x
#endif

/**************************************************************************
 *
 * @macro: M_DLL_EXPORT_LAZY (x)
 *
 * @description:
 *    a macro used to _define_ a DLL entry point in your project's private
 *    source code. Certain compilers need a special pragma or keyword to
 *    generate appropriate code when a dynamic library is built.
 *
 *    this is a variant of @M_DLL_EXPORT that can be used to avoid the
 *    need for a .DEF file when building a DLL on certain platforms
 *    (namely Windows and OS/2).
 *
 * @note:
 *    you should never use this macro directly. see the comments in
 *    @M_DLL_IMPORT for more explanations..
 *
 *    be aware that if you do not use a .DEF file on Windows, you'll be
 *    unable to safely upgrade a DLL without breaking other programs
 *    that implicetly link to it..
 *
 *    generally speaking, @M_DLL_EXPORT_LAZY should only be used when
 *    building DLLs that you fully control, not some that can be used
 *    by another application.
 */
#ifndef M_DLL_EXPORT_LAZY
#  define  M_DLL_EXPORT_LAZY(x)  extern  x
#endif

/**************************************************************************
 *
 * @macro: M__FUNC__
 *
 * @description:
 *    returns a string containing the name of the current function. This
 *    is only supported with certain compilers, by default this macro will
 *    return "unknown function".
 *
 *    this is used by log and error messages
 */
#ifndef  M__FUNC__
#  define  M__FUNC__  "unknown function"
#endif

/* */

/**************************************************************************
 *
 * @macro: M_OS
 *
 * @description:
 *    a macro that indicates the operating system or platform type
 *    that the MLib is being compiled on
 */
#define  M_OS_UNKNOWN  0
#define  M_OS_UNIX     1
#define  M_OS_WIN32    2
#define  M_OS_OS2      3
#define  M_OS_BEOS     4
#define  M_OS_NETWARE  5
#define  M_OS_MACOS    6
#define  M_OS_MACOSX   7

#ifndef  M_OS
#define  M_OS   M_OS_UNKNOWN
#endif


/**************************************************************************
 *
 * @macro: M_CPU
 *
 * @description:
 *    a macro that indicates the processor that the MLib is being
 *    compiled on
 */
#define  M_CPU_UNKNOWN  0
#define  M_CPU_X86      1
#define  M_CPU_MIPS     2
#define  M_CPU_ALPHA    3
#define  M_CPU_SPARC    4
#define  M_CPU_POWERPC  5
#define  M_CPU_68K      6

#ifndef  M_CPU
#define  M_CPU  M_CPU_UNKNOWN
#endif

/* */

#endif /* __MLIB_CONFIG_SYSTEM_H__ */

