#ifndef __MLIB_CORE_H__
#define __MLIB_CORE_H__

#ifndef __MLIB_H__
#error  "do not include this file directly, use <mlib.h> instead"
#endif


M_BEGIN_HEADER

 /*************************************************************************
  *
  * @section: core
  *
  * @title: Core
  *
  * @abstract: MLib Core instance
  *
  * @description:
  *   an MLib core object is used to model a single instance of the MLib in
  *   a given process. Normal programs should only need one MLib core,
  *   independent of the number of libraries using it..
  *
  * @order:
  *   M_Core
  *   M_CoreRec
  *
  *
  */

 /*************************************************************************
  *
  * @struct: M_CoreRec
  *
  * @description:
  *    a core object is used to model a given context/instance of the MLib.
  *    there shouldn't be more than one core per program/process using
  *    the MLib (except maybe for debugging reasons).
  *
  * @fields:
  *    memory    :: handle to main memory manager for all allocations
  *                 performed within this context
  *
  *    impl      :: this pointer is ignored by the MLib core. it can be
  *                 used to store implementation-specific data
  * @also:
  *    @M_CoreRec
  */
  typedef struct M_CoreRec_
  {
    M_Memory    memory;
    M_Pointer   impl;

  } M_CoreRec;


  typedef const struct M_Core_ClassRec_*   M_Core_Class;

 /*************************************************************************
  *
  * @function: m_core_new
  *
  * @description:
  *   this function creates a new MLib core object from a given
  *   memory manager. Don't call it directly, use @mlib_init instead
  *
  * @input:
  *    memory    :: optional memory manager handler. can be NULL
  *
  *    clazz     :: optional core object class. can be NULL.
  *
  *    impl_data :: implementation-specific pointer
  *
  * @output:
  *    amcore    :: handle to new core object. NULL in case of error
  *
  * @return:
  *    error code. 0 means success
  *
  * @note:
  *    if a memory manager handle is provided, the function @m_memory_ref
  *    is called to increment its reference count.
  *
  *    otherwise, a new standard/default memory manager is created through
  *    the function @m_memory_new
  *
  *    note that a valid "clazz" handle is required if you want to use
  *    a thread-safe MLib core. Otherwise, the core assumes that only
  *    a single thread will access it..
  */
  MLIB_API(M_Error)    m_core_new( M_Memory      memory,
                                   M_Core_Class  clazz,
                                   M_Pointer     impl_data,
                                   M_Core       *amcore );


 /*************************************************************************
  *
  * @function: m_core_ref
  *
  * @description:
  *   increment the internal reference count within a MLib core object
  *
  * @input:
  *    mcore :: handle to target core object
  */
  MLIB_API(void)       m_core_ref( M_Core  mcore );


 /*************************************************************************
  *
  * @function: m_core_unref
  *
  * @description:
  *   decrement the internal reference count within a MLib core object.
  *   if it reaches 0 or below, the core is destroyed..
  *
  * @input:
  *    mcore :: handle to target context object
  */
  MLIB_API(void)       m_core_unref( M_Core  mcore );


 /********************************************************************
  *
  * @function: m_core_push
  *
  * @description:
  *    push a new item on the current thread's cleanup stack.
  *    this function generates an exception is there isn't enough
  *    memory to allocate a stack slot.
  *
  * @input:
  *    mcore        :: handle to MLib core
  *    item         :: pointer to pushed item
  *    clenaup_func :: function called in case of exception cleanup
  *    cleanup_data :: user-provided pointer for the cleanup function
  *
  * @throws: @m_err_memory_alloc
  *
  * @note:
  *    client applications should not call this function, unless they
  *    know exactly what they do. Rather, try to use the context
  *    specific functions like @m_memory_push, @m_object_push,
  *    @m_object_array_push, etc.. as they're significnatly easier to use
  *
  */
  MLIB_API(void)       m_core_push( M_Core         mcore,
                                    M_Pointer      item,
                                    M_CleanupFunc  cleanup_func,
                                    M_Pointer      cleanup_data );


 /********************************************************************
  *
  * @function: m_core_pop
  *
  * @description:
  *    pop a given item from the current thread's clean-up stack.
  *
  * @input:
  *    mcore   :: handle to MLib core
  *    item    :: element to pop from the top of the stack
  *    action  :: action to be performed on pop
  *
  * @note:
  *    this function will _panic_ (i.e. abort your program) if the
  *    top-most element of the stack doesn't correspond to the item
  *    passed as the 2nd parameter.
  */
  MLIB_API(void)       m_core_pop ( M_Core       mcore,
                                    M_Pointer    item,
                                    M_PopAction  action );


 /********************************************************************
  *
  * @function: m_core_throw
  *
  * @description:
  *    throws an exception in current thread
  *
  * @input:
  *    mcore     :: handle to MLib core
  *    exception :: exception to throw
  */
  MLIB_API(void)       m_core_throw( M_Core       mcore,
                                     M_Exception  exception );


 /********************************************************************
  *
  * @function: m_core_rethrow
  *
  * @description:
  *    re-throws an exception. this function should only be used within a
  *    @m_xcatch .. @m_xend block.
  *
  * @input:
  *    mcore   :: handle to MLib core
  *    error   :: exception to raise.
  *
  * @note:
  *    it's possible to raise a different exception than the one
  *    that was caught in the "catch" statements. In this case,
  *    the "old" exception is released, and the new one is used..
  */
  MLIB_API(void)       m_core_rethrow( M_Core       mcore,
                                       M_Exception  exception );


 /*************************************************************************
  *
  * @functype: M_FatalHandlerFunc
  *
  * @description:
  *    a function called when a fatal exception is raised. An exception
  *    is considered fatal when it isn't handled by the program. The
  *    fatal exception handler can be set in any thread by first
  *    calling @m_core_set_fatal_handler.
  *
  * @input:
  *    mcore       :: MLib core handle
  *    exception   :: exception
  *    data        :: a user-provided generic pointer that is passed
  *                   to the fatal exception handler. this is normally
  *                   thread-specific
  */
  typedef void  (*M_FatalHandlerFunc)( M_Core       mcore,
                                       M_Exception  exception,
                                       M_Pointer    data );

 /********************************************************************
  *
  * @function: m_core_set_fatal_handler
  *
  * @description:
  *    set the fatal exception handler for the current thread.
  *
  * @input:
  *    mcore         :: handle to MLib core
  *    fatal_handler :: pointer to fatal exception handler function
  *    fatal_data    :: thread-specific user-provided data passed
  *                     to the handler in case of fatal exception
  *
  * @note:
  *    an out-of-memory exception can be raised by this function
  *    (mainly because there isn't enough memory to allocate a new
  *     thread data record within the MLib context)
  *
  *    in this case, the handler will be called directly.
  *
  *    do not try to do stupid things in the fatal handler like jumping
  *    to another thread. Its purpose is to allow you to write a
  *    semi-intelligent message before completely quiting the current
  *    thread or program..
  *
  *    you can call this function repeatedly to reset the fatal handler
  *    if you want to.
  *
  *    if no default fatal handler is set, the program will simply
  *    trap in the case of an un-handled exception.
  */
  MLIB_API(void)  m_core_set_fatal_handler(
                            M_Core              mcore,
                            M_FatalHandlerFunc  fatal_handler,
                            M_Pointer           fatal_data );


 /********************************************************************
  *
  * @function: m_core_set_xhandler
  *
  * @description:
  *    pushes a new exception handler for the current thread.
  *    this function should not be called directly by applications, but
  *    by the @M_XTRY macro.
  *
  * @input:
  *    mcore    :: handle to MLib core
  *    xhandler :: handle to new exception handler block
  */
  MLIB_API(void)       m_core_set_xhandler( M_Core      mcore,
                                            M_XHandler  handler );


 /********************************************************************
  *
  * @function: m_core_unset_xhandler
  *
  * @description:
  *    unsets a given exception handler for the current thread.
  *    this function should not be called directly by applications
  *    but by the @M_XCATCH and @M_XEND macros.
  *
  * @input:
  *    mcore    :: handle to MLib core
  *    xhandler :: exception handler block handle
  *
  * @return:
  *    exception pointer in xinfo. can be NULL
  */
  MLIB_API(M_Exception)  m_core_unset_xhandler( M_Core      mcore,
                                                M_XHandler  xinfo );

  /* */

  MLIB_API(void)  m_core_assert( M_Core   mcore,
                                 const char*  filename,
                                 long         lineno,
                                 const char*  funcname );

#if defined(M_DEBUG_ERROR) || defined(M_DEBUG_LOG)
#  define  m_assert_is_core(x)  m_core_assert( x, __FILE__, __LINE__, M__FUNC__ )
#else
#  define  m_assert_is_core(x)  M_UNUSED(x)
#endif

 /* return current exception handler */
  MLIB_API(M_XContext)  m_core_find_xcontext( M_Core       mcore,
                                              const char*  name );

  typedef const struct M_Static_ClassRec_*   M_Static_Class;

  typedef void  (*M_Static_InitFunc)( M_Pointer  static_object,
                                      M_Core     mcore,
                                      M_Pointer  data );

  typedef void  (*M_Static_DoneFunc)( M_Pointer  static_object );

  typedef struct M_Static_ClassRec_
  {
    M_Size             size;
    M_UInt32           id;
    M_Static_InitFunc  init;
    M_Static_DoneFunc  done;

  } M_Static_ClassRec;


  MLIB_API(M_Pointer)  m_core_get_static( M_Core           mcore,
                                          M_Static_Class   clazz,
                                          M_Pointer        data );

#if 0
  MLIB_API(void)       m_core_lock      ( M_Core  mcore );

  MLIB_API(void)       m_core_unlock    ( M_Core  mcore );

  MLIB_API(void)       m_core_lock_push ( M_Core  mcore );

  MLIB_API(void)       m_core_pop_unlock( M_Core  mcore );
#endif

#define  M_CORE(x)    ((M_Core)(x))
#define  M_CORE_P(x)  ((M_Core*)(x))

#define  M_CORE_DUMMY_THREAD_ID  ((M_ThreadID)-1L)

  /* */

M_END_HEADER

#endif /* __MLIB_CORE_H__ */

