#ifndef __MLIB_DATE_H__
#define __MLIB_DATE_H__

#include <mlib.h>

M_BEGIN_HEADER

 /************************************************************
  *
  * @type: M_Date
  *
  * @description:
  *   handle to a @M_DateRec structure used to model a given
  *   date.
  */
  typedef struct M_DateRec_*   M_Date;

 /************************************************************
  *
  * @struct: M_DateRec
  *
  * @description:
  *   a structure used to model a given day date. Its fields
  *   are voluntarily obfuscated in order to prevent direct
  *   access to their content.
  *
  *   use accessor functions like @m_date_day, @m_date_month
  *   and @m_date_year to retrieve date values. Other functions
  *   are provided to change and compare dates
  */
  typedef struct M_DateRec_
  {
    const M_UInt32   xyzxyz;
    const M_UInt16   xzyxzy;
    const M_Byte     yzxyzx;
    const M_Byte     yxzyxz;
    const M_Byte     zxyzxy;

  } M_DateRec;


 /************************************************************
  *
  * @enum: M_DateDay
  *
  * @description:
  *   the list of values used to model days in dates
  *
  *   note that the value @M_DATE_DAY_BAD is used to indicate
  *   an invalid date record in @m_date_day
  */
  typedef enum
  {
    M_DATE_DAY_BAD       = 0,
    M_DATE_DAY_MONDAY    = 1,
    M_DATE_DAY_TUESDAY   = 2,
    M_DATE_DAY_WEDNESDAY = 3,
    M_DATE_DAY_THURSDAY  = 4,
    M_DATE_DAY_FRIDAY    = 5,
    M_DATE_DAY_SATURDAY  = 6,
    M_DATE_DAY_SUNDAY    = 7

  } M_DateDay;


 /************************************************************
  *
  * @enum: M_DateMonth
  *
  * @description:
  *   the list of values used to model months in dates
  *
  *   note that the value @M_DATE_DAY_MONTH is used to indicate
  *   an invalid date record in @m_date_month
  */
  typedef enum
  {
    M_DATE_MONTH_BAD         = 0,
    M_DATE_MONTH_JANURAY     = 1,
    M_DATE_MONTH_FEBRUARY    = 2,
    M_DATE_MONTH_MARCH       = 3,
    M_DATE_MONTH_APRIL       = 4,
    M_DATE_MONTH_MAY         = 5,
    M_DATE_MONTH_JUNE        = 6,
    M_DATE_MONTH_JULY        = 7,
    M_DATE_MONTH_AUGUST      = 8,
    M_DATE_MONTH_SEPTEMBER   = 9,
    M_DATE_MONTH_OCTOBER     = 10,
    M_DATE_MONTH_NOVEMBER    = 11,
    M_DATE_MONTH_DECEMBER    = 12

  } M_DateMonth;


 /************************************************************
  *
  * @type: M_DateYear
  *
  * @description:
  *   the type of year numbers in dates.
  *   note that 0 is an invalid value
  */
  typedef  M_Int  M_DateYear;

#define M_DATE_YEAR_BAD  0

 /************************************************************
  *
  * @type: M_DateWeek
  *
  * @description:
  *   the type of week numbers in dates.
  *   note that 0 is an invalid value
  */
  typedef  M_Int  M_DateWeek;

#define M_DATE_WEEK_BAD   0

 /************************************************************
  *
  * @function: m_date_clear
  *
  * @description:
  *   clear a date record. this makes it invalid
  *
  * @input:
  *   date :: target date handle
  */
  MLIB_API(void)
  m_date_clear( M_Date  date );


 /************************************************************
  *
  * @function: m_date_check
  *
  * @description:
  *   checks that a date record is valid
  *
  * @input:
  *   date :: target date handle
  *
  * @return:
  *   boolean. TRUE if the record is valid
  */
  MLIB_API( M_Bool )
  m_date_check( M_Date  date );


 /************************************************************
  *
  * @function: m_date_set
  *
  * @description:
  *   change all records in a date
  *
  * @input:
  *   date  :: target date handle
  *   year  :: new date year
  *   month :: new date month
  *   day   :: new date day
  *
  * @return:
  *   boolean. TRUE if the input values were valid
  */
  MLIB_API( M_Bool )
  m_date_set( M_Date       date,
              M_DateDay    day,
              M_DateMonth  month,
              M_DateYear   year );


 /************************************************************
  *
  * @function: m_date_day
  *
  * @description:
  *   return's a date's day
  *
  * @input:
  *   date  :: target date handle
  *
  * @return:
  *   date's day. M_DATE_DAY_BAD is the date record is invalid
  */
  MLIB_API( M_DateDay )
  m_date_day( M_Date  date );


 /************************************************************
  *
  * @function: m_date_month
  *
  * @description:
  *   return's a date's month
  *
  * @input:
  *   date  :: target date handle
  *
  * @return:
  *   date's month. M_DATE_DAY_BAD is the date record is invalid
  */
  MLIB_API( M_DateMonth )
  m_date_month( M_Date  date );



 /************************************************************
  *
  * @function: m_date_year
  *
  * @description:
  *   return's a date's year
  *
  * @input:
  *   date  :: target date handle
  *
  * @return:
  *   date's year. 0 if the date record is invalid
  */
  MLIB_API( M_DateYear )
  m_date_year( M_Date  date );


 /************************************************************
  *
  * @function: m_date_set_day
  *
  * @description:
  *   change a date's day
  *
  * @input:
  *   date  :: target date handle
  *   day   :: new day value
  */
  MLIB_API( void )
  m_date_set_day( M_Date     date,
                  M_DateDay  day );


 /************************************************************
  *
  * @function: m_date_set_month
  *
  * @description:
  *   change a date's month
  *
  * @input:
  *   date  :: target date handle
  *   month :: new date month
  */
  MLIB_API( void )
  m_date_set_month( M_Date       date,
                    M_DateMonth  month );


 /************************************************************
  *
  * @function: m_date_set_year
  *
  * @description:
  *   change a date's year
  *
  * @input:
  *   date  :: target date handle
  *   year  :: new date year
  */
  MLIB_API( void )
  m_date_set_year( M_Date      date,
                   M_DateYear  year );


 /************************************************************
  *
  * @function: m_date_compare
  *
  * @description:
  *   compare two dates
  *
  * @input:
  *   date1  :: first date
  *   date2  :: second date
  *
  * @return:
  *   -1, 0 or +1 depending on date's comparison
  *
  *   note that an invalid date is always considered
  *   to be _before_ a valid one. Two invalid dates are
  *   equal, independently of their exact content
  */
  MLIB_API( M_Int )
  m_date_compare( M_Date  date1,
                  M_Date  date2 );


 /************************************************************
  *
  * @function: m_date_diff_days
  *
  * @description:
  *   return the difference in days between two dates
  *
  * @input:
  *   date1  :: first date
  *   date2  :: second date
  *
  * @return:
  *   day difference, as 32-bit integer
  *
  *   an invalid date is always considered to be at the
  *   date origin
  */
  MLIB_API( M_Int32 )
  m_date_diff_days( M_Date  date1,
                    M_Date  date2 );


 /************************************************************
  *
  * @function: m_date_add_days
  *
  * @description:
  *   change a date by adding (or substrating) day numbers
  *
  * @input:
  *   date :: target date handle
  *   days :: number of days to add, as signed 32-bit integer
  *
  * @note:
  *   a date can become invalid if too many days are
  *   substracted to it..
  */
  MLIB_API( void )
  m_date_add_days( M_Date   date,
                   M_Int32  days );


 /************************************************************
  *
  * @function: m_date_week
  *
  * @description:
  *   return the week number corresponding to a given date
  *
  * @input:
  *   date :: target date handle
  *
  * @return:
  *   week number (1..53)
  */
  MLIB_API( M_Int )
  m_date_week( M_Date  date );


 /************************************************************
  *
  * @function: m_date_set_week
  *
  * @description:
  *   set a date from week and day numbers
  *
  * @input:
  *   date :: target date handle
  *   week :: week number (1..53 range)
  *   day  :: day number in the week
  *
  * @note:
  *   you need to set the date's year through @m_date_set_year
  */
  MLIB_API( void )
  m_date_set_week( M_Date      date,
                   M_DateWeek  week,
                   M_DateDay   day );
 /* */

  MLIB_API( M_Bool )
  m_date_is_leap_year( M_DateYear  year );

  MLIB_API( M_Int )
  m_date_days_in_month( M_DateMonth  month,
                        M_DateYear   year );

  MLIB_API( M_Int )
  m_date_days_in_year( M_DateYear  year );

  MLIB_API( void )
  m_date_set_day_in_year( M_Date      date,
                          M_Int32     days,
                          M_DateYear  year );

  MLIB_API( void )
  m_date_set_julian( M_Date    date,
                     M_UInt32  julian );

  MLIB_API( M_UInt32 )
  m_date_get_julian( M_Date  date );

 /* */

M_END_HEADER

#endif /* __MLIB_DATE_H__ */