/************************************************************************
 *
 *  mdebug.h - MLib debug/log sub-system
 *
 *  author: David Turner (c) 2000
 *
 ************************************************************************/

#ifndef __MLIB_DEBUG_H__
#define __MLIB_DEBUG_H__

#ifndef __MLIB_H__
#error  "do not include this file directly, use <mlib.h> instead"
#endif

M_BEGIN_HEADER

 /**************************************************************************
  *
  * @section: debugging
  *
  * @title: Debug Macros
  *
  * @abstract: Debugging support
  *
  * @description:
  *   the following macros are useful for debugging
  */

 /************************************************************************
  *
  * @function: m_panic
  *
  * @description:
  *   print an error message, then aborts the current program in case
  *   of fatal failure..
  *
  * @input:
  *    fmt  :: printf-style format specifier string
  *    ...  :: additional string format parameters
  */
  MLIB_API( void )  m_panic( const char*  fmt, ... );


 /************************************************************************
  *
  * @function: m_log
  *
  * @description:
  *   print a message to the debug log. The log is normally the standard
  *   output, but can be re-defined in certain builds
  *
  * @input:
  *    fmt  :: printf-style format specifier string
  *    ...  :: additional string format parameters
  */
  MLIB_API( void )  m_log( const char*  fmt, ... );


  /* */
#ifdef M_DEBUG_LEVEL_LOG

  MLIB_API(int)     m_log_compare( const char* domain,
                                   const char* component,
                                   int         level );

#  define M_LOG( level, varformat )                                         \
            M_BEGIN_STMNT                                                   \
              if ( m_log_compare( M_LOG_DOMAIN, M_LOG_COMPONENT, level ) )  \
                m_log varformat;                                            \
            M_END_STMNT

#  define M_ERROR( varformat )       m_log varformat

#elif defined( M_DEBUG_LEVEL_ERROR )

#  define M_LOG( level, varformat )  M_DUMMY_STMNT
#  define M_ERROR( varformat )       m_log varformat

#else  /* !M_DEBUG_LEVEL_ERROR && !M_DEBUG_LEVEL_LOG */

#  define M_LOG( level, varformat )     M_DUMMY_STMNT
#  define M_ERROR( varformat )          M_DUMMY_STMNT

#endif /* M_DEBUG_LEVEL_LOG, M_DEBUG_LEVEL_ERROR */


  /*************************************************************************/
  /*                                                                       */
  /* Define macros and functions that are common to the debug and log      */
  /* modes.                                                                */
  /*                                                                       */
  /* You need vprintf() to be able to compile mdebug.c in its default      */
  /* implementation                                                        */
  /*                                                                       */
  /*************************************************************************/

/*************************************************************************
 *
 * @macro: m_assert (condition)
 *
 * @description:
 *   this macro is used to check assertions and is only compiled when
 *   @M_DEBUG_COMPILE_ASSERTS is defined.
 *
 *   if the condition isn't true, the program is immediately aborted
 *   and an error (panic) message is printed to the error output
 */

#ifdef M_DEBUG_COMPILE_ASSERTS

#  define m_assert( condition )                                    \
          M_BEGIN_STMNT                                            \
            if ( !( condition ) )                                  \
              m_panic( "assertion '" #condition "' failed in"      \
                       " %s (%s:%d)\n", M__FUNC__, __FILE__,       \
                       __LINE__ );                                 \
          M_END_STMNT

#else

#  define m_assert( condition )                  M_DUMMY_STMNT

#endif

/*************************************************************************
 *
 * @macro: m_assert_log (condition, varformat )
 *
 * @description:
 *   a variant of @m_assert. If the condition isn't true, the "varformat"
 *   message is sent to the log, then the program panics (i.e. aborts)
 */

#ifdef M_DEBUG_COMPILE_ASSERTS

#  define m_assert_log( condition, varformat )                     \
          M_BEGIN_STMNT                                            \
            if ( !( condition ) )                                  \
            {                                                      \
              m_log( "assertion '" #condition "' failed in"        \
                     " %s (%s:%d)\n", M__FUNC__, __FILE__,         \
                     __LINE__ );                                   \
              m_panic varformat;                                   \
          M_END_STMNT

#else

#  define m_assert_log( condition, varformat )   M_DUMMY_STMNT

#endif

/*************************************************************************
 *
 * @macro: m_assert_unreached
 *
 * @description:
 *   this macro must be used to indicate that normal execution should
 *   never reach a given point in the source code. This is generally
 *   useful in the default case of switch statements, in order to
 *   enfore valid input.
 */
#ifdef M_DEBUG_COMPILE_ASSERTS

#  define  m_assert_unreached                                              \
              m_panic( "unreachable code error in %s (%s:%d)\n",           \
                       M__FUNC__, __FILE__, __LINE__ );
#else

#  define m_assert_unreached                     M_DUMMY_STMNT

#endif


/*************************************************************************
 *
 * @macro: m_precheck (condition)
 *
 * @description:
 *   this macro is used to perform pre-condition check in function entry
 *   points. If the input condition isn't true, then:
 *
 *   if @M_DEBUG_COMPILE_ASSERTS is defined, it aborts the current program
 *   as @m_assert
 *
 *   otherwise, if @M_DEBUG_COMPILE_PRECHEKS, it prints an error message
 *   then return immediately from the current function
 *
 *   otherwise, it doesn't do anything.
 */
#ifdef M_DEBUG_COMPILE_ASSERTS

#  define  m_precheck( condition )                                            \
               M_BEGIN_STMNT                                                  \
                 if (!(condition))                                            \
                   m_panic( "pre-condition '" #condition "' failed in "       \
                            "%s (%s:%d)\n", M__FUNC__, __FILE__, __LINE__ );  \
               M_END_STMNT

#elif defined M_DEBUG_COMPILE_PRECHECKS

#  define  m_precheck( condition )                                          \
            M_BEGIN_STMNT                                                   \
              if (!(condition))                                             \
              {                                                             \
                M_ERROR(( "pre-condition '" #condition "' failed in "       \
                          "%s (%s:%d)\n", M__FUNC__, __FILE__, __LINE__ )); \
                return;                                                     \
              }                                                             \
            M_END_STMNT

#else

#  define  m_precheck( condition )            M_DUMMY_STMNT

#endif

/*************************************************************************
 *
 * @macro: m_precheck_val (condition,val)
 *
 * @description:
 *   a variant of @m_precheck that returns a value in case of failure.
 *   when the input condition isn't true, then:
 *
 *   if @M_DEBUG_COMPILE_ASSERTS is defined, it aborts the current program
 *   as @m_assert
 *
 *   otherwise, if @M_DEBUG_COMPILE_PRECHEKS, it prints an error message
 *   then return immediately from the current function with value "val"
 *
 *   otherwise, it doesn't do anything.
 */

#ifdef M_DEBUG_COMPILE_ASSERTS

#  define  m_precheck_val( condition, val )    m_precheck( condition )

#elif defined(M_DEBUG_COMPILE_PRECHECKS)

#  define  m_precheck_val( condition, val )                                 \
            M_BEGIN_STMNT                                                   \
              if (!(condition))                                             \
              {                                                             \
                M_ERROR(( "pre-condition '" #condition "' failed in "       \
                          "%s (%s:%d)\n", M__FUNC__, __FILE__, __LINE__ )); \
                return val;                                                 \
              }                                                             \
            M_END_STMNT

#else /* !COMPILE_ASSERTS && !COMPILE_PRECHECKS */

#define  m_precheck_val( condition, val )   M_DUMMY_STMNT

#endif

 /* */

#define M_PANIC( varformat )     m_panic  varformat

  /*************************************************************************/
  /*                                                                       */
  /* You need two opening resp. closing parentheses!                       */
  /*                                                                       */
  /* Example: M_LOG0(( "Value is %i", foo ))                               */
  /*                                                                       */
  /*************************************************************************/

#define M_LOG0( varformat )  M_LOG( 0, varformat )
#define M_LOG1( varformat )  M_LOG( 1, varformat )
#define M_LOG2( varformat )  M_LOG( 2, varformat )
#define M_LOG3( varformat )  M_LOG( 3, varformat )
#define M_LOG4( varformat )  M_LOG( 4, varformat )
#define M_LOG5( varformat )  M_LOG( 5, varformat )
#define M_LOG6( varformat )  M_LOG( 6, varformat )
#define M_LOG7( varformat )  M_LOG( 7, varformat )



#ifdef M_DEBUG_TEST_UNITS

  /* sends a log trace to the standard error descriptor during unit testing */
  M_LOCAL void  m_log_test( const char*  fmt, ... );

#endif /* M_DEBUG_TEST_UNITS */


#if defined( _MSC_VER )      /* Visual C++ (and Intel C++) */

  /* we disable the warning `conditional expression is constant' here */
  /* in order to compile cleanly with the maximum level of warnings   */
#pragma warning( disable : 4127 )

#endif /* _MSC_VER */

M_END_HEADER

#endif /* __MLIB_DEBUG_H__ */
