/***************************************************************************
 *
 *  mexcept.h - MLib exception sub-system
 *
 *  Copyright 2000 - David Turner
 *
 *
 ***************************************************************************/

#ifndef __MLIB_EXCEPTION_H__
#define __MLIB_EXCEPTION_H__

#ifndef __MLIB_H__
#error  "do not include this file directly, use <mlib.h> instead"
#endif

#include <setjmp.h>

M_BEGIN_HEADER

/*************************************************************************
 *
 * @section: exceptions
 *
 * @title: Exceptions
 *
 * @abstract: Structured Exception Management
 *
 * @description:
 *    The MLib implements a cleanup-stack exception handling sub-system that
 *    is described by the document at
 *    http://www.freetype.org/david/reliable-c.html
 *
 * @order:
 *    M_Exception
 *    M_ExceptionRec
 *
 *    M_XTRY_CORE
 *    M_XTRY
 *    M_XCATCH
 *    M_XCATCH_ANY
 *    M_XEND
 *    M_XTHROW_CORE
 *    M_XTHROW
 *    M_XRETHROW_CORE
 *    M_XRETHROW
 */



 /***************************************************************************/
 /***************************************************************************/
 /*****                                                                 *****/
 /*****                      MLIB EXCEPTION OBJECTS                     *****/
 /*****                                                                 *****/
 /***************************************************************************/
 /***************************************************************************/

 /***********************************************************************
  *
  * @type: M_ExceptionFamily
  *
  * @description:
  *   a handle to an _constant_ exception family object. An exception family
  *   is used  to describe ... a given family of exceptions !!. it also
  *   provides text that is dumped when an exception occurs..
  */
  typedef const struct M_ExceptionFamilyRec_*   M_ExceptionFamily;



 /***********************************************************************
  *
  * @struct: M_ExceptionRec
  *
  * @description:
  *   a structure used to model a given
  */
  typedef struct M_ExceptionRec_
  {
    M_ExceptionFamily    family;
    M_UInt               value;

  } M_ExceptionRec;


 /***************************************************************************/
 /***************************************************************************/
 /*****                                                                 *****/
 /*****                 MLIB EXCEPTION INFO OBJECTS                     *****/
 /*****                                                                 *****/
 /***************************************************************************/
 /***************************************************************************/

 /***********************************************************************
  *
  * @type: M_XHandler
  *
  * @description:
  *   a handle to an exception handler structure, see @M_XHandlerRec. This
  *   type is not opaque, but application should ignore it..
  */
  typedef struct M_XHandlerRec_*  M_XHandler;


 /*********************************************************************
  *
  * @type: M_XContext
  *
  * @description:
  *    a handle to a thread-specific exception context object. Used to hold
  *    a cleanup stack, current and fatal execption handlers, as well as a
  *    few other things..
  *
  *    this type is opaque
  *
  * @also:
  *    @M_XContextRec
  */
  typedef struct M_XContextRec_*   M_XContext;


 /*********************************************************************
  *
  * @type: M_XMark
  *
  * @description:
  *    an opaque handle type. it is used to store the current size
  *    of the cleanup stack in exception handlers
  *
  * @also:
  *    @M_XHandlerRec
  */
  typedef const struct M_XMarkRec_*      M_XMark;


 /***************************************************************
  *
  * @struct: M_XHandlerRec
  *
  * @description:
  *    a structure used to hold information during a M_XTRY .. M_XEND
  *    block..
  *
  * @fields:
  *    previous   :: handle to previous handler for current thread.
  *    mcore      :: handle to MLib core
  *    exception  :: returned exception. NULL if none
  *    context    :: current thread's exception context.
  *    mark       :: used to store the cleanup stack mark (i.e. size) for
  *                  the current thread when  M_XTRY is invoked.
  *    env        :: jump buffer environment
  * @also:
  *    @M_ExceptionRec, @M_ExceptionFamilyRec
  */
  typedef struct M_XHandlerRec_
  {
    M_XHandler     previous;
    M_Core         mcore;
    M_Exception    exception;
    M_XContext     xcontext;
    M_XMark        xmark;
    jmp_buf        env;

  } M_XHandlerRec;


 /***************************************************************************/
 /***************************************************************************/
 /*****                                                                 *****/
 /*****                 MLIB EXCEPTION FAMILY OBJECTS                   *****/
 /*****                                                                 *****/
 /***************************************************************************/
 /***************************************************************************/

 /* return exception description string */
  typedef M_CString   (*M_Exception_GetDescFunc)( M_Exception  exception,
                                                  M_Memory     memory );

 /* release exception description string */
  typedef void        (*M_Exception_FreeDescFunc)( M_Exception  exception,
                                                   M_CString    desc,
                                                   M_Memory     memory );

 /* exception family descriptor */
  typedef struct M_ExceptionFamilyRec_
  {
    M_CString                   name;      /* family name */
    M_Exception_GetDescFunc     desc_get;  /* get exception description string */
    M_Exception_FreeDescFunc    desc_free; /* free exception description string */

  } M_ExceptionFamilyRec;


 /***************************************************************
  *
  * @constant: mlib_exception_family
  *
  * @description:
  *    a constant used to address the family of all MLib-specific
  *    exceptions. These are:
  *
  *     @m_err_parameter_check,
  *
  *     @m_err_memory_alloc,
  *
  *     @m_err_mutex_init,
  *     @m_err_mutex_lock
  *     @m_err_mutex_unlock
  *
  *     @m_err_synlock_init
  *     @m_err_synlock_lock
  *     @m_err_synlock_unlock
  *
  *     @m_err_rwlock_init
  *     @m_err_rwlock_read_begin
  *     @m_err_rwlock_read_end
  *     @m_err_rwlock_write_begin
  *     @m_err_rwlock_write_end
  *
  *     @m_err_stream_open,
  *     @m_err_stream_seek
  *     @m_err_stream_read
  *     @m_err_stream_write
  */
  MLIB_VAR( const M_ExceptionFamilyRec )   mlib_exception_family;


 /***************************************************************************/
 /***************************************************************************/
 /*****                                                                 *****/
 /*****                    MLIB EXCEPTION MACROS                        *****/
 /*****                                                                 *****/
 /***************************************************************************/
 /***************************************************************************/


 /***************************************************************
  *
  * @macro: M_XTRY_CORE (mcore)
  *
  * @description:
  *    a macro used in association with @m_xend to specify a set
  *    of instructions that will be tried for exceptions..
  *
  *    the "context" parameter must be a valid MLib Context
  *    handle..
  */
#define  M_XTRY_CORE(mcore)                                                  \
            {                                                                \
              M_XHandlerRec  __xhandler;                                     \
                                                                             \
              m_core_set_xhandler( mcore, &__xhandler );                     \
              if ( !__xhandler.exception && setjmp( __xhandler.env ) == 0 )  \
              {


 /***************************************************************
  *
  * @macro: M_XTRY
  *
  * @description:
  *    this macro is a simpler variant of @m_xtry_core that
  *    uses an implicit "mcore" variable to address the MLib
  *    core context..
  */
#define  M_XTRY  M_XTRY_CORE(mcore)


 /***************************************************************
  *
  * @macro: M_XEND
  *
  * @description:
  *    a macro used in association with @m_xtry to specify a set
  *    of instructions that will be tried for exceptions..
  *
  */
#define  M_XEND                                                          \
              }                                                          \
              if ( __xhandler.exception )                                \
                m_core_unset_xhandler( __xhandler.mcore, &__xhandler );  \
            }


 /***************************************************************
  *
  * @macro: M_XCATCH (e)
  *
  * @description:
  *    a macro used to begin an exception-handling section of
  *    code. the section _must_ ended with a @m_xend
  */
#define  M_XCATCH( _e_ )                                                  \
              }                                                           \
              else                                                        \
              {                                                           \
                M_Exception   _e_ =                                       \
                      m_core_unset_xhandler( __xhandler.mcore, &__xhandler );


 /***************************************************************
  *
  * @macro: M_XCATCH_ANY
  *
  * @description:
  *    a macro similar to @m_xcatch, except that the exception
  *    code is not returned in the catch statement.
  */
#define  M_XCATCH_ANY                                                     \
              }                                                           \
              else                                                        \
              {                                                           \
                (void) m_core_unset_xhandler( __xhandler.mcore, &__xhandler );


 /***************************************************************
  *
  * @macro: M_XTHROW_CORE (e,mcore)
  *
  * @description:
  *    a macro used to throw an exception within a given MLib core context
  */
#define  M_XTHROW_CORE( _e_, _mcore_ )       m_core_throw( _mcore_, _e_ )


 /***************************************************************
  *
  * @macro: M_XTHROW (e)
  *
  * @description:
  *    a macro used to throw an exception, using the _implicit_
  *    "mcore" variable for the MLib core context
  */
#define  M_XTHROW( _e_ )     m_core_throw( mcore, _e_ )


 /***************************************************************
  *
  * @macro: M_XRETHROW_CORE (e,mcore)
  *
  * @description:
  *    a macro used to re-throw an exception, i.e. throw a new one
  *    when you're in a M_XCATCH .. M_XEND statement block..
  */
#define  M_XRETHROW_CORE( _e_, _mcore_ )   m_core_rethrow( _mcore_, _e_ )

 /***************************************************************
  *
  * @macro: M_XRETHROW (e)
  *
  * @description:
  *    a variant of @M_XRETHROW that uses the implicit "mcore"
  *    variable for the MLib core handle
  */
#define  M_XRETHROW( _e_ )     M_XRETHROW_CORE( _e_, mcore )

  /* */

M_END_HEADER

#endif /* __M_EXCEPT_H__ */
