#ifndef __MLIB_FILE_H__
#define __MLIB_FILE_H__

#include <mlib/date.h>

M_BEGIN_HEADER

 /*************************************************************************
  *
  * @section: files
  *
  * @title: Files
  *
  * @abstract: Locating & naming files
  *
  * @description:
  *   the MLib uses @M_File objects to model the _location_ of files on
  *   the system. Opening a file object creates a new @M_Stream that can
  *   be used to parse or update the file's contents.
  *
  *   for now, the MLib only supports two kinds of files:
  *
  *   memory-based files, which can be located in any location of memory
  *   which are named by a base address (along with a byte size)
  *
  *   ANSI-based files, which are named by a simple C string
  *   (the "filepathanme"), and accessed internally by the i/o functions
  *   provided by the ANSI C library (e.g. "fread", "fseek", etc..)
  *
  *   it is possible to define other file types by sub-classing the
  *   abstract @M_File_Class
  *
  * @order:
  *   M_File
  *
  *   m_file_new_memory
  *   m_file_new_ansi
  *   m_file_open_stream
  *
  *   m_err_file_init
  *   m_err_file_open
  *
  *   M_File_Class
  *   M_File_ClassRec
  */

 /*************************************************************************
  *
  * @type: M_File
  *
  * @description:
  *   an opaque handle to a file object, used to model the _location_ of a
  *   given file, be it in local memory, on the disk, on a remote
  *
  *   note that these are normal @M_Object objects
  */
  typedef struct M_FileRec_*   M_File;


 /*************************************************************************
  *
  * @enum: M_File_Flags
  *
  * @description:
  *   the list of bit flags used to open new files with @m_file_open
  *
  * @values:
  *   M_FILE_READ       :: set to open a file for reading
  *
  *   M_FILE_WRITE      :: set to open a file for writing
  *
  *   M_FILE_READ_WRITE :: union of @M_FILE_READ and @M_FILE_WRITE, used
  *                        to open a file for _both_ reading and writing
  *
  *   M_FILE_APPEND     :: set to open a file for appends (i.e. writing
  *                        from the end of a file..)
  */
  enum
  {
    M_FILE_READ       = 1,
    M_FILE_WRITE      = 2,
    M_FILE_READ_WRITE = M_FILE_READ | M_FILE_WRITE,
    M_FILE_APPEND     = 4
  };

 /*************************************************************************
  *
  * @type: M_File_Class
  *
  * @description:
  *   handle to a constant @M_File_ClassRec structure used to describe the
  *   class of a @M_File object.
  */
  typedef const struct M_File_ClassRec_*   M_File_Class;


 /*************************************************************************
  *
  * @type: M_File_OpenFunc
  *
  * @description:
  *   a method used to open a new file
  *
  * @input:
  *   file       :: file handle
  *   open_flags :: a set of @M_File_Flags used to determine how to open
  *                 the file (for reading, writing, etc..)
  * @return:
  *   new @M_Stream object to read from or write to the file
  *
  * @throws: any
  */
  typedef  M_Stream  (*M_File_OpenFunc)( M_File  file,
                                         M_UInt  open_flags );

 /*************************************************************************
  *
  * @type: M_File_GetNameFunc
  *
  * @description:
  *   a method used for debugging purposes only. It is used to convert
  *   a @M_File object into a C string describing the file. It could, for
  *   example, return a filepathname or a URL, depending on the file's
  *   real type
  *
  * @input:
  *   file       :: file handle
  *
  * @return:
  *   new @M_CString describing the file. It must be destroyed explicitely
  *   by the caller using @m_mem_free or @m_str_destroy
  *
  * @throws: any
  */
  typedef  M_CString (*M_File_GetNameFunc)( M_File  file );


 /*************************************************************************
  *
  * @struct: M_File_ClassRec
  *
  * @description:
  *   a structure used to model the class of a given @M_File implementation
  *
  * @fields:
  *   clazz    :: root @M_Class fields
  *   open     :: file open method
  *   get_path :: file pathanme debugging method. optional, can be NULL
  */
  typedef struct M_File_ClassRec_
  {
    M_ClassRec            clazz;
    M_File_OpenFunc       file_open;
    M_File_DebugTextFunc  file_pathname;

  } M_File_ClassRec;


 /*************************************************************************
  *
  * @function: m_file_new_memory
  *
  * @description:
  *   create a new memory-based file, i.e a file whose content is stored
  *   in memory
  *
  * @input:
  *   address       :: address of file's first byte
  *
  *   byte_size     :: file size in bytes
  *
  *   write_allowed :: boolean. if true, writing to the file is allowed
  *                    (but limited to the first "byte_size" bytes).
  *
  *   memory        :: handle to memory manager
  *
  *   push          :: boolean. if true, the new file object is pushed
  *                    on top of the cleanup stack
  * @return:
  *   new @M_File handle
  *
  * @throws: @m_err_memory_alloc
  */
  MLIB_API( M_File )     m_file_new_memory( M_Byte*    address,
                                            M_Size     byte_size,
                                            M_Bool     write_allowed,
                                            M_Memory   memory,
                                            M_Bool     push );


 /*************************************************************************
  *
  * @function: m_file_new_ansi
  *
  * @description:
  *   create a new file from a normal ANSI "filepathanme", like the one
  *   used in "fopen", etc..
  *
  * @input:
  *   filepathname :: the file's pathname
  *   mcore        :: MLib core handle
  *   push         :: boolean. if true, the new file is pushed on top of
  *                   the cleanup stack
  * @return:
  *   new @M_File handle
  *
  * @throws: @m_err_memory_alloc
  */
  M_File     m_file_new_ansi( M_CString  filepathname,
                              M_Core     mcore,
                              M_Bool     push );


 /*************************************************************************
  *
  * @function: m_file_open_stream
  *
  * @description:
  *   open a given file and create a new stream object for it..
  *
  * @input:
  *   file       :: source file handle
  *
  *   open_flags :: bit-flags determining how to open the file.
  *                 see @M_File_Flags
  *
  *   push       :: boolean. if true, the new stream will be pushed
  *                 on top of the cleanup stack
  *
  * @return:
  *   new @M_Stream object used to access the file's content
  */
  M_Stream   m_file_open_stream( M_File  file,
                                 M_UInt  open_flags,
                                 M_Bool  push );


 /*************************************************************************
  *
  * @const: m_file_class_type
  *
  * @description:
  *   the class type of file objects. You should only use it to define
  *   new file classes, or to check @M_File handles..
  */
  MLIB_VAR( M_ClassType )   m_file_class_type;


 /*************************************************************************
  *
  * @macro: M_FILE (x)
  *
  * @description:
  *   a convenient macro to convert any pointer to a @M_File handle
  */
#define  M_FILE(x)       ((M_File)(x))

 /*************************************************************************
  *
  * @macro: M_FILE_CHK (x)
  *
  * @description:
  *   a variant of @M_FILE that also checks in debug builds that the input
  *   handle really points to a file object
  */
#define  M_FILE_CHK(x)   M_FILE(m_object_check(x,m_file_class_type))



  MLIB_VAR( M_ClassType )   m_file_ansi_class_type;

  MLIB_VAR( M_ClassType )   m_file_memory_class_type;


  MLIB_API( M_Exception )       m_err_file_open;
  MLIB_API( M_Exception )       m_err_file_init;


M_END_HEADER

#endif /* __M_FILE_H__ */
