/*************************************************************************
 *
 * template for fast inlined AVL tree routines (lookup, insertion and
 * removal). You must define only ONE of the following macros before
 * including this file in your source code:
 *
 *     GEN_AVL_LOOKUP_DEF
 *     GEN_AVL_INSERTION_DEF
 *     GEN_AVL_REMOVAL_DEF
 *
 */

#include <mlib/generic/contain.i>

/**************************************************************************
 *
 * @macro: GEN_AVL_SIZE (tree)
 *
 * @description:
 *   this macro is used to return the lvalue of the number of nodes in
 *   an AVL tree
 */
#ifndef  GEN_AVL_SIZE
#define  GEN_AVL_SIZE(_tree)  ((_tree)->nodes)
#endif

/**************************************************************************
 *
 * @macro: GEN_AVL_ROOT (tree)
 *
 * @description:
 *   this macro is used to return the lvalue of root node in an AVL tree
 */
#ifndef  GEN_AVL_ROOT
#define  GEN_AVL_ROOT(_tree)   ((_tree)->root)
#endif

/**************************************************************************
 *
 * @macro: GEN_BALANCE
 *
 * @description:
 *   this macro holds the type of the "balance" field in AVL node trees
 */
#ifndef  GEN_BALANCE
#error  "GEN_BALANCE undefined in AVL tree template instantiation"
#endif

/**************************************************************************
 *
 * @macro: GEN_CACHE
 *
 * @description:
 *   this macro holds the type of the "cache" field in AVL node trees
 */
#ifndef  GEN_CACHE
#error  "GEN_CACHE undefined in AVL tree template instantiation"
#endif

/**************************************************************************
 *
 * @macro: GEN_NODE_BALANCE
 *
 * @description:
 *   this macro returns the lvalue of an AVL tree node's balance field
 */
#ifndef  GEN_NODE_BALANCE
#define  GEN_NODE_BALANCE(_node)  ((_node)->balance)
#endif

/**************************************************************************
 *
 * @macro: GEN_NODE_BALANCE
 *
 * @description:
 *   this macro returns the lvalue of an AVL tree node's cache field
 */
#ifndef  GEN_NODE_CACHE
#define  GEN_NODE_CACHE(_node)    ((_node)->cache)
#endif

/**************************************************************************
 *
 * @macro: GEN_NODE_BALANCE
 *
 * @description:
 *   this macro returns the lvalue of an AVL tree node's left child
 */
#ifndef  GEN_NODE_LEFT
#define  GEN_NODE_LEFT(_node)     ((_node)->left)
#endif

/**************************************************************************
 *
 * @macro: GEN_NODE_BALANCE
 *
 * @description:
 *   this macro returns the lvalue of an AVL tree node's right child
 */
#ifndef  GEN_NODE_RIGHT
#define  GEN_NODE_RIGHT(_node)    ((_node)->right)
#endif

/**************************************************************************
 *
 * @macro: GEN_NODE_CHILD_P
 *
 * @description:
 *   this macro returns the address of one of the childs of an AVL tree node
 */
#ifndef  GEN_NODE_CHILD_P
#define  GEN_NODE_CHILD_P(_node,_dir)            \
           ((_dir) > 0) ? &GEN_NODE_RIGHT(_node) \
                        : &GEN_NODE_LEFT (_node)
#endif

/**************************************************************************
 *
 * @macro: GEN_NODE_CACHED_P
 *
 * @description:
 *   this macro returns the address of an AVL tree node's cached child
 */
#ifndef  GEN_NODE_CACHED_P
#define  GEN_NODE_CACHED_P(_node)  \
             GEN_NODE_CHILD_P( _node, GEN_NODE_CACHE(_node) )
#endif

/**************************************************************************
 *
 * @macro: GEN_AVL_CLEAR (tree)
 *
 * @description:
 *   when defined, this macro is the name of a function used to clear
 *   an AVL tree
 */
#ifdef  GEN_AVL_CLEAR_DEF

#ifndef GEN_AVL_CLEAR
#define GEN_AVL_CLEAR    GEN_TABLE_FUNC(_clear)
#endif

GEN_DEF( void )
GEN_AVL_CLEAR( GEN_TABLE  tree )
{
  if ( tree )
  {
    GEN_NODE  stack[64], *top, child, node;

    GEN_TABLE_ASSERT(tree);

    node = GEN_AVL_ROOT(tree);
    top  = stack;
    for (;;)
    {
      child = GEN_NODE_LEFT(node);
      if (child)
      {
        *top++ = node;
        node   = child;
        continue;
      }

      child = GEN_NODE_RIGHT(node);
      if (child)
      {
        *top++ = node;
        node   = child;
        continue;
      }

      /* delete the node */
      GEN_NODE_DESTROY(node,tree);

      top--;
      if ( top < stack )
        break;

      node = *top;
    }
    GEN_AVL_SIZE(tree) = 0;
    GEN_AVL_ROOT(tree)  = NULL;
  }
}

#undef GEN_AVL_CLEAR_DEF
#undef GEN_AVL_CLEAR
#endif /* GEN_AVL_CLEAR_DEF */



/**************************************************************************
 *
 * @macro: GEN_AVL_LOOKUP (tree, key, pvalue)
 *
 * @description:
 *   when defined, this macro is the name of a function used to lookup
 *   in an AVL tree
 */
#ifndef GEN_AVL_LOOKUP
#define GEN_AVL_LOOKUP  GEN_TABLE_FUNC(_lookup)
#endif

#ifdef  GEN_AVL_LOOKUP_DEF

GEN_DEF( M_Bool )
GEN_AVL_LOOKUP( GEN_TABLE   tree,
                    GEN_KEY     key,
                    GEN_VALUE  *avalue )
{
  M_Bool             result = 0;
  GEN_NODE       node   = NULL;
  M_CompareDataFunc  key_compare;

  GEN_TABLE_ASSERT(tree);

  node        = GEN_AVL_ROOT(tree);
  key_compare = GEN_KEY_COMPARE(key,tree);

  if ( key_compare )
  {
    M_Pointer  key_data = GEN_KEY_DATA(key,tree);
    M_Int      dir;

    while (node)
    {
      dir = key_compare( GEN_NODE_KEY(node), key, key_data );
      if ( dir == 0 )
        break;

      node = *GEN_NODE_CHILD_P(node,(dir > 0));
    }
  }
  else
  {
    M_Int  dir;

    while (node)
    {
      dir = (M_Byte*)GEN_NODE_KEY(node) - (M_Byte*)key;
      if ( dir == 0 )
        break;

      node = *GEN_NODE_CHILD_P(node,(dir > 0));
    }
  }

  result = ( node != NULL );
  if ( result && avalue )
    *avalue = GEN_NODE_VALUE(node);

  return result;
}

#undef GEN_AVL_LOOKUP_DEF
#endif /* GEN_AVL_LOOKUP_DEF */




/**************************************************************************
 *
 * @macro: GEN_AVL_GET (tree,key)
 *
 * @description:
 *   when defined, this macro is the name of a function used to get
 *   a given key's value from an AVL tree
 */
#ifdef  GEN_AVL_GET_DEF

#ifndef GEN_AVL_GET
#define GEN_AVL_GET  GEN_TABLE_FUNC(_get)
#endif

GEN_DEF( GEN_VALUE )
GEN_AVL_GET( GEN_TABLE  tree,
                 GEN_KEY    key )
{
  GEN_VALUE  result;

  GEN_TABLE_ASSERT(tree);

  GEN_AVL_LOOKUP( tree, key, &result );
  return result;
}

#undef GEN_AVL_GET_DEF
#undef GEN_AVL_GET
#endif /* GEN_AVL_GET_DEF */





/**************************************************************************
 *
 * @macro: GEN_AVL_SET (tree,key)
 *
 * @description:
 *   when defined, this macro is the name of a function used to set
 *   a given key's value in an AVL tree
 */
#ifdef  GEN_AVL_SET_DEF

#ifndef GEN_AVL_SET
#define GEN_AVL_SET  GEN_TABLE_FUNC(_set)
#endif

#define  GEN_AVL_INSERTION  M_CONCAT( GEN_AVL_SET, _insert )

static void
GEN_AVL_INSERTION( GEN_NODE*   pnode,
                   GEN_NODE*   protate,
                   GEN_NODE    new_node )
{
  GEN_NODE  node, rotate;

  /* now, insert in the tree */
  GEN_NODE_BALANCE(new_node) = 0;

  pnode[0] = new_node;
  if ( pnode == protate )
    return;

  /* now adjust balances between subtree and new node */
  rotate = *protate;
  pnode  = GEN_NODE_CACHED_P(rotate);
  node   = *pnode;
  while (node != new_node)
  {
    GEN_NODE_BALANCE(node) = (GEN_BALANCE) (GEN_NODE_CACHE(node)*2-1);

    pnode = GEN_NODE_CACHED_P(node);
    node  = *pnode;
  }

  /* perform subtree adjustment / rotation */
  {
    GEN_NODE  n1, n2, n3;

    n1 = rotate;
    if ( GEN_NODE_CACHE(n1) == 0 )
    {
      /* new node was inserte to the left of the subtree */
      if ( GEN_NODE_BALANCE(n1) < 0 )
      {
        n2 = GEN_NODE_LEFT(n1);

        if ( GEN_NODE_BALANCE(n2) < 0 )
        {
          /* simple rotation */
          m_assert( GEN_NODE_BALANCE(n2) == -1 );

          GEN_NODE_LEFT(n1)  = GEN_NODE_RIGHT(n2);
          GEN_NODE_RIGHT(n2) = n1;
          *protate     = n2;

          GEN_NODE_BALANCE(n1) = 0;
          GEN_NODE_BALANCE(n2) = 0;
        }
        else
        {
          /* double rotation */
          m_assert( GEN_NODE_BALANCE(n2) == +1 );

          n3                     = GEN_NODE_RIGHT(n2);
          GEN_NODE_RIGHT(n2) = GEN_NODE_LEFT(n3);
          GEN_NODE_LEFT(n3)  = n2;

          GEN_NODE_LEFT(n1)  = GEN_NODE_RIGHT(n3);
          GEN_NODE_RIGHT(n3) = n1;

          *protate     = n3;

          GEN_NODE_BALANCE(n1) = (GEN_BALANCE)(GEN_NODE_BALANCE(n3) == -1);
          GEN_NODE_BALANCE(n2) = (GEN_BALANCE)(GEN_NODE_BALANCE(n3) == +1);
          GEN_NODE_BALANCE(n3) = 0;
        }
      }
      else
      {
        m_assert( GEN_NODE_BALANCE(n1) < 2 );

        GEN_NODE_BALANCE(n1)--;
      }
    }
    else
    {
      /* new node was inserted to the right of the subtree */
      if ( GEN_NODE_BALANCE(n1) > 0 )
      {
        n2 = GEN_NODE_RIGHT(n1);

        if ( GEN_NODE_BALANCE(n2) > 0 )
        {
          /* simple rotation */
          m_assert( GEN_NODE_BALANCE(n2) == +1 );

          GEN_NODE_RIGHT(n1) = GEN_NODE_LEFT(n2);
          GEN_NODE_LEFT (n2) = n1;

          GEN_NODE_BALANCE(n1)  = 0;
          GEN_NODE_BALANCE(n2)  = 0;

          *protate = n2;
        }
        else
        {
          m_assert( GEN_NODE_BALANCE(n2) == -1 );

          /* double rotation */
          n3                     = GEN_NODE_LEFT(n2);
          GEN_NODE_LEFT (n2) = GEN_NODE_RIGHT(n3);
          GEN_NODE_RIGHT(n3) = n2;

          GEN_NODE_RIGHT(n1) = GEN_NODE_LEFT(n3);
          GEN_NODE_LEFT(n3)  = n1;

          *protate = n3;

          GEN_NODE_BALANCE(n1) = (GEN_BALANCE) -(GEN_NODE_BALANCE(n3) == +1);
          GEN_NODE_BALANCE(n2) = (GEN_BALANCE)  (GEN_NODE_BALANCE(n3) == -1);
          GEN_NODE_BALANCE(n3) = 0;
        }
      }
      else
      {
        m_assert( GEN_NODE_BALANCE(n1) > -2 );

        GEN_NODE_BALANCE(n1)++;
      }
    }
  }
}


GEN_DEF( void )
GEN_AVL_SET( GEN_TABLE  tree,
                 GEN_KEY    key,
                 GEN_VALUE  value )
{
  GEN_NODE      *pnode, node, *protate;
  M_CompareDataFunc  key_compare = GEN_KEY_COMPARE(key,tree);

  GEN_TABLE_ASSERT(tree);

  pnode       = &GEN_AVL_ROOT(tree);
  node        = *pnode;
  protate     = pnode;
  key_compare = GEN_KEY_COMPARE(key,tree);

  if ( node )
  {
    if ( key_compare )
    {
      M_Pointer  key_data = GEN_KEY_DATA(key,tree);
      M_Int      dir;

      for (;;)
      {
        dir = key_compare( GEN_NODE_KEY(node), key, key_data );
        if ( dir == 0 )
          break;

        dir = (dir > 0);
        GEN_NODE_CACHE(node) = (GEN_CACHE)dir;

        pnode = GEN_NODE_CHILD_P(node,dir);
        node  = pnode[0];
        if ( !node )
          break;

        if ( GEN_NODE_BALANCE(node) != 0 )
          protate = pnode;
      }
    }
    else
    {
      M_Dist     dir;

      for (;;)
      {
        dir = (M_Byte*)GEN_NODE_KEY(node) - (M_Byte*)key;
        if ( dir == 0 )
          break;

        dir = (dir > 0);
        GEN_NODE_CACHE(node) = (GEN_CACHE)dir;

        pnode = GEN_NODE_CHILD_P(node,dir);
        node  = pnode[0];
        if ( !node )
          break;

        if ( GEN_NODE_BALANCE(node) != 0 )
          protate = pnode;
      }
    }
  }

  node = pnode[0];
  if ( !node )
  {
    /* allocate new node */
    GEN_NODE_NEW( node, key, value, tree );

    GEN_AVL_INSERTION( pnode, protate, node );

    GEN_AVL_SIZE(tree)++;
  }
  else  /* replace value */
  {
    GEN_NODE_UNSET(node,tree);
    GEN_NODE_SET(node,key,value,tree);
  }
}

#undef GEN_AVL_SET_DEF
#undef GEN_AVL_SET
#undef GEN_AVL_INSERTION
#endif /* GEN_AVL_SET_DEF */



/**************************************************************************
 *
 * @macro: GEN_AVL_REMOVE (tree,key)
 *
 * @description:
 *   when defined, this macro is the name of a function used to remove
 *   a given key, with its value, from an AVL tree
 */
#ifdef  GEN_AVL_REMOVE_DEF

#ifndef GEN_AVL_REMOVE
#define GEN_AVL_REMOVE  GEN_TABLE_FUNC(_remove)
#endif

GEN_DEF( void )
GEN_AVL_REMOVE( GEN_TABLE  tree,
                GEN_KEY    key )
{
  GEN_NODE       stack[64], *top, *proot, *pnode, node;
  M_CompareDataFunc  key_compare;

  GEN_TABLE_ASSERT(tree);

  proot       = &GEN_AVL_ROOT(tree);
  pnode       = proot;
  node        = pnode[0];
  top         = stack;
  key_compare = GEN_KEY_COMPARE(key,tree);

  /* find target node */
  if ( key_compare )
  {
    M_Pointer  key_data = GEN_KEY_DATA(key,tree);
    M_Int      dir;

    for (;;)
    {
      node = pnode[0];
      if ( !node )
        return;

      dir = key_compare( GEN_NODE_KEY(node), key, key_data );
      if ( dir == 0 )
        break;

      GEN_NODE_CACHE(node) = (GEN_CACHE) (dir > 0);

      m_assert( top - stack < 63 );

      *top++ = node;
      pnode  = GEN_NODE_CACHED_P(node);
    }
  }
  else
  {
    M_Dist  dir;

    for (;;)
    {
      node = pnode[0];
      if ( !node )
        return;

      dir = (M_Byte*)GEN_NODE_KEY(node) - (M_Byte*)key;
      if ( dir == 0 )
        break;

      GEN_NODE_CACHE(node) = (GEN_CACHE) (dir > 0);

      *top++ = node;
      pnode  = GEN_NODE_CACHED_P(node);
    }
  }

  m_assert( node );

  /* rebalance the tree */
  {
    GEN_NODE  n1, n2, n3;

    /* remove it from the tree graph */
    n1 = GEN_NODE_RIGHT(node);

    if ( n1 == NULL )
    {
      *pnode = GEN_NODE_LEFT(node);
    }
    else if ( GEN_NODE_LEFT(node) == NULL )
    {
      GEN_NODE_LEFT(n1)    = GEN_NODE_LEFT(node);
      GEN_NODE_BALANCE(n1) = GEN_NODE_BALANCE(node);
      GEN_NODE_CACHE(n1)   = 1;

      *pnode = n1;
      *top++ = n1;
    }
    else
    {
      GEN_NODE*  slot = top++;

      for (;;)
      {
        GEN_NODE_CACHE(n1) = 0;

        *top++ = n1;

        n2 = GEN_NODE_LEFT(n1);
        if ( GEN_NODE_LEFT(n2) == NULL )
          break;

        n1 = n2;
      }

      GEN_NODE_CACHE(n2) = 1;
      *slot = n2;

      GEN_NODE_LEFT(n2)  = GEN_NODE_LEFT(node);
      GEN_NODE_LEFT(n1)  = GEN_NODE_RIGHT(n2);
      GEN_NODE_RIGHT(n2) = GEN_NODE_RIGHT(node);
      *pnode = n2;
    }

    /* now, adjust the balance factors, eventually perform rotations */
    --top;
    while ( top >= stack )
    {
      n1  = *top--;

      /* retrieve address of parent pointer */
      if ( top >= stack )
        pnode = GEN_NODE_CACHED_P(*top);
      else
        pnode = proot;

      if ( GEN_NODE_CACHE(n1) == 0 )
      {
        /* we removed a node that was on the left sub-tree          */
        if ( GEN_NODE_BALANCE(n1) == -1 )
        {
          /* this node becomes balanced, the max subtree height */
          /* has changed so continue recursion                  */
          GEN_NODE_BALANCE(n1) = 0;
        }
        else if ( GEN_NODE_BALANCE(n1) == 0 )
        {
          /* this node was balanced, its left subtree is now smaller */
          /* than its right subtree. we can stop the recursion here  */
          /* because the max subtree height hasn't changed           */
          GEN_NODE_BALANCE(n1) = 1;
          break;
        }
        else
        {
          m_assert( GEN_NODE_BALANCE(n1) == 1 );

          /* the right subtree was already taller than the left subtree */
          /* we need to rotate the node here..                          */
          n2 = GEN_NODE_RIGHT(n1);

          m_assert( n2 );

          if ( GEN_NODE_BALANCE(n2) == -1 )
          {
            /* double rotation */
            n3                     = GEN_NODE_LEFT(n2);
            GEN_NODE_LEFT(n2)  = GEN_NODE_RIGHT(n3);
            GEN_NODE_RIGHT(n3) = n2;
            GEN_NODE_RIGHT(n1) = GEN_NODE_LEFT(n3);
            GEN_NODE_LEFT(n3)  = n1;
            *pnode = n3;

            GEN_NODE_BALANCE(n1) = (GEN_BALANCE) -(GEN_NODE_BALANCE(n3) ==  1);
            GEN_NODE_BALANCE(n2) = (GEN_BALANCE)  (GEN_NODE_BALANCE(n3) == -1);
          }
          else
          {
            /* simple rotation */
            GEN_NODE_RIGHT(n1) = GEN_NODE_LEFT(n2);
            GEN_NODE_LEFT(n2)  = n1;
            *pnode = n2;

            if (GEN_NODE_BALANCE(n2) == 0)
            {
              GEN_NODE_BALANCE(n2) = -1;
              break;
            }
            GEN_NODE_BALANCE(n1) = 0;
            GEN_NODE_BALANCE(n2) = 0;
          }
        }
      }
      else
      {
        /* we remove a node that was on the right sub-tree         */
        if ( GEN_NODE_BALANCE(n1) == +1 )
        {
          /* this node becomes balanced, the max subtree height */
          /* has changed so continue recursion                  */
          GEN_NODE_BALANCE(n1) = 0;
        }
        else if ( GEN_NODE_BALANCE(n1) == 0 )
        {
          /* this node was balanced, its right subtree is now smaller */
          /* than its left subtree. we can stop the recursion here    */
          /* because the max subtree height hasn't changed            */
          GEN_NODE_BALANCE(n1) = -1;
          break;
        }
        else
        {
          m_assert( GEN_NODE_BALANCE(n1) == -1 );

          /* the left subtree was already taller than the right one */
          /* we need to rotate the node here..                      */
          n2 = GEN_NODE_LEFT(n1);

          m_assert( n2 );

          if ( GEN_NODE_BALANCE(n2) == +1 )
          {
            /* double rotation */
            n3           = GEN_NODE_RIGHT(n2);
            GEN_NODE_RIGHT(n2) = GEN_NODE_LEFT(n3);
            GEN_NODE_LEFT(n3) = n2;
            GEN_NODE_LEFT(n1) = GEN_NODE_RIGHT(n3);
            GEN_NODE_RIGHT(n3) = n1;
            *pnode = n3;

            GEN_NODE_BALANCE(n1) = (GEN_BALANCE)  (GEN_NODE_BALANCE(n3) == -1);
            GEN_NODE_BALANCE(n2) = (GEN_BALANCE) -(GEN_NODE_BALANCE(n3) == +1);
          }
          else
          {
            /* simple rotation */
            GEN_NODE_LEFT(n1) = GEN_NODE_RIGHT(n2);
            GEN_NODE_RIGHT(n2) = n1;
            *pnode = n2;

            if (GEN_NODE_BALANCE(n2) == 0)
            {
              GEN_NODE_BALANCE(n2) = +1;
              break;
            }
            GEN_NODE_BALANCE(n1) = 0;
            GEN_NODE_BALANCE(n2) = 0;
          }
        }
      }
    }
  }

  /* destroy node */
  GEN_NODE_DESTROY(node,tree);
  GEN_AVL_SIZE(tree)--;
}

#undef GEN_AVL_REMOVE_DEF
#undef GEN_AVL_REMOVE
#endif /* GEN_AVL_REMOVE_DEF */




/**************************************************************************
 *
 * @macro: GEN_AVL_DEPTH (tree)
 *
 * @description:
 *   when defined, this macro is the name of a function used to return
 *   the depth of a given AVL tree
 */
#ifdef  GEN_AVL_DEPTH_DEF

#ifndef GEN_AVL_DEPTH
#define GEN_AVL_DEPTH  GEN_TABLE_FUNC(_depth)
#endif

#define  GEN_AVL_NODE_DEPTH  GEN_AVL_DEPTH ## _node

static M_Size
GEN_AVL_NODE_DEPTH( GEN_NODE  node )
{
  M_Size  left;
  M_Size  right;

  if ( !node )
    return 0;

  left  = GEN_AVL_NODE_DEPTH( GEN_NODE_LEFT(node) );
  right = GEN_AVL_NODE_DEPTH( GEN_NODE_RIGHT(node) );

  if ( left < right )
    left = right;

  return left + 1;
}


GEN_DEF( M_Size )
GEN_AVL_DEPTH( GEN_TABLE  tree )
{
  GEN_TABLE_ASSERT( tree );

  return GEN_AVL_NODE_DEPTH( GEN_AVL_ROOT(tree) );
}

#undef GEN_AVL_DEPTH_DEF
#undef GEN_AVL_DEPTH
#undef GEN_AVL_NODE_DEPTH
#endif /* GEN_AVL_DEPTH_DEF */



/**************************************************************************
 *
 * @macro: GEN_AVL_ITERATE (tree)
 *
 * @description:
 *   when defined, this macro is the name of a function used to iterate
 *   over all nodes of a given AVL tree
 */
#ifdef  GEN_AVL_ITERATE_DEF

#ifndef GEN_AVL_ITERATE
#define GEN_AVL_ITERATE  GEN_TABLE_FUNC(_iterate)
#endif

GEN_DEF( void )
GEN_AVL_ITERATE( GEN_TABLE    tree
                 GEN_PARAMS_ITER_ )
{
  GEN_NODE  stack[64], *top, node, child;


  GEN_TABLE_ASSERT(tree);

  if ( iterate_func )
  {
    top    = stack;
    node   = GEN_AVL_ROOT(tree);

    if ( node == NULL )
      return;

    for ( ;; )
    {
      child = GEN_NODE_LEFT(node);
      if (child)
      {
        *top++ = node;
        node   = child;
        continue;
      }

      GEN_NODE_ITERATE(node,tree);

      child = GEN_NODE_RIGHT(node);
      if (child)
      {
        node = child;
        continue;
      }

      if ( top-- == stack )
        break;

      node = *top;
    }
  }
}

#undef GEN_AVL_ITERATE_DEF
#undef GEN_AVL_ITERATE
#endif /* GEN_AVL_ITERATE_DEF */

/* no, undefine all generic macros */
#undef GEN_AVL_ROOT
#undef GEN_AVL_SIZE

#undef GEN_NODE_LEFT
#undef GEN_NODE_RIGHT
#undef GEN_NODE_CACHE
#undef GEN_NODE_BALANCE
#undef GEN_NODE_CHILD_P
#undef GEN_NODE_CACHED_P
#undef GEN_CACHE
#undef GEN_BALANCE

#undef GEN_AVL_LOOKUP

#include <mlib/generic/undefs.i>
