/****************************************************************************
 *
 *  this file is used by all table templates, it defines/checks several
 *  macros related to containers.
 *
 *  before including this file, you _absolutely_ need to define the
 *  following macros:
 *
 *   GEN_TABLE  :: the type of the container table
 *   GEN_NODE   :: the type of the container nodes
 *   GEN_KEY    :: the type of container node keys
 *   GEN_VALUE  :: the type of containter node values
 *
 *  besides, you'll most probably want to define these as well:
 *
 *   GEN_TABLE_MEMORY
 *   GEN_TABLE_POOL
 *
 *   GEN_NODE_KEY
 *   GEN_NODE_VALUE
 *
 *   GEN_KEY_COMPARE
 *   GEN_KEY_DATA
 *
 *   GEN_NODE_NEW
 *   GEN_NODE_UNSET
 *   GEN_NODE_SET
 *   GEN_NODE_DESTROY
 *
 *  read the comments below for more details..
 */


/**************************************************************************
 *
 * @macro: GEN_TABLE
 *
 * @description:
 *   this macro holds the type of the hash table
 */
#ifndef GEN_TABLE
#error  "GEN_TABLE macro not defined in table template instantiation"
#endif


/**************************************************************************
 *
 * @macro: GEN_TABLE_PREFIX
 *
 * @description:
 *   this macro gives the _name_ prefix of table functions. for example,
 *   define it to "m_hash_table" to be able to define functions like
 *   "m_hash_table_set", "m_hash_table_get", etc..
 */
#ifndef  GEN_TABLE_PREFIX
#error  "GEN_TABLE_PREFIX macro not defined in table template instantiation"
#endif



#define  GEN_TABLE_FUNC(suffix)  M_CONCAT(GEN_TABLE_PREFIX,suffix)


/**************************************************************************
 *
 * @macro: GEN_TABLE_ASSERT
 *
 * @description:
 *   this macro is used to check a container type
 */
#ifndef GEN_TABLE_ASSERT
#define GEN_TABLE_ASSERT(table)  M_UNUSED(table)
#endif

/**************************************************************************
 *
 * @macro: GEN_TABLE_MEMORY
 *
 * @description:
 *   this macro holds the memory manager of the container table
 */
#ifndef GEN_TABLE_MEMORY
#define GEN_TABLE_MEMORY(_table)    (_table)->memory
#endif

/**************************************************************************
 *
 * @macro: GEN_TABLE_POOL
 *
 * @description:
 *   this macro holds the memory pool used to alloc/free nodes.
 *   by default, this is the container's memory allocator
 */
#ifndef GEN_TABLE_POOL
#define GEN_TABLE_POOL(_table)      GEN_TABLE_MEMORY(_table)
#endif


/**************************************************************************
 *
 * @macro: GEN_TABLE_ALLOC (block,size,table)
 *
 * @description:
 *   this macro is used to allocate a memory block within a given table
 */
#ifndef GEN_TABLE_ALLOC
#define GEN_TABLE_ALLOC(_block,_size,_table)  \
          _block = m_mem_alloc( _size, GEN_TABLE_MEMORY(_table) )
#endif


/**************************************************************************
 *
 * @macro: GEN_TABLE_QALLOC (block,size,table)
 *
 * @description:
 *   this macro is used to allocate a memory block within a given table
 *   if doesn't raise an exception in case of memory shortage, and the
 *   new block isn't necessarily zero-ed
 */
#ifndef GEN_TABLE_QALLOC
#define GEN_TABLE_QALLOC(_block,_size,_table)  \
          _block = m_mem_qalloc( _size, GEN_TABLE_MEMORY(_table) )
#endif


/**************************************************************************
 *
 * @macro: GEN_TABLE_ALLOC (block,old_size,new_size,table)
 *
 * @description:
 *   this macro is used to re-allocate a memory block within a given table
 */
#ifndef GEN_TABLE_REALLOC
#define GEN_TABLE_REALLOC(_block,_old_size,_new_size,_table)           \
          m_mem_realloc( (M_Pointer*)&(_block), _old_size, _new_size,  \
                         GEN_TABLE_MEMORY(_table) )
#endif

/**************************************************************************
 *
 * @macro: GEN_TABLE_FREE (block,table)
 *
 * @description:
 *   this macro is used to free a memory block that was allocated
 *   through GEN_TABLE_ALLOC or GEN_TABLE_QALLOC
 */
#ifndef GEN_TABLE_FREE
#define GEN_TABLE_FREE(_block,_table)  \
          m_mem_free( _block, GEN_TABLE_MEMORY(_table) )
#endif


/**************************************************************************
 *
 * @macro: GEN_PARAMS_ITER_
 *
 * @description:
 *   a very special macro used to _declare_ parameters in iteration
 *   functions. These are normally used by the GEN_NODE_ITERATE
 *   function too..
 */
#ifndef GEN_PARAMS_ITER_
#define GEN_PARAMS_ITER_   , M_IterateDataFunc   iterate_func, \
                             M_Pointer           iterate_data
#endif


/**************************************************************************
 *
 * @macro: GEN_NODE
 *
 * @description:
 *   this macro holds the type of the container nodes. It must always be
 *   defined before including this file
 */
#ifndef GEN_NODE
#error  "GEN_NODE not defined in container template instantation"
#endif


/**************************************************************************
 *
 * @macro: GEN_NODE_KEY (node)
 *
 * @description:
 *   this macro is used to retrieve the "key" value of a given hash node
 */
#ifndef GEN_NODE_KEY
#define GEN_NODE_KEY(_node)   ((_node)->key)
#endif


/**************************************************************************
 *
 * @macro: GEN_NODE_VALUE (node)
 *
 * @description:
 *   this macro is used to retrieve the "value" of a given container node
 */
#ifndef GEN_NODE_VALUE
#define GEN_NODE_VALUE(_node)  ((_node)->value)
#endif


/**************************************************************************
 *
 * @macro: GEN_NODE_NEW (node,key,value,table)
 *
 * @description:
 *   this macro creates a new container node
 */
#ifndef GEN_NODE_NEW
#define GEN_NODE_NEW(_node,_key,_value,_table)             \
             M_BEGIN_STMNT                                 \
               M_Memory  memory = GEN_TABLE_POOL(_table);  \
                                                           \
               M_NEW( _node );                             \
               GEN_NODE_KEY(_node)   = _key;               \
               GEN_NODE_VALUE(_node) = _value;             \
             M_END_STMNT
#endif


/**************************************************************************
 *
 * @macro: GEN_NODE_SET (node,key,value,table)
 *
 * @description:
 *   this macro is used to set the value of a given hash node
 */
#ifndef GEN_NODE_SET
#define GEN_NODE_SET(_node,_key,_val,_table)  (_node)->value = _val
#endif


/**************************************************************************
 *
 * @macro: GEN_NODE_UNSET (node,table)
 *
 * @description:
 *   this macro is used to unset the value of a given hash node
 */
#ifndef GEN_NODE_UNSET
#define GEN_NODE_UNSET(_node,_table)   M_DUMMY_STMNT
#endif


/**************************************************************************
 *
 * @macro: GEN_NODE_DESTROY (node,table)
 *
 * @description:
 *   this macro is used to destroy a given container node
 */
#ifndef GEN_NODE_DESTROY
#define GEN_NODE_DESTROY(_node,_table)                         \
              M_BEGIN_STMNT                                    \
                M_Memory  memory = GEN_TABLE_POOL(_table);     \
                GEN_NODE_UNSET(_node,_table);                  \
                M_FREE(_node);                                 \
              M_END_STMNT
#endif



/**************************************************************************
 *
 * @macro: GEN_NODE_ITERATE (node,table)
 *
 * @description:
 *   this macro is used to iterate over all nodes of a given container
 *   the default implementation assumes that the iteration callback is
 *   named "iterate_func", of type @M_IterateDataFunc, and that the
 *   iteration user data is in "iterate_data"
 */
#ifndef GEN_NODE_ITERATE
#define GEN_NODE_ITERATE(_node,_table)                 \
               iterate_func( GEN_NODE_KEY(_node),      \
                             GEN_NODE_VALUE(_node),    \
                             iterate_data )
#endif


/**************************************************************************
 *
 * @macro: GEN_KEY
 *
 * @description:
 *   this macro holds the type of the "key" of hash nodes. By default,
 *   a simple pointer
 */
#ifndef GEN_KEY
#define GEN_KEY    M_Pointer
#endif

/**************************************************************************
 *
 * @macro: GEN_KEY_COMPARE (key,table)
 *
 * @description:
 *   this macro returns a @M_CompareDataFunc function pointer used to compare
 *   hash node keys to the lookup key
 */
#ifndef GEN_KEY_COMPARE
#define GEN_KEY_COMPARE(key,table)   NULL
#endif


/**************************************************************************
 *
 * @macro: GEN_KEY_DATA (key,table)
 *
 * @description:
 *   this macro returns a @M_Pointer value used by the key comparator function
 *   returned by @GEN_KEY_COMPARE
 */
#ifndef GEN_KEY_DATA
#define GEN_KEY_DATA(key,table)  NULL
#endif



/**************************************************************************
 *
 * @macro: GEN_VALUE
 *
 * @description:
 *   this macro holds the type of the "value" of hash nodes.
 *   by default, a simple pointer
 */
#ifndef GEN_VALUE
#define GEN_VALUE   M_Pointer
#endif



/**************************************************************************
 *
 * @macro: GEN_DEF (result)
 *
 * @description:
 *   this macro is used to define a template instance function. It
 *   defaults to "static"
 */
#ifndef GEN_DEF
#define GEN_DEF(x)    static x
#endif


