/****************************************************************************
 *
 *  template file for external hashing functions
 *
 *  before including this file, you _absolutely_ need to define the
 *  following macros:
 *
 *   GEN_TABLE  :: the type of the hash table
 *   GEN_NODE   :: the type of the hash nodes
 *   GEN_KEY    :: the type of hash node keys
 *   GEN_VALUE  :: the type of hash node values
 *
 *  besides, you'll most probably want to define these as well:
 *
 *   GEN_TABLE_MEMORY
 *   GEN_TABLE_POOL
 *   GEN_HASH_BUCKETS
 *   GEN_HASH_SIZE
 *   GEN_HASH_COUNT
 *
 *   GEN_NODE_KEY
 *   GEN_NODE_VALUE
 *   GEN_NODE_LINK
 *
 *   GEN_KEY_HASH
 *   GEN_KEY_COMPARE
 *   GEN_KEY_DATA
 *
 *   GEN_NODE_NEW
 *   GEN_NODE_UNSET
 *   GEN_NODE_SET
 *   GEN_NODE_DESTROY
 *
 *
 *  read the comments below for more details..
 */

#include <mlib/generic/contain.i>

/**************************************************************************
 *
 * @macro: GEN_HASH_BUCKETS (table)
 *
 * @description:
 *   this macro holds the buckets table of the hash table. must be a lvalue
 */
#ifndef GEN_HASH_BUCKETS
#define GEN_HASH_BUCKETS(table)   (table)->buckets
#endif

/**************************************************************************
 *
 * @macro: GEN_HASH_SIZE (table)
 *
 * @description:
 *   this macro holds the number of buckets in the hash table. must be a lvalue
 */
#ifndef GEN_HASH_SIZE
#define GEN_HASH_SIZE(table)      (table)->size
#endif

/**************************************************************************
 *
 * @macro: GEN_HASH_COUNT
 *
 * @description:
 *   this macro holds the number of nodes in the hash table. must be a lvalue
 */
#ifndef GEN_HASH_COUNT
#define GEN_HASH_COUNT(table)     (table)->nodes
#endif



/**************************************************************************
 *
 * @macro: GEN_NODE_LINK (node)
 *
 * @description:
 *   this macro is used to retrieve the "link" pointer from a given hash
 *   node. must be a lvalue
 */
#ifndef GEN_NODE_LINK
#define GEN_NODE_LINK(node)   ((node)->link)
#endif

/**************************************************************************
 *
 * @macro: GEN_NODE_LINK_ADDR (node)
 *
 * @description:
 *   this macro is used to retrieve the address of the "link" pointer
 *   from a given hash node. must be a lvalue
 *
 *   the default value should work in most cases.
 */
#ifndef GEN_NODE_LINK_ADDR
#define GEN_NODE_LINK_ADDR(node)  &GEN_NODE_LINK(node)
#endif

/**************************************************************************
 *
 * @macro: GEN_NODE_HASH (node)
 *
 * @description:
 *   this macro is used to retrieve the hash value of a given hash node
 */
#ifndef GEN_NODE_HASH
#define GEN_NODE_HASH(node)   ((node)->hash)
#endif



/**************************************************************************
 *
 * @macro: GEN_KEY_HASH (key,table)
 *
 * @description:
 *   this macro computes the hash value corresponding to a given key
 */
#ifndef GEN_KEY_HASH
#define GEN_KEY_HASH(key,table)   M_POINTER_TO_DIST(key)
#endif


/**************************************************************************
 *
 * @macro: GEN_HASH_COMPUTE_SIZE (count)
 *
 * @description:
 *   this macro is used to compute the "optimal" number of buckets for
 *   a given hash node count. by default, this calls "m_primes_closest"
 *   but you can re-define it when you need custom resizes..
 */
#ifndef GEN_HASH_COMPUTE_SIZE
#define GEN_HASH_COMPUTE_SIZE(count)  m_primes_find_closest(count)
#endif


/**************************************************************************
 *
 * @macro: GEN_HASH_TEST_RESIZE (table)
 *
 * @description:
 *   this macro is used to test wether a hash table needs resizing
 */
#ifndef GEN_HASH_TEST_RESIZE
#define GEN_HASH_TEST_RESIZE(table)                                 \
          ( GEN_HASH_COUNT(table)   > M_PRIMES_MIN               && \
            ( GEN_HASH_COUNT(table)*3 < GEN_HASH_SIZE (table) ||    \
              GEN_HASH_SIZE (table)*3 < GEN_HASH_COUNT(table) )  )
#endif


/**************************************************************************
 *
 * @macro: GEN_HASH_RESIZE
 *
 * @description:
 *   when defined, this macro is the name of a function that will implement
 *   the hash table resize function
 */
#ifndef GEN_HASH_RESIZE
#define GEN_HASH_RESIZE  GEN_TABLE_FUNC(_resize)
#endif

#ifdef  GEN_HASH_RESIZE_DEF

GEN_DEF(void)
GEN_HASH_RESIZE( GEN_TABLE   table )
{
  M_UFast  size, new_size;

  GEN_TABLE_ASSERT(table);

  size     = GEN_HASH_SIZE(table);
  new_size = GEN_HASH_COMPUTE_SIZE( GEN_HASH_COUNT(table) );
  if (new_size != size )
  {
    GEN_NODE*  new_buckets ;
    M_UFast    i;

    /* we don't want to raise an exception in case of memory shortage */
    /*                                                                */
    GEN_TABLE_QALLOC( new_buckets,
                          new_size * sizeof(new_buckets[0]),
                          table);
    if ( new_buckets == NULL )
      return;

    M_ZERO_ARRAY( new_buckets, new_size );

    for ( i = 0; i < size; i++ )
    {
      GEN_NODE   node, next, *pnode;

      node = GEN_HASH_BUCKETS(table)[i];
      while (node)
      {
        next  = GEN_NODE_LINK(node);
        pnode = new_buckets + ( GEN_NODE_HASH(node) % new_size );

        GEN_NODE_LINK(node) = pnode[0];
        pnode[0]            = node;

        node  = next;
      }
    }

    GEN_TABLE_FREE( GEN_HASH_BUCKETS(table), table );

    GEN_HASH_BUCKETS(table) = new_buckets;
    GEN_HASH_SIZE(table)    = new_size;
  }
}

#undef GEN_HASH_RESIZE_DEF
#endif /* GEN_HASH_RESIZE_DEF */



/**************************************************************************
 *
 * @macro: GEN_HASH_CLEAR
 *
 * @description:
 *   when defined, this macro is the name of a function that will implement
 *   the clear all nodes of a hash table
 */
#ifdef  GEN_HASH_CLEAR_DEF

#ifndef GEN_HASH_CLEAR
#define GEN_HASH_CLEAR  GEN_TABLE_FUNC(_clear)
#endif

GEN_DEF(void)
GEN_HASH_CLEAR( GEN_TABLE  table )
{
  M_UFast   i;
  M_UFast   size;
  M_Memory  memory;

  GEN_TABLE_ASSERT(table);

  size   = GEN_HASH_SIZE(table);
  memory = GEN_TABLE_MEMORY(table);

  for ( i = 0; i < size; i++ )
  {
    GEN_NODE*  pnode = GEN_HASH_BUCKETS(table) + i;
    GEN_NODE   next, node = *pnode;

    *pnode = 0;
    while ( node )
    {
      next = GEN_NODE_LINK(node);
      GEN_NODE_LINK(node) = 0;

      GEN_NODE_DESTROY(node,table);

      node = next;
    }
  }
  M_FREE( GEN_HASH_BUCKETS(table) );
  GEN_HASH_SIZE(table)  = 0;
  GEN_HASH_COUNT(table) = 0;
}

#undef GEN_HASH_CLEAR_DEF
#undef GEN_HASH_CLEAR
#endif /* GEN_HASH_CLEAR_DEF */



/**************************************************************************
 *
 * @macro: GEN_HASH_FIND_NODEPTR
 *
 * @description:
 *   when defined, this macro is the name of a function that will implement
 *   the hash node pointer lookup routine
 *
 * @note:
 *   other generic routines rely on this one, you should define
 *   GEN_HASH_FIND_NODEPTR_FUNC with the name of this function to be
 *   able to compile them too..
 */

#ifndef GEN_HASH_FIND_NODEPTR
#define GEN_HASH_FIND_NODEPTR  GEN_TABLE_FUNC(_find_nodeptr)
#endif

#ifdef  GEN_HASH_FIND_NODEPTR_DEF

GEN_DEF( GEN_NODE* )
GEN_HASH_FIND_NODEPTR( GEN_TABLE  table,
                       GEN_KEY    key )
{
  M_ULong             hash;
  GEN_NODE       *pnode, node;
  M_CompareDataFunc   key_compare;

  hash  = GEN_KEY_HASH(key,table);
  pnode = GEN_HASH_BUCKETS(table) + ( hash % GEN_HASH_SIZE(table) );

  key_compare = GEN_KEY_COMPARE(key,table);

  if ( key_compare )
  {
    M_Pointer  key_data = GEN_KEY_DATA(key,table);

    for (;;)
    {
      node = pnode[0];
      if ( !node )
        break;

      if ( GEN_NODE_HASH(node) == hash &&
           key_compare( key, GEN_NODE_KEY(node), key_data ) == 0 )
        break;

      pnode = GEN_NODE_LINK_ADDR(node);
    }
  }
  else
  {
    for ( ;; )
    {
      node = pnode[0];
      if ( !node )
        break;

      if ( GEN_NODE_KEY(node) == key )
        break;

      pnode = GEN_NODE_LINK_ADDR(node);
    }
  }
  return pnode;
}

#undef GEN_HASH_FIND_NODEPTR_DEF
#endif /* !GEN_HASH_FIND_NODEPTR_DEF */


/**************************************************************************
 *
 * @macro: GEN_HASH_SET
 *
 * @description:
 *   when defined, this macro is the name of a function that will implement
 *   the hash node pointer lookup routine
 */
#ifdef  GEN_HASH_SET_DEF

#ifndef GEN_HASH_SET
#define GEN_HASH_SET  GEN_TABLE_FUNC(_set)
#endif

GEN_DEF( void )
GEN_HASH_SET( GEN_TABLE   table,
              GEN_KEY     key,
              GEN_VALUE   value )
{
  GEN_NODE*  pnode;
  GEN_NODE   node;

  GEN_TABLE_ASSERT(table);

  pnode = GEN_HASH_FIND_NODEPTR( table, key );
  node  = pnode[0];
  if ( node )
  {
    /* replace current value */
    GEN_NODE_UNSET( node, table );
    GEN_NODE_SET( node, key, value, table );
  }
  else
  {
    GEN_NODE_NEW(node,key,value,table);

    pnode[0] = node;

    GEN_HASH_COUNT(table) ++;
  }

  if ( GEN_HASH_TEST_RESIZE(table) )
    GEN_HASH_RESIZE( table );
}

#undef GEN_HASH_SET_DEF
#undef GEN_HASH_SET
#endif /* GEN_HASH_SET_DEF */


/**************************************************************************
 *
 * @macro: GEN_HASH_TOUCH
 *
 * @description:
 *   when defined, this macro is the name of a function that will implement
 *   the hash node pointer touch routine
 */
#ifdef  GEN_HASH_TOUCH_DEF

#ifndef GEN_HASH_TOUCH
#define GEN_HASH_TOUCH  GEN_TABLE_FUNC(_touch)
#endif

GEN_DEF( M_Bool )
GEN_HASH_TOUCH( GEN_TABLE   table,
                GEN_KEY     key,
                GEN_VALUE  *pvalue )
{
  GEN_NODE*  pnode;
  GEN_NODE   node;

  GEN_TABLE_ASSERT(table);
  m_assert( pvalue );

  pnode = GEN_HASH_FIND_NODEPTR( table, key );
  node  = pnode[0];
  if ( node )
  {
    *pvalue = GEN_NODE_VALUE(node);

    return 0;
  }

  GEN_NODE_NEW(node,key,*pvalue,table);
  pnode[0] = node;

  *pvalue  = GEN_NODE_VALUE(node);

  GEN_HASH_COUNT(table)++;

  if ( GEN_HASH_TEST_RESIZE(table) )
    GEN_HASH_RESIZE( table );

  return 1;
}

#undef GEN_HASH_TOUCH_DEF
#undef GEN_HASH_TOUCH
#endif /* GEN_HASH_TOUCH_DEF */


/**************************************************************************
 *
 * @macro: GEN_HASH_GET
 *
 * @description:
 *   when defined, this macro is the name of a function that will implement
 *   a function used to retrieve a given key's value, if any
 */
#ifdef  GEN_HASH_GET_DEF

#ifndef GEN_HASH_GET
#define GEN_HASH_GET  GEN_TABLE_FUNC(_get)
#endif

GEN_DEF( GEN_VALUE )
GEN_HASH_GET( GEN_TABLE  table,
              GEN_KEY    key )
{
  GEN_NODE  *pnode, node;

  if ( table )
  {
    GEN_TABLE_ASSERT(table);

    pnode = GEN_HASH_FIND_NODEPTR( table, key );
    node  = pnode[0];
    if ( node )
      return GEN_NODE_VALUE(node);
  }
  return 0;
}

#undef GEN_HASH_GET_DEF
#undef GEN_HASH_GET
#endif /* GEN_HASH_GET_DEF */



/**************************************************************************
 *
 * @macro: GEN_HASH_LOOKUP
 *
 * @description:
 *   when defined, this macro is the name of a function that will implement
 *   a function used to test wether a given key has a value, and eventually
 *   return the latter
 */
#ifdef  GEN_HASH_LOOKUP_DEF

#ifndef GEN_HASH_LOOKUP
#define GEN_HASH_LOOKUP  GEN_TABLE_FUNC(_lookup)
#endif

GEN_DEF( M_Bool )
GEN_HASH_LOOKUP( GEN_TABLE   table,
                 GEN_KEY     key,
                 GEN_VALUE  *avalue )
{
  M_Bool     result = 0;
  GEN_NODE  *pnode, node;

  if ( table )
  {
    GEN_TABLE_ASSERT(table);

    pnode = GEN_HASH_FIND_NODEPTR( table, key );
    node  = pnode[0];
    if ( node )
    {
      result = 1;
      if ( avalue )
        *avalue = GEN_NODE_VALUE( node );
    }
  }
  return result;
}

#undef GEN_HASH_LOOKUP_DEF
#undef GEN_HASH_LOOKUP
#endif /* GEN_HASH_LOOKUP_DEF */



/**************************************************************************
 *
 * @macro: GEN_HASH_REMOVE
 *
 * @description:
 *   when defined, this macro is the name of a function that will implement
 *   a function used to remove a given key's value from a hash table
 */
#ifdef  GEN_HASH_REMOVE_DEF

#ifndef GEN_HASH_REMOVE
#define GEN_HASH_REMOVE  GEN_TABLE_FUNC(_remove)
#endif

GEN_DEF( void )
GEN_HASH_REMOVE( GEN_TABLE  table,
                 GEN_KEY    key )
{
  if (table)
  {
    GEN_NODE  *pnode, node;

    GEN_TABLE_ASSERT(table);

    pnode = GEN_HASH_FIND_NODEPTR( table, key );
    node  = *pnode;
    if (node)
    {
      /* we need to remove this node from the table */
      pnode[0]            = GEN_NODE_LINK(node);
      GEN_NODE_LINK(node) = 0;

      GEN_NODE_DESTROY( node, table );

      /* we don't resize the table here, given that we don't want */
      /* to raise any exception.. Also, most of the time, several */
      /* elements are removed from the table before a new one is  */
      /* inserted..                                               */
      GEN_HASH_COUNT(table)--;
    }

    if ( GEN_HASH_TEST_RESIZE(table) )
      GEN_HASH_RESIZE( table );
  }
}

#undef GEN_HASH_REMOVE_DEF
#undef GEN_HASH_REMOVE
#endif /* GEN_HASH_REMOVE_DEF */


/**************************************************************************
 *
 * @macro: GEN_HASH_REMOVE_SEL
 *
 * @description:
 *   when defined, this macro is the name of a function that will implement
 *   a function used to remove all keys that match a given filter
 */
#ifdef  GEN_HASH_REMOVE_SEL_DEF

#ifndef GEN_HASH_REMOVE_SEL
#define GEN_HASH_REMOVE_SEL  GEN_TABLE_FUNC(_remove_selection)
#endif

GEN_DEF( void )
GEN_HASH_REMOVE_SEL( GEN_TABLE  table
                     GEN_PARAMS_ITER_ )
{
  if ( table )
  {
    M_UFast        i, size;
    GEN_NODE*  bucket;

    GEN_TABLE_ASSERT(table);

    size   = GEN_HASH_SIZE(table);
    bucket = GEN_HASH_BUCKETS(table);

    for ( i = 0; i < size; i++, bucket++ )
    {
      GEN_NODE  *pnode, node;

      pnode = bucket;
      for (;;)
      {
        node = *pnode;
        if ( !node )
          break;

        if ( GEN_NODE_ITERATE(node,table) )
        {
          *pnode = GEN_NODE_LINK(node);
          GEN_NODE_LINK(node) = 0;

          GEN_NODE_DESTROY( node, table );
          GEN_HASH_COUNT(table)--;
        }
        else
          pnode = GEN_NODE_LINK_ADDR(node);
      }
    }

    if ( GEN_HASH_TEST_RESIZE(table) )
      GEN_HASH_RESIZE( table );
  }
}

#undef GEN_HASH_REMOVE_SEL_DEF
#undef GEN_HASH_REMOVE_SEL
#endif /* GEN_HASH_REMOVE_SEL_DEF */


/**************************************************************************
 *
 * @macro: GEN_HASH_ITERATE
 *
 * @description:
 *   when defined, this macro is the name of a function that will implement
 *   a function used to iterate over all nodes of a hash table
 */
#ifdef  GEN_HASH_ITERATE_DEF

#ifndef GEN_HASH_ITERATE
#define GEN_HASH_ITERATE  GEN_TABLE_FUNC(_iterate)
#endif

GEN_DEF( void )
GEN_HASH_ITERATE( GEN_TABLE  table
                  GEN_PARAMS_ITER_ )
{
  if ( table && iterate_func )
  {
    M_UFast        size;
    GEN_NODE*  bucket;

    GEN_TABLE_ASSERT(table);

    size   = GEN_HASH_SIZE(table);
    bucket = GEN_HASH_BUCKETS(table);

    for ( ; size > 0; size--, bucket++ )
    {
      GEN_NODE  node, next;

      node = bucket[0];
      while (node)
      {
        next = GEN_NODE_LINK(node);

        GEN_NODE_ITERATE(node,table);

        node = next;
      }
    }
  }
}

#undef GEN_HASH_ITERATE_DEF
#undef GEN_HASH_ITERATE
#endif  /* GEN_HASH_ITERATE_DEF */



/* now, undefine all macros used here !! */
#undef GEN_HASH_COMPUTE_SIZE
#undef GEN_HASH_COUNT
#undef GEN_HASH_SIZE
#undef GEN_HASH_BUCKETS

#undef GEN_NODE_LINK
#undef GEN_NODE_HASH
#undef GEN_KEY_HASH

#undef GEN_HASH_RESIZE
#undef GEN_HASH_FIND_NODEPTR

#include <mlib/generic/undefs.i>
