/****************************************************************************
 *
 * a generic quick-sort function
 *
 *   this is a non-recursive quicksort that directly processes table cells
 *
 *   please be aware that all quick-sort implementations have a O(N^2LogN)
 *   worst-case behaviour. Use a heap-sort to get O(NLogN) worst-cases
 */

#include <mlib/generic/sort.i>

/***********************************************************************
 *
 * @macro: GEN_SORT_QUICK_THRESHOLD
 *
 * @description:
 *   below this size, a sub-array will be sorted through a normal
 *   insertion sort. It's not unusual to see custom values provide
 *   an important speed boost, depending on the nature of the sorted
 *   data..
 */
#ifndef GEN_SORT_QUICK_THRESHOLD
#define GEN_SORT_QUICK_THRESHOLD  7
#endif


/***********************************************************************
 *
 * @macro: GEN_SORT_QUICK
 *
 * @description:
 *   when defined, this macro is the name of a function used to implement
 *   a quick-sort algorithm
 */
#ifdef GEN_SORT_QUICK

GEN_SORT_DEF( void )
GEN_SORT_QUICK( GEN_SORT_CELL*  cells,
                GEN_SORT_INDEX  count
                GEN_SORT_AUX_   )
{
  GEN_SORT_CELL  *base, *limit, *stack[40], **top;

  limit = cells + count;
  base  = cells;
  top   = stack;

  for (;;)
  {
    GEN_SORT_INDEX  len = (GEN_SORT_INDEX)( limit - base );
    GEN_SORT_CELL  *i, *j, *pivot;


    if ( len > GEN_SORT_QUICK_THRESHOLD )
    {
      /* we use base + len/2 as the pivot */
      pivot = GEN_SORT_PTR( base, len/2 );
      GEN_SORT_SWAP( base, pivot );

      i = GEN_SORT_PTR(base,1);
      j = GEN_SORT_PTR(limit,-1);

      /* now ensure that *i <= *base <= *j */
      if ( GEN_SORT_COMPARE( j, i ) < 0 )
        GEN_SORT_SWAP( i, j );

      if ( GEN_SORT_COMPARE( base, i ) < 0 )
        GEN_SORT_SWAP( base, i );

      if ( GEN_SORT_COMPARE( j, base ) < 0 )
        GEN_SORT_SWAP( base, j );

      for (;;)
      {
        do GEN_SORT_INCR(i); while ( GEN_SORT_COMPARE( i, base ) < 0 );
        do GEN_SORT_DECR(j); while ( GEN_SORT_COMPARE( base, j ) < 0 );

        if ( i > j )
          break;

        GEN_SORT_SWAP( i, j );
      }

      GEN_SORT_SWAP( base, j );

      /* now, push the largest sub-array */
      if ( j - base > limit - i )
      {
        top[0] = base;
        top[1] = j;
        base   = i;
      }
      else
      {
        top[0] = i;
        top[1] = limit;
        limit  = j;
      }
      top += 2;
    }
    else
    {
      /* the sub-array is small, perform insertion sort */
      j = base;
      i = GEN_SORT_PTR(j,1);

      for ( ; i < limit; j = i, GEN_SORT_INCR(i) )
      {
        for ( ; GEN_SORT_COMPARE( GEN_SORT_PTR(j,1), j ) < 0; GEN_SORT_DECR(j) )
        {
          GEN_SORT_SWAP( GEN_SORT_PTR(j,1), j );
          if ( j == base )
            break;
        }
      }
      if ( top > stack )
      {
        top  -= 2;
        base  = top[0];
        limit = top[1];
      }
      else
        break;
    }
  }
}
#endif /* GEN_SORT_QUICK */

#include <mlib/generic/undefs.i>
