#ifndef __MLIB_INTERNAL_CALC_H__
#define __MLIB_INTERNAL_CALC_H__

#include <mlib/calc.h>

M_BEGIN_HEADER

 /*************************************************************************/
 /*************************************************************************/
 /*****                                                               *****/
 /*****      manage pairs of 32-bit numbers as 64-bit integers        *****/
 /*****                                                               *****/
 /*****   IMPORTANT: the following macros all assume that 'hi' and    *****/
 /*****              'lo' are or type M_UInt32 !!                     *****/
 /*****                                                               *****/
 /*************************************************************************/
 /*************************************************************************/

/* computes '_a*_b' and puts the 64-bit result in (_hi,_lo) */
/* both '_hi' and '_lo' must be M_UInt32 lvalues..          */
/*                                                          */
#define  M_UMUL64(_hi,_lo,_a,_b)                            \
           M_BEGIN_STMNT                                    \
             M_UInt32  hi, lo, h1, h2, l1, l2, m1, m2;      \
                                                            \
             h1 = M_UINT32(_a) >> 16;                       \
             h2 = M_UINT32(_b) >> 16;                       \
             l1 = M_UINT32(_a) & M_UINT32_C(0xFFFF);        \
             l2 = M_UINT32(_b) & M_UINT32_C(0xFFFF);        \
                                                            \
             hi = h1*l1;                                    \
             m1 = h1*l2;                                    \
             m2 = h2*l1;                                    \
             m1 = m1 + m2;                                  \
             /* check carry overflow of "m1+m2" */          \
             hi = hi + (M_UINT32(m1 < m2) << 16);           \
                                                            \
             hi += (m1 >> 16);                              \
             m1  = (m1 << 16);                              \
                                                            \
             lo = l1*l2;                                    \
             lo = lo + m1;                                  \
             hi = hi + M_UINT32(lo < m1);                   \
                                                            \
             (_hi) = hi;                                    \
             (_lo) = lo;                                    \
           M_END_STMNT


/* divide a 64-bit number by a 32-bit one and return a 32-bit             */
/* value.. everything is unsigned. result is bounded to 0 .. M_UINT32_MAX */
/*                                                                        */
#define  M_UDIV64(_hi,_lo,_divisor,_quotient,_remainder)    \
           M_BEGIN_STMNT                                    \
             M_UInt32   r, q, lo, d;                        \
             M_Int      i;                                  \
                                                            \
             q  = 0;                                        \
             r  = (_hi);                                    \
             lo = (_lo);                                    \
             d  = (_divisor);                               \
                                                            \
             if ( r >= d )                                  \
             {                                              \
               /* result doesn't fit in 32-bits !! */       \
               (_quotient)  = M_UINT32_MAX;                 \
               (_remainder) = M_UINT32_MAX;                 \
             }                                              \
             else                                           \
             {                                              \
               i = 32;                                      \
               do                                           \
               {                                            \
                 r <<= 1;                                   \
                 q <<= 1;                                   \
                 r  |= lo >> 31;                            \
                                                            \
                 if ( r >= d )                              \
                 {                                          \
                   r -= d;                                  \
                   q |= 1;                                  \
                 }                                          \
                 lo <<= 1;                                  \
               }                                            \
               while ( --i );                               \
                                                            \
               (_quotient)  = q;                            \
               (_remainder) = r;                            \
             }                                              \
           M_END_STMNT


/* shift (hi,lo) to the left, parameters must be unsigned lvalues and */
/* '_shift' must be in 0..32 !!                                       */
#define  M_SHL64(_hi,_lo,_shift)                                      \
           M_BEGIN_STMNT                                              \
             M_UInt32  __lo    = (_lo);                               \
             M_Int     __shift = (M_Int)(_shift);                     \
                                                                      \
             (_hi) = ((_hi) << __shift) | (__lo >> (32-__shift));     \
             (_lo) = __lo << __shift;                                 \
           M_END_STMNT


/* shift (hi,lo) to the right, parameters must be unsigned lvalues and */
/* '_shift' must be in 0..32 !!                                        */
#define  M_SHR64(_hi,_lo,_shift)                                      \
           M_BEGIN_STMNT                                              \
             M_UInt32  __hi    = (_hi);                               \
             M_Int     __shift = (M_Int)(_shift);                     \
                                                                      \
             (_lo) = ((_lo) >> __shift) | (__hi << (32-__shift));     \
             (_hi) = __hi >> __shift;                                 \
           M_END_STMNT


/* arithmetic shift (hi,lo) to the right, parameters must be unsigned  */
/* lvalues and '_shift' must be in 0..32 !!                            */
#define  M_SAR64(_hi,_lo,_shift)                                          \
           M_BEGIN_STMNT                                                  \
             M_Int32   __hi    = M_INT32(_hi);                            \
             M_Int     __shift = (M_Int)(_shift);                         \
                                                                          \
             (_lo) = ((_lo) >> __shift) | M_UINT32(__hi << (32-__shift)); \
             (_hi) = M_UINT32(__hi >> __shift);                           \
           M_END_STMNT


/* negate a (hi,lo) pair */
#define  M_NEG64(_hi,_lo)                              \
           M_BEGIN_STMNT                               \
             M_UInt32  l = M_UINT32(-M_INT32(_lo));    \
                                                       \
             (_hi) = ~(_hi) + (l == 0);                \
             (_lo) = l;                                \
           M_END_STMNT


/* add  (hi1,lo1) to (hi,lo) */
#define  M_ADD64(_hi,_lo,_hi1,_lo1)                              \
           M_BEGIN_STMNT                                         \
             M_UInt32   __lo, __lo1;                             \
                                                                 \
             __lo1 = (_lo1);                                     \
             __lo  = (_lo) + __lo1;                              \
             (_hi) = (_hi) + (_hi1) + (__lo < __lo1);            \
             (_lo) = __lo;                                       \
           M_END_STMNT

/* substract (hi1,lo1) from (hi,lo) */
#define  M_SUB64(_hi,_lo,_hi1,_lo1)                              \
           M_BEGIN_STMNT                                         \
             M_UInt32   __lo, __lo1;                             \
                                                                 \
             __lo1 = (_lo1);                                     \
             __lo  = (_lo) - __lo1;                              \
             (_hi) = (_hi) - (_hi1) - (__lo > __lo1);            \
             (_lo) = __lo;                                       \
           M_END_STMNT


/* compare two 64-bit values */
#define  M_COMPARE64(hi1,lo1,hi2,lo2)   ( (hi1) == (hi2)         \
                                        ? M_COMPARE( lo1, lo2 )  \
                                        : M_COMPARE( hi1, hi2 ) )

 /*************************************************************************/
 /*************************************************************************/
 /*****                                                               *****/
 /*****                   manage M_Card64 numbers                     *****/
 /*****                                                               *****/
 /*************************************************************************/
 /*************************************************************************/

#define  M_CARD64_COMPARE(a,b)   ( (a)->hi == (b)->hi            \
                                 ? M_COMPARE( (a)->lo, (b)->lo ) \
                                 : M_COMPARE( (a)->hi, (b)->hi ) )

#define  M_CARD64_INIT(z,_hi,_lo)       \
             M_BEGIN_STMNT              \
               (z)->hi = M_UINT32(_hi); \
               (z)->lo = M_UINT32(_lo); \
             M_END_STMNT

/* wors only for if _shift in 0..32 */
#define  M_CARD64_SHL(z,_shift)    M_SHL64( (z)->hi, (z)->lo, _shift )

/* wors only for if _shift in 0..32 */
#define  M_CARD64_SHR(z,_shift)    M_SHR64( (z)->hi, (z)->lo, _shift )

/* wors only for if _shift in 0..32 */
#define  M_CARD64_SAR(z,_shift)    M_SAR64( (z)->hi, (z)->lo, _shift )

#define  M_CARD64_ADD(z,_hi,_lo)   M_ADD64( (z)->hi, (z)->lo, _hi, _lo )

#define  M_CARD64_SUB(z,_hi,_lo)   M_SUB64( (z)->hi, (z)->lo, _hi, _lo )

#define  M_CARD64_MUL(z,_a,_b)     M_UMUL64( (z)->hi, (z)->lo, _a, _b )

#define  M_CARD64_DIV(z,_c,_q,_r)        M_UDIV64( (z)->hi, (z)->lo, _c, _q, _r )

 /*************************************************************************/
 /*************************************************************************/
 /*****                                                               *****/
 /*****                   manage M_Calc64 numbers                     *****/
 /*****                                                               *****/
 /*************************************************************************/
 /*************************************************************************/

/* return sign of a M_Calc64 */
#define  M_CALC64_SIGN(x)   M_SIGN((x)->hi)

/* compare to M_Calc64 */
#define  M_CALC64_COMPARE(a,b)   ( (a)->hi == (a)->lo          \
                                 ? M_COMPARE((a)->lo,(b)->lo)  \
                                 : M_COMPARE((a)->hi,(b)->hi)  )

#define  M_CALC64_INIT(c,_hi,_lo)          \
           M_BEGIN_STMNT                   \
             (c)->hi = (M_Int32)(_hi);     \
             (c)->lo = (M_UInt32)(_lo);    \
           M_END_STMNT


/* wors only for if _shift in 0..32 */
#define  M_CALC64_SHL(c,_shift)   M_CARD64_SHL( M_CARD64(c), _shift )

/* wors only for if _shift in 0..32 */
#define  M_CALC64_SHR(c,_shift)   M_CARD64_SHR( M_CARD64(c), _shift )

/* wors only for if _shift in 0..32 */
#define  M_CALC64_SAR(c,_shift)   M_CARD64_SAR( M_CARD64(c), _shift )

#define  M_CALC64_NEG(c)          M_NEG64( M_CARD64(c)->hi, (c)->lo )

#define  M_CALC64_ADD(c,_hi,_lo)  M_CARD64_ADD( M_CARD64(c), M_UINT32(_hi), _lo )

#define  M_CALC64_SUB(c,_hi,_lo)  M_CARD64_SUB( M_CARD64(c), M_UINT32(_hi), _lo )

 /* */

#define  M_FIX16_SIGN(x)    (((x) > 0)*2-1)

M_END_HEADER

#endif /* __MLIB_INTERNAL_CALC_H__ */
