#ifndef __MLIB_INTERNAL_CORE_H__
#define __MLIB_INTERNAL_CORE_H__

#include <mlib/internal/ilock.h>

M_BEGIN_HEADER

 /***************************************************************************/
 /***************************************************************************/
 /*****                                                                 *****/
 /*****                      MLIB CLEANUP INTERNALS                     *****/
 /*****                                                                 *****/
 /***************************************************************************/
 /***************************************************************************/

 /********************************************************************
  *
  * @type: M_CleanupStack
  *
  * @description:
  *    handle to a cleanup stack
  */
  typedef struct M_CleanupStackRec_*    M_CleanupStack;


 /********************************************************************
  *
  * @type: M_CleanupItem
  *
  * @description:
  *    handle to a single cleanup stack element
  */
  typedef struct M_CleanupItemRec_*    M_CleanupItem;


 /********************************************************************
  *
  * @struct: M_CleanupItemRec
  *
  * @description:
  *    a single cleanup stack element/slot
  *
  * @fields:
  *    data         :: pointer to object/element on the stack
  *    cleanup_func :: the item's cleanup function
  *    cleanup_data :: the item's cleanup data
  */
  typedef struct M_CleanupItemRec_
  {
    M_Pointer      data;
    M_CleanupFunc  cleanup_func;
    M_Pointer      cleanup_data;

  } M_CleanupItemRec;


 /********************************************************************
  *
  * @type: M_CleanupChunk
  *
  * @description:
  *    handle to a cleanup stack chunk. A chunk is a fixed-size
  *    array of cleanup items.
  */
  typedef struct M_CleanupChunkRec_*   M_CleanupChunk;


/* note that 1+3*21 = 64, hence the size of a M_CleanupChunkRec will be */
/* 512 bytes on a 32-bits machine..                                      */
/*                                                                       */
#define  M_CLEANUP_CHUNK_SIZE  21

 /********************************************************************
  *
  * @struct: M_CleanupChunkRec
  *
  * @description:
  *    a clenaup chunk is used to store several contiguous stack
  *    elements in a single memory location
  *
  * @fields:
  *    items :: array of cleanup items
  *    link  :: link to previous chunk in stack or free list
  */
  typedef struct M_CleanupChunkRec_
  {
    M_CleanupItemRec  items[ M_CLEANUP_CHUNK_SIZE ];
    M_CleanupChunk    link;

  } M_CleanupChunkRec;


 /**************************************************************************
  *
  * @struct: M_CleanupStackRec
  *
  * @description:
  *   a structure used to hold a thread-specific cleanup stack
  *
  * @fields:
  *    cursor    :: current item cursor
  *    first     :: first item in current chunk
  *    last      :: last item in current chunk
  *    num_items :: total number of items in stack
  *    chunk     :: current stack chunk
  */
  typedef struct M_CleanupStackRec_
  {
    M_CleanupItem    cursor;
    M_CleanupItem    first;
    M_CleanupItem    last;
    M_UInt           num_items;
    M_CleanupChunk   chunk;

  } M_CleanupStackRec;


 /***************************************************************************/
 /***************************************************************************/
 /*****                                                                 *****/
 /*****                      MLIB EXCEPTION CONTEXTS                    *****/
 /*****                                                                 *****/
 /***************************************************************************/
 /***************************************************************************/


 /**************************************************************************
  *
  * @struct: M_XContextRec
  *
  * @description:
  *   a structure used to hold a thread-specific exception context
  *
  * @fields:
  *   mcore           :: handle to MLib core
  *   thread_id       ;; the thread's ID, as returned by the thread requester
  *
  *   cleanup_stack   :: thread's cleanup stack
  *   xhandler        :: thread's current exception handler. can be NULL
  *
  *   fatal_handler   :: current fatal exception handler
  *   fatal_data      :: data for the fatal exception handler
  */
  typedef struct M_XContextRec_
  {
    M_Core              mcore;
    M_ThreadID          thread_id;

    M_CleanupStackRec   cleanup_stack;
    M_XHandler          xhandler;

    M_FatalHandlerFunc  fatal_handler;
    M_Pointer           fatal_data;

  } M_XContextRec;

  /* */

 /* we need to export this one for future inlinings */
  MLIB_API(void)
  m_xcontext_grow_stack( M_XContext  context );


 /* we need to export this one for future inlinings */
  MLIB_API(void)
  m_xcontext_shrink_stack( M_XContext  context );


  MLIB_API(M_Error)
  m_xcontext_init( M_XContext  context,
                   M_Core      mcore,
                   M_ThreadID  id );


  MLIB_API(void)
  m_xcontext_done( M_XContext  context );

 /***************************************************************************/
 /***************************************************************************/
 /*****                                                                 *****/
 /*****                      MLIB LOCK OBJECT TYPE                      *****/
 /*****                                                                 *****/
 /***************************************************************************/
 /***************************************************************************/

 /****************************************************
  *
  * @section: core_internals
  *
  * @title: Internals
  *
  * @abstract: MLib Internals
  *
  * @description:
  *   this section contains public definitions of internal MLib structures.
  *   these are mostly useful for implementers of new core classes, in order
  *   to support custom or system-specific synchronisation and thread contexts
  *
  * @order:
  *   M_Core_Class
  *   M_Core_ClassRec
  *   M_Core_InitFunc
  *   M_Core_DoneFunc
  *
  *   M_ThreadID_GetFunc
  */

 /*************************************************************************
  *
  * @functype: M_Core_InitFunc
  *
  * @description:
  *    a function used to initialise the implementation specific part
  *    of a given MLib core object. Usually, this is used to initialise
  *    the thread subsystem
  *
  * @input:
  *    mcore         :: handle to MLib core
  *    impl_data     :: user-specified pointer. see @m_core_new
  *    xcontext_size :: size in bytes of a given thread exception context
  *
  * @return:
  *    error code. 0 means success
  *
  * @note:
  *    this function should _never_ throw an exception.
  */
  typedef M_Error      (*M_Core_InitFunc) ( M_Core     core,
                                            M_Pointer  impl_data,
                                            M_Size     xcontext_size );


 /*************************************************************************
  *
  * @functype: M_Core_DoneFunc
  *
  * @description:
  *    a function used to finalise the implementation-specific part
  *    of a given MLib core object. Usually, this is used to finalise
  *    the thread subsystem
  *
  * @input:
  *    mcore :: handle to MLib core
  */
  typedef void         (*M_Core_DoneFunc)  ( M_Core   core );


 /*************************************************************************
  *
  * @functype: M_ThreadID_GetFunc
  *
  * @description:
  *    this function is used to retrieve the current thread's ID, as
  *    well as its specific exception context when the "acontext" parameter is
  *    not NULL
  *
  * @input:
  *    mcore  :: handle to MLib core
  *
  *    create :: when true, the function should try to create a new exception
  *              context for the current thread when it doesn't already have
  *              one. Otherwise, it should simply return NULL in "ainfo"
  *
  *    acontext :: an optional pointer to a @M_XContext handle. when
  *                this parameter is not NULL, then "*acontext" will receive
  *                the address of the thread-specific exception context.
  *
  * @return:
  *    current thread's id. NULL in case of failure
  *
  * @note:
  *   implementation-specific thread subsystems are free to use any
  *   kind of container to map thread ids to xcontexts. This ensures
  *   that the best method is always used on a given platform.
  *
  *   if you need to implement this function, take note that each new
  *   thread info block must be initialised with a call to
  *   @m_thread_info_init, and finalised with @m_thread_info_done
  */
  typedef M_ThreadID  (*M_ThreadID_GetFunc)( M_Core       core,
                                             M_Bool       create,
                                             M_XContext  *acontext );



 /*************************************************************************
  *
  * @struct: M_Core_ClassRec
  *
  * @description:
  *    a structure used to hold the MLib's interface to the platform-specific
  *    thread sub-system, if any..
  *
  * @fields:
  *    core_init      :: initialise the subsystem of a given MLib
  *                      core object
  *
  *    core_done      :: finalise the subsystem of a given MLib core
  *
  *    thread_id_get  :: address of the function used to retrieve the current
  *                      thread's id and its exception context.
  *
  *    mutex_class    :: pointer to system-specific mutex routines
  *
  *    spinlock_class :: pointer to system-specific spinlock routines. can be
  *                      NULL, in which case spinlocks will be implemented as
  *                      normal mutexes.
  *
  *    rwlock_class   :: pointer to system-specific read/write lock routines.
  *                      can be NULL, in which case r/w locks will be
  *                      implemented as normal mutexes.
  *
  * @note:
  *    we do not provide thread management for now, since it's not part
  *    of our goals or requirements..
  *
  *    the 'mcore' field of each one of 'mutex_class', 'spinlock_class' and
  *    'rwlock_class' is ignored and should be set to NULL
  */
  typedef struct M_Core_ClassRec_
  {
    M_Core_InitFunc     core_init;
    M_Core_DoneFunc     core_done;

    M_ThreadID_GetFunc  thread_id_get;

    M_Lock_Class        mutex_class;
    M_Lock_Class        spinlock_class;
    M_Lock_Class        rwlock_class;

  } M_Core_ClassRec;


 /***************************************************************************/
 /***************************************************************************/
 /*****                                                                 *****/
 /*****                      MLIB CORE STATIC DATA                      *****/
 /*****                                                                 *****/
 /***************************************************************************/
 /***************************************************************************/

  typedef struct M_StaticListRec_*  M_StaticList;

  typedef struct M_StaticNodeRec_*  M_StaticNode;

  typedef struct M_StaticListRec_
  {
    M_Core        mcore;
    M_Lock        lock;
    M_StaticNode  nodes;

  } M_StaticListRec;

  typedef struct M_StaticNodeRec_
  {
    M_StaticNode    link;
    M_UInt32        id;
    M_Pointer       statik;
    M_Static_Class  clazz;

  } M_StaticNodeRec;


#if 0
  MLIB_LOCAL( M_Error )
  m_statics_init( M_StaticList  statics,
                  M_Core        mcore );

  MLIB_LOCAL( void )
  m_statics_done( M_StaticList  statics );
#endif


 /***************************************************************************/
 /***************************************************************************/
 /*****                                                                 *****/
 /*****                      MLIB CORE INTERNALS                        *****/
 /*****                                                                 *****/
 /***************************************************************************/
 /***************************************************************************/

 /**************************************************************************
  *
  * @struct: M_ICoreRec
  *
  * @description:
  *    a structure used to model the internal content of an MLib core
  *
  * @fields:
  *    root           :: root core context fields
  *
  *    magic          :: a 32-bit magic number used to validate core handles
  *
  *    core_context   :: for single-thread instances of the MLib, this field
  *                      contains a pointer to the main thread's exception
  *                      context. it is set to NULL in multi-threaded cores
  *
  *                      since it is only set within @m_core_new, and never
  *                      changed later, it can be read quickly any time to
  *                      determine if we're in a multi-threaded environment
  *                      or not..
  *
  *    core_lock       :: a spinlock used to protect the list of free chunks
  *                       as well as the core's reference count
  *
  *    thread_id_get   :: cache for core class' "thread_id_get" field
  *
  *    mutex_class     :: cache for core class' "mutex_class" field
  *
  *    spinlock_class   :: cache for core class' "spinlock_class" field.
  *
  *    rwlock_class    :: cache for core class' "rwlock_class" field
  *
  *    ref_count       :: reference count for this core
  *    core_class      :: handle to original core class
  *
  *    chunks          :: list of free cleanup chunks.
  *
  *    statics         :: list of static objects created in this core
  */
  typedef struct M_ICoreRec_
  {
    M_CoreRec               root;
    M_UInt32                magic;

    M_XContext              core_context;
    M_Lock                  core_lock;

    M_ThreadID_GetFunc      thread_id_get;
    M_Lock_ClassRec         mutex_class;
    M_Lock_ClassRec         spinlock_class;
    M_Lock_ClassRec         rwlock_class;

    M_Int                   ref_count;
    M_Core_Class            core_class;

    M_CleanupChunk          chunks;
    M_StaticListRec         statics;

  } M_ICoreRec, *M_ICore;

  /* */


#define  M_ICORE(x)    ((M_ICore)(x))
#define  M_ICORE_P(x)  ((M_ICore*)(x))


#define  m_core_xcontext(_core)  \
             ( _core ? M_ICORE(_core)->core_context : NULL )

/* returns true if the MLib core is executed in a single thread */
#define  m_core_single_thread(_core)   (m_core_xcontext(_core) != NULL)

#define  m_core_get_xcontext(_core,_context,_name)                    \
            _context = M_ICORE(_core)->core_context;                  \
            if (!_context)                                            \
            {                                                         \
              _context = m_core_find_xcontext( _core, _name );        \
            }                                                         \


#ifdef M_DEBUG_COMPILE_ASSERTS

#  define  m_cleanup_item_check( _slot, _item, _func, _function_name )        \
      M_BEGIN_STMNT                                                           \
        if ( (_slot)->data != (_item) )                                       \
          M_PANIC(( "%s: corrupted cleanup stack (invalid cleanup item).",    \
                    _function_name ));                                        \
                                                                              \
        if ( (_func) && (_slot)->cleanup_func != (_func) )                    \
          M_PANIC(( "%s: corrupted cleanup stack (invalid cleanup function).",\
                    _function_name ));                                        \
      M_END_STMNT


#  define  m_core_check( _core, _name )                                 \
       M_BEGIN_STMNT                                                    \
         if ( !(_core) || M_ICORE(_core)->magic != M_MAGIC_CORE )       \
           M_PANIC(( "%s: invalid MLib core handle", _name ));          \
       M_END_STMNT

#else /* !M_DEBUG_COMPILE_ASSERTS */

#  define  m_cleanup_item_check( _slot, _item, _func, _function_name )  \
       M_BEGIN_STMNT                                                    \
         M_UNUSED(_item);                                               \
         M_UNUSED(_slot);                                               \
       M_END_STMNT

#  define  m_core_check( _core, _name )   \
       M_BEGIN_STMNT                      \
         M_UNUSED(_core);                 \
       M_END_STMNT

#endif /* !M_DEBUG_COMPILE_ASSERTS */


/* A macro used to inline the function "m_core_push" */
#define  m_core_push_inline( _core, _item, _func, _data, _name )   \
          M_BEGIN_STMNT                                            \
            M_XContext      xcontext;                              \
            M_CleanupStack  stack;                                 \
            M_CleanupItem   slot;                                  \
                                                                   \
            m_core_check( _core, _name );                          \
            m_core_get_xcontext( _core, xcontext, _name );         \
                                                                   \
            stack = &xcontext->cleanup_stack;                      \
            slot  = stack->cursor++;                               \
                                                                   \
            slot->data         = (M_Pointer)(_item);               \
            slot->cleanup_func = (_func);                          \
            slot->cleanup_data = (_data);                          \
                                                                   \
            stack->num_items ++;                                   \
                                                                   \
            if ( slot == stack->last )                             \
              m_xcontext_grow_stack( xcontext );                   \
                                                                   \
          M_END_STMNT


/* a macro used to inline the function "m_core_pop" */
#define  m_core_pop_inline( _core, _item, _func, _action, _name )    \
          M_BEGIN_STMNT                                              \
            M_XContext      xcontext;                                \
            M_CleanupStack  stack;                                   \
            M_CleanupItem   slot;                                    \
                                                                     \
            m_core_check( _core, _name );                            \
            m_core_get_xcontext( _core, xcontext, _name );           \
                                                                     \
            stack = &xcontext->cleanup_stack;                        \
            if ( stack->cursor == stack->first )                     \
              m_xcontext_shrink_stack( xcontext );                   \
            else                                                     \
              stack->cursor--;                                       \
                                                                     \
            stack->num_items--;                                      \
                                                                     \
            slot = stack->cursor;                                    \
            m_cleanup_item_check( slot, _item, _func, _name );       \
                                                                     \
            if ((_action) == M_POP_CLEAN)                            \
              slot->cleanup_func( slot->data, slot->cleanup_data );  \
          M_END_STMNT


/* */

#define  M_MAGIC_CORE  0xc0defefe

 /* core spinlock & r/w lock synchronisation macros */

#ifdef M_CONFIG_THREAD_SAFE

#  define  M_ICORE_LOCK_ACTION_INNER( _icore, _action )   \
             (_icore)->spinlock_class. _action ( (_icore)->core_lock )

#  define  M_ICORE_LOCK_ACTION( _icore, _action )            \
             M_BEGIN_STMNT                                     \
               if ((_icore)->core_lock)                      \
                 M_ICORE_LOCK_ACTION_INNER(_icore,_action);  \
             M_END_STMNT

#else /* !THREAD_SAFE */

#  define  M_ICORE_LOCK_ACTION_INNER( _icore, _action )   \
                 M_BEGIN_STMNT                              \
                   M_UNUSED(_icore);                        \
                 M_END_STMNT

#  define  M_ICORE_LOCK_ACTION( _icore, _action )         \
                 M_BEGIN_STMNT                              \
                   M_UNUSED(_icore);                        \
                 M_END_STMNT

#endif /* !THREAD_SAFE */



#define  M_ICORE_LOCK_INNER(_icore)    \
             M_ICORE_LOCK_ACTION_INNER(_icore,enter)

#define  M_ICORE_UNLOCK_INNER(_icore)  \
             M_ICORE_LOCK_ACTION_INNER(_icore,exit)

#define  M_ICORE_LOCK(_icore)          \
             M_ICORE_LOCK_ACTION(_icore,enter)

#define  M_ICORE_UNLOCK(_icore)        \
             M_ICORE_LOCK_ACTION(_icore,exit)

#define  M_CORE_LOCK(_core)        M_ICORE_LOCK  ( M_ICORE(_core) )
#define  M_CORE_UNLOCK(_core)      M_ICORE_UNLOCK( M_ICORE(_core) )


#define  M_DEFINE_MLIB_EXCEPTION( _name, _code )               \
     static const                                              \
     M_ExceptionRec  _name ## _rec =                           \
     {                                                         \
       &mlib_exception_family,                                 \
       _code                                                   \
     };                                                        \
                                                               \
     MLIB_VARDEF(M_Exception)  _name = & _name ## _rec;

#undef  M_LOG_COMPONENT
#define M_LOG_COMPONENT  "Core"

M_END_HEADER

#endif /* __MLIB_INTERNAL_CORE_H__ */

