#ifndef __MLIB_INTERNAL_LOCK_H__
#define __MLIB_INTERNAL_LOCK_H__

#include <mlib/lock.h>

M_BEGIN_HEADER

#define  M_MAGIC_LOCK    0x6c6f636b  /* "lock" */

 /*************************************************************************
  *
  * @type: M_Lock_Class
  *
  * @description:
  *    constant handle to a @M_Lock_ClassRec structure used to model the
  *    class of a given lock object type
  */
  typedef const struct M_Lock_ClassRec_*   M_Lock_Class;


 /*************************************************************************
  *
  * @struct: M_LockRec
  *
  * @description:
  *    the root fields of a @M_Lock object
  */
  typedef struct M_LockRec_
  {
    M_Lock_Class  clazz;

  } M_LockRec;


 /*************************************************************************
  *
  * @functype: M_Lock_InitFunc
  *
  * @description:
  *    a function used to initialize a lock
  *
  * @input:
  *    lock  :: handle to target lock
  *    type  :: type of lock, see @M_LockType
  *    scope :: scope of lock, ignored for now
  *    name  :: name of lock, ignored for now
  *
  * @return:
  *    error code. 0 means success
  */
  typedef  M_Error  (*M_Lock_InitFunc)( M_Lock     lock,
                                        M_UInt     type,
                                        M_UInt     scope,
                                        M_CString  name );

 /*************************************************************************
  *
  * @functype: M_Lock_DoneFunc
  *
  * @description:
  *    finalize a given lock
  *
  * @input:
  *    lock  :: handle to target lock
  */
  typedef void      (*M_Lock_DoneFunc)( M_Lock  lock );


 /*************************************************************************
  *
  * @functype: M_Lock_EnterFunc
  *
  * @description:
  *    a function used to enter a lock through @m_lock_enter
  *
  * @input:
  *    lock  :: handle to target lock
  *
  * @return:
  *    error code. 0 means success
  */
  typedef  M_Error  (*M_Lock_EnterFunc)( M_Lock  lock );


 /*************************************************************************
  *
  * @functype: M_Lock_ExitFunc
  *
  * @description:
  *    a function used to exit a lock through @m_lock_exit
  *
  * @input:
  *    lock  :: handle to target lock
  *
  * @return:
  *    error code. 0 means success
  */
  typedef  M_Error  (*M_Lock_ExitFunc) ( M_Lock  lock );


 /*************************************************************************
  *
  * @functype: M_Lock_EnterReadFunc
  *
  * @description:
  *    a function used to enter a lock through @m_lock_enter_read
  *
  * @input:
  *    lock  :: handle to target lock
  *
  * @return:
  *    error code. 0 means success
  */
  typedef  M_Error  (*M_Lock_EnterReadFunc)( M_Lock   lock );


 /*************************************************************************
  *
  * @functype: M_Lock_EnterUpdateFunc
  *
  * @description:
  *    a function used to enter a lock through @m_lock_enter_update
  *
  * @input:
  *    lock  :: handle to target lock
  *
  * @return:
  *    error code. 0 means success
  */
  typedef  M_Error  (*M_Lock_EnterUpdateFunc)( M_Lock  lock );


 /*************************************************************************
  *
  * @functype: M_Lock_WriteUpdateFunc
  *
  * @description:
  *    a function used to enter a lock through @m_lock_write_update
  *
  * @input:
  *    lock  :: handle to target lock
  *
  * @return:
  *    error code. 0 means success
  */
  typedef  M_Error  (*M_Lock_WriteUpdateFunc)( M_Lock  lock );


 /*************************************************************************
  *
  * @struct: M_Lock_ClassRec
  *
  * @description:
  *    a structure describing the class of a given lock type
  *
  * @fields:
  *    magic        :: magic value for lock classes
  *    mcore        :: MLib core handle (optional)
  *    size         :: size of lock in bytes
  *    init         :: lock initialization
  *    done         :: lock finalization
  *    enter        :: lock enter
  *    enter_read   :: lock read enter
  *    enter_update :: lock update enter
  *    write_update :: lock update write
  *    exit         :: lock exit
  */
  typedef struct M_Lock_ClassRec_
  {
    M_UInt32                magic;
    M_Core                  mcore;
    M_Size                  size;
    M_Lock_InitFunc         init;
    M_Lock_DoneFunc         done;
    M_Lock_EnterFunc        enter;
    M_Lock_EnterReadFunc    enter_read;
    M_Lock_EnterUpdateFunc  enter_update;
    M_Lock_WriteUpdateFunc  write_update;
    M_Lock_ExitFunc         exit;

  } M_Lock_ClassRec;

 /* */

#define  m_assert_is_lock(x)  \
            m_assert( (x) && M_LOCK(x)->clazz && M_LOCK(x)->clazz->magic == M_MAGIC_LOCK )

M_END_HEADER

#endif /* __MLIB_INTERNAL_LOCK_H__ */
