#ifndef __MLIB_INTERNAL_MEMORY_H__
#define __MLIB_INTERNAL_MEMORY_H__

#include <mlib.h>

M_BEGIN_HEADER

 /***************************************************************************/
 /***************************************************************************/
 /*****                                                                 *****/
 /*****                   MLIB MEMORY MANAGER INTERNALS                 *****/
 /*****                                                                 *****/
 /***************************************************************************/
 /***************************************************************************/

 /* magic for memory managers */
#define  M_MAGIC_MEMORY       0xaafedabe

 /***************************************************************
  *
  * @type: M_Memory_Class
  *
  * @description:
  *    a handle to a _constant_ memory manager class structure.
  *
  * @also:
  *    @M_MemoryRec, @M_Memory_ClassRec
  */
  typedef const struct M_Memory_ClassRec_*    M_Memory_Class;


 /***************************************************************
  *
  * @type: M_IMemory
  *
  * @description:
  *    a handle to an internal representation of a memory manager
  *
  * @also:
  *    @M_IMemoryRec, @M_Memory_ClassRec
  */
  typedef struct M_IMemoryRec_*      M_IMemory;


 /***************************************************************
  *
  * @functype: M_Memory_AllocFunc
  *
  * @description:
  *    a memory manager method used to allocate one single block
  *    of memory..
  *
  * @input:
  *    size   :: size of memory block, in bytes, cannot be 0 !!
  *    memory :: handle to memory manager
  *
  * @return:
  *    address of newly allocated memory block. NULL in case
  *    of failure
  *
  * @also:
  *    @M_Memory_ReallocFunc, @M_Memory_FreeFunc, @M_Memory,
  *    @M_MemoryRec
  */
  typedef  M_Pointer  (*M_Memory_AllocFunc  )( M_Size    size,
                                               M_Memory  memory );

 /***************************************************************
  *
  * @functype: M_Memory_ReallocFunc
  *
  * @description:
  *    a memory manager method used to reallocate a given block
  *    of memory..
  *
  * @input:
  *    block    :: address of target memory block
  *    new_size :: new memory block's size in bytes. cannot be 0
  *    cur_size :: current memory block's size in bytes. cannot be 0
  *    memory   :: handle to memory manager
  *
  * @return:
  *    address of reallocated memory block. NULL in case of error
  *
  * @also:
  *    @M_Memory_AllocFunc, @M_Memory_FreeFunc, @M_Memory,
  *    @M_MemoryRec
  */
  typedef  M_Pointer  (*M_Memory_ReallocFunc)( M_Pointer  block,
                                               M_Size     new_size,
                                               M_Size     cur_size,
                                               M_Memory   memory );


 /***************************************************************
  *
  * @functype: M_Memory_FreeFunc
  *
  * @description:
  *    a memory manager method used to release a given block
  *    of memory..
  *
  * @input:
  *    block    :: address of target memory block. cannot be NULL !!
  *    memory   :: handle to memory manager
  *
  * @also:
  *    @M_Memory_Alloc, @M_Memory_ReallocFunc, @M_Memory,
  *    @M_MemoryRec
  */
  typedef  void       (*M_Memory_FreeFunc)   ( M_Pointer  block,
                                               M_Memory   memory );


 /***************************************************************
  *
  * @functype: M_Memory_NewFunc
  *
  * @description:
  *    a function called to create a new memory manager object
  *
  * @input:
  *    impl :: implementation-specific pointer
  *
  * @return:
  *    handle to new memory manager object (internal structure)
  *
  * @note:
  *    this function must initialise the memory manager by first
  *    calling the @m_memory_init_internal function.
  */
  typedef  M_IMemory   (*M_Memory_NewFunc)( M_Pointer  impl );


 /***************************************************************
  *
  * @functype: M_Memory_DestroyFunc
  *
  * @description:
  *    a memory manager method called when the manager is
  *    destroyed..
  *
  * @input:
  *    memory   :: handle to memory manager (internal structure)
  *
  * @also:
  *    @M_Memory, @M_MemoryRec
  */
  typedef  void       (*M_Memory_DestroyFunc)( M_IMemory   memory );


 /***************************************************************
  *
  * @enum: M_Memory_Flags
  *
  * @description:
  *    bit flag values used in a @M_MemoryRec structure
  *
  * @values:
  *   M_MEMORY_FLAG_CLEAR_BLOCKS ::
  *       indicates that the allocation routine provided by the
  *       memory manager clears the returned memory block. If unset,
  *       @m_mem_alloc and @m_mem_realloc will zero the
  *       corresponding memory blocks/portions.
  *
  *  M_MEMORY_FLAG_DISABLE_TRAP ::
  *       this flag is used internally by the library. It is ignored
  *       by nv_mem_new though..
  *
  * @also:
  *   @M_Memory_Class
  */
  typedef enum M_Memory_Flags_
  {
    M_MEMORY_FLAG_CLEAR_BLOCKS   = 1,
    M_MEMORY_FLAG_DISABLE_TRAP   = 2

  } M_Memory_Flags;



 /***************************************************************
  *
  * @struct: M_Memory_ClassRec
  *
  * @description:
  *    a simple memory allocator class structure
  *
  * @fields:
  *    alloc        :: memory allocation routine
  *    realloc      :: memory reallocation routine
  *    free         :: memory release routine
  *    mem_new      :: create a new memory manager
  *    mem_destroy  :: destroy a given memory manager
  *    mem_flags    :: memory manager flags
  *
  * @also:
  *   @M_Memory_Class,
  *   @M_Memory_AllocFunc, @M_Memory_ReallocFunc,
  *   @M_Memory_DestroyFunc, @M_Memory_Flags
  */
  typedef struct M_Memory_ClassRec_
  {
    M_Memory_AllocFunc     alloc;
    M_Memory_ReallocFunc   realloc;
    M_Memory_FreeFunc      free;
    M_Memory_NewFunc       mem_new;
    M_Memory_DestroyFunc   mem_destroy;
    M_UInt                 mem_flags;

  } M_Memory_ClassRec;


 /* */

 /***************************************************************************
  *
  * @struct: M_IMemoryRec
  *
  * @description:
  *    an MLIB internal structure used to model a given memory manager
  *
  * @fields:
  *    root      :: the root @M_MemoryRec fields
  *    magic     :: a 32-bit magic number used to validate memory manager handles
  *    alloc     :: shortcut to "clazz->alloc"
  *    free      :: shortcut to "clazz->free"
  *    realloc   :: shortcut to "clazz->realloc"
  *    flags     :: shortcut to "clazz->mem_flags"
  *    ref_count :: reference count for this manager
  *    clazz     :: memory manager class handle
  *
  *    debug     :: pointer to debugging memory allocator, if any
  *    pool      :: this pointer is not NULL for memory pools only
  *    pool_set  :: handle to the set of memory pools, if any
  */
  typedef struct M_IMemoryRec_
  {
    M_MemoryRec            root;
    M_UInt32               magic;
    M_Memory_AllocFunc     alloc;   /* shortcut for clazz->alloc   */
    M_Memory_FreeFunc      free;    /* shortcut for clazz->free    */
    M_Memory_ReallocFunc   realloc; /* shortcut for clazz->realloc */
    M_UInt                 flags;
    M_Int                  ref_count;
    M_Memory_Class         clazz;

    M_Pointer              debug;
    M_MemoryPool           pool;
    M_Pointer              pool_set;

  } M_IMemoryRec;


 /***************************************************************
  *
  * @function: m_memory_new_custom
  *
  * @description:
  *   allocate a new _custom_ memory manager
  *
  * @input:
  *   clazz     :: handle to memory manager class structure
  *   user_data :: value of the new memory manager's "user_data" field
  *
  * @return:
  *    handle to new memory manager object. NULL in case of error
  *
  * @note:
  *    the class's "alloc" func is used to allocate the memory manager
  *    object.
  *
  *    the new memory manager's reference count is set to 1. You
  *    can destroy it by using @m_mem_unref
  *
  * @also:
  *    @m_memory_ref, @m_memory_unref, @m_memory_new_std
  */
  MLIB_API(M_Memory)  m_memory_new_custom( M_Memory_Class  clazz,
                                           M_Pointer       impl );


  /* */

#define  M_IMEMORY(p)  ((M_IMemory)(p))

#define  M_MEMORY_ALLOC_INTERNAL(m,s)   \
            M_IMEMORY(m)->alloc( s, m )

#define  M_MEMORY_FREE_INTERNAL(m,s)   \
            M_IMEMORY(m)->free( s, m )

#define  m_memory_need_block_clear(_mem)  \
           (!(M_IMEMORY(_mem)->flags & M_MEMORY_FLAG_CLEAR_BLOCKS))

#define  m_memory_trap_enabled( _mem )  \
           (!(M_IMEMORY(_mem)->flags & M_MEMORY_FLAG_DISABLE_TRAP))

#define  M_ASSERT_IS_MEMORY(m)                                      \
           m_assert( (m) && M_IMEMORY(m)->magic == M_MAGIC_MEMORY )


#undef  M_LOG_COMPONENT
#define M_LOG_COMPONENT  "Memory"

  /* */

M_END_HEADER

#endif /* __MLIB_INTERNAL_MEMORY_H__ */

