#ifndef __MLIB_INTERNAL_MEMORY_POOL_H__
#define __MLIB_INTERNAL_MEMORY_POOL_H__

#include <mlib/internal/imemory.h>

M_BEGIN_HEADER

#define  M_MAGIC_MEMORY_POOL       0x6d65706f   /* "mepo" */
#define  M_MAGIC_MEMORY_POOL_SET   0x6d606f73   /* "mpos" */


/* must be a power of 2 to avoid fragmentation problems with most */
/* malloc implementations..                                       */
/*                                                                */
#define  M_MEMORY_POOL_PAGE_SIZE  2048

#define  M_MEMORY_POOL_ALIGNMENT  ((int)sizeof(void*))

#define  M_MEMORY_POOL_ALIGN(x) \
            ( ( (x) + M_MEMORY_POOL_ALIGNMENT-1 ) & -M_MEMORY_POOL_ALIGNMENT )

#define  M_MEMORY_POOL_BUCKETS  16


 /**********************************************************************
  *
  *  small technical note on memory pools:
  *
  *    each memory pool contains one or more 'pages'. Each page has exactly
  *    M_MEMORY_POOL_PAGE_SIZE bytes and is used to store the following:
  *
  *      - a set of contiguous fixed-size memory blocks
  *
  *      - a very small bitmap used during GC operations to detect
  *        "unused" pages as well as rebuild the pool's list of
  *        free blocks.
  *
  *        the bitmap is stored just after the blocks, and is accessible
  *        through 'page->limit' (it contains pool->element_count/8 bytes)
  *
  *
  *    the field 'pool->element_count' is thus generally less than
  *    'M_MEMORY_POOL_PAGE_SIZE/pool->element_size' in order to make
  *    room for the bitmap..
  *
  */


  typedef struct M_MemoryPoolPageRec_*  M_MemoryPoolPage;

  typedef struct M_MemoryPoolPageRec_
  {
    M_MemoryPoolPage   link;
    M_Byte*            base;
    M_Byte*            limit;      /* end of block page, start of bitmap */
    M_UInt             num_frees;  /* number of free blocks in this page */

  } M_MemoryPoolPageRec;

  typedef struct M_MemoryPoolRec_
  {
    M_IMemoryRec       fake;        /* fake memory allocator.. */
    M_Memory           memory;      /* parent memory allocator */
    M_UInt32           magic;
    M_Lock             lock;

    M_Pointer          element_free;   /* first free block in pool */
    M_Size             element_size;   /* padded size of elements  */
    M_Size             element_count;  /* # of elements / page     */

    M_Byte*            cursor;
    M_Byte*            limit;
    M_MemoryPoolPage   page;
    M_Size             page_size;      /* total size of page in bytes */
    M_Size             page_limit;     /* multiple of element_count   */

    M_MemoryPool       link;        /* used internally by hash table */

  } M_MemoryPoolRec;

#define  M_MEMORY_POOL(x)  ((M_MemoryPool)(x))

  typedef struct M_MemoryPoolSetRec_*  M_MemoryPoolSet;

  typedef struct M_MemoryPoolSetRec_
  {
    M_Core         mcore;
    M_UInt32       magic;
    M_Lock         synlock;                           /* common synlock for all pools */
    M_Lock         rwlock;                            /* r/w lock for hash table      */
    M_MemoryPool   pools[ M_MEMORY_POOL_BUCKETS ];    /* simple hash table            */

  } M_MemoryPoolSetRec;


#ifdef MLIB_CONFIG_OPTION_MEMORY_POOL

  MLIB_LOCAL(M_Pointer)
  m_memory_pool_alloc_page( M_MemoryPool  pool )

#endif


#define  M_MEMORY_POOL_ALLOC(_pool,_size,_result)                          \
     M_BEGIN_STMNT                                                         \
        M_Synlock   _lock = _pool->lock;                                   \
                                                                           \
        m_assert( M_MEMORY_POOL_ALIGN(size) == _pool->element_size );      \
                                                                           \
        M_SYNLOCK_LOCK(_lock);                                             \
        {                                                                  \
          _result = _pool->element_free;                                   \
          if ( _result )                                                   \
            _pool->element_free = *M_POINTER_P(_result);                   \
        }                                                                  \
        M_SYNLOCK_UNLOCK(_lock);                                           \
                                                                           \
        if ( _result == NULL )  /* no more free items in the pool */       \
          _result = m_memory_pool_alloc_page( _pool );                     \
    M_END_STMNT


#define  M_MEMORY_POOL_FREE(_pool,_block)                             \
    M_BEGIN_STMNT                                                     \
        M_Synlock  _lock = _pool->lock;                               \
                                                                      \
        M_SYNLOCK_LOCK(_lock);                                        \
        {                                                             \
          *M_POINTER_P(_block) = _pool->element_free;                 \
          _pool->element_free  = _block;                              \
        }                                                             \
        M_SYNLOCK_UNLOCK(_lock);                                      \
    M_END_STMNT


#define  m_assert_is_memory_pool(x)    \
           m_assert( (x) && (x)->magic == M_MAGIC_MEMORY_POOL )

#define  m_assert_is_memory_pool_set(x)  \
           m_assert( (x) && (x)->magic == M_MAGIC_MEMORY_POOL_SET )

  /* */

M_END_HEADER

#endif /* __MLIB_INTERNAL_MEMORY_POOL_H__ */
