#ifndef __MLIB_LHASH_H__
#define __MLIB_LHASH_H__

#include <mlib.h>

M_BEGIN_HEADER

 /**************************************************************************
  *
  * @type: M_LHash
  *
  * @description:
  *   handle to a @M_LHashRec structure used to implement dynamic hash
  *   tables.
  */
  typedef struct M_LHashRec_*   M_LHash;


 /**************************************************************************
  *
  * @type: M_LHNode
  *
  * @description:
  *   handle to a @M_LHNodeRec structure used as the "root" of all hash nodes
  *   in a linear hash table
  *
  * @note:
  *   this implementation of linear hash tables does NOT allocate the hash
  *   node itself. Instead, it's up to the user to do the allocations and
  *   releases.
  */
  typedef struct M_LHNodeRec_*    M_LHNode;


 /**************************************************************************
  *
  * @type: M_LHLookup
  *
  * @description:
  *   handle to a @M_LHNode pointer. This is returned by the @m_lhash_lookup
  *   function.
  */
  typedef M_LHNode*        M_LHLookup;


 /**************************************************************************
  *
  * @struct: M_LHNodeRec
  *
  * @description:
  *   a structure containing the root fields of all hash nodes in a linear
  *   hash table.
  *
  * @fields:
  *   link :: pointer to next node in same bucket
  *   hash :: a 32-bit unsigned value used as the node's hash value
  */
  typedef struct M_LHNodeRec_
  {
    M_LHNode   link;
    M_UInt32   hash;

  } M_LHNodeRec;


 /**************************************************************************
  *
  * @struct: M_LHashRec
  *
  * @description:
  *   a structure modelling a dyanmic linear hashing table.
  *
  * @fields:
  *   buckets    :: array of buckets (hash node lists)
  *   p          :: incremental bucket index
  *   mask       :: incremental bucket mask (i.e. max_p-1)
  *   slack      :: incremental slack value
  *   memory     :: memory manager handle. only used to allocate and
  *                 resize the buckets array.
  *
  *   node_equal :: node equality function
  *   node_data  :: optional argument for the node equality function
  */
  typedef struct M_LHashRec_
  {
    M_LHNode*        buckets;
    M_UInt           p;
    M_UInt           mask;
    M_Long           slack;
    M_Memory         memory;

    M_EqualDataFunc  node_equal;
    M_Pointer        node_data;

  } M_LHashRec;


 /**************************************************************************
  *
  * @function: m_lhash_init
  *
  * @description:
  *   initialize a linear hash table
  *
  * @input:
  *   table  :: handle to target hash table structure
  *   memory :: memory manager to use to allocate and resize the buckets
  *             array
  *
  *   node_equal :: node equality test function
  *
  * @throws:
  *   @m_err_memory_alloc if the default buckets array couldn't
  *   be allocated
  *
  * @also:
  *   @m_lhash_init_full can be used when the node equality function needs
  *   an optional argument
  */
  MLIB_API( void )
  m_lhash_init( M_LHash      table,
                M_Memory     memory,
                M_EqualFunc  node_equal );


 /**************************************************************************
  *
  * @function: m_lhash_init_full
  *
  * @description:
  *   same as @m_lhash_init, when the node equality function needs an
  *   argument
  *
  * @input:
  *   table  :: handle to target hash table structure
  *   memory :: memory manager to use to allocate and resize the buckets
  *             array
  *
  *   node_equal :: node equality test function
  *   node_data  :: optional argument to the node equality function
  *
  * @throws:
  *   @m_err_memory_alloc if the default buckets array couldn't
  *   be allocated
  *
  * @also:
  *   @m_lhash_init can be used when the node equality function doesn't need
  *   an argument
  */
  MLIB_API( void )
  m_lhash_init_full( M_LHash          table,
                     M_Memory         memory,
                     M_EqualDataFunc  node_equal,
                     M_Pointer        node_data );


 /**************************************************************************
  *
  * @function: m_lhash_done
  *
  * @description:
  *   finalize a linear hash table.
  *
  * @input:
  *   table        :: handle to target hash table structure
  *   node_destroy :: an optional node destructor
  *
  * @note:
  *   this function doesn't free the nodes, it's up to the node destructor
  *   to do that !!
  */
  MLIB_API( void )
  m_lhash_done( M_LHash        table,
                M_DestroyFunc  node_destroy );


 /**************************************************************************
  *
  * @function: m_lhash_done_full
  *
  * @description:
  *   same as @m_lhash_done, but used when the node destructor needs an
  *   additional argument
  *
  * @input:
  *   table        :: handle to target hash table structure
  *   node_destroy :: an optional node destructor
  *   node_data    :: an optional argument to the node destructor
  *
  * @note:
  *   this function doesn't free the nodes, it's up to the node destructor
  *   to do that !!
  */
  MLIB_API( void )
  m_lhash_done_full( M_LHash            table,
                     M_DestroyDataFunc  node_destroy,
                     M_CPointer         node_data );


 /**************************************************************************
  *
  * @function: m_lhash_get_size
  *
  * @description:
  *   returns the number of elements within a linear hash table
  *
  * @input:
  *   table  :: handle to target hash table structure
  *
  * @return:
  *   elements count. 0 if empty
  */
  MLIB_API( M_UInt )
  m_lhash_get_size( M_LHash  table );


 /**************************************************************************
  *
  * @function: m_lhash_lookup
  *
  * @description:
  *   lookup a given node within a hash table.
  *
  * @input:
  *   table    :: handle to source hash table
  *   key_node :: pointer to a "key" node used during the lookup. This
  *               must be set-up by the caller correctly. Its "hash" value,
  *               and its key data must be set-up correctly.
  *
  * @return:
  *   a @M_LHLookup pointer. it cannot be NULL, however it _references_ a
  *   @M_LHNode handle corresponding to the search result.
  */
  MLIB_API( M_LHLookup )
  m_lhash_lookup( M_LHash    table,
                  M_LHNode   key_node );


 /**************************************************************************
  *
  * @function: m_lhash_add
  *
  * @description:
  *   lookup a given node within a hash table.
  *
  * @input:
  *   table    :: handle to source hash table
  *   lookup   :: search lookup result, as returned by @m_lhash_lookup
  *   new_node :: address of new hash node to be inserted in the hash table
  *
  * @throws:
  *   @m_err_memory_alloc when trying to expand the buckets array
  *
  * @note:
  *   the new node must be allocated by the caller
  *
  *   this function doesn't handle collisions. It is up to the caller to
  *   ensure that '*lookup' is NULL, or that its content has been removed
  *   from the hash table before
  *
  *   the node is _always_ inserted within the table before an exception
  *   is raised.
  */
  MLIB_API( void )
  m_lhash_add( M_LHash     table,
               M_LHLookup  lookup,
               M_LHNode    new_node );


 /**************************************************************************
  *
  * @function: m_lhash_remove
  *
  * @description:
  *   remove a given node within a hash table
  *
  * @input:
  *   table    :: handle to source hash table
  *   lookup   :: search lookup result, as returned by @m_lhash_lookup
  *
  * @throws: @m_err_memory_alloc when trying to shrink the buckets array
  *
  * @note:
  *   this function doesn't free the node. It's up to the caller to do that
  *
  *   the node is _always_ owned by the table when an exception is raised
  *   during a removal
  */
  MLIB_API( void )
  m_lhash_remove( M_LHash     table,
                  M_LHLookup  lookup );

                 /* */

M_END_HEADER

#endif /* __MLIB_LHASH_H__ */
