#ifndef __MLIB_LIST_H__
#define __MLIB_LIST_H__

#include <mlib.h>

M_BEGIN_HEADER

/*************************************************************************
 *
 * <Section> lists
 *
 * <Title> Lists
 *
 * <Abstract> Simple doubly-linked lists
 *
 * <Description>
 *    The MLib contains two implementations of lists, in order to please
 *    several kinds of developers and users. Note that both implementations
 *    use the same type for the list nodes though.
 *
 *    First of all, the @M_ListRec type is used to model an easy-to-use
 *    doubly-linked list of nodes. It is implemented as a _circular_ list
 *    for efficiency and compactness
 *
 *    Second, the @M_GList type is defined similarly to the GLib "GList*"
 *    type. A M_GList is a pointer to the head of a _linear_
 *    doubly-linked list of nodes. It is NULL if the list is empty.
 */


 /************************************************************************/
 /************************************************************************/
 /*****                                                              *****/
 /*****                 LINEAR DOUBLY-LINKED LISTS                   *****/
 /*****                                                              *****/
 /************************************************************************/
 /************************************************************************/


 /***********************************************************************
  *
  * @type: M_List
  *
  * @description:
  *    handle to a simple doubly-linked list node. see @M_ListRec
  */
  typedef struct M_ListRec_*      M_List;


 /***********************************************************************
  *
  * @struct: M_ListRec
  *
  * @description:
  *    a simple doubly-linked list node structure.
  *
  * @fields:
  *    next :: handle to next node in list. NULL if last node
  *    prev :: handle to previous node in list. NULL if first node
  *    data :: generic pointer to listed element
  */
  typedef struct M_ListRec_
  {
    M_List     next;
    M_List     prev;
    M_Pointer  data;

  } M_ListRec;


 /***********************************************************************
  *
  * @functype: M_List_NodeFunc
  *
  * @description:
  *    a function used to process a given list node. For example, to
  *    delete its data member..
  *
  * @input:
  *    node      :: handle to the target list node
  *    user_data :: user-provided data
  */
  typedef void  (*M_List_NodeFunc)(  M_List     node,
                                     M_Pointer  user_data );



 /***********************************************************************
  *
  * @function: m_list_memory_pool
  *
  * @description:
  *    return a memory pool used to allocate list nodes very quickly
  *
  * @input:
  *    memory :: memory manager handle
  *
  * @return:
  *    the memory pool handle (or the input memory manager handle)
  */
#define  m_list_memory_pool(memory)   \
               m_memory_pool_find( memory, sizeof(M_ListRec) )


 /***********************************************************************
  *
  * @function: m_list_alloc
  *
  * @description:
  *    allocate a new list node
  *
  * @input:
  *    data   :: data pointer for the new node
  *    memory :: memory handle
  *
  * @return:
  *    handle to new list node
  *
  * @throws: @m_err_memory_alloc
  */
  MLIB_API(M_List)
  m_list_alloc( M_Pointer  data,
                M_Memory   memory );


 /***********************************************************************
  *
  * @function: m_list_free
  *
  * @description:
  *    release a given list node
  *
  * @input:
  *    node :: handle to target list node
  *    pool :: handle to list pool
  *
  * @note:
  *    this function doesn't destroy the node's data pointer
  */
#define  m_list_free(_node,_mem)  m_mem_free(_node,_mem)


 /***********************************************************************
  *
  * @function: m_list_append
  *
  * @description:
  *    append a new element to a list.
  *
  * @inout:
  *    plist  :: pointer to list head. cannot be NULL !!
  *
  * @input:
  *    data   :: data pointer for the new node
  *    memory :: memory manager
  *
  * @throws: @m_err_memory_alloc
  *
  * @note:
  *   this function is potentially slow because it needs to parse the
  *   whole list before appending the new node. consider using rings
  *   if you need fast appends instead..
  */
  MLIB_API(void)
  m_list_append( M_List*    plist,
                 M_Pointer  data,
                 M_Memory   memory) ;


 /***********************************************************************
  *
  * @function: m_list_prepend
  *
  * @description:
  *    prepend a new element to a list.
  *
  * @inout:
  *    plist  :: pointer to list head. cannot be NULL !!
  *
  * @input:
  *    data   :: data pointer for the new node
  *    memory :: memory manager
  *
  * @throws: @m_err_memory_alloc
  */
  MLIB_API(void)
  m_list_prepend( M_List*    plist,
                  M_Pointer  data,
                  M_Memory   memory );


 /***********************************************************************
  *
  * @function: m_list_find
  *
  * @description:
  *    find a list node with a given "data" pointer
  *
  * @input:
  *    plist  :: pointer to list head. cannot be NULL
  *    data   :: data pointer to look up in list
  *
  * @returns:
  *    list node handle corresponding to the first occurence of 'data'
  *    in the list. NULL otherwise
  */
  MLIB_API(M_List)
  m_list_find( M_List*    plist,
               M_Pointer  data );


 /***********************************************************************
  *
  * @function: m_list_remove
  *
  * @description:
  *    remove a node corresponding to the first occurence of 'data' from
  *    a given list.
  *
  * @inout:
  *    plist  :: pointer to list head. cannot be NULL
  *
  * @input:
  *    data   :: data pointer to look up in list
  *    memory :: memory manager handle
  */
  MLIB_API(void)
  m_list_remove( M_List*    plist,
                 M_Pointer  data,
                 M_Memory   memory );


 /***********************************************************************
  *
  * @function: m_list_reverse
  *
  * @description:
  *    reverse a given list
  *
  * @inout:
  *    plist  :: pointer to list head. cannot be NULL
  */
  MLIB_API(void)
  m_list_reverse( M_List*  plist );


 /***********************************************************************
  *
  * @function: m_list_length
  *
  * @description:
  *    return the number of elements in a list
  *
  * @input:
  *    plist  :: pointer to list head. cannot be NULL
  *
  * @return:
  *    number of elements in the list
  */
  MLIB_API(M_UInt)
  m_list_length( M_List*  plist );


 /***********************************************************************
  *
  * @function: m_list_nth
  *
  * @description:
  *    returns the nth's node in a list
  *
  * @input:
  *    plist  :: pointer to list head. cannot be NULL
  *    index  :: element index
  *
  * @return:
  *    n'th list node, NULL for invalid indices
  */
  MLIB_API(M_List)
  m_list_nth( M_List*  plist,
              M_UInt   index );


 /***********************************************************************
  *
  * @function: m_list_index
  *
  * @description:
  *    returns the index of a given element in a list
  *
  * @input:
  *    plist  :: pointer to list head. cannot be NULL
  *    data   :: data to lookup
  *
  * @return:
  *    "data" position in the list, or -1 if not present
  */
  MLIB_API(M_Int)
  m_list_index( M_List*    plist,
                M_Pointer  data );


 /***********************************************************************
  *
  * @function: m_list_last
  *
  * @description:
  *    return the last node in a list
  *
  * @input:
  *    plist  :: pointer to list head. cannot be NULL
  *
  * @return:
  *    last node, if any
  */
  MLIB_API(M_List)
  m_list_last( M_List*  plist );


 /***********************************************************************
  *
  * @function: m_list_first
  *
  * @description:
  *    return the first node in a list
  *
  * @input:
  *    plist  :: pointer to list head. cannot be NULL
  *
  * @return:
  *    first node, if any
  */
  MLIB_API(M_List)
  m_list_first( M_List*  plist );


 /***********************************************************************
  *
  * @function: m_list_append_node
  *
  * @description:
  *    append a node to a list
  *
  * @inout:
  *    plist  :: pointer to list head. cannot be NULL
  *
  * @input:
  *    node   :: new list node
  */
  MLIB_API(void)
  m_list_append_node( M_List*  plist,
                      M_List   node );


 /***********************************************************************
  *
  * @function: m_list_prepend_node
  *
  * @description:
  *    prepend a node to a list
  *
  * @inout:
  *    plist  :: pointer to list head. cannot be NULL
  *
  * @input:
  *    node   :: new list node
  */
  MLIB_API(void)
  m_list_prepend_node( M_List*  plist,
                       M_List   node );


 /***********************************************************************
  *
  * @function: m_list_remove_node
  *
  * @description:
  *    remove a node from a list
  *
  * @inout:
  *    plist  :: pointer to list head. cannot be NULL
  *
  * @input:
  *    node   :: new list node
  */
  MLIB_API(void)
  m_list_remove_node( M_List*  plist,
                      M_List   node );


 /***********************************************************************
  *
  * @function: m_list_foreach
  *
  * @description:
  *    apply a function on each list node
  *
  * @input:
  *    plist     :: pointer to list head. cannot be NULL
  *    node_func :: node function
  *    node_data :: user data passed to node function
  *
  * @throws: any (exceptions of node_func only)
  */
  MLIB_API(void)
  m_list_foreach( M_List*          plist,
                  M_List_NodeFunc  node_func,
                  M_Pointer        node_data );


 /***********************************************************************
  *
  * @function: m_list_free_all
  *
  * @description:
  *    free all list nodes in a list
  *
  * @inout:
  *    plist     :: pointer to list head. cannot be NULL
  *
  * @input:
  *    val_destroy :: optional node data finalizer. can be NULL
  *    val_data    :: user data passed to node finalizer
  *    memory      :: memory manager handle
  *
  * @note:
  *    the node finalizer should not free the node, only the object
  *    pointed to by the "data" field.
  */
  MLIB_API(void)
  m_list_free_all( M_List*            plist,
                   M_DestroyDataFunc  val_destroy,
                   M_Pointer          val_data,
                   M_Memory           memory );


 /***********************************************************************
  *
  * @macro: M_LIST_LOOP (list,data)
  *
  * @description:
  *    a useful macro used in association with @M_LIST_END to parse
  *    all nodes in a list and extract its "data" values
  *
  *    the first parameter is a pointer to the list head.
  *    the second parameter is a local variable name that will
  *    be used within the @M_LIST_LOOP .. @M_LIST_END statement
  *    to receive each node's "data" value..
  *
  *    here's a small example:
  *
  *    {
  *      M_List   numbers = NULL;
  *      M_Int    n;
  *
  *      for ( n = 0; n < 10; n++ )
  *        m_list_append( &list, M_POINTER_FROM_INT(n), memory );
  *
  *      M_LIST_LOOP( &list, value )
  *        printf( "%d ", M_POINTER_TO_INT(value) );
  *      M_LIST_END
  *
  *      m_list_free_all( &list, NULL, NULL, memory );
  *    }
  */
#define  M_LIST_LOOP(_plist,_data)                         \
               {                                           \
                 M_List     __list_node = *(_plist);       \
                 M_List     __list_next;                   \
                 M_Pointer  _data;                         \
                                                           \
                 while ( __list_node )                     \
                 {                                         \
                   __list_next = __list_node->next;        \
                   _data       = __list_node->data;        \


 /***********************************************************************
  *
  * @macro: M_LIST_LOOP_NODE (list,node)
  *
  * @description:
  *    a variant of @M_LIST_LOOP_NODE used to expose each list node
  *    pointer, instead of its "data" field
  */
#define  M_LIST_LOOP_NODE(_plist,_node)                    \
               {                                           \
                 M_List     __list_node = *(_plist);       \
                 M_List     __list_next;                   \
                 M_List     _node;                         \
                                                           \
                 while ( __list_node )                     \
                 {                                         \
                   __list_next = __list_node->next;        \
                   _node       = __list_node;              \


 /***********************************************************************
  *
  * @macro: M_LIST_END
  *
  * @description:
  *    a useful macro used in association with @M_LIST_LOOP to parse
  *    all nodes in a list and extract its "data" values
  */
#define  M_LIST_END                                        \
                   __list_node = __list_next;              \
                 }                                         \
               }

 /* */

 /***********************************************************************
  *
  * @function: m_list_node_skip
  *
  * @description:
  *    returns the nth's node in a list
  *
  * @input:
  *    node  :: start list node
  *    delta :: increment count (in list list nodes)
  *
  * @return:
  *    new list node, of NULL if out-of-list
  */
  MLIB_API(M_List)
  m_list_node_skip( M_List  node,
                    M_Int   delta );


 /***********************************************************************
  *
  * @function: m_list_node_last
  *
  * @description:
  *    find the last node from a given stating node in a list
  *
  * @input:
  *    node :: start list node
  *
  * @return:
  *    last node, if any
  */
  MLIB_API(M_List)
  m_list_node_last( M_List  node );



 /***********************************************************************
  *
  * @function: m_list_node_first
  *
  * @description:
  *    find the first node from a given stating node in a list
  *
  * @input:
  *    node :: start list node
  *
  * @return:
  *    first node, if any
  */
  MLIB_API(M_List)
  m_list_node_first( M_List  node );

 /* */

M_END_HEADER

#endif /* __MLIB_LIST_H__ */

