#ifndef __MLIB_LOCK_H__
#define __MLIB_LOCK_H__

#ifndef __MLIB_H__
#error  "do not include this file directly, use <mlib.h> instead"
#endif

M_BEGIN_HEADER

 /***********************************************************************
  *
  * @type: M_Lock
  *
  * @description:
  *   opaque handle to a lock object, used to synchronize access to
  *   shared resources. Several kinds of locks can be used, see @M_LockType
  *   and @M_LockScope
  *
  * @note:
  *   for now, the MLib only supports local locks, i.e. locks that
  *   operate (block) threads in the current process only.
  */
  typedef struct M_LockRec_*    M_Lock;


 /***********************************************************************
  *
  * @enum: M_LockType
  *
  * @description:
  *   a list of values that describe the types of locks supported by
  *   the MLib
  *
  * @values:
  *   M_LOCK_MUTEX ::
  *     the basic _recursive_ mutual exclusion object.
  *
  *   M_LOCK_SPIN ::
  *     a lightweight mutual exclusion object. Generally _not_ recursive.
  *     this is also called a "critical section" or "spinlock" on many
  *     other platforms
  *
  *   M_LOCK_READWRITE ::
  *     a read/write lock, used to protect data that is very often read
  *     and seldom written to. The semantics of read/write locks are the
  *     following:
  *
  *     When you only need to read some protected data, you should use
  *     @m_lock_enter_read to enter the lock.
  *
  *     When you need to read protected data, and eventually update/write to
  *     it under certain circumstances, you should use @m_lock_enter_update to
  *     enter the lock, _then_ call @m_lock_write_update before writing to it
  *     when you need to modify protected data.
  *
  *     If you need to directly change protected data, enter the lock
  *     with @m_lock_enter, this is equivalent the @m_lock_enter_update
  *     directly followed by @m_lock_write_update
  *
  *     In all cases, exit the lock with @m_lock_exit
  */
  typedef enum
  {
    M_LOCK_SPIN = 0,
    M_LOCK_MUTEX,
    M_LOCK_READWRITE

  } M_LockType;


 /***********************************************************************
  *
  * @function: m_lock_new
  *
  * @description:
  *   create a new lock
  *
  * @input:
  *   mcore :: MLib core handle
  *   type  :: type of lock
  *
  * @return:
  *   lock handle. cannot be NULL
  *
  * @throws: @m_err_memory_alloc or @m_err_lock_init
  */
  MLIB_API( M_Lock )
  m_lock_new( M_Core      mcore,
              M_LockType  type );


 /***********************************************************************
  *
  * @function: m_lock_push
  *
  * @description:
  *   push a lock on the cleanup stack. If an exception occurs, then
  *   @m_lock_exit will automatically be called
  *
  * @input:
  *   lock :: target lock handle
  *
  * @throws: @m_err_memory_alloc
  *
  * @note:
  *   you should ensure that the lock is entered !!. It is also a
  *   very bad idea to use push/pop with spin locks !!
  */
  MLIB_API( void )
  m_lock_push( M_Lock  lock );


 /***********************************************************************
  *
  * @function: m_lock_pop
  *
  * @description:
  *   pop a lock from the cleanup stack, without calling @m_lock_exit
  *
  * @input:
  *   lock :: target lock handle
  */
  MLIB_API( void )
  m_lock_pop( M_Lock  lock );


 /***********************************************************************
  *
  * @function: m_lock_destroy
  *
  * @description:
  *   destroy a given lock
  *
  * @input:
  *   lock :: target lock handle
  */
  MLIB_API( void )
  m_lock_destroy( M_Lock  lock );


 /***********************************************************************
  *
  * @function: m_lock_enter
  *
  * @description:
  *   enter/acquire a given lock.
  *
  * @input:
  *   lock :: lock handle
  *
  * @throws: @m_err_lock_enter, @m_err_bad_object, etc..
  *
  * @note:
  *   this function may block the current thread, if another @m_lock_enter,
  *   @m_lock_update or @m_lock_read call has been made.
  *
  *   when this is a read/write lock, this function acquires the lock
  *   for _writes_
  */
  MLIB_API( void )
  m_lock_enter( M_Lock  lock );


 /***********************************************************************
  *
  * @function: m_lock_exit
  *
  * @description:
  *   exit/release a given lock that was previously entered with
  *   @m_lock_enter, @m_lock_read or @m_lock_update
  *
  * @input:
  *   lock :: lock handle
  *
  * @throws: @m_err_lock_exit,  @m_err_bad_object, etc..
  */
  MLIB_API( void )
  m_lock_exit( M_Lock  lock );


 /***********************************************************************
  *
  * @function: m_lock_enter_read
  *
  * @description:
  *   this function can only be called with read/write locks
  *   (see @M_LOCK_READWRITE). It indicates that the lock should be
  *   entered in order to exclusively _read_ protected data.
  *
  *   it will block if @m_lock_write_update or @m_lock_enter has been called,
  *   until the corresponding @m_lock_exit is issued.
  *
  *   However, it will _not_ block if another @m_lock_enter_read has been
  *   issued. this allows high-speed concurrent reads of the protected data
  *
  *   you should call @m_lock_exit to exit the lock.
  *
  * @input:
  *   lock :: lock handle
  *
  * @throws: @m_err_lock_exit,  @m_err_bad_object, etc..
  */
  MLIB_API( void )
  m_lock_enter_read( M_Lock  lock );


 /***********************************************************************
  *
  * @function: m_lock_enter_update
  *
  * @description:
  *   this function can only be called with read/write locks
  *   (see @M_LOCK_READWRITE). It indicates that the lock should be
  *   entered in order to _read_ protected data and potentially _write_
  *   to it too.
  *
  *   it will block if a @m_lock_enter_update or @m_lock_enter has been
  *   issued, until the corresponding @m_lock_exit call.
  *
  *   if one or several @m_lock_read calls have been issued on the lock,
  *   this function will wait until they all exit the lock then acquire
  *   the lock.
  *
  *   you should then call @m_lock_write_update if you need to modify the
  *   protected data
  *
  *   you should call @m_lock_exit to exit the lock in all cases..
  *
  * @input:
  *   lock :: lock handle
  *
  * @throws: @m_err_lock_exit,  @m_err_bad_object, etc..
  */
  MLIB_API( void )
  m_lock_enter_update( M_Lock  lock );


 /***********************************************************************
  *
  * @function: m_lock_write_update
  *
  * @description:
  *   this function can only be called with read/write locks
  *   (see @M_LOCK_READWRITE). It indicates that you're going to modify
  *   some protected data that was previously locked through
  *   @m_lock_enter_update
  *
  *   it will block if one or more @m_lock_read have been issued, until
  *   they all exit the lock.
  *
  *   you should call @m_lock_exit to exit the lock.
  *
  * @input:
  *   lock :: lock handle
  *
  * @throws: @m_err_lock_exit,  @m_err_bad_object, etc..
  */
  MLIB_API( void )
  m_lock_write_update( M_Lock  lock );

 /* */

  MLIB_VAR( M_Exception )   m_err_lock_init;
  MLIB_VAR( M_Exception )   m_err_lock_enter;
  MLIB_VAR( M_Exception )   m_err_lock_exit;

#define  M_LOCK(x)   ((M_Lock)(x))

#ifdef M_CONFIG_THREAD_SAFE
#  define  M_LOCK_ACTION( _lock, _action )   M_BEGIN_STMNT          \
                                               if ( _lock )         \
                                                 _action( _lock );  \
                                             M_END_STMNT
#else
#  define  M_LOCK_ACTION( _lock, _action )   M_UNUSED(_lock)
#endif

#define  M_LOCK_ENTER(x)         M_LOCK_ACTION( x, m_lock_enter )
#define  M_LOCK_EXIT(x)          M_LOCK_ACTION( x, m_lock_exit )
#define  M_LOCK_ENTER_READ(x)    M_LOCK_ACTION( x, m_lock_enter_read )
#define  M_LOCK_ENTER_UPDATE(x)  M_LOCK_ACTION( x, m_lock_enter_update )
#define  M_LOCK_WRITE_UPDATE(x)  M_LOCK_ACTION( x, m_lock_write_update )
#define  M_LOCK_PUSH(x)          M_LOCK_ACTION( x, m_lock_push )
#define  M_LOCK_POP(x)           M_LOCK_ACTION( x, m_lock_pop )

#ifdef M_CONFIG_THREAD_SAFE
#  define  M_LOCK_NEW(_mcore,_type)    m_lock_new( _mcore, _type )
#else
#  define  M_LOCK_NEW(_mcore,_type)    NULL
#endif

#ifdef M_CONFIG_THREAD_SAFE
#  define  M_LOCK_DESTROY(x)    m_lock_destroy(x)
#else
#  define  M_LOCK_DESTROY(x)    M_DUMMY_STMNT
#endif


 /* */

M_END_HEADER

#endif /* __MLIB_LOCK_H__ */
