/**************************************************************************
 *
 *  MLib macros
 *
 *    this file contains various macro definitions used by the MLib
 *
 **************************************************************************/

#ifndef __MLIB_MACROS_H__
#define __MLIB_MACROS_H__

#ifndef __MLIB_H__
#error  "do not include this file directly, use <mlib.h> instead"
#endif

/*************************************************************************
 *
 * @section: macros
 *
 * @title: Macros
 *
 * @abstract: Useful macro definitions
 *
 * @description:
 *    Macros of the MLib
 *
 */

/*************************************************************************/
/*                                                                       */
/*  USEFUL MACROS                                                        */
/*                                                                       */
/*************************************************************************/


/* The ANSI C pre-processor has strange rules regarding */
/* macro concatenation and substitution which forces us */
/* to use two levels of substitution to get a correct   */
/* result..                                             */
/*                                                      */

/**************************************************************
 *
 * @macro: M_STRDEF (x)
 *
 * @description:
 *    stringify the macro's argument _after_ substituing it,
 *    unlike the normal "#" C-preprocessor directive. As an
 *    example:
 *
 *    {
 *      #define  FOO       xyz
 *      #define  FOO1      M_STRDEF(FOO)
 *      #define  FOO2      #FOO
 *    }
 *
 *    will define FOO1 as "xyz" and FOO2 as "FOO"..
 */

#define  M_STRDEF_INNER(x)     #x
#define  M_STRDEF(x)           M_STRDEF_INNER(x)


/**************************************************************
 *
 * @macro: M_CONCAT (x,y)
 *
 * @description:
 *    concatenate the macro's arguments _after_ substituing
 *    them, unlike the normal "##" C-preprocessor directive.
 *    As an example:
 *
 *    {
 *      #define  FOO       xyz
 *      #define  BAR       123
 *      #define  FOOBAR1   M_CONCAT(FOO,BAR)
 *      #define  FOOBAR2   FOO ## BAR
 *    }
 *
 *    will define FOOBAR1 as "xyz123" and FOOBAR2 as "FOOBAR"
 *
 */

#define  M_CONCAT_INNER(x,y)   x##y
#define  M_CONCAT(x,y)         M_CONCAT_INNER(x,y)


/* C++ compatibility definitions */

/****************************************************************
 *
 * @macro: M_BEGIN_HEADER
 *
 * @description:
 *    this macro should be used to begin declarations in every
 *    public header file, in association with a @M_END_HEADER
 *    at the end of the file. The role of these two macros is
 *    to ensure the proper inclusion of the file from a C++
 *    program..
 *
 * @also:
 *    @M_END_HEADER
 */

#ifdef __cplusplus
#  define  M_BEGIN_HEADER    extern "C" {
#else
#  define  M_BEGIN_HEADER    /* nothing */
#endif

/****************************************************************
 *
 * @macro: M_END_HEADER
 *
 * @description:
 *    this macro should be used to end declarations in every
 *    public header file, in association with a @M_BEGIN_HEADER
 *    at the start of the file. The role of these two macros is
 *    to ensure the proper inclusion of the file from a C++
 *    program...
 *
 * @also:
 *   @M_BEGIN_HEADER
 */

#ifdef __cplusplus
#  define  M_END_HEADER    }
#else
#  define  M_END_HEADER    /* nothing */
#endif

/****************************************************************
 *
 * @macro: M_EXTERN (x)
 *
 * @description:
 *    this macro is primarily used to declare an external C
 *    function or variable that might be called or accessed
 *    from other programs, especially C++ ones..
 */

#ifdef __cplusplus
#  define  M_EXTERN(x)       extern "C" x
#else
#  define  M_EXTERN(x)       extern x
#endif

/****************************************************************
 *
 * @macro: M_BEGIN_STMNT
 *
 * @description:
 *    this macro is used to define other macros containing
 *    multiple statements, in association with @M_END_STMNT.
 *    they ensure that the new defined macro is always
 *    seen as a single statement by the C compiler, and
 *    can be followed by a semi-column.
 *
 *    As an example, consider the following macro definitions
 *
 *    {
 *      #define  MULTI1(x,y)  (x) = 0; (y) = 0
 *
 *      #define  MULTI2(x,y)   M_BEGIN_STMNT  \
 *                               (x) = 0;     \
 *                               (y) = 0;     \
 *                             M_END_STMNT
 *    }
 *
 *    using MULTI1 will result in incorrect behaviour and/or
 *    compilation in cases like:
 *
 *    {
 *      if (z)
 *        MULTI1(x1,y1);
 *      else
 *        MULTI1(x2,y2);
 *    }
 *
 *    the problem doesn't happen with MULTI2
 *
 */

#define    M_BEGIN_STMNT     do {

/****************************************************************
 *
 * @macro: M_END_STMNT
 *
 * @description:
 *    this macro is used to define other macros containing
 *    multiple statements, in association with @M_BEGIN_STMNT.
 *    they ensure that the new defined macro is always
 *    seen as a single statement by the C compiler, and
 *    can be followed by a semi-column.
 *
 *    see the example described with @M_BEGIN_STMNT
 *
 */

#define    M_END_STMNT       } while (0)

/****************************************************************
 *
 * @macro: M_DUMMY_STMNT
 *
 * @description:
 *    this macro is used to define other macros that should
 *    be considered as a statement even if they do not include
 *    code.
 *
 *    it's mostly used with conditional compilation in order
 *    to disable certain statements (e.g. logs and assertions)
 *
 */

#define    M_DUMMY_STMNT     do { } while (0)


/****************************************************************
 *
 * @macro: M_UNUSED (x)
 *
 * @description:
 *    this macro is used to indicate that a given function
 *    parameter or static variable is unused. Its purpose is to
 *    eliminate compiler warnings when using high warning
 *    levels.
 *
 *    it is typicaly used with conditional compilation as well
 *    as witthin callbacks that not always use all the parameters
 *    passed to the function..
 *
 * @note:
 *    We could have defined this macro as "(void)(x)", however a
 *    few ANSI compilers complain about this statement..
 */

#define    M_UNUSED(x)       (x)=(x)


/****************************************************************
 *
 * @macro: M_UNUSED_CONST (x)
 *
 * @description:
 *    this macro is used to indicate that a given _constant_ function
 *    parameter or static variable is unused. Its purpose is to
 *    eliminate compiler warnings when using high warning
 *    levels.
 *
 *    it is typicaly used with conditional compilation as well
 *    as witthin callbacks that not always use all the parameters
 *    passed to the function..
 */

#define    M_UNUSED_CONST(x)       (void)(x)


/* */

/* no documentation for M_LOCAL & M_LOCALDEF for now -- need to   */
/* find a suitable explanation of single object compilation first */
/*                                                                */
#ifdef     M_MAKE_OPTION_SINGLE
#  define  M_LOCAL(x)        static x
#  define  M_LOCALDEF(x)     /* nothing */
#else
#  define  M_LOCAL(x)        extern x
#  define  M_LOCALDEF(x)     extern x
#endif


/****************************************************************
 *
 * @macro: M_SET_ERROR (f)
 *
 * @description:
 *    this macro is used to define other macros. Its purpose is
 *    to perform a function call given by the "f" parameter and
 *    store its result in an _implicit_ variable named *error*.
 *
 *    it returns a boolean which is set to TRUE whenever an
 *    error occured, i.e. the value of *error* is not 0.
 *
 *    a typical developer should never use this macro directly..
 */

#define    M_SET_ERROR(x)  ((error = (x)) != 0)


/*************************************************************************/
/*                                                                       */
/*  PROJECT-SPECIFIC macros                                              */
/*                                                                       */
/*************************************************************************/

/****************************************************************
 *
 * @macro: MLIB_API (x)
 *
 * @description:
 *    this macro is used to _declare_ an MLib function.
 *    it defaults to @M_EXTERN but can also be defined as
 *    @M_DLL_EXPORT (when building the MLib DLL), or
 *    @M_DLL_IMPORT (when using the MLib as a DLL)
 */
#ifdef MLIB_BUILD_DLL
#  define  MLIB_API(x)     M_DLL_EXPORT(x)
#elif defined(MLIB_USE_DLL)
#  define  MLIB_API(x)     M_DLL_IMPORT(x)
#else
#  define  MLIB_API(x)     M_EXTERN(x)
#endif

/****************************************************************
 *
 * @macro: MLIB_APIDEF (x)
 *
 * @description:
 *    this macro is used to _define_ an MLib function.
 *    it defaults to @M_EXTERN
 */
#ifdef MLIB_BUILD_DLL
#  define  MLIB_APIDEF(x)     x
#elif defined(MLIB_USE_DLL)
#  define  MLIB_APIDEF(x)     M_EXTERN(x)  /* should not happen */
#else
#  define  MLIB_APIDEF(x)     M_EXTERN(x)
#endif

/****************************************************************
 *
 * @macro: MLIB_VAR (x)
 *
 * @description:
 *    this macro is used to _declare_ an MLib _constant_ variable.
 *    it defaults to @M_EXTERN but can also be defined as
 *    @M_DLL_EXPORT (when building the MLib DLL) or
 *    @M_DLL_IMPORT (when using the MLib as a DLL)
 */
#ifdef MLIB_BUILD_DLL
#  define  MLIB_VAR(x)     M_DLL_EXPORT(x)
#elif defined(MLIB_USE_DLL)
#  define  MLIB_VAR(x)     M_DLL_IMPORT(x)
#else
#  define  MLIB_VAR(x)     M_EXTERN(x)
#endif

/****************************************************************
 *
 * @macro: MLIB_VARDEF (x)
 *
 * @description:
 *    this macro is used to _define_ an MLib function variable.
 *    it defaults to @M_EXTERN
 */
#ifdef MLIB_BUILD_DLL
#  define  MLIB_VARDEF(x)     x
#elif defined(MLIB_USE_DLL)
#  define  MLIB_VARDEF(x)     x  /* should not happen */
#else
#  define  MLIB_VARDEF(x)     x
#endif

  /* */

#ifdef MLIB_MAKE_OPTION_SINGLE
#  define  MLIB_LOCAL(x)   static x
#else
#  define  MLIB_LOCAL(x)   MLIB_API(x)
#endif

#ifdef MLIB_MAKE_OPTION_SINGLE
#  define  MLIB_LOCALDEF(x)  static  x
#else
#  define  MLIB_LOCALDEF(x)  MLIB_APIDEF(x)
#endif


/****************************************************************
 *
 * @macro: M_MIN (x,y)
 *
 * @description:
 *    macro to return the generic minimum of two items
 */
#define  M_MIN(x,y)   ( (x) < (y) ?  (x) : (y) )

/****************************************************************
 *
 * @macro: M_MAX (x,y)
 *
 * @description:
 *    macro to return the generic maximum of two items
 */
#define  M_MAX(x,y)   ( (x) > (y) ?  (x) : (y) )

/****************************************************************
 *
 * @macro: M_ABS (x)
 *
 * @description:
 *    macro to return the generic absolute value of an item
 */
#define  M_ABS(x)     ( (x) < 0   ? -(x) : (x) )


/****************************************************************
 *
 * @macro: M_SIGN (x)
 *
 * @description:
 *    returns -1, 0 or 1 depending on the sign of expression 'x'
 */
#define  M_SIGN(x)    ( (x) > 0 ? +1 : ( (x) < 0  ? -1 : 0 ) )

/****************************************************************
 *
 * @macro: M_COMPARE (x,y)
 *
 * @description:
 *    returns -1, 0 or 1 depending on the comparison of 'x' and 'y'
 */
#define  M_COMPARE(x,y)  ( (x) == (y) ? 0 : ( (x) > (y) ? 1 : -1 ) )


 /* */

/************************************************************************
 *
 * @macro: M_UNROLL4 (statement,count)
 *
 * @description:
 *    this macro lets you unroll loops 4-fold. This has the potential
 *    to increase their performance on modern processors, but never forget
 *    that this is mostly a processor and compiler specific issue
 *
 *    here's a simple example:
 *
 *    {
 *      M_Int  my_str_compare( M_CString  a,
 *                             M_CString  b )
 *      {
 *        M_CString  s = a;
 *        M_CString  t = b;
 *        M_Int      i = 0;
 *
 *        M_UNROLL4( {
 *
 *          if ( !(i = (*s - *t)) )
 *            break;
 *
 *          if ( !*t )
 *            break;
 *
 *          s++;
 *          t++;
 *
 *        }, size );
 *
 *        return i;
 *      }
 *    }
 *
 * @note:
 *    be careful that you won't be able to trace within the unrolled loop
 *    with a debugger, so only use this when you're certain that your loop
 *    is bug-free !!
 *
 */
#define  M_UNROLL4( _statement, _count )               \
          M_BEGIN_STMNT                                \
            M_Int   __count      = (_count);           \
                                                       \
            if ( __count > 0 )                         \
            {                                          \
              M_Int   __loop_count = (__count+3) >> 2; \
                                                       \
              switch ( __count & 3 )                   \
              {                                        \
                case 0: do { _statement;               \
                case 3:      _statement;               \
                case 2:      _statement;               \
                case 1:      _statement;               \
                default:     ;                         \
                } while ( --__loop_count > 0 );        \
              }                                        \
            }                                          \
          M_END_STMNT


/************************************************************************
 *
 * @macro: M_UNROLL8 (statement,count)
 *
 * @description:
 *    this macro lets you unroll loops 8-fold. see @M_UNROLL4 too
 */
#define  M_UNROLL8( _statement, _count )               \
          M_BEGIN_STMNT                                \
            M_Int   __count      = (_count);           \
                                                       \
            if ( __count > 0 )                         \
            {                                          \
              M_Int   __loop_count = (__count+7) >> 3; \
                                                       \
              switch ( __count & 7 )                   \
              {                                        \
                case 0: do { _statement;               \
                case 7:      _statement;               \
                case 6:      _statement;               \
                case 5:      _statement;               \
                case 4:      _statement;               \
                case 3:      _statement;               \
                case 2:      _statement;               \
                case 1:      _statement;               \
                default:     ;                         \
                } while ( --__loop_count > 0 );        \
              }                                        \
            }                                          \
          M_END_STMNT


/************************************************************************
 *
 * @macro: M_UNROLL2 (statement,count)
 *
 * @description:
 *    this macro lets you unroll loops 2-fold. see @M_UNROLL4 for
 *    more details
 */
#define  M_UNROLL2( _statement, _count )               \
          M_BEGIN_STMNT                                \
            M_Int   __count      = (_count);           \
                                                       \
            if ( __count > 0 )                         \
            {                                          \
              M_Int   __loop_count = (__count+1) >> 1; \
                                                       \
              switch ( __count & 1 )                   \
              {                                        \
                case 0: do { _statement;               \
                case 1:      _statement;               \
                default:     ;                         \
                } while ( --__loop_count > 0 );        \
              }                                        \
            }                                          \
          M_END_STMNT


/************************************************************************
 *
 * @macro: M_UNROLL0 (statement,count)
 *
 * @description:
 *    this macro can be used like @M_UNROLL4, @M_UNROLL8 or @M_UNROLL2
 *    but doesn't unroll the loop. This is useful to experiment with
 *    loop unrolling without modifying too much of the source code
 *
 *    of course, you won't be able to trace through the statemnt with
 *    most debuggers..
 */
#define  M_UNROLL0( _statement, _count )             \
          M_BEGIN_STMNT                              \
            M_Int   __loop_count = (_count);         \
                                                     \
            if ( __loop_count > 0 ) do               \
            {                                        \
              _statement;                            \
            }                                        \
            while ( --__loop_count > 0 );            \
          M_END_STMNT

 /* */

#define  M_REPEAT4( _statement )            \
           M_BEGIN_STMNT                    \
             for (;;)                       \
             {                              \
               _statement;                  \
               _statement;                  \
               _statement;                  \
               _statement;                  \
             }                              \
           M_END_STMNT


#define  M_REPEAT2( _statement )            \
           M_BEGIN_STMNT                    \
             for (;;)                       \
             {                              \
               _statement;                  \
               _statement;                  \
             }                              \
           M_END_STMNT


#define  M_REPEAT1( _statement )            \
           M_BEGIN_STMNT                    \
             for (;;)                       \
             {                              \
               _statement;                  \
             }                              \
           M_END_STMNT




#define  M_MAGIC_DEATH  0xdeadbeef

#endif /* __MLIB_MACROS_H__ */

