#ifndef __MLIB_MEMORY_H__
#define __MLIB_MEMORY_H__

#ifndef __MLIB_H__
#error  "do not include this file directly, use <mlib.h> instead"
#endif

#include <string.h>

M_BEGIN_HEADER

/*************************************************************************
 *
 * <Section> memory
 *
 * <Title> Memory
 *
 * <Abstract> Dealing with memory
 *
 * <Description>
 *    The MLib memory management interface
 *
 * <Order>
 *   M_Memory
 *   m_memory_new
 *   m_memory_destroy
 *   m_memory_ref
 *   m_memory_unref
 *
 *   M_MEM_SET
 *   M_MEM_COPY
 *   M_MEM_MOVE
 *
 *   m_mem_alloc
 *   m_mem_realloc
 *   m_mem_free
 *   m_mem_push
 *   m_mem_pop
 *   m_mem_pop_free
 *   m_alloc_push
 *
 *   M_ALLOC
 *   M_FREE
 *   M_REALLOC
 *   M_NEW
 *   M_NEW_ARRAY
 *   M_RENEW_ARRAY
 *
 *   M_ALLOC_PUSH
 *   M_POP
 *   M_POP_FREE
 *
 *   m_mem_qalloc
 *   m_mem_free
 *   m_mem_qrealloc
 */

 /********************************************************************/
 /********************************************************************/
 /*****                                                          *****/
 /*****                USEFUL MEMORY-RELATED MACROS              *****/
 /*****                                                          *****/
 /********************************************************************/
 /********************************************************************/


 /***************************************************************
  *
  * @macro: M_MEM_SET ( destination, byte, count )
  *
  * @description:
  *    a convenient macro to call the ANSI "memset"
  */
#define    M_MEM_SET( dest, byte, count )    \
              memset( (char*)(dest), (char)(byte), count )


 /***************************************************************
  *
  * @macro: M_MEM_COPY ( destination, source, count )
  *
  * @description:
  *    a convenient macro to call the ANSI "memcpy"
  */
#define    M_MEM_COPY( dest, source, count )   \
              memcpy( (char*)(dest), (char*)(source), count )


 /***************************************************************
  *
  * @macro: M_MEM_MOVE ( destination, source, count )
  *
  * @description:
  *    a convenient macro to call the ANSI "memmove"
  */
#define    M_MEM_MOVE( dest, source, count )   \
              memmove( (char*)(dest), (char*)(source), count )


 /***************************************************************
  *
  * @macro: M_MEM_ZERO( destination, count )
  *
  * @description:
  *    a convenient macro to call the ANSI "bzero", used to clear
  *    memory..
  */
#define    M_MEM_ZERO( dest, count )   \
              M_MEM_SET( (char*)(dest), 0, count )


 /***************************************************************
  *
  * @macro: M_ZERO (ptr)
  *
  * @description:
  *    a convenient macro used to clear the memory addressed by
  *    a _typed_ pointer..
  */
#define  M_ZERO(p)       M_MEM_ZERO( (p), sizeof(*(p)) )


 /***************************************************************
  *
  * @macro: M_ZERO_ARRAY (ptr,count)
  *
  * @description:
  *    a convenient macro used to clear an array addressed by a
  *    _typed_ pointer.
  *
  *   'count' is the number of items in the array. the size of each
  *   item is computed with 'sizeof(*ptr)'
  */
#define  M_ZERO_ARRAY(p,count)   M_MEM_ZERO( (p), (count)*sizeof(*(p)) )


 /***************************************************************
  *
  * @macro: M_COPY_ARRAY ( destination, source, count )
  *
  * @description:
  *    a convenient macro used to copy arrays using _typed_ pointers.
  *    'count' is the number of items in the array. The size of each
  *    item is given by 'sizeof(*source)'
  */
#define  M_COPY_ARRAY(dest,src,count)   \
                         M_MEM_COPY( dest, src, (count)*sizeof(*(src)) )


 /***************************************************************
  *
  * @macro: M_MOVE_ARRAY ( destination, source, count )
  *
  * @description:
  *    a convenient macro used to move arrays using _typed_ pointers.
  *    'count' is the number of items in the array. The size of each
  *    item is given by 'sizeof(*source)'
  */
#define  M_MOVE_ARRAY(dest,src,count)   \
                         M_MEM_MOVE( dest, src, (count)*sizeof(*(src)) )



 /********************************************************************/
 /********************************************************************/
 /*****                                                          *****/
 /*****                MEMORY MANAGER DEFINITIONS                *****/
 /*****                                                          *****/
 /********************************************************************/
 /********************************************************************/


 /***************************************************************
  *
  * @struct: M_MemoryRec
  *
  * @description:
  *    a memory manager object structure
  *
  * @fields:
  *    mcore     :: handle to root MLib core object, can be NULL.
  *
  *    impl      :: this pointer is ignored by the MLib, and can be
  *                 used to place a pointer to implementation-dependent
  *                 data.
  *
  * @note:
  *    each new manager created with @m_memory_new will have its "mcore"
  *    field set to 0. An application should _never_ set or change it.
  *
  *    instead, the function @m_core_new (and @mlib_init) will set it to
  *    the appropriate value..
  *
  * @also:
  *    @M_Memory, @M_Memory_Flag, @M_Memory_AllocFunc, @M_Memory_ReallocFunc,
  *    @M_Memory_FreeFunc, @M_Memory_DestroyFunc,
  *
  *    @m_mem_alloc, @m_mem_realloc, @m_mem_free,
  *    @m_mem_destroy,
  *
  *    @m_mem_new_std
  */
  typedef struct M_MemoryRec_
  {
    M_Core         mcore;
    M_Pointer      impl;

  } M_MemoryRec;


 /***************************************************************
  *
  * @function: m_memory_new
  *
  * @description:
  *   allocate a new _standard_ memory manager that uses the ANSI
  *   malloc/realloc/free
  *
  * @return:
  *    handle to new memory manager object. NULL in case of error
  *
  * @note:
  *    "m_memory_new_custom" can also be used to create a custom
  *    memory manager. it is defined in <mlib/internal/imemory.h>
  */
  MLIB_API(M_Memory)  m_memory_new( void );


 /***************************************************************
  *
  * @function: m_memory_ref
  *
  * @description:
  *    increment the reference count of a given memory manager.
  *
  * @input:
  *    memory   :: handle to source memory manager
  *
  * @also:
  *    @M_Memory, @M_MemoryRec, @m_memory_new_std
  */
  MLIB_API(void)       m_memory_ref( M_Memory   memory );


 /***************************************************************
  *
  * @function: m_memory_unref
  *
  * @description:
  *    decrement the reference count of a given memory manager
  *    if the count reaches 0 or below, the manager is destroyed
  *
  * @input:
  *    memory   :: handle to source memory manager.
  *
  * @also:
  *    @M_Memory, @M_MemoryRec, @m_memory_new
  */
  MLIB_API(void)       m_memory_unref( M_Memory   memory );


 /***************************************************************
  *
  * @function: m_memory_assert
  *
  * @description:
  *    this function checks a memory manager handle, and panic
  *    if it is invalid..
  *
  * @input:
  *    memory   :: handle to source memory manager.
  */
  MLIB_API(void)       m_memory_assert( M_Memory     memory,
                                        const char*  file,
                                        long         lineno,
                                        const char*  funcname );

 /***************************************************************
  *
  * @macro: m_assert_is_memory (memory)
  *
  * @description:
  *    check a memory handle
  */
#if defined(M_DEBUG_ERROR) || defined(M_DEBUG_LOG)

#  define m_assert_is_memory(m)  \
              m_memory_assert( m, __FILE__, __LINE__, M__FUNC__ )

#else

#  define m_assert_is_memory(m)  M_UNUSED(m)

#endif


 /********************************************************************/
 /********************************************************************/
 /*****                                                          *****/
 /*****                MEMORY MANAGER INTERFACE                  *****/
 /*****                                                          *****/
 /********************************************************************/
 /********************************************************************/

 /***************************************************************
  *
  * @exception: m_err_memory_alloc
  *
  * @description:
  *    this exception is raised by _many_ functions whenever a
  *    call to @m_mem_alloc or @m_mem_realloc failed due to lack
  *    of available memory.
  *
  *    note that it can also appear in a call to @m_core_push and
  *    its variants (i.e. @m_mem_push, @m_str_push, etc..)
  */
  MLIB_VAR( M_Exception )  m_err_memory_alloc;



 /***************************************************************
  *
  * @function: m_mem_alloc
  *
  * @description:
  *    allocates a single memory block.
  *
  *    throws an exception in case of memory shortage, and always
  *    return a zero-ed block.
  *
  * @input:
  *    memory   :: handle to source memory manager
  *    size     :: size of new block in bytes
  *
  * @return:
  *    a pointer to the new memory block. NULL iff size == 0
  *
  * @throws:
  *    when there is not enough memory available to satisfy the request,
  *    this function will throw @m_err_memory_alloc if the memory manager
  *    is being used by a MLib core. Otherwise, it will simply panic
  *    and abort the current program..
  *
  * @note:
  *    the returned memory block is always zero-filled !!
  *    use @m_mem_qalloc if you don't want exceptions and zero-fill
  *
  *    @m_mem_alloc_push can also be used to allocate a new block and
  *    push it on the cleanup stack
  *
  * @also:
  *    @M_Memory, @M_MemoryRec, @m_mem_realloc, @m_mem_free,
  */
  MLIB_API(M_Pointer)    m_mem_alloc( M_Size     size,
                                      M_Memory   memory );



 /***************************************************************
  *
  * @function: m_mem_realloc
  *
  * @description:
  *    reallocates a given memory block
  *
  * @inout:
  *    apointer :: address of memory block. unchanged in
  *                case of error
  * @input:
  *    cur_size :: current size of block in bytes
  *    new_size :: new size of block in bytes
  *    memory   :: handle to source memory manager
  *
  * @throws: m_err_memory_alloc in case of shortage
  *
  * @note:
  *   note that this function _requires_ the old block size
  *
  *   if you're expanding a memory block, its new content will
  *   always be zero-filled..
  *
  * @also:
  *    @M_Memory, @M_MemoryRec, @m_mem_alloc, @m_mem_free
  */
  MLIB_API(void)    m_mem_realloc( M_Pointer *apointer,
                                   M_Size     cur_size,
                                   M_Size     new_size,
                                   M_Memory   memory );


 /***************************************************************
  *
  * @function: m_mem_free
  *
  * @description:
  *    a function used to free a block of memory from a given memory
  *    manager.
  *
  * @input:
  *    block   :: address of target block. can be NULL
  *    memory  :: handle to source memory manager
  */
  MLIB_API(void)       m_mem_free( M_Pointer  block,
                                   M_Memory   memory );



 /***************************************************************
  *
  * @function: m_mem_dup
  *
  * @description:
  *    duplicate a given block of memory (allocate a new one, copy
  *    old content to it)
  *
  * @input:
  *    block    :: source address
  *    size     :: length in bytes to copy
  *    memory   :: handle to source memory manager
  *    push     :: boolean. it true, the new block is pushed on
  *                top of the cleanup stack
  *
  * @return:
  *    new memory block
  *
  * @also:
  *    @M_Memory, @M_MemoryRec, @m_mem_alloc, @m_mem_realloc, @m_mem_free
  */
  MLIB_API(M_Pointer)  m_mem_dup( M_Pointer  block,
                                  M_Size     size,
                                  M_Memory   memory,
                                  M_Bool     push );


 /***************************************************************
  *
  * @function: m_mem_batch_alloc
  *
  * @description:
  *    allocate several memory blocks at once and throw an exception
  *    if this fails.
  *
  * @input:
  *    memory   :: handle to source memory manager
  *    count    :: memory blocks number
  *    sizes    :: array of memory block sizes
  *
  * @output:
  *    pointers :: array of memory block pointer addresses.
  *
  * @throws: @m_err_mem_alloc
  *
  * @note:
  *    this function only returns when all memory blocks could
  *    be allocated.
  *
  *    otherwise, an exception is raised (and blocks that were
  *    allocated before the memory shortage are freed).
  */
  MLIB_API(void)  m_mem_batch_alloc( M_UInt      count,
                                     M_Size*     sizes,
                                     M_Pointer*  pointers,
                                     M_Memory    memory );


 /***************************************************************
  *
  * @function: m_mem_batch_free
  *
  * @description:
  *    release several memory blocks at once
  *
  * @input:
  *    count    :: memory blocks number
  *    pointers :: array of memory block pointers
  *    memory   :: handle to source memory manager
  *
  * @note:
  *    all memory block pointers are set to NULL after the call
  */
  MLIB_API(void)  m_mem_batch_free( M_UInt      count,
                                    M_Pointer*  pointers,
                                    M_Memory    memory );


 /***************************************************************
  *
  * @function: m_mem_qalloc
  *
  * @description:
  *    a function used to quickly allocate a block of memory
  *    from a given memory manager.
  *
  *    This function never throws an exception, and the returned
  *    block is _not_ zero-ed
  *
  * @input:
  *    size    :: size of new block. MUST BE > 0
  *    memory  :: handle to source memory manager
  *
  * @return:
  *    address of new block, of NULL
  *
  * @note:
  *    you are _strongly_ encouraged to use @m_mem_alloc instead since
  *    it will raise an exception in case of memory shortage, or even
  *    macros like @M_ALLOC
  */
  MLIB_API(M_Pointer)  m_mem_qalloc( M_Size    size,
                                     M_Memory  memory );


 /***************************************************************
  *
  * @function: m_mem_batch_qalloc
  *
  * @description:
  *    try to quicly allocate several memory blocks at once.
  *
  *    when this function exits, either all or none of the blocks will
  *    be allocated.
  *
  * @input:
  *    memory   :: handle to source memory manager
  *    count    :: memory blocks number
  *    sizes    :: array of memory block sizes
  *
  * @output:
  *    pointers :: array of memory block pointer addresses.
  *                in case of memory shortage, all pointers will be
  *                set to NULL
  *
  * @return:
  *    boolean. TRUE iff all blocks could be allocated
  */
  MLIB_API(M_Error) m_mem_batch_qalloc( M_UInt      count,
                                        M_Size*     sizes,
                                        M_Pointer*  pointers,
                                        M_Memory    memory );


 /***************************************************************
  *
  * @function: m_mem_push
  *
  * @description:
  *    push a given block of memory on top of the current thread's
  *    cleanup stack. It will be freed in case of exception.
  *    use @m_mem_pop to remove it from the stack (and eventually
  *    free it too).
  *
  * @input:
  *    block   :: handle to target block
  *    memory  :: handle to memory manager
  *
  * @note:
  *    this function does nothing if "block" is NULL
  *
  *    this function will panic if the memory manager isn't used by
  *    a MLib context (i.e. "memory->mcore" must not be NULL)
  */
  MLIB_API(void)  m_mem_push ( M_Pointer   block,
                               M_Memory    memory );


 /***************************************************************
  *
  * @function: m_mem_pop
  *
  * @description:
  *    pops a previously pushed memory block from the clean-up
  *    stack. the block isn't destroyed
  *
  * @input:
  *    block   :: memory block address
  *    memory  :: memory handler
  *    action  :: action to perform on pop
  *
  * @note:
  *    this function does nothing if "block" is NULL
  */
  MLIB_API(void)   m_mem_pop  ( M_Pointer    block,
                                M_Memory     memory,
                                M_PopAction  action );


 /***************************************************************
  *
  * @function: m_mem_alloc_push
  *
  * @description:
  *    creates a new memory block and push it on the cleanup
  *    stack.
  *
  * @input:
  *    memory  :: handle to memory manager
  *    size    :: size of new block in bytes
  *
  * @return:
  *    pointer to new memory block. cannot be NULL
  *
  * @raise:
  *    m_err_out_of_memory
  *
  * @note:
  *    this function will panic if the memory manager isn't linked
  *    to a MLib context
  *
  * @also:
  *    use @m_mem_pop_free to release the block, or @m_mem_pop to
  *    pop it from the cleanup stack
  */
  MLIB_API(M_Pointer)  m_mem_alloc_push( M_Size    size,
                                         M_Memory  memory );


 /********************************************************************/
 /********************************************************************/
 /*****                                                          *****/
 /*****                MEMORY MANAGER MACROS                     *****/
 /*****                                                          *****/
 /********************************************************************/
 /********************************************************************/

 /***************************************************************
  *
  * @macro: M_ALLOC ( size )
  *
  * @description:
  *    allocates a new memory block, using the _implicit_ memory
  *    manager handle named "memory" (it must be defined prior to
  *    the macro call).
  *
  * @return:
  *    address of new memory block, just like @m_mem_alloc
  */
#define    M_ALLOC(_size)    m_mem_alloc( _size, memory )



 /***************************************************************
  *
  * @macro: M_REALLOC ( ptr, cur_size, new_size )
  *
  * @description:
  *    reallocates a given memory block whose address is
  *    in the "ptr" variable. Returns an error code or
  *    throws an exception
  *
  * @note:
  *    this macro is more convenient to use than @m_mem_realloc
  *    because it uses an _implicit_ memory manager handle named
  *    "memory" (that must be defined prior to the macro call)
  *    and automatically performas a typecast of "&ptr" to
  *    (M_Pointer*)
  *
  *    note that this macro doesn't return a value, unlike @M_ALLOC
  *
  * @also:
  *    @m_mem_realloc
  */
#define    M_REALLOC(p,cur_size,new_size)  \
             m_mem_realloc( (M_Pointer*)&(p), cur_size, new_size, memory )



 /***************************************************************
  *
  * @macro: M_FREE ( ptr )
  *
  * @description:
  *    reallocates a given memory block whose address is
  *    in the "ptr" variable.
  *
  * @return:
  *    error code. 0 means success
  *
  * @note:
  *    this macro is more convenient than @m_mem_free
  *    because it uses an _implicit_ memory manager handle named
  *    "memory" and automatically performs a typecast of "&ptr"
  *    to (M_Pointer*)
  *
  *    this also means that "ptr" will always be NULL after the
  *    M_FREE
  *
  * @also:
  *    @m_mem_alloc, @m_mem_realloc, @M_ALLOC, @M_REALLOC, @M_NEW
  */
#define    M_FREE(p)   M_BEGIN_STMNT                             \
                         if (p)                                  \
                         {                                       \
                           m_mem_free( M_POINTER(p), memory );   \
                           (p) = NULL;                           \
                         }                                       \
                       M_END_STMNT


 /***************************************************************
  *
  * @macro: M_BATCH_ALLOC ( count, sizes, pointers )
  *
  * @description:
  *    this macro is used as a short-cut to @m_mem_batch_alloc
  */
#define    M_BATCH_ALLOC(_count,_sizes,_pointers)   \
             m_mem_batch_alloc( (_count), (_sizes), (_pointers), memory )


 /***************************************************************
  *
  * @macro: M_NEW ( ptr )
  *
  * @description:
  *    this macro is used to allocate a new object. the type of
  *    "ptr" is used automatically to determine the size in bytes
  *    of the new object
  *
  * @also:
  *    @M_ALLOC, @M_FREE
  */
#define    M_NEW(p)     M_NEW_ARRAY( p, 1 )


 /***************************************************************
  *
  * @macro: M_NEW_ARRAY ( ptr, count )
  *
  * @description:
  *    this macro is used to allocate a new table of "count" objects.
  *    the type of "ptr" is used automatically to determine the size
  *    in bytes of each element
  *
  * @return:
  *    error code. 0 means success
  *
  * @also:
  *    @M_NEW, @M_ALLOC, @M_FREE
  */
#define    M_NEW_ARRAY(p,count)                            \
               M_BEGIN_STMNT                               \
                 (p) = M_ALLOC( (count)*sizeof(*(p)) );    \
               M_END_STMNT


 /***************************************************************
  *
  * @macro: M_RENEW_ARRAY ( ptr, old_count, new_count )
  *
  * @description:
  *    this macro is used to reallocate a table of "old_count" elements
  *    into a table of "new_count" elements.
  *
  *    the type of "ptr" is used automatically to determine the size
  *    in bytes of each element
  *
  * @also:
  *    @M_NEW, @M_NEW_ARRAY, @M_REALLOC, @M_FREE
  */
#define    M_RENEW_ARRAY(p,old,new) \
               M_REALLOC(p,(old)*sizeof(*(p)),(new)*sizeof(*(p)))


 /***************************************************************
  *
  * @macro: M_ALLOC_PUSH ( size )
  *
  * @description:
  *    a convention macro that calls @m_mem_alloc_push with an
  *    _implicit_ 'memory' manager variable.
  *
  * @return:
  *    address of new memory block, just like @m_mem_alloc_push
  */

#define  M_ALLOC_PUSH(_size)   m_mem_alloc_push( _size, memory )


 /***************************************************************
  *
  * @macro: M_NEW_PUSH ( ptr, size )
  *
  * @description:
  *    this macro is used to allocate a new "object" then push it
  *    on the cleanup stack. the type of "ptr" is used automatically
  *    to determine the size in bytes of the new object.
  */

#define  M_NEW_PUSH(p)    M_NEW_ARRAY_PUSH( p, 1 )


 /***************************************************************
  *
  * @macro: M_NEW_ARRAY_PUSH ( ptr, size )
  *
  * @description:
  *    this macro is used to allocate a new array of objects, then
  *    push it on the cleanup stack. the type of "ptr" is used
  *    automatically to determine the size in bytes of each array
  *    element.
  */
#define  M_NEW_ARRAY_PUSH(p,count)                           \
             M_BEGIN_STMNT                                   \
               (p) = M_ALLOC_PUSH( (count)*sizeof(*(p)) );   \
             M_END_STMNT


 /***************************************************************
  *
  * @macro: M_MEM_POP ( ptr )
  *
  * @description:
  *    pop a memory block from the cleanup stack. Do _not_ release
  *    it.
  *
  * @note:
  *    this macro should only be used as a shortcut to @m_mem_pop.
  *    it cannot pop other types of objects from the cleanup stack..
  */
#define  M_MEM_POP(p)        m_mem_pop( p, memory, TRUE )


 /***************************************************************
  *
  * @macro: M_MEM_POP_FREE ( ptr )
  *
  * @description:
  *    pops a memory block from the cleanup stack, then release it.
  *
  * @note:
  *    "ptr" will always be set to NUMM after the call..
  */
#define  M_MEM_POP_FREE(p)                   \
           M_BEGIN_STMNT                     \
             m_mem_pop( p, memory, FALSE );  \
             (p) = NULL;                     \
           M_END_STMNT



  /* */

#define  M_QALLOC(_p,_size)                         \
           M_BEGIN_STMNT                            \
             (_p) = m_mem_qalloc( _size, memory );  \
           M_END_STMNT

#define  M_QNEW(p)    M_QALLOC( p, sizeof(*(p)) )

  /* */

M_END_HEADER

#endif /* __MLIB_MEMORY_H__ */

