#ifndef __MLIB_MEMORY_POOL_H__
#define __MLIB_MEMORY_POOL_H__

#include <mlib.h>

M_BEGIN_HEADER

 /*************************************************************************
  *
  * @section: memory_pools
  *
  * @title: Memory Pools
  *
  * @abstract: Fast fixed-size memory allocators
  *
  * @description:
  *   an MLib memory pool is a fast memory allocator for fixed-size blocks.
  *
  *   note that the reason why memory pools are so fast is due to two facts:
  *   first, blocks are allocated in batches from the heap. Second, once
  *   allocated, a block will never be released to the heap until the pool
  *   itself is destroyed.
  *
  *   never forget that modern implementations of the standard "malloc"
  *   and "free" functions are already very fast. Using memory pools
  *   is really handy when you want to absolutely reduce fragmentation,
  *   and need to perform _lots_ of consecutive allocs and frees..
  *
  *   the MLib uses memory pools to manage list nodes, hash nodes, tree nodes
  *   and other frequently used fixed-sized objects..
  */

 /*************************************************************************
  *
  * @type: M_MemoryPool
  *
  * @description:
  *   handle to a memory pool object. A memory pool is used to allocate
  *   and release memory blocks of the same size extremely rapidly. This
  *   is useful to manage things like list nodes, hash nodes, etc..
  */
  typedef struct M_MemoryPoolRec_*   M_MemoryPool;


 /*************************************************************************
  *
  * @function: m_memory_pool_find
  *
  * @description:
  *   find a memory pool that stores memory blocks of a given size
  *
  * @input:
  *   memory       :: memory manager handle
  *   element_size :: element size for the memory pool
  *
  * @return:
  *   memory pool handle, can be NULL
  */
  MLIB_API(M_Memory)
  m_memory_pool_find( M_Memory  memory,
                      M_Size    element_size );


 /*************************************************************************
  *
  * @function: m_memory_pool_collect
  *
  * @description:
  *   perform garbage collection on all memory pools managed by the
  *   MLib's main memory manager
  *
  * @input:
  *   memory :: memory manager handle
  */
  MLIB_API( void )
  m_memory_pool_collect( M_Memory  memory );

  /* */

M_END_HEADER

#endif /* __MLIB_MEMORY_POOL_H__ */
