#ifndef __MLIB_OBJECT_H__
#define __MLIB_OBJECT_H__

#ifndef __MLIB_H__
#error  "do not include this file directly, use <mlib.h> instead"
#endif

M_BEGIN_HEADER

 /************************************************************************
  *
  * @section: objects
  *
  * @title: Objects
  *
  * @abstract: Simple Dynamic Object Sub-System
  *
  * @description:
  *   the MLib provides a simple dynamic object model that applications
  *   can use to increase code re-use.
  *
  *   Each MLib object is a class instance. each class is a dynamically
  *   allocated structure described by a constant "class type" record
  *
  * @order:
  *   M_Object
  *   M_SObject
  *   M_ObjectRec
  *   M_SObjectRec
  *
  *   M_Class
  *
  *   M_ClassType
  *   M_ClassTypeRec
  *
  *   M_Class_InitFunc
  *   M_Class_DoneFunc
  *   M_Class_Flags
  *
  *   M_Object_InitFunc
  *   M_Object_CopyFunc
  *   M_Object_IncrFunc
  *   M_Object_DoneFunc
  *
  *   M_ClassRec
  *
  */

 /************************************************************************/
 /************************************************************************/
 /*****                                                              *****/
 /*****                IMPORTANT TYPES AND STRUCTURES                *****/
 /*****                                                              *****/
 /************************************************************************/
 /************************************************************************/

 /************************************************************************
  *
  * @type: M_Object
  *
  * @description:
  *   handle to a @M_ObjectRec structure, models an object (i.e. a
  *   class instance).
  */
  typedef struct M_ObjectRec_*            M_Object;


 /************************************************************************
  *
  * @type: M_SObject
  *
  * @description:
  *   handle to a @M_SObjectRec structure, used to model a "small" object
  *   (i.e. a class instance without any reference count).
  */
  typedef struct M_SObjectRec_*           M_SObject;


 /************************************************************************
  *
  * @type: M_Class
  *
  * @description:
  *   handle to a constant @M_ClassRec structure, used to model a class
  *   to the object system. Note that classes are _objects_ too and are
  *   allocated by the object sub-system.
  *
  *   classes are described by @M_ClassType handles.
  */
  typedef struct M_ClassRec_*       M_Class;


 /************************************************************************
  *
  * @type: M_ClassType
  *
  * @description:
  *   handle to a constant @M_ClassTypeRec structure, used to describe a
  *   given class to the object system.
  */
  typedef const struct M_ClassTypeRec_*   M_ClassType;


 /************************************************************************
  *
  * @functype: M_Object_InitFunc
  *
  * @description:
  *   a function used to initialise a given object (i.e. class instance)
  *
  * @input:
  *   object :: handle to target @M_Object
  *
  * @throws: any exception
  */
  typedef void  (*M_Object_InitFunc)( M_Object   object );


 /************************************************************************
  *
  * @functype: M_Object_DoneFunc
  *
  * @description:
  *   a function used to finalize a given object (i.e. class instance)
  *
  * @input:
  *   object :: handle to target @M_Object
  *
  * @result:
  *   boolean. true when the parent object finalizer must _not_ be called
  */
  typedef M_Bool  (*M_Object_DoneFunc)( M_Object   object );


 /************************************************************************
  *
  * @functype: M_Object_CopyFunc
  *
  * @description:
  *   a function used when an object is copied. @m_object_copy will
  *   start by copying the root object instance, then call this method
  *   in order to manage child pointers / references
  *
  *   note that this will never be called for instances of classes that
  *   have the @M_CLASS_FLAG_SINGLETON flag set.
  *
  * @input:
  *   source :: handle to source object
  *   copy   :: handle to copy
  *
  * @note:
  *   this function should _only_ modify the copy, never the source
  *
  * @throws: any exception
  */
  typedef void     (*M_Object_CopyFunc)( M_Object   source,
                                         M_Object   copy );


 /************************************************************************
  *
  * @functype: M_Object_IncrFunc
  *
  * @description:
  *   a function used to increment or decrement the reference count of a
  *   given object.
  *
  *   this can be used to implement locking/unlocking during the reference
  *   count updates of certain classes.
  *
  * @input:
  *   object    :: handle to target object
  *   increment :: count increment. can only be 1 or -1
  *
  * @return:
  *   boolean. if true, indicates that the object must be destroyed.
  *
  * @note:
  *   the return value is only tested during a @m_object_unref call, which
  *   means when "increment" == -1. If the object's reference count is not
  *   0 at this time, the program will panic !!
  *
  * @throws: any exception
  */
  typedef M_Bool   (*M_Object_IncrFunc)( M_Object  object,
                                         M_Int     increment );


 /************************************************************************
  *
  * @functype: M_Class_InitFunc
  *
  * @description:
  *   initialize a new class object
  *
  * @input:
  *   clazz :: handle to target class
  *
  * @note:
  *   before this function is called, the content of the parent's class
  *   is copied into the current one.
  *
  * @throws: any exception
  */
  typedef void     (*M_Class_InitFunc)( M_Class    clazz );


 /************************************************************************
  *
  * @functype: M_Class_DoneFunc
  *
  * @description:
  *   finalizer a given class
  *
  * @input:
  *   clazz :: handle to target class
  *
  * @return:
  *   boolean. true if we don't need to call the parent class's finalizer
  */
  typedef M_Bool  (*M_Class_DoneFunc)( M_Class   clazz );


 /************************************************************************
  *
  * @enum: M_Class_Flags
  *
  * @description:
  *   bit flags used in class type descriptors
  *
  * @values:
  *  M_CLASS_FLAG_NO_COPY ::
  *    this bit flag is set when class instances should not be copiable
  *    directly with a call to @m_object_copy. this doesn't prevent
  *    other copying functions to exist though.. but calling @m_object_copy
  *    will result in a fatal exception
  *
  *  M_CLASS_FLAG_SINGLETON ::
  *    this bit flag is set for class instances that are singleton. In this
  *    case, calling @m_object_copy will call @m_object_ref then return a
  *    pointer to the target object.
  *
  *  M_CLASS_FLAG_ABSTRACT ::
  *    this bit flag is used to indicate an abstract flag, that should never
  *    be implemented
  *
  *  M_CLASS_FLAG_SMALL ::
  *    this bit flag is used to indicate a class whose instances are
  *    "small" objects (see @M_SObject), with no reference count..
  */
  enum
  {
    M_CLASS_FLAG_NO_COPY      = 1,
    M_CLASS_FLAG_SINGLETON    = 2,
    M_CLASS_FLAG_ABSTRACT     = 4,
    M_CLASS_FLAG_SMALL        = 8
  };


 /************************************************************************
  *
  * @struct: M_ClassTypeRec
  *
  * @description:
  *   a structure used to describe a given class to the object system.
  *   class types must be unique, and they're used to allocate and
  *   manage @M_Class objects.
  *
  * @fields:
  *   super        :: pointer to super-class type
  *   name         :: class's name (as normal C string)
  *
  *   class_size   :: size of class in bytes
  *   class_init   :: class initialiser
  *   class_done   :: class finalizer
  *
  *   object_size  :: size of class instance in bytes
  *   object_init  :: object initializer function
  *   object_copy  :: object copying function
  *   object_ref   :: object reference counting function
  *   object_done  :: object finalizer function
  */
  typedef struct M_ClassTypeRec_
  {
    M_ClassType         super;
    M_CString           name;

    M_Size              class_size;
    M_UInt              class_flags;
    M_Class_InitFunc    class_init;
    M_Class_DoneFunc    class_done;

    M_Size              object_size;
    M_Object_InitFunc   object_init;
    M_Object_CopyFunc   object_copy;
    M_Object_IncrFunc   object_incr;
    M_Object_DoneFunc   object_done;

  } M_ClassTypeRec;

 /* */
#define  M_CLASS_TYPE(x)  ((M_ClassType)(x))

 /************************************************************************/
 /************************************************************************/
 /*****                                                              *****/
 /*****                           OBJECTS                            *****/
 /*****                                                              *****/
 /************************************************************************/
 /************************************************************************/

 /************************************************************************
  *
  * @struct: M_SObjectRec
  *
  * @description:
  *   a structure used to describe a given "small" class instance.
  *   see @M_ObjectRec for normal objects
  *
  * @fields:
  *   clazz     :: handle to object's class
  *
  * @note:
  *   the class type corresponding to these instances must have the
  *   the @M_CLASS_FLAG_SMALL bit-field set in the 'class_flags' field
  *   of their @M_ClassTypeRec structure..
  */
  typedef struct M_SObjectRec_
  {
    M_Class  clazz;

  } M_SObjectRec;


 /************************************************************************
  *
  * @struct: M_ObjectRec
  *
  * @description:
  *   a structure used to describe a given class instance.
  *
  * @fields:
  *   clazz     :: handle to object's class
  *   ref_count :: object's reference count
  */
  typedef struct M_ObjectRec_
  {
    M_Class   clazz;
    M_Int     ref_count;

  } M_ObjectRec;

 /* */
#ifdef M_DEBUG_COMPILE_ASSERTS
#  define  M_OBJECT_CHECK_TYPE(o,t)  (m_object_assert_is_a(o,t,M__FUNC__,__LINE__,__FILE__))
#else
#  define  M_OBJECT_CHECK_TYPE(o,t)  (o)
#endif

  MLIB_API(M_Pointer)
  m_object_assert_is_a( M_Pointer    object,
                        M_ClassType  class_type,
                        const char*  function_name,
                        long         line_number,
                        const char*  source_file );


#define  m_assert_is_object(o)  (void)M_OBJECT_CHECK_TYPE(o,NULL)


 /************************************************************************
  *
  * @macro: M_OBJECT (o)
  *
  * @description:
  *   a convenient macro used to convert a pointer to a @M_Object.
  */
#define  M_OBJECT(x)    ((M_Object)(x))


 /************************************************************************
  *
  * @macro: M_OBJECT_CHK (o)
  *
  * @description:
  *   a convenient macro used to convert a pointer to a @M_Object.
  *   on debug builds, it will panic if the parameter is not an object
  *   handle..
  */
#define  M_OBJECT_CHK(o)    M_OBJECT(M_OBJECT_CHECK_TYPE(o,NULL))


 /************************************************************************
  *
  * @function: m_new
  *
  * @description:
  *   create a new object (i.e. class instance), from a given class
  *
  * @input:
  *   class :: handle to object's class
  *
  * @return:
  *   handle to new object. cannot be NULL
  *
  * @throws: any exception (e.g. @m_err_memory_alloc)
  *
  * @note:
  *   the new object's reference count is set to 1
  *
  *   you can use @m_class_from_type to retrieve a class handle
  *   from a given class type..
  */
  MLIB_API(M_Pointer)
  m_new( M_Class   clazz );


  MLIB_API(M_Pointer)
  m_new_full( M_Class    clazz,
              M_Pointer  idata );

 /************************************************************************
  *
  * @function: m_new_from_type
  *
  * @description:
  *   create a new object (i.e. class instance), given its class type
  *
  * @input:
  *   mcore      :: MLib core handle
  *   class_type :: class type handle
  *   class_data :: used when creating the
  *
  *
  * @throws: any exception (e.g. @m_err_memory_alloc)
  *
  * @note:
  *   the new object's reference count is set to 1
  *
  *   this function is slightly slower that @m_object_new, but doesn't
  *   require a pre-determined class handle.
  */
  MLIB_API(M_Pointer)
  m_new_from_type( M_Core       mcore,
                   M_ClassType  class_type,
                   M_Pointer    class_data );


 /************************************************************************
  *
  * @function: m_ref
  *
  * @description:
  *   increment a given object's reference count
  *
  * @input:
  *   object :: handle to target @M_Object
  *
  * @note:
  *   this function will panic if it is called on a @M_SObject instance
  */
  MLIB_API(void)
  m_ref( M_Pointer object );


 /************************************************************************
  *
  * @function: m_unref
  *
  * @description:
  *   decrement a given object's reference count. it it reaches 0, the
  *   object is destroyed automatically..
  *
  * @input:
  *   object :: handle to target @M_Object
  *
  * @note:
  *   this function will always destroy @M_SObject instances
  */

  MLIB_API(void)
  m_unref( M_Pointer object );


 /************************************************************************
  *
  * @function: m_copy
  *
  * @description:
  *   copy a given object.
  *
  * @input:
  *   object :: handle to target @M_Object
  *
  * @return:
  *   object copy
  *
  * @throws: any exception (e.g. @m_err_memory_alloc)
  *
  * @note:
  *   the copy's reference count is set to 1 (except for small objects)
  */

  MLIB_API(M_Pointer)
  m_copy( M_Pointer object );



 /************************************************************************
  *
  * @function: m_push
  *
  * @description:
  *   push an object on top of the cleanup stack
  *
  * @input:
  *   object :: handle to target @M_Object
  *
  * @throws: @m_err_memory_alloc
  */
  MLIB_API(void)
  m_push( M_Pointer  object );


 /************************************************************************
  *
  * @function: m_pop
  *
  * @description:
  *   pops an object from the cleanup stack. do not unref/destroy it
  *
  * @input:
  *   object :: handle to target @M_Object
  */
  MLIB_API(void)
  m_pop( M_Pointer  object );


 /************************************************************************
  *
  * @function: m_pop_unref
  *
  * @description:
  *   pops an object from the cleanup stack, then unref it
  *
  * @input:
  *   object :: handle to target @M_Object
  */
  MLIB_API(void)
  m_pop_unref( M_Pointer  object );


 /************************************************************************
  *
  * @function: m_get_memory
  *
  * @description:
  *   retrieve the current memory manager handle from an object
  *
  * @input:
  *   object :: handle to target @M_Object
  *
  * @return:
  *   memory manager handle
  */
  MLIB_API(M_Memory)
  m_get_memory( M_Pointer object );


 /************************************************************************
  *
  * @function: m_get_core
  *
  * @description:
  *   retrieve the current MLib core handle from an object
  *
  * @input:
  *   object :: handle to target @M_Object
  *
  * @return:
  *   MLib core handle
  */
  MLIB_API(M_Core)
  m_get_core( M_Pointer object );


 /************************************************************************
  *
  * @function: m_is_a
  *
  * @description:
  *   checks wether an object is an instance of a given class type
  *
  * @input:
  *   object     :: handle to target @M_Object
  *   class_type :: class type to compare to
  *
  * @return:
  *   boolean. true if the object's class is the one described by
  *   "class_info", or one of its children..
  */
  MLIB_API(M_Bool)
  m_is_a( M_Pointer    object,
          M_ClassType  class_type );


 /************************************************************************
  *
  * @function: m_is_instance_of
  *
  * @description:
  *   checks wether an object is an instance of a given class
  *
  * @input:
  *   object     :: handle to target @M_Object
  *   class_type :: class type to compare to
  *
  * @return:
  *   boolean. true if the object's class is the one described by
  *   "class_info", or one of its children..
  */
  MLIB_API(M_Bool)
  m_is_instance_of( M_Pointer  object,
                    M_Class    clazz );

 /************************************************************************
  *
  * @function: m_object_check
  *
  * @description:
  *   checks at runtime that a pointer is really an object handle
  *   on debug builds, this will abort the program in case of error
  *
  * @input:
  *   object     :: handle to target @M_Object
  */
  MLIB_API(M_Bool)
  m_object_check( M_Pointer  object );


 /************************************************************************
  *
  * @function: m_get_class
  *
  * @description:
  *   retrieve's an object's class
  *
  * @input:
  *   object     :: handle to target @M_Object
  *
  * @return:
  *   the object's class record
  */
#define  m_get_class(o)    M_OBJECT(o)->clazz

/* */




 /************************************************************************/
 /************************************************************************/
 /*****                                                              *****/
 /*****                           CLASSES                            *****/
 /*****                                                              *****/
 /************************************************************************/
 /************************************************************************/

 /************************************************************************
  *
  * @type: M_ClassInfo
  *
  * @description:
  *   opque handle to a @M_ClassInfo structure. used internally by the
  *   object sub-system to store internal typermation about the class
  */
  typedef struct M_ClassInfoRec_*         M_ClassInfo;



 /************************************************************************
  *
  * @struct: M_ClassRec
  *
  * @description:
  *   a structure used to describe a given class. Note that classes are
  *   themselves objects..
  *
  * @fields:
  *   object      :: root object fields
  *   magic       :: magic number used to check classes at runtime
  *   type        :: handle to class type
  *   super       :: handle to super class
  *   memory      :: memory manager handle
  *   mcore       :: MLib core handle
  *
  *   idata       :: this pointer is changed on each instance creation
  *                  see @m_new
  *
  *   info        :: handle to internal data for this class (managed by
  *                  the class pool, unavailable to applications and classes)
  *
  *   object_init :: object initializer..
  *   object_done :: object finalizer for class's instances
  *   object_copy :: object copier for class's instances
  *   object_ref  :: object reference counter for class's instances, if any
  *
  * @note:
  *   normally, the "object_done" and "object_copy" fields are initialised
  *   from the "class_type" content. However, the class initialiser is capable
  *   of changing these functions if it wants to..
  */
  typedef struct M_ClassRec_
  {
    M_SObjectRec        object;
    M_UInt32            magic;
    M_ClassType         type;
    M_Class             super;
    M_Memory            memory;
    M_Core              mcore;
    M_Pointer           idata;
    M_Pointer           info;

    M_Object_InitFunc   object_init;
    M_Object_DoneFunc   object_done;
    M_Object_CopyFunc   object_copy;
    M_Object_IncrFunc   object_incr;

  } M_ClassRec;

 /* */
#ifdef M_DEBUG_COMPILE_ASSERTS
#  define  M_CLASS_CHECK_TYPE(x,t)    (m_class_assert_is_a(M_CLASS(x),t,M__FUNC__,__LINE__,__FILE__))
#else
#  define  M_CLASS_CHECK_TYPE(x,t)    (x)
#endif


  MLIB_API(M_Pointer)
  m_class_assert_is_a( M_Pointer    clazz,
                       M_ClassType  class_type,
                       const char*  function_name,
                       long         line_number,
                       const char*  source_file );

 /************************************************************************
  *
  * @macro: M_CLASS (c)
  *
  * @description:
  *   a convenient macro to convert a pointer to a M_Class.
  */
#define  M_CLASS(x)  ((M_Class)(x))


 /************************************************************************
  *
  * @macro: M_CLASS_CHK (c)
  *
  * @description:
  *   a convenient macro to convert a pointer to a M_Class. On debug builds
  *   this also perform runtime info checks (which will abort the program
  *   if the pointer is not a valid class handle)
  */
#define  M_CLASS_CHK(x)  M_CLASS( M_CLASS_CHECK_TYPE(x,NULL) )


 /************************************************************************
  *
  * @macro: m_assert_is_class (c)
  *
  * @description:
  *   checks at runtime that a given pointer is a class handle
  */
#define  m_assert_is_class(x)  M_CLASS_CHECK_TYPE(x,NULL)


 /************************************************************************
  *
  * @function: m_class_from_type
  *
  * @description:
  *   retrieves the class handle corresponding to a given class type
  *
  * @input:
  *   class_type :: class type handle
  *   mcore      :: MLib code handle
  *
  * @return:
  *   class handle, cannot be NULL
  *
  * @throws: any (e.g. @m_err_memory_alloc..)
  *
  * @note:
  *   it's usually a good idea to store the result to use it with
  *   @m_object_new
  */
  MLIB_API(M_Class)
  m_class_from_type( M_ClassType  class_type,
                     M_Core       mcore );


 /************************************************************************
  *
  * @function: m_class_flush
  *
  * @description:
  *   this function is used to remove a given class from the class pool,
  *   and should seldom be used.
  *
  * @input:
  *   clazz  :: target @M_Class handle
  *
  * @note:
  *   If one or more instances of the class exist, this function will _panic_
  *   and abort the program..
  *
  *   this function should only be used when the class's implementation
  *   isn't available anymore, for example when unloading a dynamic module that
  *   implements a given class, @m_class_flush must be called before
  *   removing its code from memory
  */
  MLIB_API(void)
  m_class_flush( M_Class  clazz );


 /************************************************************************
  *
  * @function: m_class_is_a
  *
  * @description:
  *   checks that a given class corresponds to a given class type
  *
  * @input:
  *   class  :: handle to class
  *   type   :: handle to class type used in the comparison
  *
  * @return:
  *   boolean, true if the class's info, or one of its super infos, correspond
  *   to the class type.
  */
  MLIB_API(M_Bool)
  m_class_is_a( M_Class      clazz,
                M_ClassType  class_type );


 /************************************************************************
  *
  * @function: m_class_is_sub_of
  *
  * @description:
  *   checks that a given class corresponds to a given class type.
  *   if not, aborts the program..
  *
  * @input:
  *   class  :: handle to class
  *   parent :: handle to parent class used in the comparison
  *
  * @return:
  *   boolean. true if "class" is "parent" or any of its childs in the
  *   class hierarchy tree
  */

  MLIB_API(M_Bool)
  m_class_is_sub_of( M_Class  clazz,
                     M_Class  parent );

 /* */

M_END_HEADER

#endif /* __MLIB_OBJECT_H__ */
