#ifndef __MLIB_RING_H__
#define __MLIB_RING_H__

#include <mlib.h>

M_BEGIN_HEADER

 /************************************************************************/
 /************************************************************************/
 /*****                                                              *****/
 /*****               CIRCULAR DOUBLY-LINKED LISTS                   *****/
 /*****                                                              *****/
 /************************************************************************/
 /************************************************************************/

 /***********************************************************************
  *
  * @type: M_Ring
  *
  * @description:
  *    handle to a simple doubly-linked ring node. see @M_RingRec
  */
  typedef struct M_RingRec_*      M_Ring;


 /***********************************************************************
  *
  * @struct: M_RingRec
  *
  * @description:
  *    a simple doubly-linked ring node structure.
  *
  * @fields:
  *    next :: handle to next node in list. NULL if last node
  *    prev :: handle to previous node in list. NULL if first node
  *    data :: generic pointer to listed element
  */
  typedef struct M_RingRec_
  {
    M_Ring     next;
    M_Ring     prev;
    M_Pointer  data;

  } M_RingRec;


 /***********************************************************************
  *
  * @functype: M_Ring_NodeFunc
  *
  * @description:
  *    a function used to process a given ring node. For example, to
  *    delete its data member..
  *
  * @input:
  *    node      :: handle to the target ring node
  *    user_data :: user-provided data
  */
  typedef void  (*M_Ring_NodeFunc)(  M_Ring     node,
                                     M_Pointer  user_data );


 /***********************************************************************
  *
  * @function: m_ring_memory_pool
  *
  * @description:
  *    return a memory pool used to allocate ring nodes very quickly
  *
  * @input:
  *    memory :: memory manager handle
  *
  * @return:
  *    the memory pool handle (or the input memory manager handle)
  */
#define  m_ring_memory_pool(memory)   \
               m_memory_find_pool( memory, sizeof(M_RingRec) )


 /***********************************************************************
  *
  * @function: m_ring_alloc
  *
  * @description:
  *    allocate a new ring node
  *
  * @input:
  *    data   :: data pointer for the new node
  *    memory :: memory handle
  *
  * @return:
  *    handle to new ring node
  *
  * @throws: @m_err_memory_alloc
  */
  MLIB_API(M_Ring)
  m_ring_alloc( M_Pointer  data,
                M_Memory   memory );


 /***********************************************************************
  *
  * @function: m_ring_free
  *
  * @description:
  *    release a given ring node
  *
  * @input:
  *    node    :: handle to target ring node
  *    pool :: handle to list pool
  *
  * @note:
  *    this function doesn't destroy the node's data pointer
  */
#define  m_ring_free(_node,_mem)  m_mem_free(_node,_mem)


 /***********************************************************************
  *
  * @function: m_ring_append
  *
  * @description:
  *    append a new element to a list.
  *
  * @inout:
  *    pring  :: pointer to list head. cannot be NULL !!
  *
  * @input:
  *    data   :: data pointer for the new node
  *    memory :: memory manager
  *
  * @throws: @m_err_memory_alloc
  *
  * @note:
  *   this function is potentially slow because it needs to parse the
  *   whole list before appending the new node. consider using rings
  *   if you need fast appends instead..
  */
  MLIB_API(void)
  m_ring_append( M_Ring*    pring,
                 M_Pointer  data,
                 M_Memory   memory) ;


 /***********************************************************************
  *
  * @function: m_ring_prepend
  *
  * @description:
  *    prepend a new element to a list.
  *
  * @inout:
  *    pring  :: pointer to list head. cannot be NULL !!
  *
  * @input:
  *    data   :: data pointer for the new node
  *    memory :: memory manager
  *
  * @throws: @m_err_memory_alloc
  */
  MLIB_API(void)
  m_ring_prepend( M_Ring*    pring,
                  M_Pointer  data,
                  M_Memory   memory );


 /***********************************************************************
  *
  * @function: m_ring_find
  *
  * @description:
  *    find a ring node with a given "data" pointer
  *
  * @input:
  *    pring  :: pointer to list head. cannot be NULL
  *    data   :: data pointer to look up in list
  *
  * @returns:
  *    ring node handle corresponding to the first occurence of 'data'
  *    in the list. NULL otherwise
  */
  MLIB_API(M_Ring)
  m_ring_find( M_Ring*    pring,
               M_Pointer  data );


 /***********************************************************************
  *
  * @function: m_ring_remove
  *
  * @description:
  *    remove a node corresponding to the first occurence of 'data' from
  *    a given list.
  *
  * @inout:
  *    pring  :: pointer to list head. cannot be NULL
  *
  * @input:
  *    data   :: data pointer to look up in list
  *    memory :: memory manager handle
  */
  MLIB_API(void)
  m_ring_remove( M_Ring*    pring,
                 M_Pointer  data,
                 M_Memory   memory );


 /***********************************************************************
  *
  * @function: m_ring_reverse
  *
  * @description:
  *    reverse a given list
  *
  * @inout:
  *    pring  :: pointer to list head. cannot be NULL
  */
  MLIB_API(void)
  m_ring_reverse( M_Ring*  pring );


 /***********************************************************************
  *
  * @function: m_ring_length
  *
  * @description:
  *    return the number of elements in a list
  *
  * @input:
  *    pring  :: pointer to list head. cannot be NULL
  *
  * @return:
  *    number of elements in the list
  */
  MLIB_API(M_UInt)
  m_ring_length( M_Ring*  pring );


 /***********************************************************************
  *
  * @function: m_ring_nth
  *
  * @description:
  *    returns the nth's node in a list
  *
  * @input:
  *    pring  :: pointer to list head. cannot be NULL
  *    index  :: element index
  *
  * @return:
  *    n'th ring node, NULL for invalid indices
  */
  MLIB_API(M_Ring)
  m_ring_nth( M_Ring*  pring,
              M_UInt   index );


 /***********************************************************************
  *
  * @function: m_ring_index
  *
  * @description:
  *    returns the index of a given element in a list
  *
  * @input:
  *    plist  :: pointer to list head. cannot be NULL
  *    data   :: data to lookup
  *
  * @return:
  *    "data" position in the list, or -1 if not present
  */
  MLIB_API(M_Int)
  m_ring_index( M_Ring*    plist,
                M_Pointer  data );


 /***********************************************************************
  *
  * @function: m_ring_last
  *
  * @description:
  *    return the last node in a list
  *
  * @input:
  *    plist  :: pointer to list head. cannot be NULL
  *
  * @return:
  *    last node, if any
  */
  MLIB_API(M_Ring)
  m_ring_last( M_Ring*  plist );


 /***********************************************************************
  *
  * @function: m_ring_first
  *
  * @description:
  *    return the first node in a list
  *
  * @input:
  *    plist  :: pointer to list head. cannot be NULL
  *
  * @return:
  *    first node, if any
  */
  MLIB_API(M_Ring)
  m_ring_first( M_Ring*  plist );



 /***********************************************************************
  *
  * @function: m_ring_append_node
  *
  * @description:
  *    append a node to a list
  *
  * @inout:
  *    pring  :: pointer to list head. cannot be NULL
  *
  * @input:
  *    node   :: new ring node
  */
  MLIB_API(void)
  m_ring_append_node( M_Ring*  pring,
                      M_Ring   node );


 /***********************************************************************
  *
  * @function: m_ring_prepend_node
  *
  * @description:
  *    prepend a node to a list
  *
  * @inout:
  *    pring  :: pointer to list head. cannot be NULL
  *
  * @input:
  *    node   :: new ring node
  */
  MLIB_API(void)
  m_ring_prepend_node( M_Ring*  pring,
                       M_Ring   node );


 /***********************************************************************
  *
  * @function: m_ring_remove_node
  *
  * @description:
  *    remove a node from a list
  *
  * @inout:
  *    pring  :: pointer to list head. cannot be NULL
  *
  * @input:
  *    node   :: new ring node
  */
  MLIB_API(void)
  m_ring_remove_node( M_Ring*  pring,
                      M_Ring   node );


 /***********************************************************************
  *
  * @function: m_ring_foreach
  *
  * @description:
  *    apply a function on each ring node
  *
  * @input:
  *    pring     :: pointer to list head. cannot be NULL
  *    node_func :: node function
  *    node_data :: user data passed to node function
  *
  * @throws: any (exceptions of node_func only)
  */
  MLIB_API(void)
  m_ring_foreach( M_Ring*          pring,
                  M_Ring_NodeFunc  node_func,
                  M_Pointer        node_data );


 /***********************************************************************
  *
  * @function: m_ring_free_all
  *
  * @description:
  *    free all ring nodes in a list
  *
  * @inout:
  *    pring     :: pointer to list head. cannot be NULL
  *
  * @input:
  *    val_destroy :: optional node data finalizer. can be NULL
  *    val_data    :: user data passed to finalizer
  *    memory      :: memory manager handle
  *
  * @note:
  *    the node finalizer should not free the node, only the object
  *    pointed to by the "data" field.
  */
  MLIB_API(void)
  m_ring_free_all( M_Ring*            pring,
                   M_DestroyDataFunc  val_destroy,
                   M_Pointer          val_data,
                   M_Memory           memory );


 /***********************************************************************
  *
  * @macro: M_RING_LOOP (list,data)
  *
  * @description:
  *    a useful macro used in association with @M_RING_END to parse
  *    all nodes in a list and extract its "data" values
  *
  *    the first parameter is a pointer to the list head.
  *    the second parameter is a local variable name that will
  *    be used within the @M_RING_LOOP .. @M_RING_END statement
  *    to receive each node's "data" value..
  *
  *    here's a small example:
  *
  *    {
  *      M_Ring   numbers = NULL;
  *      M_Int    n;
  *
  *      for ( n = 0; n < 10; n++ )
  *        m_ring_append( &list, M_POINTER_FROM_INT(n), memory );
  *
  *      M_RING_LOOP( &list, value )
  *        printf( "%d ", M_POINTER_TO_INT(value) );
  *      M_RING_END
  *
  *      m_ring_free_all( &list, NULL, NULL, memory );
  *    }
  *
  */
#define  M_RING_LOOP(_pring,_data)                              \
               {                                                \
                 M_Ring     __ring_first = *(_pring);           \
                 M_Ring     __ring_node  = __ring_first;        \
                                                                \
                 if ( __ring_first ) do                         \
                 {                                              \
                   M_Ring     __ring_next = __ring_node->next;  \
                   M_Pointer  _data       = __ring_node->data;


 /***********************************************************************
  *
  * @macro: M_RING_LOOP_NODE (ring,node)
  *
  * @description:
  *    a variant of @M_RING_LOOP_NODE used to expose each ring node
  *    pointer, instead of its "data" field
  */
#define  M_RING_LOOP_NODE(_pring,_node)                         \
               {                                                \
                 M_Ring     __ring_first = *(_pring);           \
                 M_Ring     __ring_node  = __ring_first;        \
                                                                \
                 if ( __ring_first ) do                         \
                 {                                              \
                   M_Ring     __ring_next = __ring_node->next;  \
                   M_Ring     _node       = __ring_node;



 /***********************************************************************
  *
  * @macro: M_RING_END
  *
  * @description:
  *    a useful macro used in association with @M_RING_LOOP to parse
  *    all nodes in a list and extract its "data" values
  */
#define  M_RING_END                                        \
                   __ring_node = __ring_next;              \
                 }                                         \
                 while ( __ring_node != __ring_first );    \
               }

 /* */

 /***********************************************************************
  *
  * @function: m_ring_node_skip
  *
  * @description:
  *    returns the nth's node in a list
  *
  * @input:
  *    node  :: start list node
  *    delta :: increment count (in list list nodes)
  *
  * @return:
  *    new ring node, of NULL if out-of-list
  */
  MLIB_API(M_Ring)
  m_ring_node_skip( M_Ring  node,
                    M_Int   delta );

 /* */

M_END_HEADER

#endif /* __MLIB_RING_H__ */
