#ifndef __MLIB_SLIST_H__
#define __MLIB_SLIST_H__

#include <mlib.h>

M_BEGIN_HEADER

 /***********************************************************************
  *
  * @type: M_SList
  *
  * @description:
  *    handle to a singly-linked list node. see @M_SListRec
  */
  typedef struct M_SListRec_*      M_SList;


 /***********************************************************************
  *
  * @struct: M_SListRec
  *
  * @description:
  *    a simple singly-linked list node structure.
  *
  * @fields:
  *    next :: handle to next node in list. NULL if last node
  *    data :: generic pointer to listed element
  */
  typedef struct M_SListRec_
  {
    M_SList    next;
    M_Pointer  data;

  } M_SListRec;


 /***********************************************************************
  *
  * @functype: M_SList_NodeFunc
  *
  * @description:
  *    a function used to process a given list node. For example, to
  *    delete its data member..
  *
  * @input:
  *    node      :: handle to the target list node
  *    user_data :: user-provided data
  */
  typedef void  (*M_SList_NodeFunc)( M_SList     node,
                                     M_Pointer  user_data );


 /***********************************************************************
  *
  * @function: m_slist_memory_pool
  *
  * @description:
  *    return a memory pool used to allocate list nodes very quickly
  *
  * @input:
  *    memory :: memory manager handle
  *
  * @return:
  *    the memory pool handle (or the input memory manager handle)
  */
#define  m_slist_memory_pool(memory)   \
               m_memory_find_pool( memory, sizeof(M_SListRec) )


 /***********************************************************************
  *
  * @function: m_slist_alloc
  *
  * @description:
  *    allocate a new list node
  *
  * @input:
  *    data   :: data pointer for the new node
  *    memory :: memory handle
  *
  * @return:
  *    handle to new list node
  *
  * @throws: @m_err_memory_alloc
  */
  MLIB_API(M_SList)
  m_slist_alloc( M_Pointer  data,
                 M_Memory   memory );


 /***********************************************************************
  *
  * @function: m_slist_free
  *
  * @description:
  *    release a given list node
  *
  * @input:
  *    node    :: handle to target list node
  *    pool :: handle to list pool
  *
  * @note:
  *    this function doesn't destroy the node's data pointer
  */
#define  m_slist_free(_node,_mem)  m_mem_free(_node,_mem)


 /***********************************************************************
  *
  * @function: m_slist_append
  *
  * @description:
  *    append a new element to a list.
  *
  * @inout:
  *    plist  :: pointer to list head. cannot be NULL !!
  *
  * @input:
  *    data   :: data pointer for the new node
  *    memory :: memory manager
  *
  * @throws: @m_err_memory_alloc
  *
  * @note:
  *   this function is potentially slow because it needs to parse the
  *   whole list before appending the new node. consider using rings
  *   if you need fast appends instead..
  */
  MLIB_API(void)
  m_slist_append( M_SList*   plist,
                  M_Pointer  data,
                  M_Memory   memory) ;


 /***********************************************************************
  *
  * @function: m_slist_prepend
  *
  * @description:
  *    prepend a new element to a list.
  *
  * @inout:
  *    plist  :: pointer to list head. cannot be NULL !!
  *
  * @input:
  *    data   :: data pointer for the new node
  *    memory :: memory manager
  *
  * @throws: @m_err_memory_alloc
  */
  MLIB_API(void)
  m_slist_prepend( M_SList*    plist,
                   M_Pointer  data,
                   M_Memory   memory );


 /***********************************************************************
  *
  * @function: m_slist_find
  *
  * @description:
  *    find a list node with a given "data" pointer
  *
  * @input:
  *    plist  :: pointer to list head. cannot be NULL
  *    data   :: data pointer to look up in list
  *
  * @returns:
  *    list node handle corresponding to the first occurence of 'data'
  *    in the list. NULL otherwise
  */
  MLIB_API(M_SList)
  m_slist_find( M_SList*   plist,
                M_Pointer  data );


 /***********************************************************************
  *
  * @function: m_slist_remove
  *
  * @description:
  *    remove a node corresponding to the first occurence of 'data' from
  *    a given list.
  *
  * @inout:
  *    plist  :: pointer to list head. cannot be NULL
  *
  * @input:
  *    data   :: data pointer to look up in list
  *    memory :: memory manager handle
  */
  MLIB_API(void)
  m_slist_remove( M_SList*   plist,
                  M_Pointer  data,
                  M_Memory   memory );


 /***********************************************************************
  *
  * @function: m_slist_reverse
  *
  * @description:
  *    reverse a given list
  *
  * @inout:
  *    plist  :: pointer to list head. cannot be NULL
  */
  MLIB_API(void)
  m_slist_reverse( M_SList*  plist );


 /***********************************************************************
  *
  * @function: m_slist_length
  *
  * @description:
  *    return the number of elements in a list
  *
  * @input:
  *    plist  :: pointer to list head. cannot be NULL
  *
  * @return:
  *    number of elements in the list
  */
  MLIB_API(M_UInt)
  m_slist_length( M_SList*  plist );


 /***********************************************************************
  *
  * @function: m_slist_nth
  *
  * @description:
  *    returns the nth's node in a list
  *
  * @input:
  *    plist  :: pointer to list head. cannot be NULL
  *    index  :: element index
  *
  * @return:
  *    n'th list node, NULL for invalid indices
  */
  MLIB_API(M_SList)
  m_slist_nth( M_SList*  plist,
               M_UInt    index );


 /***********************************************************************
  *
  * @function: m_slist_index
  *
  * @description:
  *    returns the index of a given element in a list
  *
  * @input:
  *    plist  :: pointer to list head. cannot be NULL
  *    data   :: data to lookup
  *
  * @return:
  *    "data" position in the list, or -1 if not present
  */
  MLIB_API(M_Int)
  m_slist_index( M_SList*   plist,
                 M_Pointer  data );


 /***********************************************************************
  *
  * @function: m_slist_last
  *
  * @description:
  *    return the last node in a list
  *
  * @input:
  *    plist  :: pointer to list head. cannot be NULL
  *
  * @return:
  *    last node, if any
  */
  MLIB_API(M_SList)
  m_slist_last( M_SList*  plist );


 /***********************************************************************
  *
  * @function: m_slist_first
  *
  * @description:
  *    return the first node in a list
  *
  * @input:
  *    plist  :: pointer to list head. cannot be NULL
  *
  * @return:
  *    first node, if any
  */
  MLIB_API(M_SList)
  m_slist_first( M_SList*  plist );


 /***********************************************************************
  *
  * @function: m_slist_append_node
  *
  * @description:
  *    append a node to a list
  *
  * @inout:
  *    plist  :: pointer to list head. cannot be NULL
  *
  * @input:
  *    node   :: new list node
  */
  MLIB_API(void)
  m_slist_append_node( M_SList*  plist,
                       M_SList   node );


 /***********************************************************************
  *
  * @function: m_slist_prepend_node
  *
  * @description:
  *    prepend a node to a list
  *
  * @inout:
  *    plist  :: pointer to list head. cannot be NULL
  *
  * @input:
  *    node   :: new list node
  */
  MLIB_API(void)
  m_slist_prepend_node( M_SList*  plist,
                        M_SList   node );


 /***********************************************************************
  *
  * @function: m_slist_remove_node
  *
  * @description:
  *    remove a node from a list
  *
  * @inout:
  *    plist  :: pointer to list head. cannot be NULL
  *
  * @input:
  *    node   :: new list node
  */
  MLIB_API(void)
  m_slist_remove_node( M_SList*  plist,
                       M_SList   node );


 /***********************************************************************
  *
  * @function: m_slist_foreach
  *
  * @description:
  *    apply a function on each list node
  *
  * @input:
  *    plist     :: pointer to list head. cannot be NULL
  *    node_func :: node function
  *    node_data :: user data passed to node function
  *
  * @throws: any (exceptions of node_func only)
  */
  MLIB_API(void)
  m_slist_foreach( M_SList*          plist,
                   M_SList_NodeFunc  node_func,
                   M_Pointer         node_data );


 /***********************************************************************
  *
  * @function: m_slist_free_all
  *
  * @description:
  *    free all list nodes in a list
  *
  * @inout:
  *    plist     :: pointer to list head. cannot be NULL
  *
  * @input:
  *    val_destroy :: optional node data finalizer. can be NULL
  *    val_data    :: user data passed to node finalizer
  *    memory      :: memory manager handle
  *
  * @note:
  *    the node finalizer should not free the node, only the object
  *    pointed to by the "data" field.
  */
  MLIB_API(void)
  m_slist_free_all( M_SList*          plist,
                    M_DestroyDataFunc val_destroy,
                    M_Pointer         val_data,
                    M_Memory          memory );


 /***********************************************************************
  *
  * @macro: M_SLIST_LOOP (list,data)
  *
  * @description:
  *    a useful macro used in association with @M_SLIST_END to parse
  *    all nodes in a list and extract its "data" values
  *
  *    the first parameter is a pointer to the list head.
  *    the second parameter is a local variable name that will
  *    be used within the @M_SLIST_LOOP .. @M_SLIST_END statement
  *    to receive each node's "data" value..
  *
  *    here's a small example:
  *
  *    {
  *      M_SList   numbers = NULL;
  *      M_Int    n;
  *
  *      for ( n = 0; n < 10; n++ )
  *        m_slist_append( &list, M_POINTER_FROM_INT(n), memory );
  *
  *      M_SLIST_LOOP( &list, value )
  *        printf( "%d ", M_POINTER_TO_INT(value) );
  *      M_SLIST_END
  *
  *      m_slist_free_all( &list, NULL, NULL, memory );
  *    }
  */
#define  M_SLIST_LOOP(_plist,_data)                         \
               {                                            \
                 M_SList     _slist_node = *(_plist);       \
                 M_SList     _slist_next;                   \
                 M_Pointer  _data;                          \
                                                            \
                 while ( _slist_node )                      \
                 {                                          \
                   _slist_next = _slist_node->next;         \
                   _data       = _slist_node->data;         \


 /***********************************************************************
  *
  * @macro: M_SLIST_LOOP_NODE (list,node)
  *
  * @description:
  *    a variant of @M_SLIST_LOOP_NODE used to expose each list node
  *    pointer, instead of its "data" field
  */
#define  M_SLIST_LOOP_NODE(_plist,_node)                    \
               {                                            \
                 M_SList     _slist_node = *(_plist);       \
                 M_SList     _slist_next;                   \
                 M_SList     _node;                         \
                                                            \
                 while ( _slist_node )                      \
                 {                                          \
                   _slist_next = _slist_node->next;         \
                   _node       = _slist_node;               \


 /***********************************************************************
  *
  * @macro: M_SLIST_END
  *
  * @description:
  *    a useful macro used in association with @M_SLIST_LOOP to parse
  *    all nodes in a list and extract its "data" values
  */
#define  M_SLIST_END                                       \
                   _slist_node = _slist_next;              \
                 }                                         \
               }

 /* */

 /***********************************************************************
  *
  * @function: m_slist_node_skip
  *
  * @description:
  *    returns the nth's node in a list
  *
  * @input:
  *    node  :: start list node
  *    delta :: increment count (in list nodes), must be >= 0
  *
  * @return:
  *    new list node, of NULL if out-of-list
  */
  MLIB_API(M_SList)
  m_slist_node_skip( M_SList  node,
                     M_UInt   delta );


 /***********************************************************************
  *
  * @function: m_slist_node_last
  *
  * @description:
  *    find the last node from a given stating node in a list
  *
  * @input:
  *    node :: start list node
  *
  * @return:
  *    last node, if any
  */
  MLIB_API(M_SList)
  m_slist_node_last( M_SList  node );

 /* */


M_END_HEADER

#endif /* __MLIB_SLIST_H__ */
