#ifndef __MLIB_STREAM_H__
#define __MLIB_STREAM_H__

#include <mlib/buffer.h>

M_BEGIN_HEADER

 /************************************************************************
  *
  * @section: streams
  *
  * @title: Streams
  *
  * @abstract: input/output streams
  *
  * @description:
  *   the @M_Stream type is used to model input/output streams that can
  *   be used to read or write data to any kind of file, socket or buffer
  *
  *
  * @order:
  *   M_Stream
  *   m_stream_new
  *   m_stream_new_memory
  *   m_stream_new_ansi
  *
  *   m_stream_pos
  *   m_stream_seek
  *   m_stream_skip
  *   m_stream_read
  *   m_stream_read_some
  *   m_stream_write
  *   m_stream_write_some
  *
  *   M_Frame
  *   m_stream_open_frame
  *   m_stream_close_frame
  */

 /************************************************************************/
 /************************************************************************/
 /*****                                                              *****/
 /*****                         STREAM API                           *****/
 /*****                                                              *****/
 /************************************************************************/
 /************************************************************************/

 /************************************************************************
  *
  * @type: M_Stream
  *
  * @description:
  *    a handle to a @M_StreamRec used to model a given MLib stream.
  */
  typedef struct M_StreamRec_*   M_Stream;


 /************************************************************************
  *
  * @type: M_StreamSeek
  *
  * @description:
  *    a list of values used to indicate the anchor used by @m_stream_seek
  *
  * @values:
  *    M_SEEK_BEGIN :: seek from the beginning of the stream
  *    M_SEEK_CUR   :: seek from the current stream's position
  *    M_SEEK_END   :: seek from the end of the stream
  */
  typedef enum
  {
    M_SEEK_BEGIN = 0,
    M_SEEK_CUR,
    M_SEEK_END

  } M_StreamSeek;


#if 0
 /************************************************************************
  *
  * @function: m_stream_open
  *
  * @description:
  *    create a new stream from a given @M_File object
  *
  * @input:
  *    file       :: file handle
  *    open_flags :: flags used to indicate how the stream is going
  *                  to be used (for reading, writing or appending)
  *
  * @return:
  *   new stream object handle
  *
  * @throws: any. especially if the file cannot be accessed
  *
  * @note:
  *   you can also use @m_stream_new_memory and @m_stream_new_ansi
  *   if you don't want to create a file object
  */
  MLIB_API( M_Stream )  m_stream_open( M_File   file,
                                       M_UInt   open_flags );
#endif

 /************************************************************************
  *
  * @function: m_stream_read
  *
  * @description:
  *    read bytes from a stream. unlike @m_stream_read_some, this function
  *    will throw a @m_err_stream_read exception if less than
  *    'count' bytes were read from the stream
  *
  * @input:
  *    stream  :: source stream handle
  *    dest    :: destination address
  *    count   :: number of bytes to read
  *
  * @throws: any, especially @m_err_stream_read
  */
  MLIB_API( void )      m_stream_read( M_Stream  stream,
                                       M_ByteP   dest,
                                       M_Size    count );


 /************************************************************************
  *
  * @function: m_stream_read_some
  *
  * @description:
  *    read bytes from a stream
  *
  * @input:
  *    stream  :: source stream handle
  *    dest    :: destination address
  *    buffer  :: buffer where the bytes will be stored
  *
  * @return:
  *    number of bytes really read. can be less than 'count'
  *
  * @throws: any.
  *
  * @note:
  *   this function doesn't throw an exception if less than 'count' bytes
  *   were read from the stream. If you don't want to check the return
  *   value, you'd probably want to use @m_stream_read instead.
  */
  MLIB_API( M_Size )    m_stream_read_some( M_Stream   stream,
                                            M_ByteP    dest,
                                            M_Size     count );



 /************************************************************************
  *
  * @function: m_stream_write
  *
  * @description:
  *    write bytes to a stream. unlike @m_stream_write_some, this function
  *    will throw a @m_err_stream_write exception if less than 'count'
  *    bytes could be written to the stream
  *
  * @input:
  *    stream  :: target stream handle
  *    src     :: source address
  *    count   :: number of bytes to write
  *
  * @throws: any.
  */
  MLIB_API( void )      m_stream_write( M_Stream   stream,
                                        M_CByteP   src,
                                        M_Size     count );


 /************************************************************************
  *
  * @function: m_stream_write_some
  *
  * @description:
  *    write bytes to a stream
  *
  * @input:
  *    stream  :: target stream handle
  *    src     :: source address
  *    count   :: number of bytes to write
  *
  * @return:
  *    number of bytes really written. can be less than 'count'
  *
  * @throws: any.
  *
  * @note:
  *   this function doesn't throw an exception if less than 'count' bytes
  *   were written to the stream. If you don't want to check the return
  *   value, you'd probably want to use @m_stream_write instead.
  */
  MLIB_API( M_Size )    m_stream_write_some( M_Stream  stream,
                                             M_CByteP  src,
                                             M_Size    count );


 /************************************************************************
  *
  * @function: m_stream_pos
  *
  * @description:
  *    returns the stream's current position
  *
  * @input:
  *    stream  :: target stream handle
  *
  * @return:
  *    current stream position if available, 0 otherwise
  */
  MLIB_API( M_Size )    m_stream_pos( M_Stream  stream );


 /************************************************************************
  *
  * @function: m_stream_seek
  *
  * @description:
  *    change a stream's current position
  *
  * @input:
  *    stream  :: target stream handle
  *    pos     :: new position in stream relative to 'anchor'
  *    anchor  :: a value of type @M_Stream_Seek which determines
  *               the origin of the seek
  *
  * @throws: any. certain stream implementations do not support all
  *          stream anchor types..
  */
  MLIB_API( void )      m_stream_seek( M_Stream      stream,
                                       M_Dist        pos,
                                       M_StreamSeek  anchor );


 /************************************************************************
  *
  * @function: m_stream_skip
  *
  * @description:
  *    skip a given number of bytes in a stream
  *
  * @input:
  *    stream  :: target stream handle
  *    delta   :: number of bytes to skip from current position
  *
  * @throws: any.
  */
  MLIB_API( void )     m_stream_skip( M_Stream  stream,
                                      M_Dist    delta );


 /************************************************************************
  *
  * @function: m_stream_open_frame
  *
  * @description:
  *    reads a frame of bytes from a stream and returns a pointer to its
  *    address in memory
  *
  *    if the stream is memory-based, this simply is a pointer within the
  *    stream itself. Otherwise, the function allocates a memory block in
  *    the heap and reads data in it with @m_stream_read
  *
  * @input:
  *    stream  :: frame stream handle
  *    size    :: size of frame in bytes
  *
  * @return:
  *    frame pointer (i.e. pointer to the first frame byte)
  *
  * @throws: any. nested frame openings are prohibited..
  *
  * @note:
  *   you must call @m_stream_close_frame to discard the frame. You must
  *   also close all frames of a given stream before calling @m_stream_done
  */
  MLIB_API( M_Frame )   m_stream_open_frame( M_Stream  stream,
                                             M_Size    size );


 /************************************************************************
  *
  * @function: m_stream_close_frame
  *
  * @description:
  *    close a frame previously opened with @m_stream_open_frame. This
  *    will eventually discard the frame's heap memory block
  *
  * @input:
  *    stream  :: frame stream handle
  *    frame   :: frame pointer
  */
  MLIB_API( void )      m_stream_close_frame( M_Stream  stream,
                                              M_Frame   frame );


 /************************************************************************
  *
  * @function: m_stream_new_memory
  *
  * @description:
  *    create a new memory-based stream
  *
  * @input:
  *    address    :: memory stream address
  *    size       :: byte size
  *    open_flags :: opening flags
  *    memory     :: memory manager handle
  *
  * @return:
  *   new stream handle
  *
  * @throws: any
  */
  MLIB_API( M_Stream )  m_stream_new_memory( M_CByteP   address,
                                             M_Size     size,
                                             M_UInt     open_flags,
                                             M_Memory   memory );


 /************************************************************************
  *
  * @function: m_stream_new_ansi
  *
  * @description:
  *    create a new ANSI stream
  *
  * @input:
  *    filepathname :: file pathname (standard C string)
  *    open_flags   :: opening flags
  *    memory       :: memory manager handle
  *
  * @return:
  *   new stream handle
  *
  * @throws: any
  */
  MLIB_API( M_Stream )  m_stream_new_ansi( M_CString  filepathname,
                                           M_UInt     open_flags,
                                           M_Memory   memory );


 /************************************************************************/
 /************************************************************************/
 /*****                                                              *****/
 /*****                         FRAMES API                           *****/
 /*****                                                              *****/
 /************************************************************************/
 /************************************************************************/


 /************************************************************************
  *
  * @type: M_Frame
  *
  * @description:
  *    handle to a frame structure. see @M_FrameRec
  *
  *    a frame is used to model and access a series of bytes from a given
  *    stream while reducing the number of buffers and copies. See the
  *    descriptions for @M_FrameType, @m_frame_open_input,
  *    @m_frame_open_persistent, @m_frame_open_output and @m_frame_open_full
  *    for more details..
  */
  typedef M_FrameRec_*   M_Frame;


 /************************************************************************
  *
  * @enum: M_FrameType
  *
  * @description:
  *    the list of known frame types. Each type correspond to a different
  *    use of frames and the @m_frame_open .. @m_frame_close functions
  *
  * @values:
  *   M_FRAME_TYPE_INPUT      ::
  *    a temporary frame used to read small structures from a stream and
  *    decode them quickly. These typically exist within a @m_frame_open_input
  *    ... @m_frame_close block where no other stream operations occurs.
  *    the caller is then free to parse the raw data directly within the
  *    block.
  *
  *
  *   M_FRAME_TYPE_PERSISTENT ::
  *    a persistent frame used to read raw data from a stream with
  *    minimal buffers/copying. If the stream is already memory-mapped,
  *    the data's raw address is returned. Otherwise, it is copied into
  *    a dynamically-allocated memory block that will be released when
  *    @m_frame_close is called..
  *
  *
  *   M_FRAME_TYPE_OUTPUT     ::
  *    a temporary frame used to _write_ small structures to a stream
  *    (the exact opposite of an input frame). They exist within
  *    @m_frame_open_output ... @m_frame_close blocks where callers can encode
  *    their data within the frame's buffer.
  *
  *    note that @m_frame_open_output actually reserves room to write
  *    the 'size' bytes (either in a stream buffer or in the heap), and
  *    that @m_frame_close actually flushes them to the stream.
  */
  typedef M_FrameType_
  {
    M_FRAME_TYPE_INPUT      = 0,
    M_FRAME_TYPE_PERSISTENT = 1,
    M_FRAME_TYPE_OUTPUT     = 2
  
  } M_FrameType;


 /************************************************************************
  *
  * @functype: M_Frame_CloseFunc
  *
  * @description:
  *    a function used to close a frame. This is used internally by the
  *    MLib.
  *
  * @input:
  *    close_data :: generic data passed to the close function
  *    frame      :: handle to frame
  *
  * @throws: this function should never throw an exception
  */
  typedef void  (*M_Frame_CloseFunc)( M_Pointer  close_data,
                                      M_Frame    frame );

 /************************************************************************
  *
  * @struct: M_FrameRec
  *
  * @description:
  *    the frame structure, used to hold all state related to a frame
  *    in user-space
  *
  * @fields:
  *    address    :: frame address in memory
  *    close      :: frame close function
  *    close_data :: frame close function argument
  */
  typedef struct M_FrameRec_
  {
    M_ByteP            address;
    M_Frame_CloseFunc  close;
    M_Pointer          close_data;
  
  } M_FrameRec;


 /************************************************************************
  *
  * @func: m_frame_open_input
  *
  * @description:
  *    this function is used to access the next 'size' bytes of a given
  *    'stream'. It fills a @M_FrameRec structure appropriately and return
  *    the address in memory of the data
  *
  * @input:
  *    frame      :: target frame handle
  *    size       :: size in bytes of frame
  *    stream     :: stream handle
  *
  * @return:
  *    address of frame bytes in memory. Same as 'frame->address'
  *
  * @throws: any
  *
  * @note:
  *   you should NEVER try to MODIFY the frame bytes, and you should NEVER
  *   perform any STREAM operations until this frame is closed. you should
  *   _always_ call @m_frame_close when you've finished parsing the frame
  *   bytes.
  *
  *   See the notes for @m_frame_open_full for more important details..
  *
  *   if the frame data is already in memory (e.g. when using memory-mapped
  *   files or buffered streams), this function will set 'frame->address'
  *   to their address (and the caller will access them directly).
  *
  *   otherwise, a new memory block is allocated in the heap, the bytes
  *   are read into it, and its address is set in 'frame->address'. When
  *   @m_frame_close is called, it will automatically release the
  *   memory block.
  *
  *   use @m_frame_open_full if you want to avoid dynamic memory allocations
  */
  MLIB_API( const M_ByteP )  m_frame_open_input(( M_Frame      frame,
                                                  M_Size       size,
                                                  M_Stream     stream );


 /************************************************************************
  *
  * @func: m_frame_open_persistent
  *
  * @description:
  *    this function is used to extract a bytes from a stream. It fills
  *    a @M_FrameRec structure appropriately.
  *
  * @input:
  *    frame      :: target frame handle
  *    size       :: size in bytes of frame
  *    stream     :: stream handle
  *
  * @return:
  *    address of frame bytes in memory. Same as 'frame->address'
  *
  * @throws: any
  *
  * @note:
  *   you CAN perform STREAM operations after calling this function. However,
  *   you should ALWAYS close the frame before its parent stream.
  *
  *   See the notes for @m_frame_open_full for more important details..
  *
  *   if the frame data is already in memory _and_ persists (e.g. when
  *   using memory-mapped files but _not_ buffered streams),  this function
  *   will set 'frame->address' to their address (and the caller will access
  *   them directly).
  *
  *   otherwise, a new memory block is allocated in the heap, the bytes
  *   are read into it, and its address is set in 'frame->address'. When
  *   @m_frame_close is called, it will automatically release the
  *   memory block.
  *
  *   use @m_frame_open_full if you want to avoid dynamic memory allocations
  */
  MLIB_API( const M_ByteP )  m_frame_open_persistent( M_Frame   frame,
                                                      M_Size    size,
                                                      M_Stream  stream );


 /************************************************************************
  *
  * @func: m_frame_open_output
  *
  * @description:
  *    this function is used to extract a bytes from a stream. It fills
  *    a @M_FrameRec structure appropriately.
  *
  * @input:
  *    frame      :: target frame handle
  *    size       :: size in bytes of frame
  *    stream     :: stream handle
  *
  * @return:
  *    address of frame bytes in memory. Same as 'frame->address'
  *
  * @throws: any
  *
  * @note:
  *   you CAN perform STREAM operations after calling this function. However,
  *   you should ALWAYS close the frame before its parent stream.
  *
  *   See the notes for @m_frame_open_full for more important details..
  *
  *   if the frame data is already in memory _and_ persists (e.g. when
  *   using memory-mapped files but _not_ buffered streams),  this function
  *   will set 'frame->address' to their address (and the caller will access
  *   them directly).
  *
  *   otherwise, a new memory block is allocated in the heap, the bytes
  *   are read into it, and its address is set in 'frame->address'. When
  *   @m_frame_close is called, it will automatically release the
  *   memory block.
  *
  *   use @m_frame_open_full if you want to avoid dynamic memory allocations
  */
  MLIB_API( M_ByteP )        m_frame_open_output( M_Frame   frame,
                                                  M_Size    size,
                                                  M_Stream  stream );



 /************************************************************************
  *
  * @func: m_frame_open_full
  *
  * @description:
  *    the most generic way to open a new frame
  *
  * @input:
  *    frame      :: target frame handle
  *    type       :: the frame's type
  *    size       :: size in bytes of frame
  *    buff       :: static buffer address, optional
  *    stream     :: stream handle
  *   
  *
  * @return:
  *    address of frame bytes in memory. Same as 'frame->address'.
  *    this is typed as a 'const M_ByteP', even for output frames.
  *
  * @throws: any
  *
  * @note:
  *   if you want to avoid dynamic memory allocations in @m_frame_open_input
  *   or @m_frame_open_output, you can pass the address of a user-allocated
  *   memory block (of at least 'size' bytes) in the 'buff' parameter. It
  *   will not be freed by a call to @m_frame_close
  *
  *   if 'buff' is NULL, this function will allocate a new memory block in
  *   the heap whenever necessary instead.
  */
  MLIB_API( const M_ByteP )  m_frame_open_full( M_Frame      frame,
                                                M_FrameType  type,
                                                M_Size       size,
                                                M_ByteP      buff,
                                                M_Stream     stream );


 /************************************************************************
  *
  * @func: m_frame_close
  *
  * @description:
  *    this function closes an opened frame. When a memory block was allocated
  *    in the heap, it will release it automatically..
  *
  * @input:
  *    frame      :: target frame handle
  */
  MLIB_API( void )           m_frame_close( M_Frame  frame );


 /************************************************************************/
 /************************************************************************/
 /*****                                                              *****/
 /*****                     STREAM CLASS IMPLEMENTATION              *****/
 /*****                                                              *****/
 /************************************************************************/
 /************************************************************************/

 /************************************************************************
  *
  * @functype: M_Stream_ReadFunc
  *
  * @description:
  *   a function used to read bytes from a stream
  *
  * @input:
  *   stream  :: stream handle
  *   count   :: number of bytes to read
  *   buffer  :: target buffer where bytes will be stored
  *
  * @return ::
  *   number of bytes read, can be less than 'count'
  *
  * @throws: any. However, this function should _not_ raise if
  *          less than 'count' bytes were available from the stream
  */
  typedef M_Size   (*M_Stream_ReadFunc)( M_Stream  stream,
                                         M_ByteP   dest,
                                         M_Size    count );


 /************************************************************************
  *
  * @functype: M_Stream_WriteFunc
  *
  * @description:
  *   a function used to write bytes to a stream
  *
  * @input:
  *   stream  :: stream handle
  *   count   :: number of bytes to write
  *   buffer  :: source buffer where the bytes come from
  *
  * @return ::
  *   number of bytes written, can be less than 'count'
  *
  * @throws: any. However, this function should _not_ raise if
  *          less than 'count' bytes were written to the stream
  */
  typedef M_Size   (*M_Stream_WriteFunc)( M_Stream        stream,
                                          M_CByteP        src,
                                          M_Size          count );


 /************************************************************************
  *
  * @functype: M_Stream_SeekFunc
  *
  * @description:
  *   a function used to change the current position within a stream
  *
  * @input:
  *   stream  :: stream handle
  *   pos     :: new position, relative to anchor
  *   anchor  :: stream anchor
  *
  * @throws: any.
  */
  typedef void     (*M_Stream_SeekFunc)( M_Stream      stream,
                                         M_Dist        pos,
                                         M_StreamSeek  anchor );


 /************************************************************************
  *
  * @functype: M_Stream_PosFunc
  *
  * @description:
  *   returns a stream's current position
  *
  * @input:
  *   stream  :: stream handle
  *
  * @return:
  *   stream's current position, if available (0 otherwise)
  */
  typedef M_Size   (*M_Stream_PosFunc) ( M_Stream  stream );


 /************************************************************************
  *
  * @functype: M_Stream_InitBufferFunc
  *
  * @description:
  *   initialize a buffer to read from or write to a stream
  *
  * @input:
  *   stream  :: stream handle
  *   buffer  :: target buffer handle
  *   writing :: boolean indicating wether this is used to read
  *              or write to the stream..
  */
  typedef void     (*M_Stream_InitBufferFunc)( M_Stream   stream,
                                               M_Buffer   buffer,
                                               M_Bool     writing );


 /************************************************************************
  *
  * @enum: M_StreamControl
  *
  * @description:
  *   a list of control operation opcodes for @M_Stream_ControlFunc
  *
  * @values:
  *   M_STREAM_CONTROL_FLUSH ::
  *     used to indicate that the stream must be flushed
  *
  *   M_STREAM_CONTROL_TO_READ ::
  *     used to indicate that the stream is going to be used for
  *     read operations. This should also flush the stream when
  *     necessary
  *
  *   M_STREAM_CONTROL_TO_WRITE ::
  *     used to indicate that the stream is going to be used for
  *     write operations.
  *
  *   M_STREAM_CONTROL_EOF ::
  *     used to indicate that we want to test wether the end of stream
  *     has been reached for this stream.
  */
  typedef enum
  {
    M_STREAM_CONTROL_FLUSH    = 1,
    M_STREAM_CONTROL_TO_READ  = 2,
    M_STREAM_CONTROL_TO_WRITE = 3,
    M_STREAM_CONTROL_EOF      = 4

  } M_StreamControl;


 /************************************************************************
  *
  * @functype: M_Stream_ControlFunc
  *
  * @description:
  *   a function used to perform a stream control operation. Similar to
  *   unix 'ioctl', though simpler..
  *
  * @input:
  *   stream       :: stream handle
  *   control      :: control operation. see @M_StreamControl
  *   control_data :: control-specific data
  *
  * @return:
  *   result depends on control operation. Most of them ignore it, except
  *   for M_STREAM_CONTROL_EOF, where it is interpreted as the EOF flag..
  *
  * @throws: any
  */
  typedef M_Int  (*M_Stream_ControlFunc)( M_Stream         stream,
                                          M_StreamControl  control,
                                          M_Pointer        control_data );

 /* */

 /************************************************************************
  *
  * @enum: M_StreamFlags
  *
  * @description:
  *   a list of bit flags used in @M_Stream_Class structures to describe
  *   the stream quickly
  *
  * @values:
  *   M_STREAM_FLAG_INPUT   :: the stream can be used for reading
  *
  *   M_STREAM_FLAG_OUTPUT  :: the stream can be used for writing
  *
  *   M_STREAM_FLAG_LINEAR  :: indicate that the stream content is directly
  *                            accessible in memory as a linear memory block.
  *                            this is the case of ROM-based files or
  *                            memory-mapped ones..
  *
  *   M_STREAM_FLAG_WRITING :: indicates that the last stream operation
  *                            was a write
  */
  typedef enum
  {
    M_STREAM_FLAG_INPUT   = 1,
    M_STREAM_FLAG_OUTPUT  = 2,
    M_STREAM_FLAG_LINEAR  = 4,
    M_STREAM_FLAG_WRITING = 8

  } M_StreamFlags;


 /************************************************************************
  *
  * @type: M_Stream_Class
  *
  * @description:
  *   handle to a constant @M_Stream_ClassRec structure used to model
  *   a single MLib stream
  */
  typedef struct M_Stream_ClassRec_*  M_Stream_Class;


 /************************************************************************
  *
  * @struct: M_Stream_ClassRec
  *
  * @description:
  *   a structure used to model the @M_Class of a given @M_Stream object
  *
  * @fields:
  *   clazz       :: root class fields
  *
  *   flags       :: stream class flags. see @M_StreamFlags
  *
  *   seek        :: stream seek method, can be NULL
  *   pos         :: stream pos method, can be NULL
  *
  *   read        :: stream read method
  *   write       :: stream write method
  *
  * @note:
  *   certain streams do not support seek operations. in this case,
  *   value of both 'seek' and 'pos' should be NULL
  */
  typedef struct M_Stream_ClassRec_
  {
    M_ClassRec               clazz;

    M_Stream_SeekFunc        seek;
    M_Stream_PosFunc         pos;
    M_Stream_ReadFunc        read;
    M_Stream_WriteFunc       write;
    M_Stream_ControlFunc     control;
    M_Stream_InitBufferFunc  init_buffer;

  } M_Stream_ClassRec;


 /************************************************************************
  *
  * @type: M_StreamInfo
  *
  * @description:
  *   opquae handle to a structure managed by the MLib to hold important
  *   information related to streams.
  */
  typedef struct M_StreamInfoRec_*   M_StreamInfo;


 /************************************************************************
  *
  * @struct: M_StreamRec
  *
  * @description:
  *   a structure used to hold the public fields of a @M_Stream object
  *
  * @fields:
  *   object :: parent object fields
  *   size   :: size of stream, when available. 0 otherwise
  *   buffer :: optional @M_Buffer handle for the stream
  *   flags  :: stream flags;
  */
  typedef struct M_StreamRec_
  {
    M_ObjectRec   object;
    M_Size        size;
    M_StreamInfo  info;

  } M_StreamRec;

 /* */

 MLIB_VAR( M_ClassType )   m_stream_class_type;

#define  M_STREAM(x)      ((M_Stream)(x))

#define  M_STREAM_CHK(x)  M_STREAM(m_objet_is_a(x,m_stream_class_type))

#define  m_assert_is_stream(x)  m_assert( M_STREAM_CHK(x) )

#define  M_STREAM_GET_CLASS(s)  ((M_Stream_Class) M_OBJECT(s)->clazz)

 /* */

M_END_HEADER

#endif /* __MLIB_STREAM_H__ */
