#ifndef __MLIB_STRING_LIST_H__
#define __MLIB_STRING_LIST_H__

#include <mlib.h>

M_BEGIN_HEADER

 /*************************************************************************
  *
  * @type: M_StrList
  *
  * @description:
  *   handle to a @M_StrListRec structure used to model a list of strings
  *
  *   each list implements a simple and fast array of string pointers
  *   that can easily be accessed, searched and modified
  *
  *   each list has an "owner state" that determines wether it owns the
  *   strings in its array.
  */
  typedef struct M_StrListRec_*   M_StrList;


 /*************************************************************************
  *
  * @type: M_StrList_Owner
  *
  * @description:
  *   a list of "owner" states for a string list. When a list owns its
  *   strings, the latter are freed when the list is destroyed
  *
  * @values:
  *   M_STRLIST_NONE ::
  *     used to indicate that the list doesn't own its strings.
  *
  *   M_STRLIST_KEEP ::
  *     used to indicate that the list does own its strings, and that
  *     the list will take ownership of each new string added to it
  *     through @m_strlist_add, @m_strlist_insert, etc..
  *
  *   M_STRLIST_COPY ::
  *     same as @M_STRLIST_KEEP, except that _copies_ of new strings
  *     will be added to the list through @m_strlist_add, @m_strlist_insert,
  *     etc..
  */
  typedef enum
  {
    M_STRLIST_NONE = 0,
    M_STRLIST_KEEP,
    M_STRLIST_COPY

  } M_StrList_Owner;


 /*************************************************************************
  *
  * @struct: M_StrListRec
  *
  * @description:
  *   a structure used to model a list of strings
  *
  * @fields:
  *   strings :: array of @M_String pointers
  *   count   :: number of strings in list
  */
  typedef struct M_StrListRec_
  {
    M_String*   strings;
    M_UFast     count;

  } M_StrListRec;


 /*************************************************************************
  *
  * @function: m_strlist_new
  *
  * @description:
  *   create a new (empty) string list
  *
  * @fields:
  *   memory  :: memory manager handle
  *   size    :: initial "size" (capacity) of the string array. 0 by default
  *   owner   :: initial ownership of list
  *
  * @throws: @m_err_memory_alloc
  */
  MLIB_API( M_StrList )
  m_strlist_new( M_Memory         memory,
                 M_UInt           size,
                 M_StrList_Owner  owner );


 /*************************************************************************
  *
  * @function: m_strlist_destroy
  *
  * @description:
  *   destroy a given string list. If it is "owner", the strings will be
  *   destroyed too..
  *
  * @fields:
  *   list  :: string list handle
  */
  MLIB_API( void )
  m_strlist_destroy( M_StrList  list );


 /*************************************************************************
  *
  * @function: m_strlist_push
  *
  * @description:
  *   push a string list on top of the cleanup stack
  *
  * @fields:
  *   list  :: string list handle
  *
  * @throws: @m_err_memory_alloc
  */
  MLIB_API(void)
  m_strlist_push( M_StrList  list );


 /*************************************************************************
  *
  * @function: m_strlist_pop
  *
  * @description:
  *   pop a string list from the top of the cleanup stack. do not destroy it
  *
  * @fields:
  *   list  :: string list handle
  */
  MLIB_API(void)
  m_strlist_pop( M_StrList  list );


 /*************************************************************************
  *
  * @function: m_strlist_clear
  *
  * @description:
  *   clear a given list (empty it !)
  *
  * @fields:
  *   list  :: string list handle
  */
  MLIB_API(void)
  m_strlist_clear( M_StrList  list );


 /*************************************************************************
  *
  * @function: m_strlist_copy
  *
  * @description:
  *   copy a given string list
  *
  * @fields:
  *   source  :: source string list handle
  *   owner   :: owner state of list copy. see @M_StrList_Owner
  *
  * @return:
  *   handle to new string list
  *
  * @throws: @m_err_memory_alloc
  */
  MLIB_API(M_StrList)
  m_strlist_copy( M_StrList        source,
                  M_StrList_Owner  owner );


 /*************************************************************************
  *
  * @function: m_strlist_ensure
  *
  * @description:
  *   ensures that a string list's array is capable of holding 'count'
  *   strings. This is only needed if you start tweaking the list contents
  *   by hand
  *
  * @fields:
  *   list  :: string list handle
  *   count :: new minimal capacity to ensure
  *
  * @throws: @m_err_memory_alloc
  */
  MLIB_API(void)
  m_strlist_ensure( M_StrList  list,
                    M_UInt     count );


 /*************************************************************************
  *
  * @function: m_strlist_set_owner
  *
  * @description:
  *   ensure that a string list owns its strings. If the list was created
  *   with no ownership, this function will duplicate each list entry
  *
  * @fields:
  *   list  :: string list handle
  *   owner :: new owner state. You cannot use @M_STRLIST_NONE here !!
  *
  * @note:
  *   if the list isn't owner before the call, its strings will be
  *   automatically duplicated
  *
  * @throws: @m_err_memory_alloc
  */
  MLIB_API(void)
  m_strlist_set_owner( M_StrList        list,
                       M_StrList_Owner  owner );


 /*************************************************************************
  *
  * @function: m_strlist_add
  *
  * @description:
  *   append a new string to a given string list.
  *
  * @fields:
  *   list    :: string list handle
  *   string  :: pointer to new string
  *
  * @note:
  *   if the list's owner state is @M_STRLIST_NONE, then string pointer
  *   is added to the list, but it's the caller's responsability to free it
  *   when it becomes un-needed
  *
  *   if the owner state is @M_STRLIST_KEEP, the list takes ownership
  *   of the input string (the caller mustn't free it)
  *
  *   in the owner state is @M_STRLIST_COPY, a copy of the input string is
  *   added to the list (the caller must free the input list)
  *
  *   see @m_strlist_set_owner to change the owner state
  *
  * @throws: @m_err_memory_alloc
  */
  MLIB_API( void )
  m_strlist_add( M_StrList  list,
                 M_String   string );


 /*************************************************************************
  *
  * @function: m_strlist_insert
  *
  * @description:
  *   insert a new string into a given string list
  *
  * @fields:
  *   list    :: string list handle
  *   index   :: new string's index in the list
  *   string  :: pointer to new string
  *
  * @note:
  *   see notes in @m_strlist_add
  *
  * @throws: @m_err_memory_alloc
  */
  MLIB_API( void )
  m_strlist_insert( M_StrList  list,
                    M_Int      index,
                    M_String   string );


 /*************************************************************************
  *
  * @function: m_strlist_remove
  *
  * @description:
  *   remov a given string in a string list
  *
  * @fields:
  *   list    :: string list handle
  *   index   :: string index in the list
  *
  * @note:
  *   if the list owns its strings, the relevant string will be freed.
  */
  MLIB_API( void )
  m_strlist_remove( M_StrList  list,
                    M_Int      index );


 /*************************************************************************
  *
  * @function: m_strlist_add_slice
  *
  * @description:
  *   append several strings to a string list
  *
  * @fields:
  *   list    :: string list handle
  *   count   :: number of new strings
  *   strings :: an array of new strings
  *
  * @note:
  *   see notes in @m_strlist_add
  *
  * @throws: @m_err_memory_alloc
  */
  MLIB_API( void )
  m_strlist_add_slice( M_StrList  list,
                       M_Int      count,
                       M_String*  strings );


 /*************************************************************************
  *
  * @function: m_strlist_insert_slice
  *
  * @description:
  *   insert several strings to a string list
  *
  * @fields:
  *   list    :: string list handle
  *   index   :: insertion index
  *   count   :: number of new strings
  *   strings :: an array of new strings
  *
  * @note:
  *   see notes in @m_strlist_add
  *
  * @throws: @m_err_memory_alloc
  */
  MLIB_API( void )
  m_strlist_insert_slice( M_StrList  list,
                          M_Int      index,
                          M_Int      count,
                          M_String*  strings );


 /*************************************************************************
  *
  * @function: m_strlist_remove_slice
  *
  * @description:
  *   insert several strings to a string list
  *
  * @fields:
  *   list    :: string list handle
  *   index   :: insertion index
  *   count   :: number of new strings
  *   strings :: an array of new strings
  *
  * @note:
  *   see notes in @m_strlist_add
  *
  * @throws: @m_err_memory_alloc
  */
  MLIB_API( void )
  m_strlist_remove_slice( M_StrList  list,
                          M_Int      index,
                          M_Int      count );

 /* */

M_END_HEADER

#endif /* __MLIB_STRING_LIST_H__ */
