#ifndef __MLIB_STRUCTURE_H__
#define __MLIB_STRUCTURE_H__

#include <mlib.h>

M_BEGIN_HEADER

  /* format of an 8-bit struct_op value = [ xxxxx | e | s ] */
  /* s is set to 1 if the value is signed,                 */
  /* e is set to 1 if the value is little-endian           */
  /* xxxxx is a command                                    */

#define M_STRUCT_OP_SHIFT         2
#define M_STRUCT_OP_SIGNED        1
#define M_STRUCT_OP_LITTLE        2
#define M_STRUCT_OP_COMMAND( x )  ( x >> M_STRUCT_OP_SHIFT )

#define M_MAKE_STRUCT_OP( command, little, sign ) \
          ( ( command << M_STRUCT_OP_SHIFT ) | ( little << 1 ) | sign )

#define M_STRUCT_OP_END   0
#define M_STRUCT_OP_START 1  /* start a new struct     */
#define M_STRUCT_OP_INT8  2  /* read 1-byte value     */
#define M_STRUCT_OP_INT16 3  /* read 2-byte value     */
#define M_STRUCT_OP_INT32 4  /* read 4-byte value     */
#define M_STRUCT_OP_INT24 5  /* read 3-byte value     */
#define M_STRUCT_OP_BYTES 6  /* read a bytes sequence */


 /**********************************************************************
  *
  * @enum: M_Struct_Op
  *
  * @description:
  *   a enumeration used to list the various @M_Struct_Field types.
  *
  * @fields:
  *  m_struct_op_end ::
  *      this value should _always_ appear at the end of a structure layout
  *      table. It is used to indicate the end of the table
  *
  *  m_struct_op_start ::
  *      this value can appear as the first element of a structure layout
  *      table. The corresponding M_Struct_Field is used to indicate the
  *      total amount of bytes to read from or write to memory during
  *      marshalling/unmarshalling. This is important when writing or
  *      reading to streams. It is optional though..
  *
  *  m_struct_op_byte ::
  *      this value corresponds to a field stored as a single unsigned byte
  *
  *  m_struct_op_schar ::
  *      this value corresponds to a field stored as a single signed byte
  *
  *  m_struct_op_uint16_be ::
  *      this value corresponds to a field stored as an unsigned 2-byte short
  *      with big endianess (a.k.a. "network order")
  *
  *  m_struct_op_uint16_le ::
  *      this value corresponds to a field stored as an unsigned 2-byte short
  *      with little endianess
  *
  *  m_struct_op_int16_be ::
  *      this value corresponds to a field stored as a signed 2-byte short
  *      with big endianess (a.k.a. "network order")
  *
  *  m_struct_op_int16_le ::
  *      this value corresponds to a field stored as a signed 2-byte short
  *      with little endianess
  *
  *  m_struct_op_uint32_be ::
  *      this value corresponds to a field stored as an unsigned 4-byte long
  *      with big endianess (a.k.a. "network order")
  *
  *  m_struct_op_uint32_le ::
  *      this value corresponds to a field stored as an unsigned 4-byte long
  *      with little endianess
  *
  *  m_struct_op_int32_be ::
  *      this value corresponds to a field stored as a signed 4-byte long
  *      with big endianess (a.k.a. "network order")
  *
  *  m_struct_op_int32_le ::
  *      this value corresponds to a field stored as a signed 4-byte long
  *      with little endianess
  *
  *  m_struct_op_uint24_be ::
  *      this value corresponds to a field stored as an unsigned 3-byte long
  *      with big endianess (a.k.a. "network order")
  *
  *  m_struct_op_uint24_le ::
  *      this value corresponds to a field stored as an unsigned 3-byte long
  *      with little endianess
  *
  *  m_struct_op_int24_be ::
  *      this value corresponds to a field stored as a signed 3-byte long
  *      with big endianess (a.k.a. "network order")
  *
  *  m_struct_op_in24_le ::
  *      this value corresponds to a field stored as a signed 3-byte long
  *      with little endianess
  *
  *  m_struct_op_bytes ::
  *      this value corresponds to a sequence of bytes with fixed size
  *      the byte count is contained in the corresponding M_Struct_Field
  *      element
  *
  *  m_struct_op_skip ::
  *      this value corresponds to a sequence of bytes that should be
  *      skipped when reading fields from memory. It is generally used
  *      to ignore fields when reading certain file formats
  *      the byte count is contained in the corresponding M_Struct_Field
  *      element
  */
  typedef enum  M_Struct_Op_
  {
    m_struct_op_end        = 0,
    m_struct_op_start      = M_MAKE_STRUCT_OP( M_STRUCT_OP_START, 0, 0 ),

    m_struct_op_byte       = M_MAKE_STRUCT_OP( M_STRUCT_OP_INT8,  0, 0 ),
    m_struct_op_schar      = M_MAKE_STRUCT_OP( M_STRUCT_OP_INT8,  0, 1 ),

    m_struct_op_uint16_be  = M_MAKE_STRUCT_OP( M_STRUCT_OP_INT16, 0, 0 ),
    m_struct_op_int16_be   = M_MAKE_STRUCT_OP( M_STRUCT_OP_INT16, 0, 1 ),
    m_struct_op_uint16_le  = M_MAKE_STRUCT_OP( M_STRUCT_OP_INT16, 1, 0 ),
    m_struct_op_int16_le   = M_MAKE_STRUCT_OP( M_STRUCT_OP_INT16, 1, 1 ),

    m_struct_op_uint32_be  = M_MAKE_STRUCT_OP( M_STRUCT_OP_INT32, 0, 0 ),
    m_struct_op_uint32_le  = M_MAKE_STRUCT_OP( M_STRUCT_OP_INT32, 0, 1 ),
    m_struct_op_int32_be   = M_MAKE_STRUCT_OP( M_STRUCT_OP_INT32, 1, 0 ),
    m_struct_op_int32_le   = M_MAKE_STRUCT_OP( M_STRUCT_OP_INT32, 1, 1 ),

    m_struct_op_uint24_be  = M_MAKE_STRUCT_OP( M_STRUCT_OP_INT24, 0, 0 ),
    m_struct_op_uint24_le  = M_MAKE_STRUCT_OP( M_STRUCT_OP_INT24, 0, 1 ),
    m_struct_op_int24_be   = M_MAKE_STRUCT_OP( M_STRUCT_OP_INT24, 1, 0 ),
    m_struct_op_int24_le   = M_MAKE_STRUCT_OP( M_STRUCT_OP_INT24, 1, 1 ),

    m_struct_op_bytes      = M_MAKE_STRUCT_OP( M_STRUCT_OP_BYTES, 0, 0 ),
    m_struct_op_skip       = M_MAKE_STRUCT_OP( M_STRUCT_OP_BYTES, 0, 1 )

  } M_Struct_Op;


 /************************************************************************
  *
  * @struct: M_StructField
  *
  * @description:
  *    a handle to a _constant_ @M_StructFieldRec, used to describe a
  *    stored structure field
  */
  typedef const struct M_StructFieldRec_*   M_StructField;

 /************************************************************************
  *
  * @struct: M_StructFieldRec
  *
  * @description:
  *    a simple structure used to model the layout of a single
  *    field in a given structure. Note that each field has a
  *    "memory size" and a "structure size".
  *
  *    the first is the size in bytes used to pack the field in memory
  *    (see @m_struct_write_fields).
  *
  *    the second is the field size in its structure
  *
  *    M_Struct_Field elements are grouped in constant tables named
  *    "structure layout" tables.
  *
  * @fields:
  *    value  :: a @M_Struct_Op value used to describe the field's memory
  *              size and type
  *
  *    size   :: the structure field size
  *
  *    offset :: the field offset relative to its structure's start
  *
  * @note;
  *    defining M_Struct_Field object directly is highly un-recommended.
  *    we strongly suggest you to use the M_STRUCT_XXX macros instead
  *    as they're easy to use and prevent misuse.
  */
  typedef struct  M_StructFieldRec_
  {
    M_Byte       value;
    M_Byte       size;
    M_UShort     offset;

  } M_StructFieldRec;


  /* */

#define  M_STRUCT_FIELD_SIZE( s, f )  \
           ((M_Byte)M_FIELD_SIZE( M_STRUCTURE, f ))

#define  M_STRUCT_FIELD_OFFSET( s, f )  \
           ((M_UShort)M_FIELD_OFFSET( M_STRUCTURE, f ))


#define M_STRUCT_FIELD( struct_op, field )  \
          {                                 \
            struct_op,                      \
            M_STRUCT_FIELD_SIZE( field ),   \
            M_STRUCT_FIELD_OFFSET( field )  \
          }

#define M_MAKE_EMPTY_FIELD( struct_op )  { struct_op, 0, 0 }

 /***********************************************************************
  *
  * @macro: M_STRUCT_START (size)
  *
  * @description:
  *    this macro is used to define a new @M_Struct_Field in a structure
  *    layout table.
  *
  *    its role is to give the total size in bytes of the stored structure
  *    it should only appear at the start of a given layout table, though
  *    is is optional.
  */
#define M_STRUCT_START( size )   { m_struct_op_start, 0, size }

 /***********************************************************************
  *
  * @macro: M_STRUCT_END
  *
  * @description:
  *    this macro is used to define a new @M_Struct_Field in a structure
  *    layout table.
  *
  *    it should _ALWAYS_ appear as the last element of a layout table,
  *    as it indicates its end.
  */
#define M_STRUCT_END             { m_struct_op_end,   0, 0 }

 /***********************************************************************
  *
  * @macro: M_STRUCT_INT32 (field)
  *
  * @description:
  *    this macro is used to define a new @M_Struct_Field in a structure
  *    layout table. it describes a field that is stored as a 4-byte signed
  *    word in memory with big endianess
  *
  *    the field is taken from the structure defined by the _implicit_
  *    @M_STRUCTURE macro
  */
#define M_STRUCT_INT32( f )       M_STRUCT_FIELD( m_struct_op_int32_be, f )

 /***********************************************************************
  *
  * @macro: M_STRUCT_UINT32 (field)
  *
  * @description:
  *    this macro is used to define a new @M_Struct_Field in a structure
  *    layout table. it describes a field that is stored as a 4-byte unsigned
  *    word in memory with big endianess
  *
  *    the field is taken from the structure defined by the _implicit_
  *    @M_STRUCTURE macro
  */
#define M_STRUCT_UINT32( f )      M_STRUCT_FIELD( m_struct_op_uint32_be, f )

 /***********************************************************************
  *
  * @macro: M_STRUCT_INT24 (field)
  *
  * @description:
  *    this macro is used to define a new @M_Struct_Field in a structure
  *    layout table. it describes a field that is stored as a 3-byte signed
  *    word in memory with big endianess
  *
  *    the field is taken from the structure defined by the _implicit_
  *    @M_STRUCTURE macro
  */
#define M_STRUCT_INT24( f )       M_STRUCT_FIELD( m_struct_op_int24_be, f )

 /***********************************************************************
  *
  * @macro: M_STRUCT_UINT24 (field)
  *
  * @description:
  *    this macro is used to define a new @M_Struct_Field in a structure
  *    layout table. it describes a field that is stored as a 3-byte unsigned
  *    word in memory with big endianess
  *
  *    the field is taken from the structure defined by the _implicit_
  *    @M_STRUCTURE macro
  */
#define M_STRUCT_UINT24( f )      M_STRUCT_FIELD( m_struct_op_uint24_be, f )

 /***********************************************************************
  *
  * @macro: M_STRUCT_INT16 (field)
  *
  * @description:
  *    this macro is used to define a new @M_Struct_Field in a structure
  *    layout table. it describes a field that is stored as a 2-byte signed
  *    word in memory with big endianess
  *
  *    the field is taken from the structure defined by the _implicit_
  *    @M_STRUCTURE macro
  */
#define M_STRUCT_INT16( f )       M_STRUCT_FIELD( m_struct_op_int16_be, f )

 /***********************************************************************
  *
  * @macro: M_STRUCT_UINT16 (field)
  *
  * @description:
  *    this macro is used to define a new @M_Struct_Field in a structure
  *    layout table. it describes a field that is stored as a 2-byte unsigned
  *    word in memory with big endianess
  *
  *    the field is taken from the structure defined by the _implicit_
  *    @M_STRUCTURE macro
  */
#define M_STRUCT_UINT16( f )      M_STRUCT_FIELD( m_struct_op_uint16_be, f )

 /***********************************************************************
  *
  * @macro: M_STRUCT_UINT8 (field)
  *
  * @description:
  *    this macro is used to define a new @M_Struct_Field in a structure
  *    layout table. it describes a field that is stored as a 1-byte unsigned
  *    word in memory
  *
  *    the field is taken from the structure defined by the _implicit_
  *    @M_STRUCTURE macro
  */
#define M_STRUCT_UINT8( f )       M_STRUCT_FIELD( m_struct_op_uint8, f )

 /***********************************************************************
  *
  * @macro: M_STRUCT_INT8 (field)
  *
  * @description:
  *    this macro is used to define a new @M_Struct_Field in a structure
  *    layout table. it describes a field that is stored as a 1-byte signed
  *    word in memory
  *
  *    the field is taken from the structure defined by the _implicit_
  *    @M_STRUCTURE macro
  */
#define M_STRUCT_INT8( f )        M_STRUCT_FIELD( m_struct_op_int8, f )


 /***********************************************************************
  *
  * @macro: M_STRUCT_INT32_LE (field)
  *
  * @description:
  *    this macro is used to define a new @M_Struct_Field in a structure
  *    layout table. it describes a field that is stored as a 4-byte signed
  *    word in memory with little endianess
  *
  *    the field is taken from the structure defined by the _implicit_
  *    @M_STRUCTURE macro
  */
#define M_STRUCT_INT32_LE( f )    M_STRUCT_FIELD( m_struct_op_int32_le, f )

 /***********************************************************************
  *
  * @macro: M_STRUCT_UINT32_LE (field)
  *
  * @description:
  *    this macro is used to define a new @M_Struct_Field in a structure
  *    layout table. it describes a field that is stored as a 4-byte unsigned
  *    word in memory with little endianess
  *
  *    the field is taken from the structure defined by the _implicit_
  *    @M_STRUCTURE macro
  */
#define M_STRUCT_UINT32_LE( f )   M_STRUCT_FIELD( m_struct_op_uint32_le, f )

 /***********************************************************************
  *
  * @macro: M_STRUCT_INT16_LE (field)
  *
  * @description:
  *    this macro is used to define a new @M_Struct_Field in a structure
  *    layout table. it describes a field that is stored as a 2-byte signed
  *    word in memory with little endianess
  *
  *    the field is taken from the structure defined by the _implicit_
  *    @M_STRUCTURE macro
  */
#define M_STRUCT_INT16_LE( f )    M_STRUCT_FIELD( m_struct_op_int16_le, f )

 /***********************************************************************
  *
  * @macro: M_STRUCT_UINT16_LE (field)
  *
  * @description:
  *    this macro is used to define a new @M_Struct_Field in a structure
  *    layout table. it describes a field that is stored as a 2-byte unsigned
  *    word in memory with little endianess
  *
  *    the field is taken from the structure defined by the _implicit_
  *    @M_STRUCTURE macro
  */
#define M_STRUCT_UINT16_LE( f )   M_STRUCT_FIELD( m_struct_op_uint16_le, f )

 /***********************************************************************
  *
  * @macro: M_STRUCT_BYTES (field,count)
  *
  * @description:
  *    this macro is used to define a new @M_Struct_Field in a structure
  *    layout table, corresponding to a sequence of "count" contiguous
  *    bytes
  *
  *    the field is taken from the structure defined by the _implicit_
  *    @M_STRUCTURE macro
  */
#define M_STRUCT_BYTES( field, count )     \
          {                                \
            m_struct_op_bytes,             \
            count,                         \
            M_STRUCT_FIELD_OFFSET( field ) \
          }

 /***********************************************************************
  *
  * @macro: M_STRUCT_SKIP_INT32
  *
  * @description:
  *    this macro is used to define a new @M_Struct_Field in a structure
  *    layout table, corresponding to an ignored 4-byte long in memory.
  */
#define M_STRUCT_SKIP_INT32       { m_struct_op_int32_be, 0, 0 }

 /***********************************************************************
  *
  * @macro: M_STRUCT_SKIP_INT24
  *
  * @description:
  *    this macro is used to define a new @M_Struct_Field in a structure
  *    layout table, corresponding to an ignored 3-byte word in memory.
  */
#define M_STRUCT_SKIP_INT24       { m_struct_op_int24_be, 0, 0 }

 /***********************************************************************
  *
  * @macro: M_STRUCT_SKIP_INT16
  *
  * @description:
  *    this macro is used to define a new @M_Struct_Field in a structure
  *    layout table, corresponding to an ignored 2-byte word in memory.
  */
#define M_STRUCT_SKIP_INT16       { m_struct_op_int16_be, 0, 0 }

 /***********************************************************************
  *
  * @macro: M_STRUCT_SKIP_INT8
  *
  * @description:
  *    this macro is used to define a new @M_Struct_Field in a structure
  *    layout table, corresponding to an ignored 1-byte word in memory.
  */
#define M_STRUCT_SKIP_INT8        { m_struct_op_int8,     0, 0 }

 /***********************************************************************
  *
  * @macro: M_STRUCT_SKIP_BYTES (count)
  *
  * @description:
  *    this macro is used to define a new @M_Struct_Field in a structure
  *    layout table, corresponding to a ignored sequence of "count" bytes
  *    in memory.
  */
#define M_STRUCT_SKIP_BYTES( count )  { m_struct_op_skip, count, 0 }


 /***************************************************************************
  *
  * @function: m_struct_read_fields
  *
  * @description:
  *    read a structure's fields from memory (unmarshalling)
  *
  * @input:
  *    fields :: pointer to table of @M_Struct_Field elements used
  *              to describe the fields layout
  *
  *    object :: address of target object
  *
  * @inout:
  *    buffer :: address of memory buffer when the structure is read.
  *              the pointer is incremented by this function
  *
  */
  MLIB_API(void)   m_struct_read_fields( M_StructField  fields,
                                         M_Pointer      object,
                                         M_Byte*       *buffer );


 /***************************************************************************
  *
  * @function: m_struct_write_fields
  *
  * @description:
  *    write a structure's fields to memory (marshalling)
  *
  * @input:
  *    fields :: pointer to table of @M_Struct_Field elements used
  *              to describe the fields layout
  *
  *    object :: address of source object
  *
  * @inout:
  *    abuffer :: address of memory buffer where the structure is written.
  *               the pointer is incremented by this function
  *
  */
  MLIB_API(void)   m_struct_write_fields( M_StructField  fields,
                                          M_Pointer      object,
                                          M_Byte*       *abuffer );

  /* */

M_END_HEADER

#endif /* __MLIB_STRUCTURE_H__ */
