#ifndef __MLIB_TIME_H__
#define __MLIB_TIME_H__

#include <mlib/date.h>

M_BEGIN_HEADER

 /******************************************************************
  *
  * @type: M_Time
  *
  * @description:
  *   handle to a @M_TimeRec structure used to model a given
  *   time.
  */
  typedef struct M_TimeRec_*     M_Time;


 /******************************************************************
  *
  * @struct: M_TimeRec
  *
  * @description:
  *   structure used to model a given time. Note that its fields are
  *   voluntarily obfuscated to prevent direct access..
  *
  * @fields:
  *   xyzxyzxyz ::
  *   xzyxzyxzy ::
  */
  typedef struct M_TimeRec_
  {
    const M_UInt32  xyzxyzxyz;
    const M_UInt32  xzyxzyxzy;

  } M_TimeRec;


 /******************************************************************
  *
  * @function: m_time_now
  *
  * @description:
  *   return the current UTC time
  *
  * @output:
  *   time  :: handle to @M_TimeRec structure used to hold the current
  *            time..
  */
  MLIB_API( void )
  m_time_now( M_Time  time );


 /******************************************************************
  *
  * @function: m_time_compare
  *
  * @description:
  *   compare two time values
  *
  * @input:
  *   time1 :: handle to first time structure
  *   time2 :: handle to second time structure
  *
  * @return:
  *   -1, 0 or +1 depending on the time values
  */
  MLIB_API( M_Int )
  m_time_compare( const M_Time  time1,
                  const M_Time  time2 );


 /******************************************************************
  *
  * @function: m_time_to_ansi
  *
  * @description:
  *   convert a @M_TimeRec value to ANSI 'time_t'
  *
  * @input:
  *   time :: handle to source time value
  *
  * @return:
  *   ANSI 'time_t' value
  */
  MLIB_API( time_t )
  m_time_to_ansi( M_Time  time );


 /******************************************************************
  *
  * @function: m_time_from_ansi
  *
  * @description:
  *   convert an ANSI 'time_t' into a @M_TimeRec value
  *
  * @input:
  *   time      :: handle to target time value
  *   ansi_time :: ANSI time value
  */
  MLIB_API( void )
  m_time_from_ansi( M_Time  time,
                    time_t  anti_time );


 /******************************************************************
  *
  * @function: m_time_to_date
  *
  * @description:
  *   convert a time value into a date and seconds period
  *
  * @input:
  *   time      :: handle to target time value
  *
  * @output:
  *   date      :: target @M_DateRec structure
  *   pseconds  :: number of seconds since midnight of date. can be NULL
  *   pusec     :: number of microseconds passed 'pseconds'. can be NULL
  */
  MLIB_API( void )
  m_time_to_date( M_Time   time,
                  M_Date   date,
                  M_Int32 *pseconds,
                  M_Int32 *pusec );



 /******************************************************************
  *
  * @function: m_time_from_date
  *
  * @description:
  *   convert a time value into a date and seconds period
  *
  * @input:
  *   time      :: handle to target time value
  *   date      :: handle to source date. cannot be NULL !!
  *   seconds   :: number of seconds since midnight of date. can be NULL
  *   usec      :: number of microseconds passed 'seconds'. can be NULL
  */
  MLIB_API( M_Bool )
  m_time_from_date( M_Time   time,
                    M_Date   date,
                    M_Int32  seconds,
                    M_Int32  usec );


 /********************************************************************
  *
  * @type: M_DateTime
  *
  * @description:
  *   handle to a @M_DateTime structure used to model a full date and time
  */
  typedef struct M_DateTimeRec_*     M_DateTime;


 /********************************************************************
  *
  * @struct:  M_DateTimeRec
  *
  * @description:
  *   a structure used to model a full date and time
  *
  * @fields:
  *   usec          :: micro seconds past 'second'
  *   second        :: seconds (0-59) past 'minute'
  *   minute        :: minute (0-59) past 'hour'
  *   hour          :: hour past midnight
  *
  *   day           :: date day
  *   month         :: date month
  *   year          :: date year
  *
  *   week_day      :: week day (0 == sunday)
  *   year_day      :: year day (1 == January 1st)
  *   is_daylight   :: is it a daylight saving time ?
  *   gmt_offset    :: seconds east of GMT
  */
  typedef struct M_DateTimeRec_
  {
    const M_Int32      usec;
    const M_Int        sec;
    const M_Int        min;
    const M_Int        hour;

    const M_DateDay    day;
    const M_DateMonth  month;
    const M_DateYear   year;

    const M_Int        week_day;
    const M_Int        year_day;
    const M_Bool       is_daylight;
    const M_Int32      gmt_offset;

  } M_DateTimeRec;


 /* */
 
  MLIB_APIDEF( void )
  m_date_time_now( M_DateTime  date_time );

  MLIB_APIDEF( M_Int )
  m_date_time_compare( const M_DateTime  date_time1,
                       const M_DateTime  date_time2 );

  MLIB_APIDEF( void )
  m_date_time_import( M_DateTime  date_time,
                      M_Time      time );


  MLIB_APIDEF( void )
  m_date_time_export( M_DateTime  date_time,
                      M_Time      time );

  MLIB_APIDEF( void )
  m_date_time_export_date( M_DateTime  date_time,
                           M_Date      date );

  MLIB_APIDEF( void )
  m_date_time_import_date( M_DateTime  date_time,
                           M_Date      date,
                           M_Int32     seconds );

 /* */

M_END_HEADER

#endif /* __MLIB_TIME_H__ */
