#ifndef __MLIB_TREE_H__
#define __MLIB_TREE_H__

#include <mlib.h>

M_BEGIN_HEADER

 /*************************************************************************
  *
  * @section: trees
  *
  * @title: Trees
  *
  * @abstract: Fast ordered containers
  *
  * @description:
  *   an MLib tree is a generic container that implements
  *   a simple ordered dictionary of (key,value) pairs, where
  *   keys and values can be of any pointer type.
  *
  *   See the @M_HashTable type for fast unordered containers.
  *   Trees are implemented as balanced AVL trees, using a fast
  *   non-recursive implementation.
  */

 /**********************************************************************
  *
  * @type: M_Tree
  *
  * @description:
  *   handle to a generic tree container. A tree is a sorted map of
  *   (key,value) pairs
  */
  typedef struct M_TreeRec_*    M_Tree;


 /**********************************************************************
  *
  * @function: m_tree_new
  *
  * @description:
  *    create a new (empty) generic tree
  *
  * @input:
  *    memory      :: memory manager handle
  *    key_compare :: simple key comparison function
  *
  * @return:
  *    handle to new tree object
  *
  * @throws: @m_err_memory_alloc
  */
  MLIB_API(M_Tree)     m_tree_new( M_Memory       memory,
                                   M_CompareFunc  key_compare );

 /**********************************************************************
  *
  * @function: m_tree_new_full
  *
  * @description:
  *    create a new (empty) generic tree, with more specific parameters
  *
  * @input:
  *    memory  :: memory manager handle
  *
  *    key_compare :: key comparison function
  *    key_destroy :: key destructor, can be NULL
  *    key_data    :: key-specific data pointer
  *    val_destroy :: value destructor, can be NULL
  *    val_data    :: value-specific data pointer
  *
  * @return:
  *    handle to new tree object
  *
  * @throws: @m_err_memory_alloc
  *
  * @note:
  *    the key destructor is called whenever a non-NULL key is removed from
  *    the tree. the key data pointer is passed to both the comparison
  *    function and the destructor
  *
  *    the value destructor is called whenever a non-NULL value is removed or
  *    replaced from the tree. it receives the value data pointer as its
  *    second argument
  *
  *    the desctructors should _never_ raise an exception. the
  *    comparison function can raise an exception, but should
  *    be avoided whenever necessary
  */
  MLIB_API(M_Tree)     m_tree_new_full( M_Memory           memory,
                                        M_CompareDataFunc  key_compare,
                                        M_DestroyDataFunc  key_destroy,
                                        M_Pointer          key_data,
                                        M_DestroyDataFunc  val_destroy,
                                        M_Pointer          val_data );


 /**********************************************************************
  *
  * @function: m_tree_destroy
  *
  * @description:
  *    destroy a given tree (after clearing it)
  *
  * @input:
  *    tree :: handle to target tree
  */
  MLIB_API(void)       m_tree_destroy( M_Tree  tree );


 /**********************************************************************
  *
  * @function: m_tree_get_size
  *
  * @description:
  *    return the size (i.e. number of nodes) in a tree
  *
  * @input:
  *    tree :: handle to target tree
  *
  * @return:
  *    number of nodes in the tree
  */
  MLIB_API(M_ULong)    m_tree_get_size( M_Tree  tree );


 /**********************************************************************
  *
  * @function: m_tree_get_depth
  *
  * @description:
  *    return the depth of a given tree
  *
  * @input:
  *    tree :: handle to target tree
  *
  * @return:
  *    depth of the tree
  */
  MLIB_API(M_UInt)     m_tree_get_depth( M_Tree  tree );


 /**********************************************************************
  *
  * @function: m_tree_get
  *
  * @description:
  *    retrieve the value associated to a given key.
  *
  * @input:
  *    tree :: handle to target tree
  *    key  :: lookup key
  *
  * @return:
  *    value associated to key. NULL if the key isn't part of the tree
  */
  MLIB_API(M_Pointer)  m_tree_get( M_Tree     tree,
                                   M_Pointer  key );


 /**********************************************************************
  *
  * @function: m_tree_lookup
  *
  * @description:
  *    tests wether a key is part of a tree, and return its value
  *
  * @input:
  *    tree :: handle to target tree
  *    key  :: lookup key
  *
  * @output:
  *    avalue :: key value, if any. (this parameter is optional)
  *
  * @return:
  *    boolean. set iff the key is set in the tree
  */
  MLIB_API(M_Bool)     m_tree_lookup( M_Tree      tree,
                                      M_Pointer   key,
                                      M_Pointer  *avalue );


 /**********************************************************************
  *
  * @function: m_tree_set
  *
  * @description:
  *    set or replace the value corresponding to a given key in the tree
  *
  * @input:
  *    tree  :: handle to target tree
  *    key   :: lookup key
  *    value :: new value for this key
  *
  * @throws: @m_err_memory_alloc
  */
  MLIB_API(void)       m_tree_set( M_Tree     tree,
                                   M_Pointer  key,
                                   M_Pointer  value );


 /**********************************************************************
  *
  * @function: m_tree_remove
  *
  * @description:
  *    removes a given key, and its associated value, from the tree
  *
  * @input:
  *    tree  :: handle to target tree
  *    key   :: lookup key
  */
  MLIB_API(void)       m_tree_remove( M_Tree     tree,
                                      M_Pointer  key );

  /* */

M_END_HEADER

#endif /* __MLIB_TREE_H__ */
