/* basic data type definitions */
#ifndef __MLIB_TYPES_H__
#define __MLIB_TYPES_H__

#ifndef __MLIB_H__
#error  "do not include this file directly, use <mlib.h> instead"
#endif

/* yes, we require ANSI C header files !! */
#include <stddef.h>

M_BEGIN_HEADER


/******************************************************************
 *
 * @section: types
 *
 * @title: Basic Types
 *
 * @abstract: Basic Data Types
 *
 * @description:
 *    This section contains the definitions of basic data types defined
 *    by the MLib
 */

 /*********************************************************************/
 /*********************************************************************/
 /*****                                                           *****/
 /*****                    BASIC DATA TYPES                       *****/
 /*****                                                           *****/
 /*********************************************************************/
 /*********************************************************************/


 /***************************************************************
  *
  * @type: M_Byte
  *
  * @description:
  *    an unsigned byte. Equivalent to unsigned char
  *    though a bit more convenient to use..
  *
  * @also:
  *   the @M_ByteP type defines pointers to bytes, and @M_CByteP defines
  *   pointers to constant bytes..
  */
  typedef unsigned char   M_Byte;


 /***************************************************************
  *
  * @type: M_SChar
  *
  * @description:
  *    a signed byte. Equivalent to signed char
  *    though a bit more convenient to use..
  */
  typedef signed char     M_SChar;


 /***************************************************************
  *
  * @type: M_Char
  *
  * @description:
  *    a byte character. Equivalent to char but defined
  *    for completeness.
  *
  * @also:
  *   the @M_String type defines pointers to chars, and @M_CString
  *   defines pointers to constant chars.
  */
  typedef char            M_Char;


 /***************************************************************
  *
  * @type: M_Short
  *
  * @description:
  *    a two-byte signed word. Equivalent to short
  *    but defined for completeness.
  */
  typedef short           M_Short;


 /***************************************************************
  *
  * @type: M_UShort
  *
  * @description:
  *    a two-byte unsigned word. Equivalent to unsigned short
  *    though a bit more convenient to use.
  */
  typedef unsigned short  M_UShort;


 /***************************************************************
  *
  * @type: M_Int
  *
  * @description:
  *    an integer. Equivalent to int but defined for
  *    completeness
  */
  typedef int             M_Int;


 /***************************************************************
  *
  * @type: M_UInt
  *
  * @description:
  *    an unsigned word. Equivalent to unsigned int
  *    though a bit more convenient to use
  */
  typedef unsigned int    M_UInt;


 /***************************************************************
  *
  * @type: M_Fast
  *
  * @description:
  *    a signed integer word that is at least 32-bits. This is
  *    often more convenient to use than @M_Long which may be
  *    be 64-bits on some platforms (and slow)
  */
#if M_SIZEOF_INT < 4
  typedef long            M_Fast;
#else
  typedef int             M_Fast;
#endif

 /***************************************************************
  *
  * @type: M_UFast
  *
  * @description:
  *    an unsigned integer word that is at least 32-bits. This is
  *    often more convenient to use than @M_ULong which may be
  *    be 64-bits on some platforms (and slow)
  */
#if M_SIZEOF_INT < 4
  typedef unsigned long   M_UFast;
#else
  typedef unsigned int    M_UFast;
#endif

 /***************************************************************
  *
  * @type: M_Long
  *
  * @description:
  *    a long integer. Equivalent to long but defined for
  *    completeness
  */
  typedef long            M_Long;

 /***************************************************************
  *
  * @type: M_ULong
  *
  * @description:
  *    an unsigned long integer. Equivalent to unsigned long
  *    though a bit more convenient to use
  */
  typedef unsigned long   M_ULong;


 /***************************************************************
  *
  * @type: M_Int8
  *
  * @description:
  *    an 8-bit signed integer. Equivalent to @M_SChar though
  *    defined for clarity.
  */
  typedef M_SChar         M_Int8;

 /***************************************************************
  *
  * @type: M_UInt8
  *
  * @description:
  *    an 8-bit unsigned integer. Equivalent to @M_Byte though
  *    defined for clarity.
  */
  typedef M_Byte          M_UInt8;

 /***************************************************************
  *
  * @type: M_Int16
  *
  * @description:
  *    a 16-bit signed integer. Equivalent to @M_Short though
  *    defined for clarity.
  */
  typedef short           M_Int16;

 /***************************************************************
  *
  * @type: M_UInt16
  *
  * @description:
  *    a 16-bit unsigned integer. Equivalent to @M_UShort though
  *    defined for clarity.
  */
  typedef unsigned short  M_UInt16;


 /***************************************************************
  *
  * @type: M_Int32
  *
  * @description:
  *    a 32-bit signed integer on all platforms
  *
  * @also:
  *   @M_UInt32, @M_INT32, @M_INT32_C, @M_INT32_FMT
  *   @M_INT32_MIN, @M_INT32_MAX, @M_INT32_MSB
  */
#if M_SIZEOF_INT == 4
  typedef int             M_Int32;
#elif M_SIZEOF_LONG == 4
  typedef long            M_Int32;
#else
# error "no 32-bits type found !!"
#endif

 /***************************************************************
  *
  * @type: M_UInt32
  *
  * @description:
  *    a 32-bit unsigned integer on all platforms
  *
  * @also:
  *   @M_Int32, @M_UINT32, @M_UINT32_C, @M_UINT32_FMT
  *   @M_UINT32_MIN, @M_UINT32_MAX, @M_UINT32_MSB
  */
#if M_SIZEOF_INT == 4
  typedef unsigned int    M_UInt32;
#elif M_SIZEOF_LONG == 4
  typedef unsigned long   M_UInt32;
#else
# error "no 32-bits type found !!"
#endif

 /***************************************************************
  *
  * @type: M_Int64
  *
  * @description:
  *    a 64-bit signed integer. only available on certain platforms
  *
  * @also:
  *   @M_UInt64, @M_INT64, @M_INT64_C, @M_INT64_FMT
  *   @M_INT64_MIN, @M_INT64_MAX, @M_INT64_MSB
  *   @M_INT64_LO, @M_INT64_HI, @M_INT64_MAKE
  */
#ifdef M_CONFIG_HAS_INT64
  typedef M_INT64_T        M_Int64;
#endif

 /***************************************************************
  *
  * @type: M_UInt64
  *
  * @description:
  *    a 64-bit unsigned integer. only available on certain platforms
  *
  * @also:
  *   @M_Int64, @M_UINT64, @M_UINT64_C, @M_UINT64_FMT
  *   @M_UINT64_MAX, @M_UINT64_MSB
  *   @M_UINT64_LO, @M_UINT64_HI, @M_UINT64_MAKE
  */
#ifdef M_CONFIG_HAS_INT64
  typedef unsigned M_INT64_T  M_UInt64;
#endif


 /***************************************************************
  *
  * @type: M_Size
  *
  * @description:
  *    an _unsigned_ type large enough to hold memory sizes.
  */
  typedef size_t          M_Size;


#if 0
 /***************************************************************
  *
  * @type: M_SSize
  *
  * @description:
  *    a _signed_ version of @M_Size
  */
  typedef signed M_Size   M_SSize;
#endif

 /***************************************************************
  *
  * @type: M_Dist
  *
  * @description:
  *    the largest signed integer available on the current
  *    platform to express pointer differences.
  */
  typedef ptrdiff_t       M_Dist;


 /***************************************************************
  *
  * @type: M_String
  *
  * @description:
  *    a pointer to a character string. Equivalent to 'char *'
  *    though a bit more convenient..
  */
  typedef char*           M_String;


 /***************************************************************
  *
  * @type: M_CString
  *
  * @description:
  *    a pointer to a constant character string. Equivalent to
  *    'const char *' though a bit more convenient..
  */
  typedef const char*     M_CString;


 /***************************************************************
  *
  * @type: M_Error
  *
  * @description:
  *    an integer error code. 0 means success, everything else
  *    indicates failure..
  */
  typedef int             M_Error;


 /***************************************************************
  *
  * @type: M_Bool
  *
  * @description:
  *    a boolean, whose value can only be TRUE or FALSE
  */
  typedef int             M_Bool;


 /***************************************************************
  *
  * @macro: M_BOOL(x)
  *
  * @description:
  *    a macro used to convert a value to a boolean
  */
#define  M_BOOL(x)   ( (x) != 0 )


 /***************************************************************
  *
  * @type: M_Pointer
  *
  * @description:
  *    a typeless pointer. Equivalent to void* though a bit
  *    more convenient to use
  */
  typedef void*           M_Pointer;


 /***************************************************************
  *
  * @type: M_CPointer
  *
  * @description:
  *    a typeless constant pointer. Equivalent to const void*
  *    though a bit more convenient to use
  */
  typedef const void*     M_CPointer;


 /*********************************************************************/
 /*********************************************************************/
 /*****                                                           *****/
 /*****                    FORMATTING                             *****/
 /*****                                                           *****/
 /*********************************************************************/
 /*********************************************************************/

 /***************************************************************
  *
  * @macro: M_INT32 (value)
  *
  * @description:
  *    a macro used to cast anything to a @M_Int32. If the value is
  *    constant, use @M_INT32_C instead
  */
#define  M_INT32(x)    ((M_Int32)(x))


 /***************************************************************
  *
  * @macro: M_INT32_C (constant)
  *
  * @description:
  *    this macro is used to specify a 32-bit constant within C source code.
  *    for example:
  *
  *    {
  *       M_Int32  x = M_INT32_C(0x10000);
  *    }
  */
#if M_SIZEOF_INT >= 4
#  define  M_INT32_C(x)     ( x )
#else
#  define  M_INT32_C(x)     ( x ## L )
#endif


 /***************************************************************
  *
  * @macro: M_INT32_FMT
  *
  * @description:
  *    this macro returns a string used in printf-like format strings
  *    to name 32-bit integers.. Here's an example:
  *
  *  {
  *    M_Int32  x = ..... ;
  *
  *    printf( "x is " M_INT32_FMT "\n", x );
  *  }
  *
  *  on most 32-bit machines, it defaults to "%d"
  */
#define  M_INT32_FMT       "%" M_INT32_FMT_0


 /***************************************************************
  *
  * @macro: M_INT32_FMT2 (prefix)
  *
  * @description:
  *    this macro returns a string used in printf-like format strings
  *    to name 32-bit integers.. it includes a _prefix_ that
  *    is inserted between the "%" and the type suffix
  *
  *    for example, the two following statements are equivalent
  *    on most 32-bit machines:
  *
  *    {
  *      printf( "x => " M_INT32_FMT2(12) "\n", x );
  *      printf( "x => %12d\n", x );
  *    }
  *
  *    note that the prefix is automatically stringified by the macro
  */
#define  M_INT32_FMT2(s)      "%" M_STRINGIFY(s) M_INT32_FMT_0


 /***************************************************************
  *
  * @macro: M_INT32_FMT_0
  *
  * @description:
  *    this macro returns a string used in printf-like format strings
  *    to name 32-bit integers.. however, it doesn't have a '%' in it
  *
  *    most users will use @M_INT32_FMT or @M_INT32_FMT2
  *
  *  {
  *    M_Int32  x = ..... ;
  *
  *    printf( "x is %" M_INT32_FMT_0 "\n", x );
  *  }
  *
  *  on most 32-bit machines, it defaults to "d"
  */
#ifndef M_INT32_FMT_0
#  if M_SIZEOF_INT >= 4
#    define  M_INT32_FMT_0      "d"
#  else
#    define  M_INT32_FMT_0      "ld"
#  endif
#endif

 /***************************************************************
  *
  * @macro: M_INT32_MIN
  *
  * @description:
  *    returns the lowest 32-bit signed value. This isn't 0x80000000
  *    because this value is equal to its negative, and this causes
  *    subtle problems (bugs) in certain expressions that use
  *    @M_INT32_MIN
  */
#define  M_INT32_MIN  M_INT32_C(0x80000001)

 /***************************************************************
  *
  * @macro: M_INT32_MAX
  *
  * @description:
  *    returns the highest 32-bit signed value
  */
#define  M_INT32_MAX  M_INT32_C(0x7FFFFFFF)

 /***************************************************************
  *
  * @macro: M_INT32_MSB
  *
  * @description:
  *    returns the value corresponding to the most significant bit
  *    in a 32-bit value. note that this is negative !!
  */
#define  M_INT32_MSB  M_INT32_C(0x80000000)



 /***************************************************************
  *
  * @macro: M_UINT32 (value)
  *
  * @description:
  *    a macro used to cast anything to a @M_UInt32. If the value is
  *    constant, use @M_UINT32_C instead
  */
#define  M_UINT32(x)      ((M_UInt32)(x))


 /***************************************************************
  *
  * @macro: M_UINT32_C (constant)
  *
  * @description:
  *    same as @M_INT32_C for _unsigned_ 32-bit constants
  */
#if M_SIZEOF_INT >= 4
#define  M_UINT32_C(x)     ( x ## U )
#else
#define  M_UINT32_C(x)     ( x ## UL )
#endif


 /***************************************************************
  *
  * @macro: M_UINT32_FMT
  *
  * @description:
  *    same as @M_INT32_FMT for unsigned integers
  */
#define  M_UINT32_FMT     M_INT32_FMT

 /***************************************************************
  *
  * @macro: M_UINT32_FMT2
  *
  * @description:
  *    same as @M_INT32_FMT2 for unsigned integers
  */
#define  M_UINT32_FMT2    M_INT32_FMT2


 /***************************************************************
  *
  * @macro: M_UINT32_MAX
  *
  * @description:
  *    returns the highest 32-bit unsigned value
  */
#define  M_UINT32_MAX     M_UINT32_C(0xFFFFFFFF)


 /***************************************************************
  *
  * @macro: M_UINT32_MSB
  *
  * @description:
  *    returns a value corresponding to the most significant bit
  *    of a 32-bit unsigned integer
  */
#define  M_UINT32_MSB     M_UINT32_C(0x80000000)

 /* */

#define  M_UINT32_HI(x)   (M_UINT32(x) >> 16)
#define  M_UINT32_LO(x)   (M_UINT32(x) & M_UINT32_C(0xFFFF))

 /* */

#ifdef M_CONFIG_HAS_INT64

 /***************************************************************
  *
  * @macro: M_INT64 (value)
  *
  * @description:
  *    macro to cast anything to a signed 64-bit integer. If
  *    the value is constant, use @M_INT64_C instead..
  */
#  define M_INT64(x)    ((M_Int64)(x))

 /***************************************************************
  *
  * @macro: M_INT64_C (constant)
  *
  * @description:
  *    same as @M_INT32_C, but for 64-bit signed integers. not
  *    available on all platforms
  */
#  ifndef M_INT64_C
#    if M_SIZEOF_LONG >= 8
#      define M_INT64_C(x)   (x)
#    else
#      define M_INT64_C(x)   ( x ## LL )
#    endif
#  endif

 /***************************************************************
  *
  * @macro: M_UINT64 (value)
  *
  * @description:
  *    macro to cast anything to an unsigned 64-bit integer. If
  *    the value is constant, use @M_INT64_C instead..
  */
#  define  M_UINT64(x)   ((M_UInt64)(x))


 /***************************************************************
  *
  * @macro: M_UINT64_C (constant)
  *
  * @description:
  *    same as @M_UINT32_C, but for 64-bit unsigned integers. not
  *    available on all platforms
  */
#  ifndef M_UINT64_C
#    if M_SIZEOF_LONG >= 8
#      define M_UINT64_C(x)  ( x ## UL )
#    else
#      define M_UINT64_C(x)  ( x ## ULL )
#    endif
#  endif


 /***************************************************************
  *
  * @macro: M_INT64_FMT
  *
  * @description:
  *    same as @M_INT32_FMT, for 64-bit integers.
  *    not available on all platforms
  */
#  define  M_INT64_FMT    "%" M_INT64_FMT_0


 /***************************************************************
  *
  * @macro: M_INT64_FMT2 (prefix)
  *
  * @description:
  *    same as @M_INT32_FMT2, for 64-bit integers.
  *    not available on all platforms
  */
#  define  M_INT64_FMT2(s)   "%" M_STRINGIFY(s) M_INT64_FMT_0


 /***************************************************************
  *
  * @macro: M_UINT64_FMT
  *
  * @description:
  *    same as @M_UINT32_FMT, for 64-bit _unsigned_ integers.
  *    not available on all platforms
  */
#  define   M_UINT64_FMT   M_INT64_FMT


 /***************************************************************
  *
  * @macro: M_INT64_FMT_0
  *
  * @description:
  *    same as @M_INT32_FMT_0, for 64-bit integer types
  *    not available on all platforms
  *
  * @note:
  *    the definition here is a default that works with most
  *    compilers..
  *
  *    compiler-specific definitions should be defined in their own
  *    configuration header (e.g. <mlib/config/compiler/visualc.h>
  *    for Visual C++ defines it as "I64"
  */
#  ifndef M_INT64_FMT_0
#    define  M_INT64_FMT_0  "ll"
#  endif

 /***************************************************************
  *
  * @macro: M_INT64_LO (value)
  *
  * @description:
  *    retrieves the low 32-bits of a 64-bit integer, result is a
  *    32-bit _unsigned_ integer
  */
#  define   M_INT64_LO(x)   ((M_UInt32) (x))


 /***************************************************************
  *
  * @macro: M_INT64_HI (value)
  *
  * @description:
  *    retrieves the high 32-bits of a 64-bit integer, result is a
  *    32-bit _signed_ integer
  */
#  define   M_INT64_HI(x)   ((M_Int32 )((x) >> 32))


 /***************************************************************
  *
  * @macro: M_UINT64_LO (value)
  *
  * @description:
  *    retrieves the low 32-bits of an 64-bit integer, result is a
  *    32-bit _unsigned_ integer
  */
#  define   M_UINT64_LO(x)  M_INT64_LO(x)


 /***************************************************************
  *
  * @macro: M_INT64_HI (value)
  *
  * @description:
  *    retrieves the high 32-bits of a 64-bit integer, result is a
  *    32-bit _unsigned_ integer
  */
#  define   M_UINT64_HI(x)  ((M_UInt32)((x) >> 32))


 /***************************************************************
  *
  * @macro: M_INT64_MAKE (hi,lo)
  *
  * @description:
  *    make a new 64-bit value from two 32-bit "high" and "low" values
  */
#  define   M_INT64_MAKE(hi,lo)   \
               (((M_Int64 )(M_Int32 )(hi) << 32) | (M_UInt32)(lo))


 /***************************************************************
  *
  * @macro: M_UINT64_MAKE (hi,lo)
  *
  * @description:
  *    make a new unsigned 64-bit value from two 32-bit "high" and
  *    "low" values
  */
#  define   M_UINT64_MAKE(hi,lo)  \
               (((M_UInt64)(M_UInt32)(hi) << 32) | (M_UInt32)(lo))


 /***************************************************************
  *
  * @macro: M_UINT64_MAX
  *
  * @description:
  *    return the highest 64-bit unsigned integer value
  */
#  define   M_UINT64_MAX   ((M_UInt64)-1)


 /***************************************************************
  *
  * @macro: M_INT64_MAX
  *
  * @description:
  *    return the highest 64-bit signed integer value
  */
#  define   M_INT64_MAX    M_INT64_C(0x7FFFFFFFFFFFFFFF)


 /***************************************************************
  *
  * @macro: M_INT64_MIN
  *
  * @description:
  *    return the lowest 64-bit signed integer value. Note that
  *    this is not 0x8000000000000000 since this value is equal
  *    to its own negative, which creates problems in certain
  *    expression who need to use @M_INT64_MIN
  */
#  define   M_INT64_MIN    M_INT64_C(0x8000000000000001)


 /***************************************************************
  *
  * @macro: M_INT64_MSB
  *
  * @description:
  *    return the value corresponding to the most significant bit of
  *    a 64-bit integer
  */
#  define   M_INT64_MSB    M_INT64_C(0x8000000000000000)


 /***************************************************************
  *
  * @macro: M_UINT64_MSB
  *
  * @description:
  *    same as @M_INT64_MSB, returns an unsigned constant
  */
#  define   M_UINT64_MSB   M_UINT64_C(0x8000000000000000)

 /* */

#endif /* M_CONFIG_HAS_INT64 */


 /****************************************************************
  *
  * @macro: M_SIZE_MAX
  *
  * @description:
  *    the greatest value of a @M_Size variable
  */
#define  M_SIZE_MAX   ((M_Size)~0)


 /****************************************************************
  *
  * @macro: M_NUM_ELEMENTS (table)
  *
  * @description:
  *    a macro used to return the number of elements in a
  *    static table.
  */
#define  M_NUM_ELEMENTS(t)    (sizeof(t)/sizeof(t[0]))


 /****************************************************************
  *
  * @macro: M_FIELD_OFFSET (type,field)
  *
  * @description:
  *    returns the offset in bytes of a given field in an
  *    object of type 'type'.
  */
#define  M_FIELD_OFFSET(_type,_field)    offsetof(_type,_field)


 /****************************************************************
  *
  * @macro: M_FIELD_SIZE (type,field)
  *
  * @description:
  *    returns the size in bytes of a given field in an
  *    object of type '*ptype'.
  *
  * @note:
  *    we need to use a dummy constant variable in this macro
  *    because something like:
  *
  *    {
  *      sizeof( ((_type*)0)->_field )
  *    }
  *
  *    is not portable among all compilers/platforms..
  */
#define  M_FIELD_SIZE(_type,_field)     \
               sizeof( ((_type*)&_mlib_dummy_byte)->_field )


 /****************************************************************
  *
  * @macro: M_POINTER (x)
  *
  * @description:
  *    a convenient macro used to convert anything to a pointer
  */
#define  M_POINTER(x)    ((M_Pointer)(x))


 /****************************************************************
  *
  * @macro: M_POINTER_P (x)
  *
  * @description:
  *    a convenient macro used to convert anything to a pointer
  *    pointer
  */
#define  M_POINTER_P(x)  ((M_Pointer*)(x))


 /****************************************************************
  *
  * @macro: M_POINTER_TO_DIST (x)
  *
  * @description:
  *    a convenient macro used to convert a pointer to a @M_Dist
  *    integer (without potential bit loss)
  *
  * @note:
  *    you should only use this macro with pointer values created
  *    with @M_POINTER_FROM_DIST. Any other use might result in
  *    unexpected results on some platforms
  */
#define  M_POINTER_TO_DIST(x)    ((M_Dist)(x))


 /****************************************************************
  *
  * @macro: M_POINTER_FROM_DIST (x)
  *
  * @description:
  *    a convenient macro used to convert a @M_Dist into a pointer.
  *
  * @note:
  *    only use @M_POINTER_TO_DIST to retrieve the value embedded
  *    in the pointer..
  */
#define  M_POINTER_FROM_DIST(x)  ((M_Pointer)(M_Dist)(x))


 /****************************************************************
  *
  * @macro: M_POINTER_TO_UINT (x)
  *
  * @description:
  *    a convenient macro used to convert a pointer to a @M_UInt
  *    integer.
  *
  * @note:
  *    you should only use this macro with pointer values created
  *    with @M_POINTER_FROM_UINT. Any other use might result in
  *    unexpected results on some platforms
  */
#define  M_POINTER_TO_UINT(x)    ((M_UInt)(x))


 /****************************************************************
  *
  * @macro: M_POINTER_FROM_UINT (x)
  *
  * @description:
  *    a convenient macro used to convert a @M_Dist into a pointer.
  *
  * @note:
  *    only use @M_POINTER_TO_UINT to retrieve the value embedded
  *    in the pointer.
  */
#define  M_POINTER_FROM_UINT(x)  ((M_Pointer)(M_UInt)(x))


 /****************************************************************
  *
  * @macro: M_POINTER_TO_INT (x)
  *
  * @description:
  *    a convenient macro used to convert a pointer to a @M_Int
  *    integer (without potential bit loss)
  *
  * @note:
  *    you should only use this macro with pointer values created
  *    with @M_POINTER_FROM_INT. Any other use might result in
  *    unexpected results on some platforms
  */
#define  M_POINTER_TO_INT(x)    ((M_Int)(M_Pointer)(x))


 /****************************************************************
  *
  * @macro: M_POINTER_FROM_INT (x)
  *
  * @description:
  *    a convenient macro used to convert a @M_Int into a pointer.
  *
  * @note:
  *    only use @M_POINTER_TO_INT to retrieve the value embedded
  *    in the pointer.
  */
#define  M_POINTER_FROM_INT(x)  ((M_Pointer)(M_Int)(x))


 /****************************************************************
  *
  * @macro: M_CPOINTER
  *
  * @description:
  *    a convenient macro used to convert anything to a constant
  *    pointer
  */
#define  M_CPOINTER(x)    ((M_CPointer)(x))


 /****************************************************************
  *
  * @macro: M_CPOINTER_P (x)
  *
  * @description:
  *    a convenient macro used to convert anything to a pointer to
  *    a constant pointer
  */
#define  M_CPOINTER_P(x)  ((M_CPointer*)(x))

 /* */

 /****************************************************************
  *
  * @type: M_ByteP
  *
  * @description:
  *    a type used to model a pointer to bytes (@M_Byte)
  */
typedef  M_Byte*        M_ByteP;

 /****************************************************************
  *
  * @type: M_CByteP
  *
  * @description:
  *    a type used to model a pointer to constant bytes
  */
typedef  const M_Byte*  M_CByteP;

 /****************************************************************
  *
  * @macro: M_BYTEP
  *
  * @description:
  *    a convenient macro used to convert any pointer to a @M_ByteP
  */
#define  M_BYTEP(x)   ((M_ByteP)(x))

 /****************************************************************
  *
  * @macro: M_CBYTEP
  *
  * @description:
  *    a convenient macro used to convert any pointer to a @M_CByteP
  */
#define  M_CBYTEP(x)  ((M_CByteP)(x))


 /*********************************************************************/
 /*********************************************************************/
 /*****                                                           *****/
 /*****                   FUNCTION CALLBACKS TYPES                *****/
 /*****                                                           *****/
 /*********************************************************************/
 /*********************************************************************/


 /****************************************************************
  *
  * @functype: M_DestroyFunc
  *
  * @description:
  *    a type describing a function used to destroy (or unref) a
  *    single object
  *
  * @input:
  *    item :: handle to target element
  */
  typedef void  (*M_DestroyFunc)( M_CPointer  item );


 /****************************************************************
  *
  * @functype: M_DestroyDataFunc
  *
  * @description:
  *    a variant of @M_DestroyFunc used to pass item-specific
  *    data to the destructor
  *
  * @input:
  *    item      :: handle to target element
  *    item_data :: item-specific data
  */
  typedef void  (*M_DestroyDataFunc)( M_CPointer  item,
                                      M_CPointer  item_data );


 /****************************************************************
  *
  * @functype: M_EqualFunc
  *
  * @description:
  *    a type describing a simple equality test function
  *
  * @input:
  *    a :: first element
  *    b :: second element
  *
  * @return:
  *   TRUE iff the two elements are equal
  */
  typedef M_Bool  (*M_EqualFunc)( M_CPointer  a,
                                  M_CPointer  b );


 /****************************************************************
  *
  * @functype: M_EqualDataFunc
  *
  * @description:
  *    a type describing a simple equality test function using
  *    extra data
  *
  * @input:
  *    a    :: first element
  *    b    :: second element
  *    data :: generic data pointer
  *
  * @return:
  *   TRUE iff the two elements are equal
  */
  typedef M_Bool  (*M_EqualDataFunc)( M_CPointer  a,
                                      M_CPointer  b,
                                      M_CPointer  data );


 /****************************************************************
  *
  * @functype: M_CompareFunc
  *
  * @description:
  *    a type describing a simple comparison function
  *
  * @input:
  *    a :: first element
  *    b :: second element
  *
  * @return:
  *   -1, 0 or +1 if a < b, a==b or a > b, respectively
  */
  typedef M_Int  (*M_CompareFunc)( M_CPointer   a,
                                   M_CPointer   b );


 /****************************************************************
  *
  * @functype: M_CompareDataFunc
  *
  * @description:
  *    a type describing a comparison function using extra data
  *
  * @input:
  *    a    :: first element
  *    b    :: second element
  *    data :: extra data used by the comparison
  *
  * @return:
  *   -1, 0 or +1 if a < b, a==b or a > b, respectively
  */
  typedef M_Int  (*M_CompareDataFunc)( M_CPointer   a,
                                       M_CPointer   b,
                                       M_CPointer   data );


 /****************************************************************
  *
  * @functype: M_HashFunc
  *
  * @description:
  *    type describing a simple hashing function
  *
  * @input:
  *    key :: input key
  *
  * @return:
  *   key hash
  */
  typedef M_Size  (*M_HashFunc)( M_CPointer   key );


 /****************************************************************
  *
  * @functype: M_HashDataFunc
  *
  * @description:
  *    type describing a hashing function using several parameters
  *
  * @input:
  *    key       :: input key
  *    user_data :: user-provided data
  *
  * @return:
  *   key hash
  */
  typedef M_Size  (*M_HashDataFunc)( M_CPointer   key,
                                     M_CPointer   user_data );


 /****************************************************************
  *
  * @functype: M_IterateFunc
  *
  * @description:
  *    type describing a function used to iterate over containers
  *    (i.e. hash tables, trees, etc..).
  *
  * @input:
  *    key   :: iteration node's "key"
  *    value :: iteration node's "value" (if any, NULL otherwise)
  *
  * @return:
  *   boolean. used only in "selection removals" to indicate that the
  *   key/node should be removed from the container
  */
  typedef M_Bool  (*M_IterateFunc)( M_CPointer  key,
                                    M_CPointer  value );


 /****************************************************************
  *
  * @functype: M_IterateDataFunc
  *
  * @description:
  *    a variant of @M_IterateFunc that also takes a user-provided
  *    pointer
  *
  * @input:
  *    key   :: iteration node's "key"
  *    value :: iteration node's "value" (if any, NULL otherwise)
  *    data  :: user-provided pointer
  *
  * @return:
  *   boolean. used only in "selection removals" to indicate that the
  *   key/node should be removed from the container
  */
  typedef M_Bool  (*M_IterateDataFunc)( M_Pointer  key,
                                        M_Pointer  value,
                                        M_Pointer  data );

 /*********************************************************************/
 /*********************************************************************/
 /*****                                                           *****/
 /*****                   SYNCHRONISATION TYPES                   *****/
 /*****                                                           *****/
 /*********************************************************************/
 /*********************************************************************/

 /*********************************************************************/
 /*********************************************************************/
 /*****                                                           *****/
 /*****                       MLIB CORE TYPES                     *****/
 /*****                                                           *****/
 /*********************************************************************/
 /*********************************************************************/

 /*****************************************************
  *
  * @section: core
  */


 /***************************************************************
  *
  * @type: M_Core
  *
  * @description:
  *    a handle to a given top-level MLib core context object.
  *    There must be only one core context per process..
  *
  * @also:
  *   @M_CoreRec, @mlib_init, @m_core_new, @m_core_destroy,
  */
  typedef struct M_CoreRec_*      M_Core;



 /********************************************************************
  *
  * @enum: M_PopAction
  *
  * @description:
  *    a list of values used to indicate what to do when popping an
  *    item from the cleanup stack. used @m_core_pop as well as many
  *    variants
  *
  * @values:
  *   M_POP_KEEP ::
  *     pop the item from the stack. don't clean it
  *
  *   M_POP_CLEAN ::
  *     pop the item from the stack and "destroy" it by calling its
  *     cleanup routine immediately
  */
  typedef enum
  {
    M_POP_KEEP  = 0,
    M_POP_CLEAN = 1

  } M_PopAction;



 /**************************************************************************
  *
  * @type: M_ThreadID
  *
  * @description;
  *    a thread id as seen by the MLib. This is a generic typeless pointer
  *    used to discriminate between threads in a running process.
  *
  * @also:
  *    thread ids are returned by a @M_Core_GetThreadFunc function
  */
  typedef M_CPointer    M_ThreadID;



 /*****************************************************
  *
  * @section: memory
  */


 /***************************************************************
  *
  * @type: M_Memory
  *
  * @description:
  *    a handle to a given memory manager.
  *
  * @also:
  *   @M_Memory Rec, @m_memory_new, @m_memory_destroy,
  */
  typedef struct M_MemoryRec_*      M_Memory;


 /*****************************************************
  *
  * @section: exceptions
  */

 /***************************************************************
  *
  * @type: M_Exception
  *
  * @description:
  *    a handle to a given @M_ExceptionRec record, used to model
  *    a given run-time exception.
  *
  * @note:
  *    @M_Exception is always a pointer to a _constant_ exception
  *    record.
  */
  typedef const struct M_ExceptionRec_*     M_Exception;



 /********************************************************************
  *
  * @functype: M_CleanupFunc
  *
  * @description:
  *    a function pointer to a cleanup routine for a given item
  *    type
  *
  * @input:
  *    item         :: pointer to item to cleanup
  *    cleanup_data :: cleanup-type specific data
  */
  typedef void (*M_CleanupFunc)( M_Pointer  item,
                                 M_Pointer  cleanup_data );


 /**************************************************************************
  *
  * @exception: m_err_parameter_check
  *
  * @description;
  *    this exception is thrown when certain functions receive an invalid
  *    parameter (for example, a NULL handle).
  *
  *    it is normally _only_ compiled in debug builds of the MLib
  */
  MLIB_VAR( M_Exception )  m_err_parameter_check;

  /* */

/* we really shouldn't need to do that !! */
#ifndef NULL
#  define  NULL  0
#endif

#ifndef TRUE
#  define  TRUE (!0)
#endif

#ifndef FALSE
#  define  FALSE  0
#endif

#undef  MIN
#define MIN(x,y)   ( (x) <= (y) ? (x) : (y) )

#undef  MAX
#define MAX(x,y)   ( (x) >= (y) ? (x) : (y) )


  /* a dummy constant byte used by the @m_field_size macro */
  MLIB_VAR(const M_Byte)   _mlib_dummy_byte;

  /* */

M_END_HEADER

#endif /* __MLIB_TYPES_H__ */
