#ifndef __MLIB_VERSION_H__
#define __MLIB_VERSION_H__

#include <mlib.h>

M_BEGIN_HEADER

 /****************************************************************
  *
  * @macro: MLIB_VERSION_MAJOR
  *
  * @description:
  *    the MLib's major version number, corresponding to the header
  *    files used during compilation of a MLib client program/library
  *
  *    the major version number is used to indicate an interface version
  *    changing is expected to break source and binary compatibility
  *
  * @note:
  *    use @mlib_get_version to retrieve the build-time version number,
  *    which might be different in certain cases, and is more relevant
  */
#define  MLIB_VERSION_MAJOR   0


 /****************************************************************
  *
  * @macro: MLIB_VERSION_MINOR
  *
  * @description:
  *    the MLib's minor version number, corresponding to the header
  *    files used during compilation of a MLib client program/library
  *
  *    the minor version number is used to indicate minor interface
  *    changes, most notably the addition of new types and functions
  *
  *    NOTE THAT EVEN MINOR NUMBERS ARE RESERVED FOR DEVELOPMENTS VERSIONS
  *    OF THE LIBRARY..
  *
  *    it shouldn't break binary compatibility, but sometimes break
  *    source compatibility (though this can often be prevented by
  *    the use of macros)
  *
  * @note:
  *    use @mlib_get_version to retrieve the build-time version number,
  *    which might be different in certain cases, and is more relevant
  */
#define  MLIB_VERSION_MINOR   1     /* still development/experimental */


 /****************************************************************
  *
  * @macro: MLIB_VERSION_PATCH
  *
  * @description:
  *    the MLib's patch version number, corresponding to the header
  *    files used during compilation of a MLib client program/library
  *
  *    the patch version is used to indicate important bug fixes. changing
  *    it shouldn't break either binary or source compatibility..
  *
  *    it shouldn't break binary compatibility, but sometimes break
  *    source compatibility (though this can often be prevented by
  *    the use of macros)
  *
  * @note:
  *    use @mlib_get_version to retrieve the build-time version number,
  *    which might be different in certain cases, and is more relevant
  */
#define  MLIB_VERSION_PATCH   0


 /****************************************************************
  *
  * @macro: MLIB_VERSION_STRING
  *
  * @description:
  *    A string giving the MLib's version number, for example "0.1.0"
  *
  * @note:
  *    use @mlib_get_version_string to retrieve the build-time version string,
  *    which might be different in certain cases, and is more relevant
  */
#define  MLIB_VERSION_STRING  M_VERSION_STRDEF(        \
                                MLIB_VERSION_MAJOR,    \
                                MLIB_VERSION_MINOR,    \
                                MLIB_VERSION_PATCH )

 /****************************************************************
  *
  * @macro: M_VERSION_STRDEF (major,minor,patch)
  *
  * @description:
  *    A macro use to generate a version string from major, minor and
  *    patch numbers
  */
#define  M_VERSION_STRDEF(major,minor,patch)   \
                           M_STRDEF(major) "." \
                           M_STRDEF(minor) "." \
                           M_STRDEF(patch)


 /****************************************************************
  *
  * @type: M_VersionRec
  *
  * @description:
  *    a structure used to model a given exploded version number
  *
  * @fields:
  *    major :: major version number
  *    minor :: minor version number
  *    patch :: patch version number
  *
  * @note:
  *   by convention, even minor numbers are reserved for developer
  *   (i.e. unstable) releases..
  */
  typedef struct M_VersionRec_
  {
    M_UInt   major;
    M_UInt   minor;
    M_UInt   patch;

  } M_VersionRec, *M_Version;



 /****************************************************************
  *
  * @function: m_version_compare
  *
  * @description:
  *    compare two version records and returns -1, 0 or +1 depending
  *    on their "differences"
  *
  * @input:
  *   v1 :: handle to first version
  *   v2 :: handle to second version
  *
  * @return:
  *   -1, 0 or +1 depending on versions content
  */
  MLIB_API( M_Int )
  m_version_compare( M_Version  v1,
                     M_Version  v2 );


 /****************************************************************
  *
  * @function: m_version_equal
  *
  * @description:
  *    return TRUE is two versions are equal
  *
  * @input:
  *   v1 :: handle to first version
  *   v2 :: handle to second version
  *
  * @return:
  *   -1, 0 or +1 depending on versions content
  */
  MLIB_API( M_Bool )
  m_version_equal( M_Version  v1,
                   M_Version  v2 );


 /****************************************************************
  *
  * @function: m_version_check_binary
  *
  * @description:
  *    return TRUE if version "v2" is binary-compatible with version
  *    "v1". This means that they have the same major number, and that
  *    "v1.minor" is less then "v2.minor"
  *
  * @input:
  *   v1 :: handle to first version
  *   v2 :: handle to second version
  *
  * @return:
  *   TRUE iff "v2" is binary-compatible with "v1"
  */
  MLIB_API( M_Bool )
  m_version_check_binary( M_Version  v1,
                          M_Version  v2 );


 /****************************************************************
  *
  * @function: m_version_check_source
  *
  * @description:
  *    return TRUE if version "v2" is source-compatible with version
  *    "v1". This means that they have the same major and minor
  *    numbers (independent of patch number)
  *
  * @input:
  *   v1 :: handle to first version
  *   v2 :: handle to second version
  *
  * @return:
  *   -1, 0 or +1 depending on versions content
  */
  MLIB_API( M_Bool )
  m_version_check_source( M_Version  v1,
                          M_Version  v2 );


 /****************************************************************
  *
  * @function: mlib_get_version
  *
  * @description:
  *    retrieve build-time version of the MLib
  *
  * @output:
  *    version :: version structure. see @M_VersionRec
  */
  MLIB_API( void )
  mlib_get_version( M_Version  version );


 /****************************************************************
  *
  * @function: mlib_get_version_string
  *
  * @description:
  *    retrieve build-time version string of the MLib
  *
  * @return:
  *    version string (constant, no need to free it)
  */
  MLIB_API( M_CString )
  mlib_get_version_string( void );

/* */

M_END_HEADER

#endif /* __MLIB_VERSION_H__ */
