#include <mlib/buffer.h>

  /* cleanup function used within m_buffer_get_line0 */
  static void
  m_buffer_get_line_cleanup( M_String*  astring,
                             M_Memory   memory )
  {
    M_String  str = *astring;

    if ( str )
      m_mem_free( str, memory );
  }




 /* IMPORTANT TECHNICAL DETAIL:                                         */
 /*   when the text line is completely contained within the buffer,     */
 /*   this function will not use the cleanup stack. Otherwise, a        */
 /*   dynamically growing memory block is used to extract the text      */
 /*   from multiple buffer feeds..                                      */
 /*                                                                     */
  MLIB_LOCALDEF( M_String )
  m_buffer_get_line0( M_Buffer  buffer,
                      M_Memory  memory,
                      M_Size   *alen,
                      M_Bool    include_eol )
  {
    M_ByteP   p, end;
    M_Byte    c;
    M_Bool    skip_eol;
    M_Size    len, size, new_size;
    M_String  result = NULL;

    m_assert_is_buffer( buffer );

    if ( alen )
      *alen = 0;

    len    = 0;
    size   = 0;
    result = NULL;

    skip_eol = 0;

    for (;;)
    {
      p   = buffer->cur;
      end = buffer->limit;
      c   = 1;

      while ( p < end )
      {
        c = p[0];

        /* when 'include_eol' is FALSE, encountering a '\r' or a '\n' */
        /* means the end of line                                      */
        if ( c == '\0' || ( !include_eol && ( c == '\r' || c == '\n' ) ) )
        {
          skip_eol = 1;
          goto EndOfLine;
        }

        p++;
      }

      /* we reached the buffer's limit, save the current characters */
      p   = buffer->cur;
      len = (M_Size)( end - p );
      if ( len > 0 )
      {
        new_size = size + len;

        if ( size == 0 )
          m_core_push( memory->mcore, &result,
                       (M_CleanupFunc) m_buffer_get_line_cleanup,
                       memory );

        M_REALLOC( result, size, new_size );
        M_MEM_COPY( result + size, p, len );
        size = new_size;
      }

      /* then feed the buffer, requesting at least 1 new character */
      if ( buffer->feed( buffer, 1 ) == 0 )
      {
        p   = buffer->cur;
        end = buffer->limit;

        m_assert( p == end );
        break;
      }
    }

  EndOfLine:
    /* we reached the end of line or the end of file, save characters */
    len      = (M_Size)( p - buffer->cur );
    new_size = size + len + 1;

    if ( size == 0 )
    {
      /* if no previous characters were saved, we don't need to use */
      /* the cleanup stack..                                        */
      result = M_ALLOC( new_size );
    }
    else
    {
      M_REALLOC( result, size, new_size );
      m_core_pop( memory->mcore, &result, M_POP_KEEP );
    }

    M_MEM_COPY( result + size, buffer->cur, len );

    size         = new_size-1;
    result[size] = 0;


    /* skip the EOL character(s) when needed */
    if ( skip_eol )
    {
      c           = p[0];
      buffer->cur = ++p;

      /* for '\r', we need to check wether a '\n' follows */
      if ( c == '\r' )
      {
        if ( p < end )
          c = p[0];
        else
        {
          if ( buffer->feed( buffer, 1 ) >= 1 )
            c = buffer->cur[0];

          p   = buffer->cur;
          end = buffer->limit;
        }

        if ( c == '\n' )
        {
          /* we necessarily have 'p < end' here !! */
          m_assert( p < end );

          buffer->cur = p+1;
        }
      }
    }

    if ( alen )
      *alen = size;

    return result;
  }


