#include <mlib.h>
#include <mlib/internal/icore.h>


  MLIB_VARDEF(const M_ExceptionFamilyRec)
  mlib_exception_family =
  {
    "MLib",
    0,
    0
  };


  MLIB_APIDEF( void )
  m_core_assert( M_Core       mcore,
                 const char*  filename,
                 long         lineno,
                 const char*  funcname )
  {
    if ( mcore != NULL || M_ICORE(mcore)->magic != M_MAGIC_CORE )
      M_PANIC(( "%s: invalid core handle in '%s' (%s:%ld)\n",
                "mlib.assert",
                funcname,
                filename,
                lineno ));
  }


 /***************************************************************************/
 /***************************************************************************/
 /*****                                                                 *****/
 /*****                   MLIB CLEANUP HANDLING                         *****/
 /*****                                                                 *****/
 /***************************************************************************/
 /***************************************************************************/

  /* storage of cleanup chunks */

  static M_CleanupChunk
  m_cleanup_chunk_new( M_ICore  core )
  {
    M_CleanupChunk  chunk;

    M_ICORE_LOCK( core );
    {
      chunk = core->chunks;
      if (chunk)
      {
        core->chunks = chunk->link;
        chunk->link  = 0;
      }
    }
    M_ICORE_UNLOCK( core );

    if (!chunk)
      chunk = (M_CleanupChunk) m_mem_qalloc(
                 sizeof(*chunk), core->root.memory );


    return chunk;
  }


  static void
  m_cleanup_chunk_destroy( M_ICore          core,
                           M_CleanupChunk  chunk )
  {
    M_ICORE_LOCK( core );
    {
      chunk->link  = core->chunks;
      core->chunks = chunk;
    }
    M_ICORE_UNLOCK( core );
  }


 /***************************************************************************/
 /***************************************************************************/
 /*****                                                                 *****/
 /*****                 MLIB EXCEPTION CONTEXT MANAGEMENT               *****/
 /*****                                                                 *****/
 /***************************************************************************/
 /***************************************************************************/

  MLIB_APIDEF(void)
  m_xcontext_grow_stack( M_XContext   xcontext )
  {
    M_Core          core  = xcontext->mcore;
    M_CleanupStack  stack = &xcontext->cleanup_stack;
    M_CleanupChunk  chunk;

    chunk = m_cleanup_chunk_new( M_ICORE(core) );
    if (!chunk)
    {
      M_ERROR(( "%s: out of memory. aborting thread\n",
                "mlib.xcontext.grow_stack" ));
      m_core_throw( core, m_err_memory_alloc );
    }

    chunk->link  = stack->chunk;
    stack->chunk = chunk;
    stack->first = stack->cursor = chunk->items;
    stack->last  = chunk->items + M_CLEANUP_CHUNK_SIZE - 1;
  }


  MLIB_APIDEF(void)
  m_xcontext_shrink_stack( M_XContext  xcontext )
  {
    M_Core          core  = xcontext->mcore;
    M_CleanupStack  stack = &xcontext->cleanup_stack;
    M_CleanupChunk  chunk;

    chunk = stack->chunk;

    m_assert( chunk );

    if ( !chunk || !chunk->link )
      M_PANIC(( "%s: empty cleanup stack !!",
                "mlib.xcontext.shink_stack" ));

    stack->chunk = chunk->link;
    m_cleanup_chunk_destroy( M_ICORE(core), chunk );
    chunk = stack->chunk;

    stack->first  = chunk->items;
    stack->last   = chunk->items + M_CLEANUP_CHUNK_SIZE - 1;
    stack->cursor = stack->last;
  }



  MLIB_APIDEF(M_Error)
  m_xcontext_init( M_XContext   xcontext,
                   M_Core       core,
                   M_ThreadID   id )
  {
    M_CleanupChunk  chunk;
    M_CleanupStack  stack;
    M_ICore         icore = M_ICORE(core);

    m_assert( xcontext && id );
    m_assert_is_core( core );

    chunk = m_cleanup_chunk_new( icore );
    if (!chunk) return M_ERR_MEMORY_ALLOC;

    stack = &xcontext->cleanup_stack;

    M_MEM_SET( xcontext, 0, sizeof(*xcontext) );

    xcontext->mcore     = core;
    xcontext->thread_id = id;

    /* initialise cleanup stack */
    chunk->link      = NULL;
    stack->chunk     = chunk;
    stack->first     = stack->cursor = chunk->items;
    stack->last      = chunk->items + M_CLEANUP_CHUNK_SIZE - 1;
    stack->num_items = 0;

    return 0;
  }


  MLIB_APIDEF(void)
  m_xcontext_done( M_XContext  xcontext )
  {
    M_CleanupChunk  *pchunk;
    M_CleanupStack   stack;
    M_ICore          core;

    m_assert( xcontext );
    m_assert_is_core( xcontext->mcore );

    stack = &xcontext->cleanup_stack;
    core  = M_ICORE(xcontext->mcore);

    /* get rid of all cleanup chunks from this xcontext */
    pchunk = &stack->chunk;
    if ( pchunk[0] )
    {
      while ( pchunk[0] )
        pchunk = &(pchunk[0]->link);

      M_ICORE_LOCK(core);
      {
        pchunk[0]    = core->chunks;
        core->chunks = stack->chunk;
        stack->chunk = NULL;
      }
      M_ICORE_UNLOCK(core);
    }

    M_MEM_SET( xcontext, 0, sizeof(*xcontext) );
  }


  /* storage of xcontext data blocks */

  MLIB_APIDEF(M_XContext)
  m_core_find_xcontext( M_Core     mcore,
                        M_CString  function_name )
  {
    M_XContext  xcontext;
    M_ICore     internal = M_ICORE(mcore);

    m_assert_is_core( mcore );

    xcontext = internal->core_context;
    if ( !xcontext && !internal->thread_id_get( mcore, 0, &xcontext ) )
    {
      /* we will panick here */
      M_PANIC(( "%s: %s\n", function_name,
         "unable to find current xcontext's exception context."
         " Must abort for safety !!\n" ));
    }
    return xcontext;
  }


  static M_XContext
  m_core_find_xcontext_with_create( M_Core  core )
  {
    M_XContext  xcontext;
    M_ICore     internal = M_ICORE(core);

    m_assert_is_core( core );

    xcontext = internal->core_context;
    if ( !xcontext && !internal->thread_id_get( core, 1, &xcontext ) )
    {
      M_ERROR(( "%s: not enough memory to allocate new exception"
                "context !!\n",
                "mlib.core.find_xcontext_with_create" ));
    }
    return xcontext;
  }



 /*************************************************************************/
 /*************************************************************************/
 /*****                                                               *****/
 /*****                   EXCEPTION HANDLING                          *****/
 /*****                                                               *****/
 /*************************************************************************/
 /*************************************************************************/


/* define a dummy, non-NULL xcontext id */
#define  M_THREAD_ID_DUMMY   ((M_ThreadID)(M_Pointer)1)

  MLIB_APIDEF(void)
  m_core_set_fatal_handler( M_Core              core,
                            M_FatalHandlerFunc  fatal_handler,
                            M_Pointer           fatal_data )
  {
    M_XContext  xcontext;

    m_assert_is_core( core );

    /* try to find current xcontext's exception context */
    xcontext = m_core_find_xcontext_with_create( core );
    if (!xcontext)
    {
      /* there is not enough memory to allocate a new xcontext */
      /* we call the fatal handler _directly_..                */
      fatal_handler( core, m_err_memory_alloc, fatal_data );
    }

    /* otherwise, set the xcontext's fatal handler fields */
    xcontext->fatal_handler = fatal_handler;
    xcontext->fatal_data    = fatal_data;
  }


  /* set exception handler for current xcontext - we must also create a */
  /* new xcontext if necessary..                                      */
  MLIB_APIDEF(void)
  m_core_set_xhandler( M_Core      core,
                       M_XHandler  xhandler )
  {
    M_XContext  xcontext;

    m_assert_is_core( core );
    m_assert( xhandler );

    xhandler->mcore = core;

    /* find current xcontext's data block. create one if necessary */
    xcontext = m_core_find_xcontext_with_create( core );
    if (!xcontext)
    {
      xhandler->exception  = m_err_memory_alloc;
      xhandler->xcontext   = NULL;
      xhandler->previous   = NULL;
    }
    else
    {
      xhandler->exception  = NULL;
      xhandler->xcontext   = xcontext;
      xhandler->xmark      = (M_XMark) xcontext->cleanup_stack.num_items;
      xhandler->previous   = xcontext->xhandler;
      xcontext->xhandler   = xhandler;
    }
  }


  MLIB_APIDEF(void)
  m_core_throw( M_Core       core,
                M_Exception  exception )
  {
    M_XContext   xcontext;
    M_XHandler   xhandler;

    m_assert_is_core(core);

    /* retrieve current xcontext's info block */
    xcontext = m_core_find_xcontext( core, "mlib.core.throw" );

    /* now check that an exception handler was set in the current xcontext */
    if ( !xcontext->xhandler )
    {
      /* if no exception handler was set for this xcontext, we need to */
      /* invoke the fatal handler !!                                   */
      if (!xcontext->fatal_handler)
        M_PANIC(( "%s: unhandled exception %p occured in thread %p "
                  "with no fatal handler. I must abort for safety !!",
                  "mlib.core.throw", exception, xcontext->thread_id ));

      xcontext->fatal_handler( core, exception, xcontext->fatal_data );
    }

    /* ok, perform normal exception cleanup before long-jumping */
    xhandler = xcontext->xhandler;

    /* we need to cleanup the current xcontext stack */
    {
      M_CleanupStack  stack = &xcontext->cleanup_stack;
      M_CleanupItem   slot;

      while ( stack->num_items > M_POINTER_TO_UINT(xhandler->xmark) )
      {
        if ( stack->cursor == stack->first )
          m_xcontext_shrink_stack( xcontext );
        else
          stack->cursor--;

        slot = stack->cursor;
        stack->num_items--;

        slot->cleanup_func( slot->data, slot->cleanup_data );
      }
    }

    /* record error, and unset current handler */
    xhandler->exception = exception;
    xhandler->xcontext  = NULL;
    xcontext->xhandler  = xhandler->previous;

    /* we can now "throw" through "longjmp" */
    longjmp( xhandler->env, 1 );
  }


  MLIB_APIDEF(M_Exception)
  m_core_unset_xhandler( M_Core      core,
                         M_XHandler  xhandler )
  {
    M_XContext   xcontext;
    M_Exception  result;

    m_assert_is_core( core );
    m_assert( xhandler );

    M_UNUSED(core);

    result   = xhandler->exception;
    xcontext = xhandler->xcontext;

    if ( xcontext )
    {
      if ( !xcontext->xhandler )
        M_PANIC(( "%s: no exception handler was "
                  "previously set in current exception context. !!"
                  "mlib.core.unset_xhandler" ));

      if ( xhandler != xcontext->xhandler )
        M_PANIC(( "%s: invalid exception handler parameter !!"
                  "mlib.core.unset_xhandler" ));

      if ( xcontext->cleanup_stack.num_items != 0 )
        M_PANIC(( "%s: cleanup stack isn't empty !!",
                  "mlib.core.unset_xhandler" ));

      xcontext->xhandler  = xhandler->previous;
      xhandler->xcontext  = NULL;
    }

    xhandler->exception = NULL;
    xhandler->previous  = NULL;
    xhandler->xmark     = NULL;

    return result;
  }



  MLIB_APIDEF(void)
  m_core_push( M_Core         core,
               M_Pointer      item,
               M_CleanupFunc  cleanup_func,
               M_Pointer      cleanup_data )
  {
    M_XContext  xcontext;

    m_assert_is_core( core );
    m_assert( item && cleanup_func );

    xcontext = m_core_find_xcontext( core, "mlib.core.push" );

    m_core_push_inline( core, item, cleanup_func, cleanup_data,
                           "mlib.core.push" );
  }


  MLIB_APIDEF(void)
  m_core_pop ( M_Core       core,
               M_Pointer    item,
               M_PopAction  action )
  {
    M_XContext  xcontext;

    m_assert_is_core(core);

    xcontext = m_core_find_xcontext( core, "mlib.core.pop" );
    m_core_pop_inline( core, item, 0, action, "mlib.core.pop" );
  }



 /*************************************************************************/
 /*************************************************************************/
 /*****                                                               *****/
 /*****                   MANAGING THE POOL OF STATICS                *****/
 /*****                                                               *****/
 /*************************************************************************/
 /*************************************************************************/


  MLIB_APIDEF( M_Error )
  m_statics_init( M_StaticList  statics,
                  M_Core        mcore )
  {
    M_Error  error = 0;

    statics->mcore = mcore;
    statics->nodes = NULL;
    statics->lock  = NULL;

#ifdef M_CONFIG_THREAD_SAFE
    /* allocate lock by hand */
    if ( !m_core_single_thread( mcore ) )
    {
      M_ICore   icore  = M_ICORE(mcore);
      M_Memory  memory = mcore->memory;
      M_Lock    lock;

      lock = m_mem_qalloc( icore->spinlock_class.size, memory );
      if ( lock )
      {
        lock->clazz = (M_Lock_Class) &icore->spinlock_class;

        error = icore->spinlock_class.init( lock, M_LOCK_SPIN, 0, 0 );
        if ( error )
          m_mem_free( lock, memory );
        else
          statics->lock = lock;
      }
      else
        error = M_ERR_MEMORY_ALLOC;
    }
#endif

    return error;
  }


  static void
  m_static_node_done( M_StaticNode  node,
                      M_Memory      memory )
  {
    if ( node->statik )
    {
      if ( node->clazz->done )
        node->clazz->done( node->statik );

      M_FREE( node->statik );
    }
    M_FREE( node->statik );
  }


  static void
  m_statics_done( M_StaticList  statics )
  {
    M_Memory      memory;
    M_Core        mcore;
    M_StaticNode  node, next;

    if ( statics )
    {
      M_Lock  lock;

      mcore = statics->mcore;
      lock  = statics->lock;

      m_assert_is_core( mcore );
      memory = mcore->memory;

      M_LOCK_ENTER( lock );
      {
        node = statics->nodes;
        while (node)
        {
          next = node->link;
          m_static_node_done( node, memory );
          node = next;
        }
        statics->nodes = NULL;
      }
      M_LOCK_EXIT( lock );
    }
  }


  static M_Pointer
  m_statics_get( M_StaticList    statics,
                 M_Static_Class  clazz,
                 M_Pointer       data )
  {
    M_Core        mcore;
    M_Lock        lock;
    M_StaticNode  node;
    M_ICore       icore;
    M_Pointer     statik = 0;
    M_Memory      memory;

    mcore  = statics->mcore;
    m_assert_is_core( mcore );

    memory = mcore->memory;
    lock   = statics->lock;
    icore  = M_ICORE(mcore);

    m_assert( clazz && clazz->id && clazz->init );

    M_LOCK_ENTER_UPDATE( lock );
    {
      /* look for the static in our current list.. */
      for ( node = statics->nodes; node; node = node->link )
        if ( node->id == clazz->id )
        {
          /* we found it, check that we have the same "info" field */
          if ( node->clazz != clazz )
            M_PANIC(( "mlib.core.set_static: duplicate static registration" ));

          break;
        }

      if (!node)
      {
        /* we need to create a new static node */
        M_LOCK_WRITE_UPDATE(lock);
        M_LOCK_PUSH( lock );

        M_NEW(node);

        m_core_push( mcore, node, (M_CleanupFunc) m_static_node_done, memory );

        node->id     = clazz->id;
        node->clazz  = clazz;
        node->statik = M_ALLOC( clazz->size );

        if ( clazz->init )
          clazz->init( node->statik, mcore, data );

        m_core_pop( mcore, node, M_POP_KEEP );

        node->link           = icore->statics.nodes;
        icore->statics.nodes = node;

        M_LOCK_POP( lock );
      }
    }
    M_LOCK_EXIT( lock );

    statik = node->statik;

    return statik;
  }


  MLIB_APIDEF( M_Pointer )
  m_core_get_static( M_Core          mcore,
                     M_Static_Class  clazz,
                     M_Pointer       data )
  {
    m_assert_is_core( mcore );
    m_assert( clazz != NULL );

    return m_statics_get( &M_ICORE(mcore)->statics, clazz, data );
  }



 /***************************************************************************/
 /***************************************************************************/
 /*****                                                                 *****/
 /*****                   MLIB CORE OBJECT CREATION                     *****/
 /*****                                                                 *****/
 /***************************************************************************/
 /***************************************************************************/


 /* initialise the MLib core spinlock, do not throw an exception */
  static M_Error
  m_core_lock_init( M_ICore  icore )
  {
    M_Memory   memory = icore->root.memory;
    M_Lock     lock;
    M_Error    error = M_ERR_MEMORY_ALLOC;

    lock = m_mem_qalloc( icore->spinlock_class.size, memory );
    if (lock)
    {
      lock->clazz = (M_Lock_Class) &icore->spinlock_class;

      error = icore->spinlock_class.init(lock,M_LOCK_SPIN,0,0);
      if (!error)
        icore->core_lock = lock;
      else
      {
        M_ERROR(( "%s: could not initialise core spinlock\n", "mlib.core.new" ));
        m_mem_free( lock, memory );
      }
    }
    return error;
  }


  MLIB_APIDEF(M_Error)
  m_core_new( M_Memory       memory,
              M_Core_Class   clazz,
              M_Pointer      impl,
              M_Core        *acore )
  {
    M_Error     error;
    M_ICore     core;

    /* create new memory manager if needed */
    if (!memory)
      memory = m_memory_new();
    else
    {
      if ( memory->mcore )
        M_PANIC(( "%s: memory manager already in use by another MLib core",
                  "mlib.core.new" ));

      m_memory_ref( memory );
    }

    *acore = 0;

    /* create new MLib core object in memory */
    core = m_mem_qalloc( sizeof(*core), memory );
    if (!core)
    {
      error = M_ERR_MEMORY_ALLOC;
      goto Exit;
    }

    M_ZERO( core );

    core->root.memory  = memory;
    core->root.impl    = impl;
    core->magic        = M_MAGIC_CORE;
    core->core_class   = clazz;
    core->ref_count    = 1;

    if ( !clazz )
    {
      /* this MLib core is only used by a single xcontext  */
      /* we will allocate the exception context manually */

      M_XContext  xcontext;

      xcontext = m_mem_qalloc( sizeof(*xcontext), memory );
      if (!xcontext)
      {
        M_ERROR(( "%s: out of memory\n", "mlib.core.new" ));
        error = M_ERR_MEMORY_ALLOC;
        goto Fail;
      }

      /* no need for locks */
      core->core_context      = xcontext;
      core->core_lock         = NULL;

      error = m_xcontext_init( xcontext, M_CORE(core), M_CORE_DUMMY_THREAD_ID );
      if (error) goto Fail;
    }
    else
    {
      /* this MLib core is used by several xcontexts at once                   */
      /* each xcontext exception context is created by the xcontext sub-system */
      /* allocate the global spinlock manually                                  */

      core->thread_id_get =    clazz->thread_id_get;

      core->mutex_class       = *clazz->mutex_class;
      core->mutex_class.magic = M_MAGIC_LOCK;
      core->mutex_class.mcore = M_CORE(core);

      if ( !clazz->spinlock_class )
        core->spinlock_class = core->mutex_class;
      else
      {
        core->spinlock_class = *clazz->spinlock_class;
        core->spinlock_class.magic = M_MAGIC_LOCK;
        core->spinlock_class.mcore = M_CORE(core);
      }

      if ( !clazz->rwlock_class )
        core->rwlock_class = core->mutex_class;
      else
      {
        core->rwlock_class       = *clazz->rwlock_class;
        core->rwlock_class.magic = M_MAGIC_LOCK;
        core->rwlock_class.mcore = M_CORE(core);
      }

      /* now initialise the global lock */
      error = m_core_lock_init( core );
      if (error) goto Fail;
    }

    /* initialise list of static objects */
    error = m_statics_init( &core->statics, M_CORE(core) );
    if (error) goto Fail;

    /* now, attach the memory manager to this MLib core */
    memory->mcore = *acore = M_CORE(core);

  Exit:
    *acore = (M_Core)core;
    return error;

  Fail:
    if ( core )
    {
      if ( core->core_context )
      {
        m_xcontext_done( core->core_context );
        core->core_context = NULL;
      }

      if ( core->core_lock )
      {
        core->spinlock_class.done( core->core_lock );
        m_mem_free( core->core_lock, memory );
        core->core_lock = NULL;
      }

      core->magic = M_MAGIC_DEATH;
      m_mem_free( core, memory );

      m_memory_unref(memory);
      core = NULL;
    }
    goto Exit;
  }





  static void
  m_core_destroy( M_Core  core )
  {
    M_Memory  memory;
    M_ICore   internal;

    m_assert_is_core( core );
    m_assert_is_memory( core->memory );

    memory   = core->memory;
    internal = M_ICORE(core);

    m_assert( memory->mcore == core );

    /* destroy all static objects */
    m_statics_done( &internal->statics );

    /* destroy xcontext info blocks & global lock */
    if ( internal->core_context )
    {
      m_xcontext_done( internal->core_context );
      M_FREE( internal->core_context );
    }
    else
    {
      /* the xcontext sub-system must call m_xcontext_done on each */
      /* exception context object..                              */
      internal->core_class->core_done( core );

      /* destroy global spinlock */
      if ( internal->core_lock )
      {
        internal->spinlock_class.done( internal->core_lock );
        M_FREE( internal->core_lock );
      }
    }

    /* discard all free cleanup chunks */
    {
      M_CleanupChunk  chunk;

      chunk = internal->chunks;
      if ( chunk )
      {
        M_CleanupChunk  next;

        while (chunk)
        {
          next = chunk->link;
          m_mem_free( chunk, memory );
          chunk = next;
        }

        internal->chunks = NULL;
      }
    }

    /* unlink memory manager from core */
    memory->mcore = 0;

    /* set magic. this prevents the use of destroyed cores through */
    /* dangling pointers..                                            */
    internal->magic = M_MAGIC_DEATH;

    /* free core object */
    M_FREE( core );

    /* unref memory manager */
    m_memory_unref( memory );
  }



  MLIB_APIDEF(void)
  m_core_ref( M_Core  core )
  {
    M_ICore  internal = M_ICORE(core);

    m_assert_is_core(core);

    M_ICORE_LOCK( internal );
    {
      internal->ref_count++;
    }
    M_ICORE_UNLOCK( internal );
  }



  MLIB_APIDEF(void)
  m_core_unref( M_Core  core )
  {
    M_ICore  internal = M_ICORE(core);

    m_assert_is_core(core);

    M_ICORE_LOCK( internal );

    if ( --internal->ref_count <= 0 )
    {
      internal->ref_count = 0;
      M_ICORE_UNLOCK( internal );
      m_core_destroy( core );
    }
    else
      M_ICORE_UNLOCK( internal );
  }

