#include <mlib.h>
#include <stdio.h>  /* for vprintf */
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>

  MLIB_APIDEF( void )
  m_log( const char*  fmt, ... )
  {
    va_list  ap;


    va_start( ap, fmt );
    vprintf( fmt, ap );
    va_end( ap );
  }


  MLIB_APIDEF( void )
  m_panic( const char*  fmt, ... )
  {
    va_list  ap;

    /* print panic disclaimer */
    fprintf( stderr,
   "THE SOFTWARE YOU'RE RUNNING HAS A FATAL BUG, AND HAS BEEN ABORTED\n"
   "Please contact your software vendor for updates or to report the bug,\n"
   "using the following technical description:\n" );

    va_start( ap, fmt );
    vfprintf( stderr, fmt, ap );
    va_end( ap );

    exit( 1 );
  }


#ifdef M_DEBUG_TEST_UNITS

  MLIB_APIDEF( void )
  m_log_test( const char*  fmt, ... )
  {
    va_list  ap;


    va_start( ap, fmt );
    vfprintf( stderr, fmt, ap );
    va_end( ap );
  }

#endif /* M_DEBUG_TEST_UNITS */


#if defined( M_DEBUG_LEVEL_LOG )

  typedef struct M_LogDomainRec_*      M_LogDomain;
  typedef struct M_LogComponentRec_*   M_LogComponent;


  typedef struct M_LogDomainRec_
  {
    M_LogDomain     link;
    M_CString       name;
    M_Int           undefined;
    M_LogComponent  components;

  } M_LogDomainRec;


  typedef struct M_LogComponentRec_
  {
    M_LogComponent  link;
    M_CString       name;
    M_Int           level;

  } M_LogComponentRec;


/* NOTE: We use static variables in logging debug mode !! */
/*       that's better than requesting a M_Core on each   */
/*       log..                                            */

  static M_Memory       m_log_memory      = 0;
  static M_LogDomain    m_log_domain_list = 0;


 /* destroy log domain record */
  static void
  m_log_domain_destroy( M_LogDomain  domain,
                        M_Memory     memory )
  {
    M_LogComponent  cur, next;

    /* destroy list of log components */
    cur = domain->components;
    while (cur)
    {
      next = cur->link;
      M_FREE( cur );
      cur  = next;
    }
    domain->components = 0;

    /* clear domain name - we don't need to free it */
    domain->name = 0;

    /* release domain */
    M_FREE( domain );
  }


 /* release all memory allocated by the logging sub-system */
  static void
  m_log_done( void )
  {
    if (m_log_memory)
    {
      M_LogDomain  cur, next;

      /* destroy list of log domains */
      cur = m_log_domain_list;
      while (cur)
      {
        next = cur->link;
        m_log_domain_destroy( cur, m_log_memory );
        cur  = next;
      }
      m_log_domain_list = 0;

      /* destroy memory manager */
      m_memory_unref( m_log_memory );
      m_log_memory = 0;
    }
  }



  MLIB_APIDEF(void)
  m_log_init( M_Memory  memory )
  {
    if ( !m_log_memory && !atexit( m_log_done ) )
    {
      if (!memory)
      {
        memory = m_memory_new();
        if (!memory)
          m_panic( "mlib.log.init: can't create default memory manager" );
      }
      m_log_memory = memory;
    }
  }


  static M_LogDomain
  m_log_get_domain( const char*  domain_name )
  {
    M_LogDomain* pdomain;
    M_LogDomain  domain = 0;

    /* look for the current domain log in the list */
    pdomain = &m_log_domain_list;
    for (;;)
    {
      domain = pdomain[0];
      if (!domain)
        break;

      if ( m_str_equal( domain->name, domain_name ) )
      {
        pdomain[0]        = domain->link;
        domain->link      = m_log_domain_list;
        m_log_domain_list = domain;
        return domain;
      }

      pdomain = &domain->link;
    }

    /* we didn't found it, create a new log domain record */
    {
      M_Memory  memory = m_log_memory;

      M_QNEW( domain );
      if (!domain)
        return 0;

      domain->name       = domain_name;
      domain->components = 0;
      domain->undefined  = 0;

      domain->link      = m_log_domain_list;
      m_log_domain_list = domain;
    }

    return domain;
  }


  static  M_LogComponent
  m_log_get_component( const char*  domain_name,
                       const char*  component_name )
  {
    M_LogDomain      domain;
    M_LogComponent*  pcomponent;
    M_LogComponent   component;

    domain = m_log_get_domain( domain_name );
    if ( !domain || domain->undefined )
      return 0;

    if (!domain->components)
    {
      M_Memory  memory    = m_log_memory;
      char*     p = getenv( domain->name );

      /* return immediately if the domain name is not defined */
      /* in the environment..                                 */
      if (!p)
      {
        domain->undefined = 1;
        return 0;
      }

      /* we now need to parse the environment variable   */
      /* to get the levels of each component. the format */
      /* of the string is "COMP1:x COMP2:x COMP3:x ..."  */
      /* where COMP1, COMP2, etc.. are component names   */
      /* and 'x' is a number..                           */

      while (*p)
      {
        int  len;

        /* skip leading spaces */
        p += strspn( p, " " );

        /* skip component name */
        len = strcspn( p, " :" );
        if (len == 0)
          break;

        M_QALLOC( component, sizeof(*component) + len+1 );
        if (!component)
          break;

        component->name = (const char*)(component + 1);
        M_MEM_COPY( component->name, p, len );
        ((char*)component->name)[len] = 0;

        p += len;
        if (*p == ':')
        {
          p ++;
          len = strspn( p, "0123456789" );
          component->level = atoi(p);
          p += len;
        }
        else
          component->level = 1;  /* by default !! */

        component->link    = domain->components;
        domain->components = component;
      }

      domain->undefined = ( domain->components == 0 );
    }

    /* now, try to find the current component in the domain's list */
    pcomponent = &domain->components;
    for (;;)
    {
      component = pcomponent[0];
      if (!component) break;

      if ( m_str_equal( component->name, component_name ) )
      {
        pcomponent[0]      = component->link;
        component->link    = domain->components;
        domain->components = component;
        break;
      }

      pcomponent = &component->link;
    }

    return component;
  }


  MLIB_APIDEF(int)
  m_log_compare( const char* domain_name,
                 const char* component_name,
                 int         level )
  {
    M_LogComponent  component;

    if (!m_log_memory)
      m_log_init( 0 );

    component = m_log_get_component( domain_name, component_name );
    if (component)
      return component->level >= level;

    return 0;
  }

#endif /* M_DEBUG_LEVEL_LOG */



