#include <mlib/internal/imemory.h>
#include <mlib/internal/imemdbug.h>
#include <stdlib.h>

 /*********************************************************************/
 /*                                                                   */
 /*                E X C E P T I O N   H A N D L I N G                */
 /*                                                                   */
 /*********************************************************************/

  static const M_ExceptionRec  m_err_memory_alloc_rec =
  {
    &mlib_exception_family,
    M_ERR_MEMORY_ALLOC,
  };

  MLIB_VARDEF( M_Exception )  m_err_memory_alloc = &m_err_memory_alloc_rec;


  MLIB_APIDEF( void )
  m_memory_assert( M_Memory     memory,
                   const char*  filename,
                   long         lineno,
                   const char*  funcname )
  {
    if ( memory == NULL || M_IMEMORY(memory)->magic != M_MAGIC_MEMORY )
      M_PANIC(( "%s: invalid memory manager handle in '%s' (%s:%ld)\n",
                "mlib.assert",
                funcname,
                filename,
                lineno ));
  }


 /*********************************************************************/
 /*                                                                   */
 /*                  M E M O R Y   M A N A G E R S                    */
 /*                                                                   */
 /*********************************************************************/

  static M_IMemory
  m_memory_std_new( M_Pointer  impl )
  {
    M_UNUSED(impl);
    return (M_IMemory) malloc( sizeof( M_IMemoryRec ) );
  }


  static void
  m_memory_std_destroy( M_IMemory  memory )
  {
    free( memory );
  }


  static const
  M_Memory_ClassRec  m_memory_std_class =
  {
    (M_Memory_AllocFunc)    malloc,
    (M_Memory_ReallocFunc)  realloc,
    (M_Memory_FreeFunc)     free,
    m_memory_std_new,
    m_memory_std_destroy,
    0
  };


 /* return a new memory allocator that uses the standard malloc/free */
  MLIB_APIDEF(M_Memory)
  m_memory_new( void )
  {
    return m_memory_new_custom( &m_memory_std_class, 0 );
  }



  MLIB_APIDEF(M_Memory)
  m_memory_new_custom( M_Memory_Class  clazz,
                       M_Pointer       impl )
  {
    M_IMemory    memory;

    if (!clazz)
      M_PANIC(( "invalid memory manager class pointer !!\n" ));

    memory = clazz->mem_new( impl );
    if (memory)
    {
      /* clear all flags */
      M_MEM_SET( memory, 0, sizeof(*memory) );

      memory->root.impl  = impl;
      memory->root.mcore = 0;
      memory->magic      = M_MAGIC_MEMORY;
      memory->alloc      = clazz->alloc;
      memory->free       = clazz->free;
      memory->realloc    = clazz->realloc;
      memory->clazz      = clazz;
      memory->ref_count  = 1;

      /* take care of unsetting the "disable_trap" flag if it is mistakenly */
      /* set in the user-provided manager class                             */
      memory->flags = clazz->mem_flags & ~M_MEMORY_FLAG_DISABLE_TRAP;
    }
#ifdef M_CONFIG_OPTION_DEBUG_MEMORY
    m_mem_debug_init( memory );
#endif

    return (M_Memory)memory;
  }


  MLIB_APIDEF(void)
  m_memory_ref( M_Memory  memory )
  {
    M_IMemory  internal = M_IMEMORY(memory);

    m_assert_is_memory( memory );

    internal->ref_count++;
  }


  static void
  m_memory_destroy( M_Memory   memory )
  {
    M_IMemory       internal = M_IMEMORY(memory);
    M_Memory_Class  clazz;

    m_assert_is_memory( memory );

#ifdef M_CONFIG_OPTION_DEBUG_MEMORY
    m_mem_debug_done( M_IMEMORY(memory) );
#endif

    /* clear magic, just in case someone wants to use this */
    /* memory manager later through a dangling pointer..   */
    /* this will trigger assertions..                      */
    internal->magic = M_MAGIC_DEATH;

    /* destroy the memory object */
    clazz = internal->clazz;
    if (clazz->mem_destroy)
      clazz->mem_destroy(internal);
  }



  MLIB_APIDEF(void)
  m_memory_unref( M_Memory  memory )
  {
    M_IMemory       internal = M_IMEMORY(memory);

    m_assert_is_memory( memory );

    internal->ref_count--;
    if ( internal->ref_count <= 0 )
      m_memory_destroy( memory );
  }



 /*********************************************************************/
 /*                                                                   */
 /*                M E M O R Y   M A N A G E M E N T                  */
 /*                                                                   */
 /*********************************************************************/

