#include <mlib/internal/imemdbug.h>

 /*********************************************************************/
 /*                                                                   */
 /*                D E B U G G I N G   A L L O C A T O R              */
 /*                                                                   */
 /*********************************************************************/

#ifdef M_CONFIG_OPTION_DEBUG_MEMORY

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

  typedef struct M_MemNodeRec_*   M_MemNode;
  typedef struct M_MemTableRec_*  M_MemTable;

#define  M_MEM_VAL(addr)   M_POINTER_TO_DIST(addr)


  typedef struct M_MemNodeRec_
  {
    M_Byte*    address;
    M_Long     size;     /* < 0 if the block was freed */
    M_MemNode  link;

  } M_MemNodeRec;

  typedef struct M_MemTableRec_
  {
    M_Memory    memory;
    M_ULong     size;
    M_ULong     nodes;
    M_MemNode*  buckets;

    M_ULong     alloc_total;
    M_ULong     alloc_current;
    M_ULong     alloc_max;

  } M_MemTableRec;


#define M_FILENAME( x )  ((x) ? (x) : "unknown file")


  static const M_UInt  m_mem_primes[] =
  {
    7,
    11,
    19,
    37,
    73,
    109,
    163,
    251,
    367,
    557,
    823,
    1237,
    1861,
    2777,
    4177,
    6247,
    9371,
    14057,
    21089,
    31627,
    47431,
    71143,
    106721,
    160073,
    240101,
    360163,
    540217,
    810343,
    1215497,
    1823231,
    2734867,
    4102283,
    6153409,
    9230113,
    13845163,
  };


  static M_ULong
  m_mem_closest_prime( M_ULong  num )
  {
    M_UInt  i;


    for ( i = 0;
          i < sizeof ( m_mem_primes ) / sizeof ( m_mem_primes[0] ); i++ )
      if ( m_mem_primes[i] > num )
        return m_mem_primes[i];

    return M_MEM_SIZE_MAX;
  }





  /* don't use "m_panic" in order to print additional information */
  extern void
  m_mem_debug_panic( const char*  fmt, ... )
  {
    va_list  ap;


    printf( "MLib.Debug.Memory: " );

    va_start( ap, fmt );
    vprintf( fmt, ap );
    va_end( ap );

    printf( "\n" );

    m_panic( "Fatal memory error" );
  }


  static M_Pointer
  m_mem_table_alloc( M_Size      size,
                     M_MemTable  table )
  {
    return M_IMEMORY(table->memory)->clazz->alloc( size, table->memory );
  }


  static void
  m_mem_table_free( M_Pointer   block,
                    M_MemTable  table )
  {
    M_IMEMORY(table->memory)->clazz->free( block, table->memory );
  }


  static void
  m_mem_table_resize( M_MemTable  table )
  {
    M_ULong   new_size;

    new_size = m_mem_closest_prime( table->nodes );
    if (new_size != table->size)
    {
      M_MemNode*  new_buckets ;
      M_ULong     i;

      new_buckets = m_mem_table_alloc( new_size * sizeof(M_MemNode), table );
      if ( new_buckets == NULL )
        return;

      memset( new_buckets, 0, sizeof(M_MemNode)*new_size );

      for ( i = 0; i < table->size; i++ )
      {
        M_MemNode  node, next, *pnode;
        M_ULong     hash;

        node = table->buckets[i];
        while (node)
        {
          next  = node->link;
          hash  = M_MEM_VAL(node->address) % new_size;
          pnode = new_buckets + hash;

          node->link = pnode[0];
          pnode[0]   = node;

          node  = next;
        }
      }

      if ( table->buckets )
        m_mem_table_free( table->buckets, table );

      table->buckets = new_buckets;
      table->size    = new_size;
    }
  }


  static M_MemTable
  m_mem_table_new( M_Memory  memory )
  {
    M_MemTable  table;

    table = m_mem_qalloc( sizeof(*table), memory );
    if ( table == NULL ) goto Exit;

    memset( table, 0, sizeof(*table) );

    table->memory = memory;
    table->size   = M_MEM_SIZE_MIN;
    table->nodes  = 0;

    table->buckets = m_mem_table_alloc( table->size * sizeof(M_MemNode), table );
    if ( table->buckets )
      memset( table->buckets, 0, sizeof(M_MemNode)*table->size );
    else
    {
      m_mem_free( table, memory );
      table = NULL;
    }

  Exit:
    return table;
  }



  static void
  m_mem_table_destroy( M_MemTable  table )
  {
    M_ULong  i;

    if ( table )
    {
      M_Memory  memory = table->memory;
      M_Long    leak_count = 0;
      M_ULong   leaks = 0;

      for ( i = 0; i < table->size; i++ )
      {
        M_MemNode  *pnode = table->buckets + i, next, node = *pnode;

        while (node)
        {
          next = node->link;
          node->link  = 0;

          if ( node->size > 0 )
          {
            if ( leak_count < 32 )
              printf( "leaked memory block at address %p, size %ld\n",
                       node->address, node->size );

            leak_count++;
            leaks += node->size;

            m_mem_table_free( node->address, table );
          }

          node->address = NULL;
          node->size    = 0;

          m_mem_table_free( node->address, table );
          node = next;
        }
        table->buckets[i] = 0;
      }
      m_mem_table_free( table->buckets, table );
      table->buckets = NULL;

      table->size   = 0;
      table->nodes  = 0;

      printf(
        "FreeType: total memory allocations = %ld\n", table->alloc_total );
      printf(
        "FreeType: maximum memory footprint = %ld\n", table->alloc_max );

      m_mem_free( table, memory );

      if ( leak_count > 0 )
        m_mem_debug_panic( "%ld bytes of memory leaked in %ld blocks\n",
                            leaks, leak_count );
      printf( "no FreeType memory leaks detected !!\n" );

    }
  }



  static M_MemNode*
  m_mem_table_get_nodep( M_MemTable  table,
                         M_Byte*     address )
  {
    M_ULong           hash;
    M_MemNode       *pnode, node;

    hash  = M_MEM_VAL(address);
    pnode = table->buckets + (hash % table->size);

    for (;;)
    {
      node = pnode[0];
      if (!node)
        break;

      if ( node->address == address )
        break;

      pnode = &node->link;
    }
    return pnode;
  }



  static void
  m_mem_table_set( M_MemTable  table,
                   M_Byte*     address,
                   M_ULong     size )
  {
    M_MemNode  *pnode, node;

    if (table)
    {
      pnode = m_mem_table_get_nodep( table, address );
      node  = *pnode;
      if (node)
      {
        if ( node->size < 0 )
        {
          /* this block was already freed. this means that our memory is */
          /* now completely corrupted !!                                 */
          m_mem_debug_panic( "memory heap corrupted" );
        }
        else
        {
          /* this block was already allocated. this means that our memory */
          /* is also corrupted !!                                         */
          m_mem_debug_panic( "duplicate block allocation at address "
                           "%p, size %ld\n",
                           address, size );
        }
      }

      /* we need to create a new node in this table */
      node = m_mem_table_alloc( sizeof(*node), table );
      if ( node == NULL )
        m_mem_debug_panic( "not enough memory to run memory tests" );

      node->address = address;
      node->size    = size;
      node->link    = pnode[0];

      pnode[0] = node;
      table->nodes++;

      table->alloc_total   += size;
      table->alloc_current += size;

      if ( table->alloc_current > table->alloc_max )
        table->alloc_max = table->alloc_current;

      if ( table->nodes*3 < table->size  ||
           table->size *3 < table->nodes )
        m_mem_table_resize( table );
    }
  }


  static void
  m_mem_table_remove( M_MemTable  table,
                      M_Byte*     address )
  {
    if (table)
    {
      M_MemNode  *pnode, node;

      pnode = m_mem_table_get_nodep( table, address );
      node  = *pnode;
      if (node)
      {
        if ( node->size < 0 )
          m_mem_debug_panic( "freeing memory block at %p more than once !!",
                              address );

        /* we simply invert the node's size to indicate that the node */
        /* was freed. We also change its content..                    */
        memset( address, 0xF3, node->size );

        table->alloc_current -=  node->size;
        node->size            = -node->size;
      }
      else
        m_mem_debug_panic( "trying to free unknown block at %p\n",
                            address );
    }
  }


  static M_Pointer
  m_mem_debug_alloc( M_Size    size,
                     M_Memory  memory )
  {
    M_MemTable  table  = M_IMEMORY(memory)->debug;
    M_Byte*     block;

    if ( size <= 0 )
      m_mem_debug_panic( "negative block size allocation (%ld)", size );

    block = m_mem_table_alloc( size, table );
    if ( block )
      m_mem_table_set( table, block, (M_ULong )size );

    return (M_Pointer) block;
  }


  static void
  m_mem_debug_free( M_Pointer  block,
                    M_Memory   memory )
  {
    M_MemTable  table = M_IMEMORY(memory)->debug;

    if ( block == NULL )
      m_mem_debug_panic( "trying to free NULL !!" );

    m_mem_table_remove( table, (M_Byte* )block );

    /* we never really free the block */
  }



  static M_Pointer
  m_mem_debug_realloc( M_Pointer  block,
                       M_Size     new_size,
                       M_Size     cur_size,
                       M_Memory   memory )
  {
    M_MemTable  table = M_IMEMORY(memory)->debug;
    M_MemNode   node, *pnode;
    M_Pointer   new_block;

    if ( block == NULL || cur_size == 0 )
      m_mem_debug_panic( "trying to reallocate NULL" );

    if ( new_size <= 0 )
      m_mem_debug_panic( "trying to reallocate %p to size 0 (current is %ld)",
                          block, cur_size );

   /* check 'cur_size' value */
    pnode = m_mem_table_get_nodep( table, (M_Byte* )block );
    node  = *pnode;
    if (!node)
      m_mem_debug_panic( "trying to reallocate unknown block at %p",
                          block );

    if ( node->size <= 0 )
      m_mem_debug_panic( "trying to reallocate freed block at %p",
                          block );

    if ( (M_Size)node->size != cur_size )
      m_mem_debug_panic( "invalid realloc request for %p. cur_size is "
                          "%ld instead of %ld", block, cur_size, node->size );

    new_block = m_mem_debug_alloc( new_size, memory );
    if ( new_block == NULL )
      return NULL;

    memcpy( new_block, block, cur_size < new_size ? cur_size : new_size );

    m_mem_debug_free( (M_Byte* )block, memory );

    return new_block;
  }


  MLIB_LOCALDEF(M_Bool)
  m_mem_debug_init( M_IMemory  memory )
  {
    M_MemTable  table;
    M_Bool      result = 0;

    if ( getenv( "MLIB_DEBUG_MEMORY" ) )
    {
      table = m_mem_table_new( (M_Memory)memory );
      if ( table )
      {
        memory->debug   = table;
        memory->alloc   = m_mem_debug_alloc;
        memory->realloc = m_mem_debug_realloc;
        memory->free    = m_mem_debug_free;

        result = 1;
      }
    }
    return result;
  }


  MLIB_LOCALDEF(void)
  m_mem_debug_done( M_IMemory  memory )
  {
    M_MemTable  table = memory->debug;

    if ( table )
    {
      memory->debug   = NULL;
      memory->free    = memory->clazz->free;
      memory->alloc   = memory->clazz->alloc;
      memory->realloc = memory->clazz->realloc;

      m_mem_table_destroy( table );
    }
  }

#endif /*  M_CONFIG_OPTION_DEBUG_MEMORY */


