#include <mlib.h>
#include <mlib/internal/imempool.h>
#include <mlib/internal/icore.h>


#ifdef M_CONFIG_OPTION_MEMORY_POOLS

 /************************************************************************/
 /************************************************************************/
 /*****                                                              *****/
 /*****                    MEMORY POOL INTERNALS                     *****/
 /*****                                                              *****/
 /************************************************************************/
 /************************************************************************/


 /* allocate a new page, and returns the first free element in it */
  MLIB_LOCALDEF(M_Pointer)
  m_memory_pool_alloc_page( M_MemoryPool  pool )
  {
    M_Size            sizes[2];
    M_Pointer         ptrs[2];
    M_Memory          memory = pool->memory;
    M_MemoryPoolPage  page;
    M_Pointer*        pitem  = NULL;
    M_Pointer         result = NULL;
    M_Lock            lock   = pool->lock;

    m_assert( pool->page_size > pool->element_count*pool->element_size +
                                (pool->element_count/8) );

    /* allocate a new page - it is followed by a bitmap used during GC */
    sizes[0] = sizeof(*page);
    sizes[1] = pool->page_size;

    M_BATCH_ALLOC( 2, sizes, ptrs );
    page        = ptrs[0];
    page->base  = ptrs[1];
    page->limit = page->base + pool->page_limit;

    /* now, generate free items list within the new page */
    {
      M_Int     incr  = pool->element_size;
      M_Byte*   item  = page->base + incr;

      pitem = M_POINTER_P(page->base);

      /* just in case !! */
      m_assert( item < page->limit );

      do
      {
        m_assert( (M_Byte*)pitem  >= page->base && (M_Byte*)pitem < page->limit );

        *pitem = item;
        pitem  = M_POINTER_P(item);
        item  += incr;
      }
      while ( item + incr < page->limit );

      m_assert( (M_Byte*)pitem  >= page->base && (M_Byte*)pitem < page->limit );
      *pitem = NULL;
    }

    /* add the page to the pool, and to the free list     */
    /* don't forget to lock the pool for these operations */
    M_LOCK_ENTER(lock);
    {
      page->link = pool->page;
      pool->page = page;

      /* another thread might have already released an item, or   */
      /* allocated a new pool page since our last M_LOCK_EXTI     */
      /* so don't assume that pool->element_free is still NULL..  */
      /*                                                          */
      if ( pool->element_free )
        *pitem = pool->element_free;

      pool->element_free = M_POINTER(page->base);

      /* ok, now allocate one item to return it */
      result = pool->element_free;
      m_assert( result );

      pool->element_free = *M_POINTER_P(result);
    }
    M_LOCK_EXIT(lock);

    return result;
  }



  MLIB_APIDEF( void )
  m_memory_pool_gc( M_MemoryPool  pool )
  {
    M_Lock  lock = pool->lock;

    if ( pool->element_free == NULL )
      return;

    M_LOCK_ENTER(lock);
    {
      M_Size            bitmap_len = pool->element_count/8;
      M_MemoryPoolPage  page;
      M_Byte*           block;
      M_Byte*           bitmap;
      M_Size            offset;
      M_UInt            index;

      /* first of all, clear all page bitmaps */
      page = pool->page;
      while ( page )
      {
        M_MEM_ZERO( page->limit, bitmap_len );
        page->num_frees = 0;
        page = page->link;
      }

      /* now, parse the free block list, and fill the bitmaps */
      page   = pool->page;
      bitmap = page->limit;
      block  = pool->element_free;
      while ( block )
      {
        offset = (M_Size)( block - page->base );
        if ( offset >= pool->page_limit )
        {
          page  = pool->page;
          while (page)
          {
            offset = (M_Size)( block - page->base );
            if ( offset < pool->page_limit )
              break;

            page = page->link;
          }

          if ( page == NULL )
          {
            /* a free block that is in none of our pool pages !! */
            /* we detected something really wrong !!             */
            M_PANIC(( "mlib.memory_pool.collect: detected invalid free block"
                      " in pool." ));
          }

          bitmap = page->limit;
        }

        page->num_frees ++;

        index = (M_UInt)( offset/pool->element_size );
        bitmap[ index >> 3 ] |= (M_Byte)( 0x80 >> (index & 7) );

        block = *M_POINTER_P(block);
      }


      /* release all pages that only contain free blocks */
      {
        M_MemoryPoolPage*  ppage  = &pool->page;
        M_Memory           memory = pool->memory;

        for (;;)
        {
          page = *ppage;
          if ( page == NULL )
            break;

          if ( page->num_frees == pool->element_count )
          {
            *ppage = page->link;
            M_FREE( page->base );
            M_FREE( page );
          }
          else
            ppage = &page->link;
        }
      }


      /* recompute the free block list from the page bitmaps */
      {
        M_Byte**  pblock = (M_Byte**) &pool->element_free;

        page = pool->page;
        while ( page )
        {
          M_Byte*  bitmap = page->limit;
          M_Byte*  limit  = bitmap + bitmap_len;
          M_UInt   val = 0, mask  = 0;

          block = page->base;
          for ( ;; )
          {
            if ( mask == 0 )
            {
              if ( bitmap == limit )
                break;

              mask = 0x80;
              val  = *bitmap++;
            }

            if ( val & mask )
            {
              *pblock = block;
              pblock  = (M_Byte**) block;
            }

            block += pool->element_size;
          }

          page = page->link;
        }
        *pblock = NULL;
      }
    }
    M_LOCK_EXIT(lock);
  }


 /* allocate a new pool from the current pool set      */
 /* assume that the set rwlock is in update_write mode */
 /*                                                    */
 /* do not throw an exception, returns NULL in case of */
 /* failure..                                          */
 /*                                                    */
  static M_MemoryPool
  m_memory_pool_new( M_MemoryPoolSet  set,
                     M_Size           element_size,
                     M_Size           page_size )
  {
    M_Memory      memory = set->mcore->memory;
    M_Size        element_count, page_limit;
    M_MemoryPool  pool;

    m_assert_is_memory( memory );
    m_assert( element_size > 0 );

    /* compute number of elements in pool. Read technical note */
    /* in <mlib/internal/imempool.h> for details..             */

    element_size  = M_MEMORY_POOL_ALIGN( element_size );
    page_size     = M_MEMORY_POOL_PAGE_SIZE;
    page_limit    = page_size;

    element_count = page_size / element_size;
    for (;;)
    {
      page_limit = element_count*element_size;
      if ( page_limit + (element_count/8) <= page_size )
        break;

      element_count--;
    }

    if ( element_count < 16 )
      return NULL;

    pool = m_mem_qalloc( sizeof(*pool), memory );
    if (pool)
    {
      M_ZERO( pool );
      M_MEM_COPY( &pool->fake, memory, sizeof(M_IMemoryRec) );

      pool->fake.pool = pool;

      pool->memory = memory;
      pool->magic  = M_MAGIC_MEMORY_POOL;
      pool->lock   = set->synlock;  /* share synlock among all pools */

      pool->element_size  = element_size;
      pool->element_count = element_count;
      pool->page_size     = page_size;
      pool->page_limit    = page_limit;
      pool->page          = NULL;
    }
    return pool;
  }



  static void
  m_memory_pool_destroy( M_MemoryPool  pool )
  {
    if (pool)
    {
      M_Memory  memory;

      m_assert_is_memory_pool( pool );
      memory = pool->memory;

      /* clear all pages */
      {
        M_MemoryPoolPage  page, next;

        /* we cannot check that all elements have been freed, but we */
        /* can clear all the pages before releasing them in order to */
        /* avoid problems with dangling pointers..                   */

        page = pool->page;
        while (page)
        {
          next = page->link;

          page->limit = NULL;
          page->link  = NULL;
          M_MEM_ZERO( page->base, pool->page_size );
          M_FREE( page->base );
          M_FREE( page );

          page = next;
        }
        pool->page = NULL;
      }

      pool->element_free = 0;
      pool->element_size = 0;
      pool->cursor       = 0;
      pool->limit        = 0;

      pool->lock   = NULL;
      pool->memory = 0;
      pool->magic  = M_MAGIC_DEATH;  /* will trigger asserts */
      pool->link   = NULL;

      M_FREE( pool );
    }
  }



 /************************************************************************/
 /************************************************************************/
 /*****                                                              *****/
 /*****                   MEMORY POOL STATIC DATA                    *****/
 /*****                                                              *****/
 /************************************************************************/
 /************************************************************************/




  /* destroy all memory pools */
  static void
  m_memory_pool_set_done( M_MemoryPoolSet  set )
  {
    M_MemoryPool      pool, next;
    M_UInt             n;
    M_Memory           memory;

    m_assert_is_memory_pool_set( set );

    memory = set->mcore->memory;
    for ( n = 0; n < M_MEMORY_POOL_BUCKETS; n++ )
    {
      pool = set->pools[n];
      while (pool)
      {
        next = pool->link;
        m_memory_pool_destroy( pool );
        pool = next;
      }
      set->pools[n] = NULL;
    }

    m_lock_destroy( set->rwlock );
    m_lock_destroy( set->synlock );
  }


  static void
  m_memory_pool_set_init( M_MemoryPoolSet  pool_set,
                          M_Core           mcore )
  {
    pool_set->mcore   = mcore;
    pool_set->magic   = M_MAGIC_MEMORY_POOL_SET;

    /* allocate a new synlock. it is shared by all memory pools to */
    /* synchronize allocation/frees..                              */
    /*                                                             */
    pool_set->synlock = m_lock_new( mcore, M_LOCK_MUTEX );

    /* allocate a new read/write lock. it is used to protect the   */
    /* hash table of available memory pools..                      */
    /*                                                             */
    pool_set->rwlock  = m_lock_new( mcore, M_LOCK_READWRITE );
  }


  /* describe the static list of memory pools */
  static const M_Static_ClassRec  m_memory_pool_set_info =
  {
    sizeof( M_MemoryPoolSetRec ),
    M_MAGIC_MEMORY_POOL_SET,
    (M_Static_InitFunc) m_memory_pool_set_init,
    (M_Static_DoneFunc) m_memory_pool_set_done
  };




  static M_MemoryPoolSet
  m_memory_pool_set_find( M_Memory  memory )
  {
    M_Core           mcore = memory->mcore;
    M_MemoryPoolSet  pool_set;

    /* retrieve memory pools set */
    M_CORE_LOCK( mcore );
    {
      pool_set = M_IMEMORY(memory)->pool_set;
      if (!pool_set)
      {
        /* this should happen only once per session */
        pool_set = m_core_get_static( mcore, &m_memory_pool_set_info, NULL );
        m_assert_is_memory_pool_set( pool_set );

        M_IMEMORY(memory)->pool_set = pool_set;
      }
    }
    M_CORE_UNLOCK( mcore );

    return pool_set;
  }




  MLIB_APIDEF( void )
  m_memory_pool_collect( M_Memory  memory )
  {
    M_Lock           lock;
    M_MemoryPoolSet  pool_set;

    pool_set = m_memory_pool_set_find( memory );
    lock     = pool_set->rwlock;

    /* good, now look up the list of memory pools */
    {
      M_UInt        n;
      M_MemoryPool  cur, *bucket;

      M_LOCK_ENTER_READ( lock );

      for ( n = 0; n < M_MEMORY_POOL_BUCKETS; n++ )
      {
        bucket = pool_set->pools + n;

        for ( cur = *bucket; cur; cur = cur->link )
          m_memory_pool_gc( cur );
      }

      M_LOCK_EXIT( lock );
    }
  }





  MLIB_APIDEF(M_Memory)
  m_memory_pool_find( M_Memory  memory,
                      M_Size    size )
  {
    M_MemoryPoolSet   pool_set;
    M_MemoryPool      pool;
    M_Core            mcore;
    M_Lock            lock;

    m_assert_is_memory( memory );

    mcore = memory->mcore;
    m_assert_is_core( mcore );

    pool_set = m_memory_pool_set_find( memory );

    /* adjust element size */
    size = M_MEMORY_POOL_ALIGN(size);
    lock = pool_set->rwlock;


    /* good, now look up the list of memory pools */
    {
      M_MemoryPool  cur, *bucket;

      M_LOCK_ENTER_UPDATE( lock );

      pool   = NULL;
      bucket = pool_set->pools +
               ((size/M_MEMORY_POOL_ALIGNMENT) % M_MEMORY_POOL_BUCKETS );

      for ( cur = *bucket; cur; cur = cur->link )
        if ( cur->element_size == M_MEMORY_POOL_ALIGN(size) )
        {
          pool = cur;
          break;
        }

      if ( pool == NULL )
      {
        /* we need to create a new pool for this size */
        M_LOCK_WRITE_UPDATE(lock);

        pool = m_memory_pool_new( pool_set, size,  M_MEMORY_POOL_PAGE_SIZE );
        if (pool)
        {
          pool->link = *bucket;
          *bucket    = pool;
        }
      }

      M_LOCK_EXIT(lock);

      /* only happens when we couldn't allocate a new memory pool */
      if ( !pool )
        m_core_throw( mcore, m_err_memory_alloc );
    }

    return (M_Memory)pool;
  }

#else  /* !M_CONFIG_OPTION_MEMORY_POOLS */

  MLIB_APIDEF(M_Memory)
  m_memory_pool_find( M_Memory  memory,
                      M_Size    size )
  {
    m_assert_is_memory( memory );
    M_UNUSED(size);

    return memory;
  }

  MLIB_APIDEF( void )
  m_memory_collect_all( M_Memory  memory )
  {
    M_UNUSED(memory);
  }

#endif  /* !M_CONFIG_OPTION_MEMORY_POOLS */
