#include <mlib.h>
#include <mlib/internal/iobject.h>

 /************************************************************************/
 /************************************************************************/
 /*****                                                              *****/
 /*****                    META-CLASS DEFINITION                     *****/
 /*****                                                              *****/
 /************************************************************************/
 /************************************************************************/

 /* forward declarations */
 static M_Bool    m_class_done( M_Class  clazz );

 static M_Class   m_meta_class_get( M_MetaClass  meta,
                                    M_ClassType  type );


 /* finalize a given class, assume that it was already removed */
 /* from the meta-class dictionary..                           */
  static void
  m_class_fini( M_Class   clazz )
  {
    /* call class finalizers */
    {
      M_ClassType  type = clazz->type;

      m_assert( type );  /* each class must have a class type !! */

      do
      {
        /* we call each finalizer from bottom to top. notice */
        /* that if a finalizer returns TRUE, we immediately  */
        /* stop the traversal..                              */
        if ( type->class_done && type->class_done( clazz ) )
          break;

        type = type->super;
      }
      while (type);
    }

    clazz->object_done = 0;
    clazz->object_copy = 0;
    clazz->object_incr = 0;
    clazz->object_init = 0;

    clazz->type   = NULL;
    clazz->super  = NULL;

    clazz->idata  = NULL;
    clazz->magic  = M_MAGIC_DEATH;
    clazz->memory = NULL;
    clazz->mcore  = NULL;

    clazz->object.clazz = NULL;
  }


 /* destroy a given class, this is called from the meta class */
 /* hash table routines, never directly..                     */
  static void
  m_class_destroy( M_Class   clazz )
  {
    M_Memory  memory = clazz->memory;

    m_class_fini( clazz );
    M_FREE( clazz );
  }


 /* used when someone tries to ref/unref a small object */
  static M_Bool
  m_sobject_incr( M_Object  obj,
                  M_Int     incr )
  {
    M_UNUSED(obj);

    M_PANIC(( "%s: trying to update a small object (M_SObject type)\n",
              incr > 0 ? "mlib.ref" : "mlib.unref" ));

    return 1;
  }


  static M_Class
  m_class_create( M_MetaClass  meta,
                  M_ClassType  type )
  {
    M_Class      parent = NULL;
    M_Class      clazz;
    M_UInt       flags;
    M_Memory     memory = M_CLASS(meta)->memory;
    M_Core       mcore  = M_CLASS(meta)->mcore;

    m_assert( type != NULL && type->class_size >= sizeof(M_ClassRec) );

    if ( type->super )
    {
      m_assert( type->class_size >= type->super->class_size );
      parent = (M_Class) m_meta_class_get( meta, type->super );
    }

    clazz = M_ALLOC( type->class_size );
    if ( parent )
    {
      /* copy most fields from parent class */
      M_MEM_COPY( clazz, parent, type->super->class_size );

      /* clear specific fields */
      clazz->info  = NULL;
      clazz->idata = NULL;
    }

    clazz->object.clazz = (M_Class) meta;
    clazz->memory       = memory;
    clazz->mcore        = mcore;
    clazz->magic        = M_MAGIC_CLASS;
    clazz->type         = type;
    clazz->super        = parent;

    flags = type->class_flags;

    if ( type->object_init && (flags & M_CLASS_FLAG_ABSTRACT) == 0 )
      clazz->object_init = type->object_init;

    if ( type->object_copy && (flags & M_CLASS_FLAG_NO_COPY) == 0 )
      clazz->object_copy = type->object_copy;

    if ( type->object_done )
      clazz->object_done = type->object_done;

    if ( flags & M_CLASS_FLAG_SMALL )
      clazz->object_incr = m_sobject_incr;
    else if ( type->object_incr )
      clazz->object_incr = type->object_incr;

    /* call the initialiser */
    {
      M_Class_InitFunc  class_init = NULL;
      M_ClassType       cur_type;
      M_UInt            trys;

      trys     = 128;
      cur_type = type;
      do
      {
        class_init = cur_type->class_init;
        if ( class_init )
          break;

        cur_type = cur_type->super;

        /* check for class type loops !! */
        if ( cur_type == type )
          M_PANIC(( "mlib.class.create: circular class type loop detected"
                    " from '%s'", type->name ? type->name : "unknown class" ));
        trys--;
      }
      while ( cur_type && trys > 0 );

      if ( trys == 0 )
        M_PANIC(( "mlib.class.create: class type hierarchy too deep"
                  " from '%s'", type->name ? type->name : "unknown class" ));

      /* now, call the class initialiser */
      if ( class_init )
      {
        m_core_push( mcore, clazz,
                     (M_CleanupFunc) m_class_destroy,
                     NULL );

        class_init( clazz );

        m_core_pop( mcore, clazz, M_POP_KEEP );
      }
    }
    return clazz;
  }


#include <mlib/internal/gobject.h>

#define  GEN_HASH_RESIZE        m_meta_class_resize
#define  GEN_HASH_CLEAR         m_meta_class_clear
#define  GEN_HASH_FIND_NODEPTR  m_meta_class_get_nodep
#define  GEN_HASH_TOUCH         m_meta_class_touch
#define  GEN_HASH_GET           m_meta_class_get
#define  GEN_HASH_REMOVE        m_meta_class_remove

#define  GEN_HASH_RESIZE_DEF
#define  GEN_HASH_CLEAR_DEF
#define  GEN_HASH_FIND_NODEPTR_DEF
#define  GEN_HASH_TOUCH_DEF
#define  GEN_HASH_GET_DEF
#define  GEN_HASH_REMOVE_DEF

#include <mlib/generic/hash.i>


  static M_Bool
  m_meta_class_done( M_MetaClass  meta )
  {
    m_assert_is_meta_class(meta);

    if ( meta->lock )
    {
      M_LOCK_DESTROY(meta->lock);
      meta->lock = NULL;
    }

    /* clear all remaining classes if needed */
    if ( meta->size )
      m_meta_class_clear( meta );

    meta->magic = M_MAGIC_DEATH;

    return 1;
  }


 /* inner meta class initialisation function */
  static void
  m_meta_class_init0( M_MetaClass meta )
  {
    M_Memory  memory = meta->clazz.memory;

    meta->magic = M_MAGIC_META_CLASS;
    meta->count = 0;
    meta->size  = M_PRIMES_MIN;

    M_NEW_ARRAY( meta->buckets, meta->size );

    meta->lock = M_LOCK_NEW( meta->clazz.mcore, M_LOCK_MUTEX );
  }



  static const M_ClassTypeRec  m_meta_class_type_rec =
  {
    (M_ClassType)  NULL,
    "MetaClass",

    sizeof( M_MetaClassRec ),
    M_CLASS_FLAG_ABSTRACT | M_CLASS_FLAG_NO_COPY | M_CLASS_FLAG_SMALL,
    (M_Class_InitFunc) m_meta_class_init0,
    (M_Class_DoneFunc) m_meta_class_done,

    sizeof( M_ClassRec ),
    (M_Object_InitFunc) NULL,
    (M_Object_CopyFunc) NULL,
    (M_Object_IncrFunc) NULL,
    (M_Object_DoneFunc) m_class_done
  };


  static M_ClassType  m_meta_class_type = &m_meta_class_type_rec;



  static void
  m_meta_class_init( M_MetaClass  meta,
                     M_Core       mcore )
  {
    M_Class     clazz  = (M_Class)meta;

    m_assert_is_core(mcore);

    /* these fields must be initialised to call m_class_create properly */
    clazz->memory = mcore->memory;
    clazz->mcore  = mcore;

    /* big hack, but it works !! */
    clazz = m_class_create( meta, m_meta_class_type );
  }



  static const M_Static_ClassRec   m_meta_class_static_class =
  {
    sizeof (M_MetaClassRec),
    M_META_CLASS_ID,
    (M_Static_InitFunc)  m_meta_class_init,
    (M_Static_DoneFunc)  m_meta_class_done
  };




 /************************************************************************/
 /************************************************************************/
 /*****                                                              *****/
 /*****                    CLASS MANAGEMENT                          *****/
 /*****                                                              *****/
 /************************************************************************/
 /************************************************************************/

#define  m_assert_is_clazz(x)  m_assert( (x) && (x)->magic == M_MAGIC_CLASS )


  static M_Bool
  m_class_done( M_Class  clazz )
  {
    m_assert_is_clazz( clazz );

    m_class_fini( clazz );

    return 1;
  }


  static M_Class
  m_class_find( M_ClassType  type,
                M_MetaClass  meta )
  {
    M_Class  clazz;

    m_assert_is_meta_class(meta);

    M_LOCK_ENTER(meta->lock);

    /* lookup class in meta-class dictionary, if not found, create a new */
    /* one..                                                             */
    clazz = NULL;
    m_meta_class_touch( meta, type, &clazz );

    M_LOCK_EXIT(meta->lock);

    return (M_Class) clazz;
  }


  static void
  m_class_remove( M_Class  clazz )
  {
    M_MetaClass  meta;

    m_assert_is_clazz(clazz);

    meta = (M_MetaClass) clazz->object.clazz;
    m_assert_is_meta_class(meta);

    M_LOCK_ENTER(meta->lock);

    m_meta_class_remove( meta, clazz->type );

    M_LOCK_EXIT(meta->lock);
  }


 /************************************************************************/
 /************************************************************************/
 /*****                                                              *****/
 /*****                        CLASS ROUTINES                        *****/
 /*****                                                              *****/
 /************************************************************************/
 /************************************************************************/

#define  M_IS_CLASS(c)  ((c) && M_CLASS(c)->magic == M_MAGIC_CLASS)


  MLIB_APIDEF(M_Pointer)
  m_class_assert_is_a( M_Pointer    clazz_,
                       M_ClassType  class_type,
                       const char*  function_name,
                       long         line_number,
                       const char*  source_file )
  {
    M_Class  clazz = clazz_;

    if ( !M_IS_CLASS(clazz) )
      goto Fail;

    if ( !class_type )
      return clazz_;

    do
    {
      if ( clazz->type == class_type )
        return clazz_;

      clazz = clazz->super;
    }
    while (clazz);

  Fail:
    M_PANIC(( "%s:%ld: invalid class handle in '%s'\n",
               source_file, line_number, function_name ));
    return 0;
  }



  MLIB_APIDEF(M_Class)
  m_class_from_type( M_ClassType  class_type,
                     M_Core       mcore )
  {
    M_MetaClass  meta;

    m_assert_is_core(mcore);
    m_assert( class_type );

    meta = m_core_get_static( mcore, &m_meta_class_static_class, 0 );
    m_assert_is_meta_class(meta);

    return m_class_find( class_type, meta );
  }



  MLIB_APIDEF(void)
  m_class_flush( M_Class  clazz )
  {
    m_assert_is_clazz( clazz );

    m_class_remove( clazz );
  }



  MLIB_APIDEF(M_Bool)
  m_class_is_a( M_Class      clazz,
                M_ClassType  class_type )
  {
    if ( M_IS_CLASS(clazz) )
    {
      do
      {
        if ( clazz->type == class_type )
          return 1;

        clazz = clazz->super;
      }
      while (clazz);
    }
    return 0;
  }


  MLIB_APIDEF(M_Bool)
  m_class_is_sub_of( M_Class  clazz,
                     M_Class  parent )
  {
    m_assert_is_clazz( parent );
    if ( clazz )
    {
      do
      {
        if ( clazz == parent )
          return 1;

        clazz = clazz->super;
      }
      while ( clazz != NULL );
    }
    return 0;
  }



  MLIB_APIDEF(M_Bool)
  m_class_check( M_Pointer  clazz )
  {
    return M_IS_CLASS(M_CLASS(clazz));
  }




 /************************************************************************/
 /************************************************************************/
 /*****                                                              *****/
 /*****                        OBJECT ROUTINES                       *****/
 /*****                                                              *****/
 /************************************************************************/
 /************************************************************************/

#define  M_IS_OBJECT(o) ((o) && M_IS_CLASS(M_OBJECT(o)->clazz) )


  MLIB_APIDEF(M_Pointer)
  m_object_assert_is_a( M_Pointer    object_,
                        M_ClassType  class_type,
                        const char*  function_name,
                        long         line_number,
                        const char*  source_file )
  {
    M_Object  object = object_;

    if ( !M_IS_OBJECT(object_) )
      goto Fail;

    if ( !class_type || m_class_is_a( object->clazz, class_type ) )
      return object_;

  Fail:
    M_PANIC(( "%s:%ld: pointer used in '%s' is not a valid object handle\n",
              source_file, line_number, function_name ));
    return 0;
  }


  MLIB_APIDEF(M_Pointer)
  m_new( M_Class  clazz )
  {
    return  m_new_full( clazz, NULL );
  }


  MLIB_APIDEF(M_Pointer)
  m_new_full( M_Class   clazz,
              M_Pointer idata )
  {
    M_Object  object;
    M_Memory  memory;

    m_assert_is_clazz( clazz );
    memory = clazz->memory;

    if ( clazz->type->class_flags & M_CLASS_FLAG_ABSTRACT )
      M_PANIC(( "mlib.new: trying to instantiate the abstract class '%s'\n",
                clazz->type->name ));

    object = M_ALLOC( clazz->type->object_size );

    object->clazz     = clazz;
    object->ref_count = 1;

    if (clazz->object_init)
    {
      M_Core  mcore = clazz->mcore;

      /* save idata within the class */
      clazz->idata = idata;

      m_core_push( mcore, object, (M_CleanupFunc) clazz->object_done, 0 );
      {
        clazz->object_init( object );
      }
      m_core_pop( mcore, object, M_POP_KEEP );
    }
    return object;
  }


  MLIB_APIDEF(M_Pointer)
  m_new_from_type( M_Core       mcore,
                   M_ClassType  class_type,
                   M_Pointer    idata )
  {
    M_Class  clazz;

    m_assert_is_core(mcore);

    clazz = m_class_from_type( class_type, mcore );

    return  m_new_full( clazz, idata );
  }



  MLIB_APIDEF(void)
  m_ref( M_Pointer object_ )
  {
    if (object_)
    {
      M_Object  object = object_;
      M_Class   clazz;

      m_assert_is_object(object);

      clazz = object->clazz;
      if ( clazz->object_incr )
        clazz->object_incr( M_OBJECT(object), 1 );
      else
        M_OBJECT(object)->ref_count++;
    }
  }


  MLIB_APIDEF(void)
  m_unref( M_Pointer object_ )
  {
    if (object_)
    {
      M_Bool    destroy;
      M_Object  object = object_;
      M_Class   clazz;

      m_assert_is_object(object);

      clazz = object->clazz;
      if ( clazz->object_incr )
        destroy = clazz->object_incr( M_OBJECT(object), -1 );
      else
        destroy = ( --M_OBJECT(object)->ref_count == 0 );

      if ( destroy )
      {
        M_Memory  memory;

        if ( object->ref_count != 0 )
          m_panic( "mlib.unref: custom class 'unref' malfunction\n" );

        /* call all finalizers from bottom to top      */
        /* stop the traversal as soon as one finalizer */
        /* returns TRUE..                              */
        memory = clazz->memory;

        do
        {
          if ( clazz->object_done( object ) )
            break;

          clazz = clazz->super;
        }
        while (clazz);

        object->clazz = NULL;
        M_FREE(object);
      }
    }
  }



  MLIB_APIDEF(M_Pointer)
  m_copy( M_Pointer object_ )
  {
    M_Object  copy = NULL;

    if (object_)
    {
      M_Object     object = object_;
      M_Class      clazz;
      M_ClassType  type;
      M_Memory     memory;

      m_assert_is_object(object);

      clazz  = object->clazz;
      memory = clazz->memory;
      type   = clazz->type;

      m_assert( ( type->class_flags & M_CLASS_FLAG_NO_COPY ) == 0 );

      /* never copy objects if the SINGLETON flag is set */
      if ( type->class_flags & M_CLASS_FLAG_SINGLETON )
      {
        m_ref( object );
        return object;
      }

      /* allocate new copy */
      copy = M_ALLOC( type->object_size );

      /* set reference count to 1 */
      copy->clazz     = clazz;
      copy->ref_count = 1;

      /* copy contents */
      if ( clazz->object_copy )
      {
        m_core_push( clazz->mcore, copy, (M_CleanupFunc) clazz->object_done, 0 );
        {
          clazz->object_copy( object, copy );
        }
        m_core_pop( clazz->mcore, copy, M_POP_KEEP );
      }
      else
      {
        M_MEM_COPY( copy, object, type->object_size );
        copy->clazz     = clazz;
        copy->ref_count = 1;
      }
    }
    return copy;
  }



  MLIB_APIDEF(void)
  m_push( M_Pointer  object_ )
  {
    if ( object_ )
    {
      M_Object  object = object_;
      M_Class   clazz;

      m_assert_is_object(object);
      clazz = object->clazz;

      m_core_push( clazz->mcore, object, (M_CleanupFunc) m_unref, 0 );
    }
  }



  MLIB_APIDEF(void)
  m_pop( M_Pointer  object_ )
  {
    if ( object_ )
    {
      M_Object  object = object_;
      M_Class   clazz;

      m_assert_is_object(object);
      clazz = object->clazz;

      m_core_pop( clazz->mcore, object, M_POP_KEEP );
    }
  }



  MLIB_APIDEF(void)
  m_pop_unref( M_Pointer  object_ )
  {
    if ( object_ )
    {
      M_Object  object = object_;
      M_Class   clazz;

      m_assert_is_object(object);
      clazz = object->clazz;

      m_core_pop( clazz->mcore, object, M_POP_CLEAN );
    }
  }


  MLIB_APIDEF(M_Memory)
  m_get_memory( M_Pointer  object_ )
  {
    M_Object  object = object_;

    m_assert_is_object(object);
    return object->clazz->memory;
  }


  MLIB_APIDEF(M_Core)
  m_get_core( M_Pointer  object_ )
  {
    M_Object  object = object_;

    m_assert_is_object(object);
    return object->clazz->mcore;
  }



  MLIB_APIDEF(M_Bool)
  m_is_a( M_Pointer    object_,
          M_ClassType  class_type )
  {
    M_Object     object = object_;
    M_ClassType  type;

    if ( !M_IS_OBJECT(object) )
      return 0;

    if ( !class_type )
      return 1;

    type = object->clazz->type;
    do
    {
      if ( type == class_type )
        return 1;

      type = type->super;
    }
    while (type);

    return 0;
  }



  MLIB_APIDEF(M_Bool)
  m_is_instance_of( M_Pointer  object,
                    M_Class    clazz )
  {
    m_assert_is_clazz( clazz );
    return m_is_a( object, clazz->type );
  }



  MLIB_APIDEF(M_Bool)
  m_object_check( M_Pointer  object_ )
  {
    M_Object  object = object_;

    return M_IS_OBJECT(object);
  }


  MLIB_APIDEF(void)
  m_throw( M_Pointer    object_,
           M_Exception  exception )
  {
    M_Object  object = object_;

    m_assert_is_object( object );
    m_core_throw( object->clazz->mcore, exception );
  }
