#include <mlib.h>
#include <mlib/internal.h>

  MLIB_APIDEF(void)
  m_struct_read_fields( M_StructField  fields,
                        M_Pointer      object,
                        M_Byte*       *buffer )
  {
    M_Byte*  p = *buffer;
    
    do
    {
      M_UInt32  value;
      M_Int     sign_shift;
      
      switch (fields->value)
      {
        case m_struct_op_start:  /* access a new struct - ignore here ..*/
          {
            fields++;
            continue;  /* loop ! */
          }
          
        case m_struct_op_byte:
        case m_struct_op_schar:  /* read a single byte */
          {
            value      = M_NEXT_BYTE(p);
            sign_shift = 24;
            break;
          }
          
        case m_struct_op_int16_be:
        case m_struct_op_uint16_be:  /* read a 2-byte big-endian short */
          {
            value      = M_NEXT_UINT16(p);
            sign_shift = 16;
            break;
          }

        case m_struct_op_int16_le:
        case m_struct_op_uint16_le:  /* read a 2-byte little-endian short */
          {
            value      = M_NEXT_UINT16_LE(p);
            sign_shift = 16;
            break;
          }

        case m_struct_op_int32_be:
        case m_struct_op_uint32_be:  /* read a 4-byte big-endian long */
          {
            value      = M_NEXT_UINT32(p);
            sign_shift = 0;
            break;
          }

        case m_struct_op_int32_le:
        case m_struct_op_uint32_le:  /* read a 4-byte little-endian long */
          {
            value      = M_NEXT_UINT32_LE(p);
            sign_shift = 0;
            break;
          }

        case m_struct_op_int24_be:
        case m_struct_op_uint24_be:  /* read a 3-byte big-endian long */
          {
            value      = M_NEXT_UINT24(p);
            sign_shift = 8;
            break;
          }

        case m_struct_op_int24_le:
        case m_struct_op_uint24_le:  /* read a 3-byte little-endian long */
          {
            value      = M_NEXT_UINT24_LE(p);
            sign_shift = 8;
            break;
          }

        case m_struct_op_bytes:  /* read contiguous bytes */
        case m_struct_op_skip:
          {
            if (fields->value == m_struct_op_bytes)
              M_MEM_COPY( (M_Byte*)object + fields->offset, p, fields->size );
            
            p += fields->size;
            fields++;
            continue;
          }

        default:
          /* otherwise, exit the loop */
          M_ERROR(( "%s: invalid structure field opcode %02x\n",
                    fields->value, "mlib.struct.read_fields" ));
          goto Exit;
      }
      
      /* now, compute the signed value is necessary */
      if (fields->value & M_STRUCT_OP_SIGNED)
        value   = (M_UInt32)((M_Int32)(value << sign_shift) >> sign_shift);
      
      /* finally, store the value in the object */
      {
        M_Byte*  q = (M_Byte*)object + fields->offset;
        
        
        switch (fields->size)
        {
          case 1:
            *(M_Byte*)q = (M_Byte)  value;
            break;
            
          case 2:
            *(M_UInt16*)q = (M_UInt16)value;
            break;
            
          case 4:
            *(M_UInt32*)q = (M_UInt32) value;
            break;
            
#if M_SIZEOF_LONG == 8
          case 8:
            /* for 64-bit systems only */
            *(M_ULong*)q = (M_UInt32) value;
#endif
          default:
            m_assert_unreached;
        }
      }
      
      /* go to next field */
      fields++;
    }
    while (1);

  Exit:
    *buffer = p;
  }


  MLIB_APIDEF(void)
  m_struct_write_fields( M_StructField  fields,
                         M_Pointer      object,
                         M_Byte*       *buffer )
  {
    M_Byte*  p = *buffer;
    
    do
    {
      /* exit if needed */
      switch (fields->value)
      {
        case m_struct_op_start:
          break;
          
        case m_struct_op_end:
          {
            goto Exit;
          }
          
        case m_struct_op_bytes:
          {
            M_MEM_COPY( p, (M_Byte*)object + fields->offset, fields->size );
            p += fields->size;
            break;
          }
          
        case m_struct_op_skip:
          {
            p += fields->size;
            break;
          }
          
        case m_struct_op_byte:
        case m_struct_op_schar:
        case m_struct_op_int16_be:
        case m_struct_op_int16_le:
        case m_struct_op_uint16_be:
        case m_struct_op_uint16_le:
        case m_struct_op_int24_be:
        case m_struct_op_int24_le:
        case m_struct_op_uint24_be:
        case m_struct_op_uint24_le:
        case m_struct_op_int32_be:
        case m_struct_op_int32_le:
        case m_struct_op_uint32_be:
        case m_struct_op_uint32_le:
          {
            M_UInt32  value;
            M_Int     sign_shift;
            M_Byte*   q = (M_Byte*)object + fields->offset;
          
            switch (fields->size)
            {
              case 1:
                value      = *(M_Byte*)q;
                sign_shift = 24;
                break;
                
              case 2:
                value      = *(M_UInt16*)q;
                sign_shift = 16;
                break;
                
              case 4:
                value      = *(M_UInt32*)q;
                sign_shift = 24;
                break;
                  
#if M_SIZEOF_LONG == 8
              case 8:  /* for 64-bit systems */
                value      = *(M_ULong*)q;
                sign_shift = 0;
#endif
                
              default: /* should not happen */
                m_assert_unreached;
                value      = 0;
                sign_shift = 0;
                ;
            }

            if (fields->value & M_STRUCT_OP_SIGNED)
              value = (M_UInt32)((M_Int32)(value << sign_shift) >> sign_shift);

            switch (fields->value)
            {
              case m_struct_op_byte:
              case m_struct_op_schar:
                {
                  M_WRITE_BYTE(p,value);
                  break;
                }
                
              case m_struct_op_int16_be:
              case m_struct_op_uint16_be:
                {
                  M_WRITE16_BE(p,value);
                  break;
                }
              
              case m_struct_op_int16_le:
              case m_struct_op_uint16_le:
                {
                  M_WRITE16_LE(p,value);
                  break;
                }
              
              case m_struct_op_int24_be:
              case m_struct_op_uint24_be:
                {
                  M_WRITE24_BE(p,value);
                  break;
                }
              
              case m_struct_op_int24_le:
              case m_struct_op_uint24_le:
                {
                  M_WRITE24_LE(p,value);
                  break;
                }
              
              case m_struct_op_int32_be:
              case m_struct_op_uint32_be:
                {
                  M_WRITE32_BE(p,value);
                  break;
                }
              
              case m_struct_op_int32_le:
              case m_struct_op_uint32_le:
                {
                  M_WRITE32_LE(p,value);
                  break;
                }
              
              default:
                ;
            }
            break;
          }
        
        default:
          M_ERROR(( "%s: invalid structure field opcode %02x\n",
                    fields->value, "mlib.struct.write_fields" ));
          goto Exit;
      }
      fields++;
    }
    while (1);

  Exit:  
    *buffer = p;
  }

