#include <mlib/lhash.h>

#include <stdio.h>

#define  MAX_NODES_DEFAULT 1000000

  static M_Core          core;
  static M_Memory        memory;
  static M_Error         error;
  static M_LHashRec      table;

  static M_Memory        node_pool;

  static long            max_nodes = MAX_NODES_DEFAULT;

#define  KEY(x)  M_POINTER_FROM_DIST(x)

typedef struct MyNodeRec_
{
  M_LHNodeRec  lh;
  M_Pointer    key;
  M_Pointer    value;

} MyNodeRec, *MyNode;


static MyNode
my_node_alloc( )
{
  return (MyNode) m_mem_alloc( sizeof(MyNodeRec), node_pool );
}


static void
my_node_free( MyNode  node )
{
  M_Memory  memory = node_pool;

  M_FREE( node );
}


  static M_Bool  equal( MyNode  n1, MyNode  n2 )
  {
    return n1->key == n2->key;
  }

  static M_UInt32  hash( M_Pointer  key )
  {
    return M_POINTER_TO_DIST(key);
  }


static void
m_lhash_set( M_LHash    table,
             M_Pointer  key,
             M_Pointer  value )
{
  M_LHLookup  lookup;
  MyNodeRec   key_node;
  MyNode      node;

  key_node.key     = key;
  key_node.lh.hash = hash( key );

  lookup = m_lhash_lookup( table, (M_LHNode) &key_node );
  node   = (MyNode) *lookup;
  if ( node != NULL )
  {
    node->value = value;
    return;
  }

  node  = my_node_alloc();
  *node = key_node;

  node->value = value;
  m_lhash_add( table, lookup, (M_LHNode) node );
}


static M_Pointer
m_lhash_get( M_LHash    table,
             M_Pointer  key )
{
  M_LHLookup  lookup;
  M_Pointer   result = NULL;
  MyNodeRec   key_node;
  MyNode      node;

  key_node.key     = key;
  key_node.lh.hash = hash( key );

  lookup = m_lhash_lookup( table, (M_LHNode) &key_node );
  node   = (MyNode) *lookup;
  if ( node != NULL )
    result = node->value;

  return result;
}


static void
m_lhash_unset( M_LHash    table,
               M_Pointer  key )
{
  M_LHLookup  lookup;
  M_Pointer   result = NULL;
  MyNodeRec   key_node;
  MyNode      node;

  key_node.key     = key;
  key_node.lh.hash = hash( key );

  lookup = m_lhash_lookup( table, (M_LHNode) &key_node );
  node   = (MyNode) *lookup;
  if ( node != NULL )
  {
    m_lhash_remove( table, lookup );
    my_node_free( node );
  }
}



int main( int  argc, const char** argv )
{
  M_Int  n, use_pool = 1;


  if ( argc >= 2 )
  {
    max_nodes = atoi( argv[1] );
    if ( max_nodes < 1 )
      max_nodes = 1;

    if ( argc >= 3 )
      use_pool = 0;
  }

  error = m_core_new( 0, 0, 0, &core );
  if (error)
    M_PANIC(( "%s: could not create new MLib core", "test.lhash" ));

  memory = core->memory;

  node_pool = memory;
  if ( use_pool )
    node_pool = m_memory_pool_find( memory, sizeof(MyNodeRec) );

  if ( node_pool == memory )
    M_ERROR(( "not using memory pools" ));

  /* initialize table */
  m_lhash_init( &table, memory, (M_EqualFunc) equal );

  /* insert all nodes */
  for ( n = 0; n < max_nodes; n++ )
    m_lhash_set( &table, KEY(n), KEY(n) );

  /* return table depth and check all nodes */
  M_ERROR(( "table size = %d\n",
             m_lhash_get_size( &table ) ));

  /* check lookup */
  for ( n = 0; n < max_nodes; n++ )
  {
    if ( m_lhash_get( &table, KEY(n) ) != KEY(n) )
      M_PANIC(( "%s: couldn't lookup element # %d", "test.lhash", n ));
  }

  /* replace one node with another */
  m_lhash_set( &table, KEY(1), KEY(max_nodes) );
  if ( m_lhash_get( &table, KEY(1) ) != KEY(max_nodes) )
    M_PANIC(( "%s: table replacement didn't work !!\n", "test.lhash" ));

  for ( n = 0; n < max_nodes; n++ )
  {
    if ( n != 1 && m_lhash_get( &table, KEY(n) ) != KEY(n) )
      M_PANIC(( "%s: couldn't lookup element # %d", "test.lhash", n ));
  }

  /* now, remove all nodes */
  for ( n = 0; n < max_nodes; n++ )
    m_lhash_unset( &table, KEY(n) );

  if ( m_lhash_get_size(&table) != 0 )
    M_PANIC(( "table not empty on exit" ));

  m_lhash_done( &table, NULL );

  m_core_unref( core );

  printf( "hash table test: ok !\n" );
  return 0;
}
