/* simple string management */
#ifndef __M_STRING_H__
#define __M_STRING_H__

#include <mlib.h>

M_BEGIN_HEADER

/******************************************************************
 *
 * @section: strings
 *
 * @title: Strings
 *
 * @abstract: Static C string management
 *
 * @description:
 *    This section contains many functions and macros used to manage
 *    C strings. Note that in order to prevent common but extremely
 *    annoying bugs (read: buffer overwrites), the MLib departs from
 *    the ANSI C library in several ways:
 *
 *
 *    - first of all, all bounded strings are _always_ zero terminated,
 *      unlike the ones supported by the ANSI C library.
 *
 *      on one hand, this means that a bounded string of size 'N' can hold
 *      at most 'N-1' non-zero characters (which is really rarely a problem).
 *      on the other hand, there is no direct equivalent to "strnlen",
 *      "strncasecmp", "strncpy", etc..
 *      etc..
 *
 *    - there are no APIs to support "insecure" operations. For example,
 *      using "strcpy", which copies memory without checking room in the
 *      target buffer is the source of many security exploits in all
 *      kinds of software.
 *
 *
 *    - the 'case' functions (e.g. m_str_case_compare, m_str_toupper, etc..)
 *      _only_ support ASCII and completely ignore specific locales. This
 *      is intentional since we want to _strongly_ encourage you to use
 *      Unicode (and UTF8/UTF16) for any kind of internationalisation
 *      feature you may need..
 *
 *
 *   note that the MLib also provides dynamic strings, called "ropes".
 *   you should use them whenever you need to perform string processing
 *   instead of relying exclusively on the functions there.. ropes also
 *   provide many useful operations like insertion, removal, cropping,
 *   that are not directly supported here, and that you'd better not
 *   rewrite on your own..
 */


/******************************************************************
 *
 * @macro:  M_STRING (x)
 *
 * @description:
 *    a macro to cast anything to a string handle
 *
 */
#define  M_STRING(x)    ((M_String)(x))


/******************************************************************
 *
 * @macro:  M_STRING_P (x)
 *
 * @description:
 *    a macro to cast anything to a pointer to a string handle
 *
 */
#define  M_STRING_P(x)  ((M_String*)(x))


/******************************************************************
 *
 * @macro:  M_CSTRING (x)
 *
 * @description:
 *    a macro to cast anything to a constant string handle
 *
 */
#define  M_CSTRING(x)    ((M_CString)(x))


/******************************************************************
 *
 * @macro:  M_CSTRING_P (x)
 *
 * @description:
 *    a macro to cast anything to a pointer to a constant string handle
 *
 */
#define  M_CSTRING_P(x)  ((M_CString*)(x))


/******************************************************************
 *
 * @function: m_str_compare (s1,s2)
 *
 * @description:
 *    a function used to compare two strings. can be used as a @M_CompareFunc
 *
 * @input:
 *    s1 :: first string pointer. can be NULL
 *    s2 :: second string pointer. can be NULL
 *
 * @return:
 *   -1, 0 or +1 depending on the string comparison.
 *
 * @note:
 *   a NULL pointer is considered to be strictly equivalent to an
 *   empty string
 */
  MLIB_API(M_Int)   m_str_compare( M_CString  s1,
                                   M_CString  s2 );


/******************************************************************
 *
 * @function: m_str_equal (s1,s2)
 *
 * @description:
 *    a function used to test two strings for equality. can be used as a
 *    @M_EqualFunc
 *
 * @input:
 *    s1 :: first string pointer. can be NULL
 *    s2 :: second string pointer. can be NULL
 *
 * @return:
 *   TRUE if the two strings are strictly equal
 *
 * @note:
 *   a NULL pointer is considered strictly equal to an empty string
 */
 MLIB_API(M_Bool)  m_str_equal( M_CString  s1,
                                M_CString  s2 );


/******************************************************************
 *
 * @function: m_str_compare (s1,s2)
 *
 * @description:
 *    a function used to compare two strings irrespective of case.
 *    can be used as a @M_CompareFunc
 *
 * @input:
 *    s1 :: first string pointer. can be NULL
 *    s2 :: second string pointer. can be NULL
 *
 * @return:
 *   -1, 0 or +1 depending on the string comparison.
 *
 * @note:
 *   a NULL pointer is considered to be strictly equivalent to an
 *   empty string
 *
 *   this function _only_ works for ANSI strings.
 */
  MLIB_API(M_Int)   m_str_case_compare( M_CString  s1,
                                        M_CString  s2 );


/******************************************************************
 *
 * @function: m_str_equal (s1,s2)
 *
 * @description:
 *    a function used to test two strings for equality, irrespective
 *    of case. can be used as a @M_EqualFunc
 *
 * @input:
 *    s1 :: first string pointer. can be NULL
 *    s2 :: second string pointer. can be NULL
 *
 * @return:
 *   TRUE if the two strings are strictly equal, irrespective of case
 *
 * @note:
 *   a NULL pointer is considered strictly equal to an empty string
 */
 MLIB_API(M_Bool)   m_str_case_equal( M_CString  s1,
                                      M_CString  s2 );


/******************************************************************
 *
 * @function: m_str_len (string)
 *
 * @description:
 *    return the size of a string in characters (excluding the trailing
 *    zero).
 *
 * @input:
 *   string :: input string pointer. can be NULL
 *
 * @return:
 *   string length
 *
 * @note:
 *   unlike the ANSI 'strlen', this function accepts NULL string pointers
 *   and will return 0 for them..
 */
  MLIB_API( M_Size )   m_str_len( M_CString  string );


/******************************************************************
 *
 * @function: m_str_chr
 *
 * @description:
 *    find the first occurence of either a given character or '\0'
 *    in a string and return its index. Note that this function doesn't
 *    return a pointer but an index..
 *
 * @input:
 *    string :: source string pointer
 *    c      :: character to look for
 *
 * @return:
 *    index of first occurence of character in the string, otherwise
 *    the string length
 *
 * @note:
 *    'string' cannot be NULL
 *
 *    to test success, look at 'string[ m_str_chr(string,c) ]', if
 *    it is 0, the search failed..
 */
  MLIB_API(M_Size)   m_str_chr( M_CString  string,
                                char       c );


/******************************************************************
 *
 * @function: m_str_rchr
 *
 * @description:
 *    find the last occurence of either a given character or '\0'
 *    in a string and return its index. Note that this function doesn't
 *    return a pointer but an index..
 *
 * @input:
 *    string :: source string pointer
 *    c      :: character to look for
 *
 * @return:
 *    index of last occurence of character in the string, otherwise
 *    the string length
 *
 * @note:
 *    'string' cannot be NULL
 *
 *    to test success, look at 'string[ m_str_rchr(string,c) ]', if
 *    it is 0, the search failed..
 */
  MLIB_API(M_Size)   m_str_rchr( M_CString  string,
                                 char       c );


/******************************************************************
 *
 * @function: m_str_start
 *
 * @description:
 *    returns true iff a given string starts with a given prefix
 *
 * @input:
 *    string :: the string to scan
 *    prefix :: the prefix to look up
 *
 * @return:
 *    boolean.
 */
  MLIB_API(M_Bool)   m_str_start( M_CString  string,
                                  M_CString  prefix );


/******************************************************************
 *
 * @function: m_str_new
 *
 * @description:
 *    a function used to allocate a new string
 *
 * @input:
 *    length  :: length of string in characters
 *    memory  :: memory allocator to use
 *    push    :: boolean. when set, the new string is pushed on the
 *               cleanup stack
 * @return:
 *    pointer to new string. NULL if 'length' is 0.
 *
 * @note:
 *    if length is > 0, then this function allocates 'length+1' bytes and
 *    always sets the string's buffer to 0
 *
 * @throws: m_err_out_of_memory
 *
 * @also:
 *    @m_str_copy, @m_str_destroy
 */
  MLIB_API(M_String)   m_str_new( M_Size       length,
                                  M_Memory     memory,
                                  M_Bool       push );


/******************************************************************
 *
 * @function: m_str_dup
 *
 * @description:
 *    a function used to duplicate a given string
 *
 * @input:
 *    source  :: source string
 *    memory  :: memory allocator to use for new string
 *    push    :: boolean. When set, the new string is pushed on the
 *               cleanup stack
 *
 * @return:
 *    string copy. its size is "source.length+1" bytes
 *
 * @throws: m_err_out_of_memory
 *
 * @also:
 *    @m_str_new, @m_str_destroy
 */
  MLIB_API(M_String)   m_str_dup( M_CString   source,
                                  M_Memory    memory,
                                  M_Bool      push );


/******************************************************************
 *
 * @function: m_str_free
 *
 * @description:
 *    destroy a given string
 *
 * @input:
 *    memory  :: memory allocator to use for new string
 *
 * @inout:
 *    astring :: the target string address. set to NULL on exit
 *
 * @also:
 *    @m_str_new, @m_str_dup
 */
  MLIB_API(void)   m_str_free( M_CString   string,
                               M_Memory    memory );



/******************************************************************
 *
 * @function: m_str_push
 *
 * @description:
 *    push a given string on top of the cleanup stack
 *
 * @input:
 *    string  :: string to push
 *    memory  :: memory manager to use
 *
 * @also:
 *    @m_str_pop
 */
  MLIB_API(void)   m_str_push( M_CString   string,
                               M_Memory    memory );


/******************************************************************
 *
 * @function: m_str_pop
 *
 * @description:
 *    pops a string that was previously pushed on the cleanup stack,
 *    either with @m_str_push, @m_str_new or @m_str_copy
 *
 * @input:
 *    string  :: string to pop
 *    memory  :: memory allocator to use
 *    action  :: action to perform on pop
 *
 * @also:
 *    @m_str_push
 */
  MLIB_API(void)   m_str_pop( M_CString    string,
                              M_Memory     memory,
                              M_PopAction  action );


/******************************************************************
 *
 * @function: m_str_toupper
 *
 * @description:
 *    converts a given input string to uppercase
 *
 * @input:
 *    string :: handle to source string
 */
  MLIB_API(void)    m_str_toupper( M_String  string );


/******************************************************************
 *
 * @function: m_str_tolower
 *
 * @description:
 *    converts a given input string to lowercase
 *
 * @input:
 *    string :: handle to source string
 */
  MLIB_API(void)    m_str_tolower( M_String  string );


/******************************************************************
 *
 * @function: m_str_hash
 *
 * @description:
 *    a simple function used to hash strings.
 *
 * @input:
 *    string :: handle to source string
 *
 * @return:
 *    hash value, as unsigned long
 */
  MLIB_API(M_ULong)   m_str_hash( M_CString  string );

  /* */


/******************************************************************
 *
 * @function: m_strn_copy
 *
 * @description:
 *    copy an input string into a bounded string. This function
 *    will _always_ zero-terminate the target, unlike the ANSI
 *    "strncpy"..
 *
 * @input:
 *    target      :: address of target string buffer
 *    target_size :: size of target string buffer in bytes
 *    source      :: address of source string (must be zero-terminated)
 *
 * @return:
 *    number of characters copied (excluding the trailing zero)
 *
 * @note:
 *    the bounded strings managed by the ANSI C library functions
 *    (like strncpy) are not necessarily zero-terminated. Unfortunately,
 *    this is very often the source of considerable bugs and security
 *    problems in many, many C programs..
 *
 *    to prevent developers from such frequent and predictable mistakes, the
 *    MLib string functions will _always_ zero-terminate bounded strings.
 *    This also means that a bounded string has a size of 'N' bytes, it will
 *    only be able to hold 'N-1' characters
 *
 *    on the other hand, there is no need for equivalents of "strnlen"
 *    and a few other exotic functions..
 */
  MLIB_API(M_Size)     m_strn_copy( M_String   target,
                                    M_Size     target_size,
                                    M_CString  source );


/******************************************************************
 *
 * @function: m_strn_append
 *
 * @description:
 *    append a given input string to a bounded string. the target
 *    string is always zero-terminated..
 *
 *    be careful, this function can be _slow_
 *
 * @input:
 *    target      :: address of target string buffer
 *    target_size :: size of target string buffer in bytes
 *    source      :: address of source string (must be zero-terminated)
 *
 * @return:
 *    number of characters copied (excluding the trailing zero),
 *    can be 0..
 *
 * @note:
 *    see the note for @m_strn_copy regarding bounded strings
 */
  MLIB_API(M_Size)     m_strn_append( M_String   target,
                                      M_Size     target_size,
                                      M_CString  source );


/******************************************************************
 *
 * @function: m_strn_append_c
 *
 * @description:
 *    append a single character to a bounded string. the target
 *    string is always zero-terminated..
 *
 *    be careful, this function can be _slow_
 *
 * @input:
 *    target      :: address of target string buffer
 *    target_size :: size of target string buffer in bytes
 *    c           :: character code to append.
 *
 * @return:
 *    number of characters copied (i.e. 0 or 1)
 *
 * @note:
 *    see the note for @m_strn_copy regarding bounded strings
 *
 *    if 'c' is 0, nothing will happen..
 */
  MLIB_API(M_Size)     m_strn_append_c( M_String   target,
                                        M_Size     target_size,
                                        char       c );


/******************************************************************
 *
 * @function: m_str_dup_ansi_n
 *
 * @description:
 *    a function used to duplicate a given ANSI bounded-string
 *    (one that isn't necessarily zero-terminated)
 *
 * @input:
 *    source  :: source string (ANSI bounded string)
 *    maxlen  :: maximal source string length in bytes
 *    memory  :: memory allocator to use for new string
 *    push    :: boolean. When set, the new string is pushed on the
 *               cleanup stack
 *
 * @return:
 *    string copy. its size is "source.length+1" bytes
 *
 * @throws: m_err_out_of_memory
 *
 * @also:
 *    @m_str_new, @m_str_destroy
 */
  MLIB_API(M_String)   m_str_dup_ansi_n( M_CString   source,
                                         M_Size      maxlen,
                                         M_Memory    memory,
                                         M_Bool      push );



 /* */

/******************************************************************
 *
 * @function: m_tolower
 *
 * @description:
 *    convert an ASCII character to lowercase, e.g. 'X' => 'x'
 *
 * @input:
 *    c :: ASCII character code
 *
 * @return:
 *    lower-case ASCII character code.
 *
 * @note:
 *   unlike the ANSI libc, this function completely ignores locale-specific
 *   information. You are _strongly_ encouraged to use Unicode for all
 *   internationalisation purpose anyway !!
 */
  MLIB_API(M_Char)   m_tolower( M_Char  c );


/******************************************************************
 *
 * @function: m_toupper
 *
 * @description:
 *    convert an ASCII character to uppercase, e.g. 'x' => 'X'
 *
 * @input:
 *    c :: ASCII character code
 *
 * @return:
 *    upper-case ASCII character code.
 *
 * @note:
 *   unlike the ANSI libc, this function completely ignores locale-specific
 *   information. You are _strongly_ encouraged to use Unicode for all
 *   internationalisation purpose anyway !!
 */
  MLIB_API(M_Char)   m_toupper( M_Char  c );


/******************************************************************
 *
 * @macro: M_TO_LOWER (c)
 *
 * @description:
 *    a macro to perform to compute the result of @m_tolower quickly
 */
#define  M_TO_LOWER(x)    ( (M_UInt)((x)-'A') < 26 ? (x)+('a'-'A') : (x) )


/******************************************************************
 *
 * @macro: M_TO_UPPER (c)
 *
 * @description:
 *    a macro to perform to compute the result of @m_toupper quickly
 */
#define  M_TO_UPPER(x)    ( (M_UInt)((x)-'a') < 26 ? (x)+('A'-'a') : (x) )

/* */

#define  M_IS_SPACE(x)    ( (x) == ' '  || (x) == '\t' )

#define  M_IS_EOL(x)      ( (x) == '\n' || (x) == '\r' )

#define  M_IS_WHITE(x)    ( M_IS_SPACE(x) || M_IS_EOL(x) )

#define  M_IS_LETTER(x)   ( (M_UInt)((x)-'A') < 26 || (M_UInt)((x)-'a') < 26 )

#define  M_IS_DIGIT(x)    ( (M_UInt)((x)-'0') < 10 )

#define  M_IS_HEXDIGIT(x) ( (M_UInt)((x)-'0') < 10 || \
                            (M_UInt)((x)-'A') < 6  || \
                            (M_UInt)((x)-'a') < 6  )

#define  M_IS_ALPHA(x)    ( M_IS_DIGIT(x) || M_IS_LETTER(x) || (x) == '_' )

 /* */

MLIB_TEST(string)

M_END_HEADER

#endif /* __M_STRING_H__ */
